{*****************************************************************************}
{                                                                             }
{   This file is part of the LKRecords components Library                     }
{                                                                             }
{   See the file COPYING.LGPL.txt included in this distribution,              }
{   for details about the Licence and Copyright.                              }
{                                                                             }
{   This code is distributed in the hope that it will be useful,              }
{   but WITHOUT ANY WARRANTY; without even the implied warranty of            }
{   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      }
{                                                                             }
{   Author: Leslie John Kaye :    www.leskaye.com                             }
{                                                                             }
{   Abstract:                                                                 }
{      Components relating to TLKDBEditsPanel and LKDBGridController which    }
{      provide runtime manipulation of database controls                      }
{                                                                             }
{*****************************************************************************}

unit LKDBControls;

interface

uses
  SysUtils, Windows, Messages, Classes, System.UITypes,
  Graphics, Controls, Forms, Dialogs,
  ComCtrls, DB, DBCtrls, DBGrids, LKDBImageEditDlg, Mapi, ShellApi, StrUtils,
  Variants, StdCtrls, Mask, Math, JvExComCtrls, JvDateTimePicker,
  JvDBDateTimePicker, JvExMask, JvToolEdit, JvMaskEdit, JvCheckedMaskEdit,
  JvDatePickerEdit, JvDBDatePickerEdit, JvExStdCtrls, JvRichEdit, JvDBRichEdit,
  JvDBCheckBox, JvDBImage, JvDBFindEdit, JvCombobox,
  JvDBSearchComboBox, JvEdit, JvDBSearchEdit, JvHtControls, JvDBHTLabel,
  JvDBLookupComboEdit, JvDBControls, JvDBUltimGrid, Grids, JvExDBGrids,
  JvDBGrid, JvDBLookupTreeView, JvDBTreeView, JvDBCombobox, ExtCtrls,
  JvExExtCtrls, JvExtComponent, JvDBRadioPanel, JvDBLookup, JvExControls,
  JvSpin, JvDBSpinEdit, LKDefine, LKRecordsComponents;

type

  TLKEditsLayout = (aceHorizontal, aceOneCol, aceTwoCol, aceThreeCol);

  TLKDBControlType = (dbcUnassigned,
    // VCL or JVCL
    dbcText, dbcEdit, dbcMemo, dbcImage, dbcListBox, dbcComboBox, dbcCheckbox,
    dbcRadioGroup, dbcLookUpListBox, dbcLookUpComboBox, dbcRichEdit,
    // LK
    dbcButtonEdit, dbcFilePath, dbcURL, dbcEMail, dbcImageFile, dbcDCLookup,
    // JVCL only
    dbcDatePicker, dbcDateTimePicker, dbcProgressBar, dbcSpinEdit,
    dbcRadioPanel, dbcTreeView, dbcLookupTreeViewCombo, dbcLookupTreeView,
    dbcComboEdit, dbcDateEdit, dbcTimeEdit, dbcCalcEdit, dbcMaskEdit,
    dbcLookupComboEdit, dbcSearchEdit, dbcSearchComboBox, dbcFindEdit
    );

  TLKDBButtonEdit = class(TDBEdit)
  private
    FButton: TButton;
    FOnButtonClick: TNotifyEvent;
    procedure DoButtonClick(Sender: TObject);
  protected
    procedure DefaultButtonClick(Sender: TObject); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    property Button: TButton read FButton write FButton;
  published
    property OnButtonClick: TNotifyEvent read FOnButtonClick write
        FOnButtonClick;
  end;

  TLKDBDateEdit = class(TLKDBButtonEdit)
  private
    FCalendar: TMonthCalendar;
  protected
    procedure CalendarDblClick(Sender: TObject);
    procedure DefaultButtonClick(Sender: TObject); override;
    property Calendar: TMonthCalendar read FCalendar write FCalendar;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TLKDBImageEdit = class(TLKDBButtonEdit)
  private
    FImageFilePath: string;
    FImgDlg: TDlgLKDBImageEdit;
  protected
    procedure DefaultButtonClick(Sender: TObject); override;
  public
    property ImgDlg: TDlgLKDBImageEdit read FImgDlg;
  published
    property ImageFilePath: string read FImageFilePath write FImageFilePath;
  end;

  TLKDBEMailEdit = class(TLKDBButtonEdit)
  protected
    procedure DefaultButtonClick(Sender: TObject); override;
  end;

  TLKDBUrlEdit = class(TLKDBButtonEdit)
  protected
    procedure DefaultButtonClick(Sender: TObject); override;
  end;

  TLKDBFilePathEdit = class(TLKDBButtonEdit)
  protected
    procedure DefaultButtonClick(Sender: TObject); override;
  end;

  TLKDBDcLookupEdit = class(TLKDBButtonEdit)
  private
    FLookupDataComponent: TLKDataComponent;
    FLookupDataset: TLKACTable;
    FLookupFieldName: string;
    FLookupOnExit: Boolean;
    FOnLookupFind: TNotifyEvent;
    procedure DoLookupFind;
    function GetLookupFieldName: string;
  protected
    procedure DefaultButtonClick(Sender: TObject); override;
    procedure DoExit; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property LookupDataComponent: TLKDataComponent read FLookupDataComponent
        write FLookupDataComponent;
    property LookupDataset: TLKACTable read FLookupDataset write FLookupDataset
        default tbMaster;
    property LookupFieldName: string read GetLookupFieldName write
        FLookupFieldName;
    property LookupOnExit: Boolean read FLookupOnExit write FLookupOnExit
        default True;
    property OnLookupFind: TNotifyEvent read FOnLookupFind write FOnLookupFind;
  end;

  TLKDBControl = class abstract(TCollectionItem)
  private
    FAlignment: TAlignment;
    FCaption: string;
    FControlOwner: TComponent;
    FControlParent: TWinControl;
    FControlType: TLKDBControlType;
    FDataSource: TDataSource;
    FOnEnter: TNotifyEvent;
    FOnExit: TNotifyEvent;
    FOnKeyPress: TKeyPressEvent;
    FVisible: Boolean;
    function GetFieldName: string;
    procedure SetField(AField: TField);
  protected
    FEditControl: TControl;
    FField: TField;
    FReadOnly: Boolean;
    function CreateControlsFromControlType: Boolean; virtual;
    function GetWidth: Integer; virtual;
    procedure SetAlignment(Value: TAlignment); virtual;
    procedure SetCaption(const Value: string); virtual;
    procedure SetControlParent(AParent: TWinControl); virtual;
    procedure SetControlType(const Value: TLKDBControlType); virtual;
    procedure SetControlTypeFromField;
    procedure SetDataSource(ADatasource: TDataSource); virtual;
    procedure SetEditControlDataField; virtual;
    procedure SetOnEnter(const Value: TNotifyEvent); virtual;
    procedure SetOnExit(const Value: TNotifyEvent); virtual;
    procedure SetOnKeyPress(const Value: TKeyPressEvent); virtual;
    procedure SetReadOnly(const Value: Boolean); virtual;
    procedure SetVisible(const Value: Boolean); virtual;
    procedure SetWidth(const Value: Integer); virtual;
  public
    constructor Create(Collection: TCollection); override;
    procedure DoEditOnExit(Sender: TObject); virtual;
    property Alignment: TAlignment read FAlignment write SetAlignment;
    property Caption: string read FCaption write SetCaption;
    property ControlOwner: TComponent read FControlOwner write FControlOwner;
    property ControlParent: TWinControl read FControlParent write
        SetControlParent;
    property ControlType: TLKDBControlType read FControlType write
        SetControlType;
    property DataSource: TDataSource read FDataSource write SetDataSource;
    property EditControl: TControl read FEditControl;
    property Field: TField read FField write SetField;
    property FieldName: string read GetFieldName;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly;
    property Visible: Boolean read FVisible write SetVisible;
    property Width: Integer read GetWidth write SetWidth;
    property OnEnter: TNotifyEvent read FOnEnter write SetOnEnter;
    property OnExit: TNotifyEvent read FOnExit write SetOnExit;
    property OnKeyPress: TKeyPressEvent read FOnKeyPress write SetOnKeyPress;
  end;

  TLKDBControlCollection = class abstract(TCollection)
  private
    FControlOwner: TComponent;
  protected
    FExcludeFieldList: TStringList;
    FReadOnly: Boolean;
    function GetItem(Index: Integer): TLKDBControl;
    procedure SetItem(Index: Integer; Value: TLKDBControl);
    procedure SetReadOnly(const Value: Boolean); virtual;
  public
    destructor Destroy; override;
    function Add: TLKDBControl;
    procedure Clear;
    procedure ConstructFromDatasource(const ExcludeFieldNames: string); virtual;
    procedure DoOnFormClose;
    function FormatGet: string; virtual;
    function FormatSet(const Value: string): string; virtual;
    property ControlOwner: TComponent read FControlOwner write FControlOwner;
    property Items[Index: Integer]: TLKDBControl read GetItem write SetItem;
        default;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly;
  end;

  TLKDcDBGridControl = class(TLKDBControl)
  strict private
    FGrid: TJvDBUltimGrid;
  private
    function GetGridControl: TJvDBGridControl;
    procedure SetGrid(Value: TJvDBUltimGrid);
  protected
    FGridColumn: TColumn;
    function CreateControlsFromControlType: Boolean; override;
    function GetWidth: Integer; override;
    procedure SetAlignment(Value: TAlignment); override;
    procedure SetCaption(const Value: string); override;
    procedure SetControlType(const Value: TLKDBControlType); override;
    procedure SetDisplayName(const Value: string); override;
    procedure SetGridColumn(Value: TColumn);
    procedure SetReadOnly(const Value: Boolean); override;
    procedure SetVisible(const Value: Boolean); override;
    procedure SetWidth(const Value: Integer); override;
  public
    procedure DoEditOnExit(Sender: TObject); override;
    property Grid: TJvDBUltimGrid read FGrid write SetGrid;
    property GridColumn: TColumn read FGridColumn write SetGridColumn;
    property GridControl: TJvDBGridControl read GetGridControl;
  end;

  TLKDcDBGridControlClass = class of TLKDcDBGridControl;

  TLKDBGridCollection = class(TLKDBControlCollection)
  private
    FGrid: TJvDBUltimGrid;
    procedure RebulldControls;
  protected
    function GetItem(Index: Integer): TLKDcDBGridControl;
    procedure SetItem(Index: Integer; Value: TLKDcDBGridControl);
    procedure SetReadOnly(const Value: Boolean); override;
  public
    constructor Create(ItemClass: TLKDcDBGridControlClass);
    function Add: TLKDcDBGridControl;
    procedure ConstructFromDatasource(const ExcludeFieldNames: string);
        override;
    function FormatGet: string; override;
    function FormatSet(const Value: string): string; override;
    procedure SetFocus;
    property Grid: TJvDBUltimGrid read FGrid write FGrid;
    property Items[Index: Integer]: TLKDcDBGridControl read GetItem write
        SetItem; default;
  end;

  TLKDcDBEditControl = class(TLKDBControl)
  private
    FColumn: Integer;
  protected
    FCaptionLabel: TStaticText;
    FLayout: TLKEditsLayout;
    FLeft: Integer;
    FLinesHeight: Double;
    FTabOrder: Integer;
    FTop: Integer;
    function CreateControlsFromControlType: Boolean; override;
    function GetLinesHeight: Double;
    function GetWidth: Integer; override;
    procedure SetCaption(const Value: string); override;
    procedure SetControlParent(AParent: TWinControl); override;
    procedure SetControlType(const Value: TLKDBControlType); override;
    procedure SetDisplayName(const Value: string); override;
    procedure SetLeft(const Value: Integer);
    procedure SetLinesHeight(const Value: Double);
    procedure SetReadOnly(const Value: Boolean); override;
    procedure SetTabOrder(const Value: Integer);
    procedure SetTop(const Value: Integer);
    procedure SetVisible(const Value: Boolean); override;
    procedure SetWidth(const Value: Integer); override;
  public
    constructor Create(Collection: TCollection); override;
    property CaptionLabel: TStaticText read FCaptionLabel;
    property Column: Integer read FColumn write FColumn;
    property Layout: TLKEditsLayout read FLayout write FLayout;
    property Left: Integer read FLeft write SetLeft;
    property LinesHeight: Double read GetLinesHeight write SetLinesHeight;
    property TabOrder: Integer read FTabOrder write SetTabOrder;
    property Top: Integer read FTop write SetTop;
  end;

  TLKDcDBEditControlClass = class of TLKDcDBEditControl;

  TLKDBEditCollection = class(TLKDBControlCollection)
  private
    FControlDataSource: TDataSource;
    FControlParent: TWinControl;
  protected
    function GetItem(Index: Integer): TLKDcDBEditControl;
    procedure SetControlDataSource(ADataSource: TDataSource);
    procedure SetControlParent(AParent: TWinControl);
    procedure SetItem(Index: Integer; Value: TLKDcDBEditControl);
  public
    constructor Create(ItemClass: TLKDcDBEditControlClass);
    function Add: TLKDcDBEditControl;
    procedure Clear;
    procedure ConstructFromDatasource(const ExcludeFieldNames: string);
        override;
    function FormatGet: string; override;
    function FormatSet(const Value: string): string; override;
    procedure UpdateFields;
    property ControlDataSource: TDataSource read FControlDataSource write
        SetControlDataSource;
    property ControlParent: TWinControl read FControlParent write
        SetControlParent;
    property Items[Index: Integer]: TLKDcDBEditControl read GetItem write
        SetItem; default;
  end;

  TLKDcDBEditsPanel = class(TPanel)
  private
    FColWidth0: Integer;
    FColWidth1: Integer;
    FColWidth2: Integer;
    FDisplayHeight: Integer;
    FEditControls: TLKDBEditCollection;
    FLayout: TLKEditsLayout;
    function GetCount: Integer;
    function GetDataSource: TDataSource;
    function GetReadOnly: Boolean;
    procedure SetDataSource(ADataSource: TDataSource);
    procedure SetLayout(const Value: TLKEditsLayout);
    procedure SetReadOnly(const Value: Boolean);
  protected
    function GetFormat: string;
    function GetItem(Index: Integer): TLKDcDBEditControl;
    procedure Resize; override;
    procedure SetFormat(const Value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ConstructFromDatasource(const ExcludeFieldNames: string);
    function DlgEditListPropsExecute(AOwner: TComponent): TModalResult;
    procedure DoOnFormClose;
    function LoadFormat(const AComponent: TLKSystemComponent; ASection, AnItem:
        string): Boolean;
    procedure SaveFormat(const AComponent: TLKSystemComponent; ASection,
        AnItem: string);
    procedure UpdateFields;
    property Count: Integer read GetCount;
    property DisplayHeight: Integer read FDisplayHeight;
    property Format: string read GetFormat write SetFormat;
    property Items[Index: Integer]: TLKDcDBEditControl read GetItem; default;
  published
    property ColWidth0: Integer read FColWidth0 write FColWidth0;
    property ColWidth1: Integer read FColWidth1 write FColWidth1;
    property ColWidth2: Integer read FColWidth2 write FColWidth2;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property EditControls: TLKDBEditCollection read FEditControls;
    property Layout: TLKEditsLayout read FLayout write SetLayout;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
  end;

  TLKDcDBGridController = class(TComponent)
  private
    FGrid: TJvDBUltimGrid;
    FGridColumnItems: TLKDBGridCollection;
    function GetCount: Integer;
    function GetReadOnly: Boolean;
    function GetSelectedField: TField;
    procedure SetGrid(Value: TJvDBUltimGrid);
    procedure SetReadOnly(const Value: Boolean);
    procedure SetSelectedField(Value: TField);
  protected
    function GetItem(Index: Integer): TLKDcDBGridControl;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ClearFormat(const AComponent: TLKSystemComponent; ASection,
        AnItem: string);
    procedure ConstructFromDatasource(const ExcludeFieldNames: string);
    function DlgEditListPropsExecute(AOwner: TComponent): TModalResult; virtual;
    procedure DoOnFormClose;
    function LoadFormat(const AComponent: TLKSystemComponent; ASection, AnItem:
        string): Boolean;
    procedure SaveFormat(const AComponent: TLKSystemComponent; ASection,
        AnItem: string);
    property Count: Integer read GetCount;
    property Items[Index: Integer]: TLKDcDBGridControl read GetItem; default;
    property SelectedField: TField read GetSelectedField write SetSelectedField;
  published
    property Grid: TJvDBUltimGrid read FGrid write SetGrid;
    property GridColumnItems: TLKDBGridCollection read FGridColumnItems;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
  end;


const
  FontReadOnlyColor = clNavy;
  EditReadOnlyBackgroundColor = clInfoBk;
  EditBackgroundColor = clWindow;
  FontEditColor = clBlack;

resourcestring
  //Email
  EmailInitError = 'Error initialising email';
  EmailSendError = 'Error sending mail';

procedure EditControlResize(Sender: TWinControl; ALayout: TLKEditsLayout);

procedure Register;

implementation

uses LKUtilsEditListPropsDlg;

procedure Register;
begin
  RegisterComponents('LKUtils', [TLKDBButtonEdit]);
  RegisterComponents('LKUtils', [TLKDBImageEdit]);
  RegisterComponents('LKUtils', [TLKDBEMailEdit]);
  RegisterComponents('LKUtils', [TLKDBFilePathEdit]);
  RegisterComponents('LKUtils', [TLKDBUrlEdit]);
  RegisterComponents('LKUtils', [TLKDcDBGridController]);
  RegisterComponents('LKUtils', [TLKDcDBEditsPanel]);
end;


procedure EditControlResize(Sender: TWinControl; ALayout: TLKEditsLayout);
var
  i, clipWidth, edits0Width, labelWidth, edits1Width, ColDown,
    downOffset, rightOffset, edits0Left, label1Left, edits1Left, label2left,
      edits2Left, nextTop, nextLeft, noOfLines: integer;
begin
  if (sender.width < 60) then
    exit;
  noOfLines := 0;
  colDown := 16;
  case ALayout of
    aceHorizontal:
      begin
        clipWidth := Sender.Width - 16;
        nextLeft := 8;
        NextTop := 4;
        colDown := 44;
        downOffset := 16; // offset of edit from label
        i := 0;
        noOfLines := 1;
        with Sender do
          while (i < ControlCount - 1) do
          begin
            rightOffset := 0;
            if (Controls[i + 1] is TDBEdit) then
            begin
              with TDBEdit(Controls[i + 1]) do
              begin
                width := field.DisplayWidth + 2;
                if (width > clipWidth) then
                  (width := clipWidth);
                rightOffset := width; //for following controls
              end;
            end;
            if (Controls[i + 1] is TJvDBComboBox) then
            begin
              with TJvDBComboBox(Controls[i + 1]) do
              begin
                width := field.DisplayWidth + 2;
                if (width > clipWidth) then
                  (width := clipWidth);
                rightOffset := width; //for following controls
              end;
            end;
            if (Controls[i + 1] is TDBMemo) then
            begin
              with TDBMemo(Controls[i + 1]) do
              begin
                width := field.DisplayWidth + 2;
                if (width > clipWidth) then
                  (width := clipWidth);
                rightOffset := width; //for following controls
              end;
            end;
            if (nextLeft + rightOffset > clipWidth) then
            begin
              nextLeft := 8;
              nextTop := nextTop + colDown;
              inc(noOfLines);
            end;
            with Controls[i] do //label
            begin
              Top := nextTop;
              Left := nextLeft;
            end;
            inc(i);
            with Controls[i] do //edit
            begin
              Top := nextTop + downOffset;
              Left := nextLeft;
            end;
            inc(i);
            nextLeft := nextLeft + rightOffset;
          end;
      end;
    aceOneCol:
      begin
        clipWidth := Sender.Width - 24;
        labelWidth := trunc(clipWidth * 0.35);
        edits0Width := trunc(clipWidth * 0.65);
        edits0Left := 16 + labelWidth;
        with Sender do
          for i := 0 to ControlCount - 1 do
          begin
            if Controls[i] is TLabel then
              with Controls[i] do
              begin
                left := 8;
                width := labelWidth;
                inc(noOfLines);
              end;
            if (Controls[i] is TDBCheckBox) then
              with Controls[i] do
              begin
                left := 8;
                width := labelWidth + 40;
                inc(noOfLines);
              end;
            if (Controls[i] is TDBEdit) then
              with Controls[i] do
              begin
                left := edits0Left;
                width := edits0Width;
              end;
            if (Controls[i] is TJvDBComboBox) then
              with Controls[i] do
              begin
                left := edits0Left;
                width := edits0Width;
              end;
            if (Controls[i] is TDBMemo) then
              with Controls[i] do
              begin
                left := edits0Left;
                width := edits0Width;
                inc(noOfLines);
              end;
          end;
      end;
    aceTwoCol:
      begin
        clipWidth := Sender.Width - 40;
        labelWidth := trunc(clipWidth * 0.15);
        edits0Width := trunc(clipWidth / 2);
        edits1Width := clipWidth - (2 * labelWidth + edits0Width);
        edits0Left := 16 + labelWidth;
        label1Left := edits0Left + edits0Width + 8;
        edits1Left := label1Left + labelWidth + 8;
        with Sender do
          for i := 0 to ControlCount - 1 do
          begin
            if Controls[i] is TLabel then
              with Controls[i] do
              begin
                case tag of
                  0:
                    begin
                      left := 8;
                      inc(noOfLines);
                    end
                else
                  left := label1Left;
                end; // Tag 1 = second column
                width := labelWidth;
              end;
            if (Controls[i] is TDBEdit) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      left := edits0Left;
                      width := edits0Width;
                    end;
                else
                  begin
                    left := edits1Left;
                    width := edits1Width;
                  end;
                end;
            if (Controls[i] is TJvDBComboBox) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      left := edits0Left;
                      width := edits0Width;
                    end;
                else
                  begin
                    left := edits1Left;
                    width := edits1Width;
                  end;
                end;
            if (Controls[i] is TDBMemo) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      inc(noOfLines);
                      left := edits0Left;
                      width := edits0Width;
                    end;
                else
                  begin
                    left := edits1Left;
                    width := edits1Width;
                  end;
                end;
            if (Controls[i] is TDBCheckBox) then
              with Controls[i] do
              begin
                case tag of
                  0:
                    begin
                      left := 8;
                      inc(noOfLines);
                    end;
                else
                  left := label1Left;
                end;
                width := labelWidth + 40;
              end;
          end;
      end;
    aceThreeCol:
      begin
        clipWidth := Sender.Width - 48;
        labelWidth := trunc(clipWidth * 0.10);
        edits0Width := trunc(clipWidth * 0.30);
        edits1Width := trunc(clipWidth * 0.20); // also edits2Width
        edits0Left := 16 + labelWidth;
        label1Left := edits0Left + edits0Width + 8;
        edits1Left := label1Left + labelWidth + 8;
        label2Left := edits1Left + edits1Width + 8;
        edits2Left := label2Left + labelWidth + 8;
        with Sender do
          for i := 0 to ControlCount - 1 do
          begin
            if Controls[i] is TLabel then
              with Controls[i] do
              begin
                case tag of
                  0:
                    begin
                      left := 8;
                      inc(noOfLines);
                    end;
                  1: left := label1Left;
                else
                  left := Label2Left; // Tag 2 = second column
                end;
                width := labelWidth;
              end;
            if (Controls[i] is TDBEdit) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      left := edits0Left;
                      width := edits0Width;
                    end;
                  1:
                    begin
                      left := edits1Left;
                      width := edits1Width;
                    end;
                else
                  begin
                    left := edits2Left;
                    width := edits1Width;
                  end;
                end;
            if (Controls[i] is TJvDBComboBox) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      left := edits0Left;
                      width := edits0Width;
                    end;
                  1:
                    begin
                      left := edits1Left;
                      width := edits1Width;
                    end;
                else
                  begin
                    left := edits2Left;
                    width := edits1Width;
                  end;
                end;
            if (Controls[i] is TDBMemo) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      left := edits0Left;
                      width := edits0Width;
                      inc(noOfLines);
                    end;
                  1:
                    begin
                      left := edits1Left;
                      width := edits1Width;
                    end;
                else
                  begin
                    left := edits2Left;
                    width := edits1Width;
                  end;
                end;
            if (Controls[i] is TDBCheckBox) then
              with Controls[i] do
                case tag of
                  0:
                    begin
                      left := 8;
                      width := labelWidth + 40;
                      ;
                    end;
                  1:
                    begin
                      left := label1Left;
                      width := labelWidth + 40;
                    end;
                else
                  begin
                    left := label2Left;
                    width := labelWidth + 40;
                  end;
                end;
          end;
      end;
  end;
  Sender.height := 8 + colDown * noOfLines;
end;


{
******************************* TLKDBButtonEdit ********************************
}
constructor TLKDBButtonEdit.Create(AOwner: TComponent);
var
  I: Integer;
begin
  inherited;
  ControlStyle := ControlStyle + [csCaptureMouse];
  Borderstyle := bsNone;
  Height := 15;
  i := Width - 15;
  Button := TButton.Create(AOwner);
  with Button do
  begin
    //ControlStyle := ControlStyle + [csReplicatable];
    Visible := True;
    Height := 15;
    Width := 15;
    Left := i;
    Caption := '...'; // flat := true;
    Align := alCustom;
    //Font.Style := [fsBold];
    Anchors := [akRight] + [akTop];
    Parent := Self;
    OnClick := doButtonClick;
    // ParentColor := false;
    // Color := clBtnFace;
    TabStop := False;
  end;
end;

procedure TLKDBButtonEdit.DefaultButtonClick(Sender: TObject);
begin
  Self.SetFocus;
end;

procedure TLKDBButtonEdit.DoButtonClick(Sender: TObject);
begin
  if Assigned(FOnButtonClick) then
    FOnButtonClick(Sender)
  else
    DefaultButtonClick(Sender);
end;

{
******************************** TLKDBDateEdit *********************************
}
constructor TLKDBDateEdit.Create(AOwner: TComponent);
begin
  inherited;
  Button.Caption := 'v';
  Calendar := TMonthCalendar.Create(self);
  with Calendar do
  begin
    Visible := false;
    Parent := Self.Parent;
    OnDblClick := CalendarDblClick;
  end;
end;

destructor TLKDBDateEdit.Destroy;
begin
  Calendar.Free;
  inherited Destroy;
end;

procedure TLKDBDateEdit.CalendarDblClick(Sender: TObject);
begin
  if (Calendar.Date <> Field.AsDateTime) then
  begin
    Field.Dataset.Edit;
    Field.AsDateTime := Calendar.Date;
  end;
  Calendar.Visible := False;
  Button.Caption := 'v';
end;

procedure TLKDBDateEdit.DefaultButtonClick(Sender: TObject);
var
  t, l: Integer;
begin
  t := Top + 16;
  if (Parent.Height < t + 154) then
    t := Top - 155;
  if (t < 0) then
    t := 0;
  l := Left - 32;
  if (l < 0) then
    l := 0;
  Calendar.Parent := Parent;
  with Calendar do
  begin
    Top := t;
    Left := l;
    Visible := (Button.Caption = 'v');
  end;
  with Button do
    if (Caption = 'v') then
      Caption := '^'
    else
      Caption := 'v';
  inherited DefaultButtonClick(Sender);
end;

{
******************************** TLKDBImageEdit ********************************
}
procedure TLKDBImageEdit.DefaultButtonClick(Sender: TObject);
begin
  if not Assigned(FImgDlg) then
  begin
    Application.CreateForm(TDlgLKDBImageEdit, FImgDlg);
    if Owner is TForm then
      FImgDlg.PopupParent := TForm(Owner);
    FImgDlg.Construct(FImageFilePath, Field);
  end;
  with FImgDlg do
  begin
    LoadFieldImage;
    Show;
  end;
  inherited DefaultButtonClick(Sender);
end;

{
******************************** TLKDBEMailEdit ********************************
}
procedure TLKDBEMailEdit.DefaultButtonClick(Sender: TObject);
var
  Recipients: TStringList;
  S: string;

  procedure EMail(const Originator, Subject, Content: string;
    Recipients, AttachmentFiles: TStrings);
  var
    MapiMessage: TMapiMessage;
    dwRet: Cardinal;
    MAPI_Session: Cardinal;
    MError: Cardinal;
    WndList: Pointer;
    A: AnsiString;
    T: string;
    I, FilesCount, RecipsCount: integer;
    Orig: TMapiRecipDesc;
    RecipNames: array[0..1000] of AnsiString;
    RecipAddress: array[0..1000] of AnsiString;
    Recips: array[0..1000] of TMapiRecipDesc;
    FilePaths: array[0..1000] of AnsiString;
    FileNames: array[0..1000] of AnsiString;
    Files: array[0..1000] of TMapiFileDesc;

  begin
    dwRet := MapiLogon(Application.Handle, '', '', MAPI_LOGON_UI or
      MAPI_NEW_SESSION, 0, @MAPI_Session);
    if (dwRet <> SUCCESS_SUCCESS) then
    begin
      MessageDlg(EmailInitError + ' ' + intToStr(dwRet), mtError, [mbOK], 0);
      Sysutils.abort;
    end;
    try
      RecipsCount := 0;
      if Assigned(Recipients) then
        for I := 0 to Recipients.Count - 1 do
        begin
          T := Trim(Recipients.Names[I]);
          A := AnsiString(T);
          if (T > '') then
          try
            RecipNames[RecipsCount] := A;
            RecipAddress[RecipsCount] := 'SMTP: ' + A;
            with Recips[RecipsCount] do
            begin
              ulReserved := 0;
              T := Uppercase(Recipients.Values[T]);
              if T = 'TO' then
                ulRecipClass := MAPI_TO;
              if T = 'CC' then
                ulRecipClass := MAPI_CC;
              if T = 'BCC' then
                ulRecipClass := MAPI_BCC;
              lpszName := PAnsiChar(RecipNames[RecipsCount]);
              lpszAddress := PAnsiChar(RecipAddress[RecipsCount]);
              ulEIDSize := 0;
            end;
            Inc(RecipsCount);
          except
          end;
        end;

      // attachments
      FilesCount := 0;
      if Assigned(AttachmentFiles) then
        for I := 0 to AttachmentFiles.Count - 1 do
        begin
          T := Trim(AttachmentFiles.Names[I]);
          if (T > '') then
          try
            FilePaths[FilesCount] := AnsiString(ExtractFilePath(T));
            FileNames[FilesCount] := AnsiString(ExtractFileName(T));
            with Files[RecipsCount] do
            begin
              ulReserved := 0;
              flFlags := 0;
              nPosition := ULONG($FFFFFFFF);
              lpszPathName := PAnsiChar(FilePaths[FilesCount]);
              lpszFileName := PAnsiChar(FileNames[FilesCount]);
              lpFileType := nil;
            end;
            inc(FilesCount);
          except
          end;
        end;

      with MapiMessage do
      begin
        ulReserved := 0;
        lpszSubject := PAnsiChar(AnsiString(Subject));
        lpszNoteText := PAnsiChar(AnsiString(Content));
        lpszMessageType := nil;
        lpszDateReceived := nil;
        lpszConversationID := nil;
        flFlags := 0;
        A := AnsiString(Trim(Originator));
        Orig.lpszName := PAnsiChar(A);
        Orig.ulRecipClass := MAPI_ORIG;
        Orig.lpszAddress := '';
        lpOriginator := @Orig;
        nRecipCount := RecipsCount;
        lpRecips := @Recips;
        nFileCount := FilesCount;
        lpFiles := @Files;
      end;

      WndList := DisableTaskWindows(0);
      try
        MError := MapiSendMail(MAPI_Session, Application.Handle, MapiMessage,
          MAPI_DIALOG or MAPI_LOGON_UI or MAPI_NEW_SESSION, 0);
        if MError <> 0 then
          MessageDlg(EmailSendError, mtError, [mbOK], 0);
      finally EnableTaskWindows(WndList);
      end;
    finally MapiLogOff(MAPI_Session, Application.Handle, 0, 0);
    end;
  end;

begin
  S := Field.AsString;
  if (Pos('@', S) > 0) then
  begin
    Recipients := TStringList.Create;
    try
      Recipients.Add(S);
      EMail('', '', '', Recipients, nil);
    finally Recipients.Free;
    end;
  end;
  inherited DefaultButtonClick(Sender);
end;

{
********************************* TLKDBUrlEdit *********************************
}
procedure TLKDBUrlEdit.DefaultButtonClick(Sender: TObject);
var
  URL: string;
begin
  URL := Field.AsString;
  if (Pos('//', URL) > 0) then
    ShellExecute(Self.Handle, 'open', PChar(URL), nil, nil, 1);
  inherited DefaultButtonClick(Sender);
end;

{
****************************** TLKDBFilePathEdit *******************************
}
procedure TLKDBFilePathEdit.DefaultButtonClick(Sender: TObject);
var
  S: string;
begin
  S := Field.AsString;
  if (S < '  ') then
  begin
    with TOpenDialog.Create(Owner) do
      if Execute then
      begin
        Field.Dataset.Edit;
        Field.Value := FileName;
      end;
  end
  else if FileExists(S) then
    ShellExecute(Self.Handle, 'open', PChar(S), nil, nil, 1)
  else
    Showmessage('File not found: ' + S);
  inherited DefaultButtonClick(Sender);
end;

{
****************************** TLKDBDcLookupEdit *******************************
}
constructor TLKDBDcLookupEdit.Create(AOwner: TComponent);
begin
  inherited;
  FLookupOnExit := True;
end;

procedure TLKDBDcLookupEdit.DefaultButtonClick(Sender: TObject);
begin
  Button.Enabled := False;
  try
    with Field.Dataset do
      if (State = dsBrowse) then
        Edit;
    case FLookupDataset of
      tbMaster:
        begin
          FLookupDataComponent.View(Field.Value, True, [mbOK]);
          Field.Dataset.Edit;
          Field.Value :=
            FLookupDataComponent.TblMaster.FieldByName(LookupFieldName).Value;
        end;
      tbDetail:
        begin
          FLookupDataComponent.ViewRecord(null, True, [mbOK]);
          Field.Dataset.Edit;
          Field.Value :=
            FLookupDataComponent.TblDetail.FieldByName(LookupFieldName).Value;
        end;
    end;
    DoLookupFind;
  finally Button.Enabled := True;
  end;
  inherited DefaultButtonClick(Sender);
end;

procedure TLKDBDcLookupEdit.DoExit;
var
  S: Variant;
begin
  if FLookupOnExit then
  begin
    S := Field.Value;
    if (Field.Dataset.State <> dsBrowse) and (S > ' ') then
      case FLookupDataset of
        tbMaster:
          if FLookupDataComponent.TblMaster.Locate(LookupFieldName, S, []) then
            DoLookupFind
          else DefaultButtonClick(Button);
        tbDetail:
          if FLookupDataComponent.TblDetail.Locate(LookupFieldName, S, []) then
            DoLookupFind
          else DefaultButtonClick(Button);
      end;
  end;
  inherited DoExit;
end;

procedure TLKDBDcLookupEdit.DoLookupFind;
begin
  if Assigned(FOnLookupFind) then
    FOnLookupFind(Self);
end;

function TLKDBDcLookupEdit.GetLookupFieldName: string;
begin
  if (FLookupFieldName = '') and Assigned(Field) then
    FLookupFieldName := Field.FieldName;
  Result := FLookupFieldName;
end;

{
********************************* TLKDBControl *********************************
}
constructor TLKDBControl.Create(Collection: TCollection);
begin
  inherited;
  FField := nil;
  FAlignment := taLeftJustify;
  FControlType := dbcUnassigned;
  //FVisible := true;
end;

function TLKDBControl.CreateControlsFromControlType: Boolean;
begin
  Result := False;
end;

procedure TLKDBControl.DoEditOnExit(Sender: TObject);
begin
  if Assigned(FOnExit) then
    FOnExit(Sender);
end;

function TLKDBControl.GetFieldName: string;
begin
  if Assigned(FField) then
    Result := FField.FieldName
  else
    Result := '';
end;

function TLKDBControl.GetWidth: Integer;
begin
  Result := 0;
end;

procedure TLKDBControl.SetAlignment(Value: TAlignment);
begin
  if Assigned(FEditControl) then
  begin
    if (FEditControl is TCustomEdit) then
      TCustomEdit(FEditControl).Alignment := Value
    else if (FEditControl is TJvCustomEdit) then
      TJvCustomEdit(FEditControl).Alignment := Value;
  end;
  FAlignment := Value;
end;

procedure TLKDBControl.SetCaption(const Value: string);
begin
  FCaption := Value;
end;

procedure TLKDBControl.SetControlParent(AParent: TWinControl);
begin
  FControlParent := AParent;
  if Assigned(FEditControl) then
    FEditControl.Parent := AParent;
end;

procedure TLKDBControl.SetControlType(const Value: TLKDBControlType);
begin
  if FControlType <> Value then
  begin
    FControlType := Value;
    FreeAndNil(FEditControl);
    if not Assigned(FField) then
      Exit;
    if not CreateControlsFromControlType then
      Exit;
    SetControlParent(FControlParent);
    SetDataSource(FDataSource);
    SetEditControlDataField;
    SetCaption(FCaption);
    SetReadOnly(FReadOnly);
    SetOnEnter(FOnEnter);
    SetOnExit(FOnExit);
  end;
end;

procedure TLKDBControl.SetControlTypeFromField;
begin
  if (Assigned(FField) and (ControlType = dbcUnassigned)) then
    case FField.DataType of
      ftUnknown:   ControlType := dbcUnassigned;
      ftDate,
      ftTimeStamp: ControlType := dbcDateEdit;
      ftTime:      ControlType := dbcTimeEdit;
      ftWideMemo,
      ftFmtMemo,
      ftMemo:      ControlType := dbcMemo;
      ftGraphic:   ControlType := dbcImage;
      ftSmallInt, ftBoolean: ControlType := dbcCheckbox;
    else
      ControlType := dbcEdit;
    end;
end;

procedure TLKDBControl.SetDataSource(ADatasource: TDataSource);
begin
  FDataSource := ADatasource;
  if Assigned(FEditControl) then
    case ControlType of
      dbcDatePicker: TJvDBDatePickerEdit(FEditControl).DataSource := ADatasource;
      dbcDateTimePicker: TJvDBDateTimePicker(FEditControl).DataSource := ADatasource;
      dbcRichEdit: TJvDBRichEdit(FEditControl).DataSource := ADatasource;
      dbcSpinEdit: TJvDBSpinEdit(FEditControl).DataSource := ADatasource;
      dbcLookUpListBox: TJvDBLookUpList(FEditControl).DataSource := ADatasource;
      dbcLookUpComboBox: TJvDBLookUpCombo(FEditControl).DataSource := ADatasource;
      dbcRadioPanel: TJvDBRadioPanel(FEditControl).DataSource := ADatasource;
      dbcComboBox: TDBComboBox(FEditControl).DataSource := ADatasource;
      dbcTreeView: TJvDBTreeView(FEditControl).DataSource := ADatasource;
      dbcLookupTreeViewCombo: TJvDBLookupTreeViewCombo(FEditControl).DataSource := ADatasource;
      dbcLookupTreeView: TJvDBLookupTreeView(FEditControl).DataSource := ADatasource;
      dbcComboEdit: TJvDBComboEdit(FEditControl).DataSource := ADatasource;
      dbcDateEdit: TJvDBDateEdit(FEditControl).DataSource := ADatasource;
      dbcTimeEdit: TJvDBMaskEdit(FEditControl).DataSource := ADatasource;
      dbcCalcEdit: TJvDBCalcEdit(FEditControl).DataSource := ADatasource;
      dbcMaskEdit: TJvDBMaskEdit(FEditControl).DataSource := ADatasource;
      dbcLookupComboEdit: TJvDBLookUpComboEdit(FEditControl).DataSource := ADatasource;
      dbcSearchEdit: TJvDBSearchEdit(FEditControl).DataSource := ADatasource;
      dbcSearchComboBox: TJvDBSearchComboBox(FEditControl).DataSource := ADatasource;
      dbcFindEdit: TJvDBFindEdit(FEditControl).DataSource := ADatasource;
      dbcImage: TJvDBImage(FEditControl).DataSource := ADatasource;
      dbcCheckbox: TDBCheckBox(FEditControl).DataSource := ADatasource;
      dbcText: TDBText(FEditControl).DataSource := ADatasource;
      dbcMemo: TDBMemo(FEditControl).DataSource := ADatasource;
      dbcListBox: TDBListBox(FEditControl).DataSource := ADatasource;
      dbcRadioGroup: TDBRadioGroup(FEditControl).DataSource := ADatasource;
    else if FEditControl is TDBEdit then
      TDBEdit(FEditControl).DataSource := ADatasource;
    end;
end;

procedure TLKDBControl.SetEditControlDataField;
begin
  if (Assigned(FEditControl) and Assigned(FField)) then
    case ControlType of
      dbcDCLookup: TLKDBDcLookupEdit(FEditControl).Datafield := FField.FieldName;
      dbcText: TDBText(FEditControl).Datafield := FField.FieldName;
      dbcMemo: TDBMemo(FEditControl).Datafield := FField.FieldName;
      dbcListBox: TDBListBox(FEditControl).Datafield := FField.FieldName;
      dbcRadioGroup: TDBRadioGroup(FEditControl).Datafield := FField.FieldName;
      dbcDatePicker: TJvDBDatePickerEdit(FEditControl).Datafield := FField.FieldName;
      dbcDateTimePicker: TJvDBDateTimePicker(FEditControl).Datafield := FField.FieldName;
      dbcRichEdit: TJvDBRichEdit(FEditControl).Datafield := FField.FieldName;
      dbcSpinEdit: TJvDBSpinEdit(FEditControl).Datafield := FField.FieldName;
      dbcLookUpListBox: TJvDBLookUpList(FEditControl).Datafield := FField.FieldName;
      dbcLookUpComboBox: TJvDBLookUpCombo(FEditControl).Datafield := FField.FieldName;
      dbcRadioPanel: TJvDBRadioPanel(FEditControl).Datafield := FField.FieldName;
      dbcComboBox: TDBComboBox(FEditControl).Datafield := FField.FieldName;
      dbcTreeView: TJvDBTreeView(FEditControl).Itemfield := FField.FieldName;
      dbcLookupTreeViewCombo: TJvDBLookupTreeViewCombo(FEditControl).Datafield := FField.FieldName;
      dbcLookupTreeView: TJvDBLookupTreeView(FEditControl).Datafield := FField.FieldName;
      dbcComboEdit: TJvDBComboEdit(FEditControl).Datafield := FField.FieldName;
      dbcDateEdit: TJvDBDateEdit(FEditControl).Datafield := FField.FieldName;
      dbcTimeEdit: TJvDBMaskEdit(FEditControl).Datafield := FField.FieldName;
      dbcCalcEdit: TJvDBCalcEdit(FEditControl).Datafield := FField.FieldName;
      dbcMaskEdit: TJvDBMaskEdit(FEditControl).Datafield := FField.FieldName;
      dbcLookupComboEdit: TJvDBLookUpComboEdit(FEditControl).Datafield := FField.FieldName;
      dbcSearchEdit: TJvDBSearchEdit(FEditControl).Datafield := FField.FieldName;
      dbcSearchComboBox: TJvDBSearchComboBox(FEditControl).Datafield := FField.FieldName;
      dbcFindEdit: TJvDBFindEdit(FEditControl).Datafield := FField.FieldName;
      dbcImage: TJvDBImage(FEditControl).Datafield := FField.FieldName;
      dbcCheckbox: TDBCheckBox(FEditControl).Datafield := FField.FieldName;
    else if EditControl is TDBEdit then
      TDBEdit(EditControl).Datafield := FField.FieldName;
  end;
end;

procedure TLKDBControl.SetField(AField: TField);
begin
  if (FField = AField) then
    Exit;
  FField := AField;
  if Assigned(FField) then
  begin
    SetControlTypeFromField;
    Caption := FField.DisplayName;
    Visible := FField.Visible;
    ReadOnly := FField.ReadOnly;
    SetEditControlDataField;
  end
  else
  begin
    ControlType := dbcUnassigned;
  end;
end;

procedure TLKDBControl.SetOnEnter(const Value: TNotifyEvent);
begin
  FOnEnter := Value;
  if not Assigned(FEditControl) then
    Exit;
  case ControlType of
    dbcMemo: TDBMemo(FEditControl).OnEnter := Value;
    dbcListBox: TDBListBox(FEditControl).OnEnter := Value;
    dbcRadioGroup: TDBRadioGroup(FEditControl).OnEnter := Value;
    dbcDatePicker: TJvDBDatePickeredit(FEditControl).OnEnter := Value;
    dbcDateTimePicker: TJvDBDateTimePicker(FEditControl).OnEnter := Value;
    dbcRichEdit: TJvDBRichEdit(FEditControl).OnEnter := Value;
    dbcSpinEdit: TJvDBSpinEdit(FEditControl).OnEnter := Value;
    dbcLookUpListBox: TJvDBLookUpList(FEditControl).OnEnter := Value;
    dbcLookUpComboBox: TJvDBLookUpCombo(FEditControl).OnEnter := Value;
    dbcRadioPanel: TJvDBRadioPanel(FEditControl).OnEnter := Value;
    dbcComboBox: TDBComboBox(FEditControl).OnEnter := Value;
    dbcTreeView: TJvDBTreeView(FEditControl).OnEnter := Value;
    dbcLookupTreeViewCombo: TJvDBLookupTreeViewCombo(FEditControl).OnEnter := Value;
    dbcLookupTreeView: TJvDBLookupTreeView(FEditControl).OnEnter := Value;
    dbcComboEdit: TJvDBComboEdit(FEditControl).OnEnter := Value;
    dbcDateEdit: TJvDBDateEdit(FEditControl).OnEnter := Value;
    dbcTimeEdit: TJvDBMaskEdit(FEditControl).OnEnter := Value;
    dbcCalcEdit: TJvDBCalcEdit(FEditControl).OnEnter := Value;
    dbcMaskEdit: TJvDBMaskEdit(FEditControl).OnEnter := Value;
    dbcLookupComboEdit: TJvDBLookUpComboEdit(FEditControl).OnEnter := Value;
    dbcSearchEdit: TJvDBSearchEdit(FEditControl).OnEnter := Value;
    dbcSearchComboBox: TJvDBSearchComboBox(FEditControl).OnEnter := Value;
    dbcFindEdit: TJvDBFindEdit(FEditControl).OnEnter := Value;
    dbcImage: TJvDBImage(FEditControl).OnEnter := Value;
    dbcCheckbox: TDBCheckBox(FEditControl).OnEnter := Value;
  else if (FEditControl is TDBEdit) then
    TDBEdit(FEditControl).OnEnter := Value;
  end;
end;

procedure TLKDBControl.SetOnExit(const Value: TNotifyEvent);
begin
  FOnExit := Value;
  if not Assigned(FEditControl) then
    Exit;
  case ControlType of
    dbcMemo: TDBMemo(FEditControl).OnExit := DoEditOnExit;
    dbcListBox: TDBListBox(FEditControl).OnExit := DoEditOnExit;
    dbcRadioGroup: TDBRadioGroup(FEditControl).OnExit := DoEditOnExit;
    dbcDatePicker: TJvDBDatePickerEdit(FEditControl).OnExit := DoEditOnExit;
    dbcDateTimePicker: TJvDBDateTimePicker(FEditControl).OnExit := DoEditOnExit;
    dbcRichEdit: TJvDBRichEdit(FEditControl).OnExit := DoEditOnExit;
    dbcSpinEdit: TJvDBSpinEdit(FEditControl).OnExit := DoEditOnExit;
    dbcLookUpListBox: TJvDBLookUpList(FEditControl).OnExit := DoEditOnExit;
    dbcLookUpComboBox: TJvDBLookUpCombo(FEditControl).OnExit := DoEditOnExit;
    dbcRadioPanel: TJvDBRadioPanel(FEditControl).OnExit := DoEditOnExit;
    dbcComboBox: TDBComboBox(FEditControl).OnExit := DoEditOnExit;
    dbcTreeView: TJvDBTreeView(FEditControl).OnExit := DoEditOnExit;
    dbcLookupTreeViewCombo: TJvDBLookupTreeViewCombo(FEditControl).OnExit := DoEditOnExit;
    dbcLookupTreeView: TJvDBLookupTreeView(FEditControl).OnExit := DoEditOnExit;
    dbcComboEdit: TJvDBComboEdit(FEditControl).OnExit := DoEditOnExit;
    dbcDateEdit: TJvDBDateEdit(FEditControl).OnExit := DoEditOnExit;
    dbcTimeEdit: TJvDBMaskEdit(FEditControl).OnExit := DoEditOnExit;
    dbcCalcEdit: TJvDBCalcEdit(FEditControl).OnExit := DoEditOnExit;
    dbcMaskEdit: TJvDBMaskEdit(FEditControl).OnExit := DoEditOnExit;
    dbcLookupComboEdit: TJvDBLookUpComboEdit(FEditControl).OnExit := DoEditOnExit;
    dbcSearchEdit: TJvDBSearchEdit(FEditControl).OnExit := DoEditOnExit;
    dbcSearchComboBox: TJvDBSearchComboBox(FEditControl).OnExit := DoEditOnExit;
    dbcFindEdit: TJvDBFindEdit(FEditControl).OnExit := DoEditOnExit;
    dbcImage: TJvDBImage(FEditControl).OnExit := DoEditOnExit;
    dbcCheckbox: TDBCheckBox(FEditControl).OnExit := DoEditOnExit;
  else if (FEditControl is TDBEdit) then
    TDBEdit(FEditControl).OnExit := DoEditOnExit;
  end; // case
end;

procedure TLKDBControl.SetOnKeyPress(const Value: TKeyPressEvent);
begin
  FOnKeyPress := Value;
  if not Assigned(FEditControl) then
    Exit;
  case ControlType of
    dbcMemo: TDBMemo(FEditControl).OnKeyPress := Value;
    dbcListBox: TDBListBox(FEditControl).OnKeyPress := Value;
    dbcRadioGroup: ; //TDBRadioGroup(FEditControl).OnKeyPress := Value;;
    dbcDatePicker: TJvDBDatePickerEdit(FEditControl).OnKeyPress := Value;
    dbcDateTimePicker: TJvDBDateTimePicker(FEditControl).OnKeyPress := Value;
    dbcRichEdit: TJvDBRichEdit(FEditControl).OnKeyPress := Value;
    dbcSpinEdit: TJvDBSpinEdit(FEditControl).OnKeyPress := Value;
    dbcLookUpListBox: TJvDBLookUpList(FEditControl).OnKeyPress := Value;
    dbcLookUpComboBox: TJvDBLookUpCombo(FEditControl).OnKeyPress := Value;
    dbcRadioPanel: ; //TJvDBRadioPanel(FEditControl).OnKeyPress := Value;
    dbcComboBox: TDBComboBox(FEditControl).OnKeyPress := Value;
    dbcTreeView: TJvDBTreeView(FEditControl).OnKeyPress := Value;
    dbcLookupTreeViewCombo: TJvDBLookupTreeViewCombo(FEditControl).OnKeyPress := Value;
    dbcLookupTreeView: TJvDBLookupTreeView(FEditControl).OnKeyPress := Value;
    dbcComboEdit: TJvDBComboEdit(FEditControl).OnKeyPress := Value;
    dbcDateEdit: TJvDBDateEdit(FEditControl).OnKeyPress := Value;
    dbcTimeEdit: TJvDBMaskEdit(FEditControl).OnKeyPress := Value;
    dbcCalcEdit: TJvDBCalcEdit(FEditControl).OnKeyPress := Value;
    dbcMaskEdit: TJvDBMaskEdit(FEditControl).OnKeyPress := Value;
    dbcLookupComboEdit: TJvDBLookUpComboEdit(FEditControl).OnKeyPress := Value;
    dbcSearchEdit: TJvDBSearchEdit(FEditControl).OnKeyPress := Value;
    dbcSearchComboBox: TJvDBSearchComboBox(FEditControl).OnKeyPress := Value;
    dbcFindEdit: TJvDBFindEdit(FEditControl).OnKeyPress := Value;
    dbcImage: TJvDBImage(FEditControl).OnKeyPress := Value;
    dbcCheckbox: TDBCheckBox(FEditControl).OnKeyPress := Value;
  else if (FEditControl is TDBEdit) then
    TDBEdit(FEditControl).OnKeyPress := Value;
  end;
end;

procedure TLKDBControl.SetReadOnly(const Value: Boolean);
begin
  FReadOnly := Value;
end;

procedure TLKDBControl.SetVisible(const Value: Boolean);
begin
  FVisible := Value;
end;

procedure TLKDBControl.SetWidth(const Value: Integer);
begin
  //
end;

{
**************************** TLKDBControlCollection ****************************
}
destructor TLKDBControlCollection.Destroy;
begin
  FExcludeFieldList.Free;
  inherited Destroy;
end;

function TLKDBControlCollection.Add: TLKDBControl;
begin
  Result := TLKDBControl(inherited Add);
end;

procedure TLKDBControlCollection.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    FreeAndNil(Items[I].FEditControl);
  inherited Clear;
end;

procedure TLKDBControlCollection.ConstructFromDatasource(const
    ExcludeFieldNames: string);

  procedure DelimitedStringToList(const AString, Delimiter: string; AList:
    TStrings);
  var
    start, finish: integer;
    s, t: string;
  begin
    AList.Clear;
    s := AString;
    start := 1;
    while (start <= Length(s)) do
    begin
      ;
      finish := PosEx(Delimiter, s, start);
      if (finish = 0) then
        finish := Length(s) + 1;
      t := System.Copy(s, start, finish - start);
      Trim(t);
      if (t > '') then
        AList.Add(t);
      start := finish + Length(Delimiter)
    end;
  end;

begin
  FExcludeFieldList := TStringList.Create;
  DelimitedStringToList(ExcludeFieldNames, ';', FExcludeFieldList);
  Clear;
end;

procedure TLKDBControlCollection.DoOnFormClose;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    with Items[I] do
      if EditControl is TLKDBImageEdit then
        with TLKDBImageEdit(EditControl) do
          if Assigned(ImgDlg) and ImgDlg.Visible then
            ImgDlg.Close;
end;

function TLKDBControlCollection.FormatGet: string;
begin
  Result := '';
end;

function TLKDBControlCollection.FormatSet(const Value: string): string;
begin
  //
end;

function TLKDBControlCollection.GetItem(Index: Integer): TLKDBControl;
begin
  Result := TLKDBControl(inherited GetItem(Index));
end;

procedure TLKDBControlCollection.SetItem(Index: Integer; Value: TLKDBControl);
begin
  inherited SetItem(Index, Value);
end;

procedure TLKDBControlCollection.SetReadOnly(const Value: Boolean);
var
  I: Integer;
begin
  FReadOnly := Value;
  if Value then
    for I := 0 to Count - 1 do
      Items[I].ReadOnly := Value;
end;

{
****************************** TLKDcDBGridControl ******************************
}
function TLKDcDBGridControl.CreateControlsFromControlType: Boolean;
var
  N: string;
begin
  Result := (Assigned(GridColumn) and Assigned(FField));
  if not Result then
    Exit;
  case ControlType of
    dbcUnassigned: Exit;
    dbcDCLookup:
      begin
        N := Grid.Name + FField.FieldName;
        FEditControl := TLKDBDcLookupEdit.Create(ControlOwner);
        with TLKDBDcLookupEdit(FEditControl) do
        begin
          ParentColor := True;
          Height := 15;
          if (FField.tag = 1) then
            with font do
              style := style + [fsBold];
        end;
      end;
    dbcCheckbox:
      begin
        FEditControl := TDBCheckBox.Create(ControlOwner);
        with TDBCheckBox(FEditControl) do
        begin
          Alignment := taLeftJustify;
          Caption := '';
          AllowGrayed := True;
          Height := 15;
          Width := 15;
          if (FField.Datatype in [ftSmallInt, ftBoolean]) then
          begin
            ValueChecked := '1';
            ValueUnchecked := '0';
          end;
        end;
      end;
    dbcMemo, dbcRadioGroup:
      begin
      end;
    dbcEdit: case FField.DataType of
        ftString, ftWideString:
          begin
          end;
        ftInteger, ftWord, ftFloat, ftCurrency, ftBCD:
          begin
          end;
        ftSmallInt, ftBoolean:
          begin
          end;
      end;
    dbcButtonEdit:
      begin
        FEditControl := TLKDBButtonEdit.Create(ControlOwner);
        with TLKDBButtonEdit(FEditControl) do
        begin
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcImageFile:
      begin
        FEditControl := TLKDBImageEdit.create(ControlOwner);
        with TLKDBImageEdit(FEditControl) do
        begin
          ParentColor := True;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcFilePath:
      begin
        FEditControl := TLKDBFilePathEdit.Create(ControlOwner);
        with TLKDBFilePathEdit(FEditControl) do
        begin
          ParentColor := True;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcURL:
      begin
        FEditControl := TLKDBUrlEdit.Create(ControlOwner);
        with TLKDBUrlEdit(FEditControl) do
        begin
          ParentColor := True;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcEMail:
      begin
        FEditControl := TLKDBEMailEdit.Create(ControlOwner);
        with TLKDBEMailEdit(FEditControl) do
        begin
          ParentColor := True;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcTimeEdit:
      begin
        FEditControl := TJvDBMaskEdit.Create(ControlOwner);
        with TJvDBMaskEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          EditMask := '!90:00;1;_';
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcDateEdit:
      begin
        FEditControl := TJvDBDateEdit.Create(ControlOwner);
        with TJvDBDateEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcDatePicker:
      begin
        FEditControl := TJvDBDatePickerEdit.Create(ControlOwner);
        with TJvDBDatePickerEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcDateTimePicker:
      begin
        FEditControl := TJvDBDateTimePicker.Create(ControlOwner);
        with TJvDBDateTimePicker(FEditControl) do
        begin
          Height := 21;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcRichEdit:
      begin
        FEditControl := TJvDBRichEdit.Create(ControlOwner);
        with TJvDBRichEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcSpinEdit:
      begin
        FEditControl := TJvDBSpinEdit.Create(ControlOwner);
        with TJvDBSpinEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcLookUpListBox:
      begin
        FEditControl := TJvDBLookUpList.Create(ControlOwner);
        with TJvDBLookUpList(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcLookUpComboBox:
      begin
        FEditControl := TJvDBLookUpCombo.Create(ControlOwner);
        with TJvDBLookUpCombo(FEditControl) do
        begin
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcRadioPanel:
      begin
        FEditControl := TJvDBRadioPanel.Create(ControlOwner);
        with TJvDBRadioPanel(FEditControl) do
        begin
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcComboBox:
      begin
        FEditControl := TDBComboBox.Create(ControlOwner);
        with TDBComboBox(FEditControl) do
        begin
          Height := 22;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcTreeView:
      begin
        FEditControl := TJvDBTreeView.Create(ControlOwner);
        with TJvDBTreeView(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcLookupTreeViewCombo:
      begin
        FEditControl := TJvDBLookupTreeViewCombo.Create(ControlOwner);
        with TJvDBLookupTreeViewCombo(FEditControl) do
        begin
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcLookupTreeView:
      begin
        FEditControl := TJvDBLookupTreeView.Create(ControlOwner);
        with TJvDBLookupTreeView(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcComboEdit:
      begin
        FEditControl := TJvDBComboEdit.Create(ControlOwner);
        with TJvDBComboEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcCalcEdit:
      begin
        FEditControl := TJvDBCalcEdit.Create(ControlOwner);
        with TJvDBCalcEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcMaskEdit:
      begin
        FEditControl := TJvDBMaskEdit.Create(ControlOwner);
        with TJvDBMaskEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcLookupComboEdit:
      begin
        FEditControl := TJvDBLookUpComboEdit.Create(ControlOwner);
        with TJvDBLookUpComboEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcSearchEdit:
      begin
        FEditControl := TJvDBSearchEdit.Create(ControlOwner);
        with TJvDBSearchEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcSearchComboBox:
      begin
        FEditControl := TJvDBSearchComboBox.Create(ControlOwner);
        with TJvDBSearchComboBox(FEditControl) do
        begin
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcFindEdit:
      begin
        FEditControl := TJvDBFindEdit.Create(ControlOwner);
        with TJvDBFindEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcImage:
      begin
        FEditControl := TJvDBImage.Create(ControlOwner);
        with TJvDBImage(FEditControl) do
        begin
          BorderStyle := bsNone;
          Height := 15;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    else
      Result := inherited CreateControlsFromControlType;
  end;
end;

procedure TLKDcDBGridControl.DoEditOnExit(Sender: TObject);
begin
  Grid.CloseControl;
  inherited DoEditOnExit(Sender);
end;

function TLKDcDBGridControl.GetGridControl: TJvDBGridControl;
begin
  if Assigned(GridColumn) then
    Result := Grid.EditControls.ControlByField(FField.FieldName)
  else
    Result := nil;
end;

function TLKDcDBGridControl.GetWidth: Integer;
begin
  if Assigned(GridColumn) then
    Result := GridColumn.Width
  else
    Result := 0;
end;

procedure TLKDcDBGridControl.SetAlignment(Value: TAlignment);
begin
  inherited SetAlignment(Value);
  if Assigned(GridColumn) then
    with GridColumn do
    begin
      Alignment := Value;
      Title.Alignment := Value;
    end;
end;

procedure TLKDcDBGridControl.SetCaption(const Value: string);
begin
  inherited SetCaption(Value);
  if Assigned(FGridColumn) then
    FGridColumn.DisplayName := Value;
end;

procedure TLKDcDBGridControl.SetControlType(const Value: TLKDBControlType);
begin
  inherited SetControlType(Value);
  if Assigned(FEditControl) and Assigned(FField) then
  begin
    if not Assigned(GridControl) then
      with Grid.EditControls.Add do
        FieldName := FField.FieldName;
    FEditControl.Name := Grid.Name + FField.FieldName;
    with GridControl do
    begin
      ControlName := FEditControl.Name;
      FitCell := fcCellSize;
    end;
  end;
end;

procedure TLKDcDBGridControl.SetDisplayName(const Value: string);
begin
  if Assigned(FGridColumn) then
    GridColumn.DisplayName := Value;
end;

procedure TLKDcDBGridControl.SetGrid(Value: TJvDBUltimGrid);
begin
  If Assigned(Value) then
    DataSource := Value.DataSource;
  FGrid := Value;
end;

procedure TLKDcDBGridControl.SetGridColumn(Value: TColumn);
begin
  FGridColumn := Value;
  if Assigned(Value) then
  begin
    Field := Value.Field;
    if Assigned(Field) then
      case Field.Datatype of
        ftBlob, ftMemo: Visible := False;
      end;
    Caption := Value.Title.Caption;
  end;
end;

procedure TLKDcDBGridControl.SetReadOnly(const Value: Boolean);
begin
  inherited SetReadOnly(Value);
  if not Assigned(GridColumn) then
    Exit;
  GridColumn.ReadOnly := Value;
  // Tabstop := not FReadOnly;
  with GridColumn do
    if Value then
    begin
      Color := clInfoBk;
      Font.Color := clNavy;
    end
    else
    begin
      Color := clWindow; //Parent.Color;
      Font.Color := clBlack;
    end;
end;

procedure TLKDcDBGridControl.SetVisible(const Value: Boolean);
begin
  inherited SetVisible(Value);
  if not Assigned(GridColumn) then
    Exit;
  GridColumn.Visible := Value;
end;

procedure TLKDcDBGridControl.SetWidth(const Value: Integer);
begin
  if not Assigned(GridColumn) then
    Exit;
  GridColumn.Width := Value;
end;

{
***************************** TLKDBGridCollection ******************************
}
constructor TLKDBGridCollection.Create(ItemClass: TLKDcDBGridControlClass);
begin
  inherited Create(ItemClass);
end;

function TLKDBGridCollection.Add: TLKDcDBGridControl;
begin
  Result := TLKDcDBGridControl(inherited Add);
end;

procedure TLKDBGridCollection.ConstructFromDatasource(const ExcludeFieldNames:
    string);
var
  I: Integer;
begin
  inherited ConstructFromDatasource(ExcludeFieldNames);
  if not Assigned(Grid) then
    Exit;
  with Grid do
  begin
    if not Assigned(Datasource.Dataset) then
      Exit;
    Columns.RebuildColumns;
    with Datasource.Dataset do
      for I := FieldCount - 1 downto 0 do
      begin
        if (Fields[I].tag = 1) then
          with Columns[I].Font do
            Style := Style + [fsBold];
      end;
  end;
  RebulldControls;
  SetReadOnly(FReadOnly);
end;

function TLKDBGridCollection.FormatGet: string;
var
  i: Integer;
  s: string;

  procedure NextValue(Value: string);
  begin
    Result := Result + Value + ';';
  end;

begin
  Result := '';
  if not Assigned(Grid) then
    Exit;
  with Grid.Columns do
    for i := 0 to count - 1 do
      with Items[i] do
      begin
        NextValue(FieldName);
        // NextValue(IntToStr(Ord(ControlType)));
        // NextValue(IntToStr(Ord(Alignment)));
        // NextValue(IntToStr(Ord(ButtonStyle)));
        // NextValue(IntToStr(Ord(Color)));
        // NextValue(DisplayName);
        // NextValue(IntToStr(DropDownRows));
        // NextValue(BooleanToYesNo(Expanded));
        // NextValue(IntToStr(Ord(Font.Charset)));
        // NextValue(IntToStr(Ord(Font.Color)));
        // NextValue(IntToStr(Font.Height));
        // NextValue(Font.Name);
        // s := '';
        // if fsBold in Font.Style then s := s + 'b';
        // if fsItalic in Font.Style then s := s + 'i';
        // if fsUnderline in Font.Style then s := s + 'u';
        // if fsStrikeOut in Font.Style then s := s + 's';
        // NextValue(s);
        // s := ListToDelimitedString('|',PickList); NextValue(s);
        // NextValue(BooleanToYesNo(ReadOnly));
        if Visible then s := 'Y' else s := 'N';
        NextValue(s[1]);
        NextValue(IntToStr(Width));
        NextValue(''); // spare
      end;
end;

function TLKDBGridCollection.FormatSet(const Value: string): string;
var
  S: string;
  ThisFieldName: string;
  ThisCol: Integer;
  Found: Boolean;
  I: Integer;

  function NextValue: string;
  var
    I: integer;
  begin
    I := pos(';', S);
    Result := System.Copy(S, 1, I - 1);
    System.Delete(S, 1, I);
  end;

begin
  Result := Value;
  if (Value = '') or not Assigned(Grid) then
    Exit;
  S := Value;
  ThisCol := 0;
  while (pos(';', S) > 0) do
  begin
    ThisFieldName := NextValue;
    Found := false;
    with Grid.Columns do
      for I := ThisCol to Count - 1 do
        if (Items[I].FieldName = ThisFieldName) then
          with Items[I] do
          begin
            Found := true;
            //  ControlType := TLKDBControlType(StrToInt(NextValue));
            //  Alignment := TAlignment(StrToInt(NextValue));
            //  ButtonStyle := TColumnButtonStyle(StrToInt(NextValue));
            //  Color := TColor(StrToInt(NextValue));
            //  DisplayName := NextValue;
            //  DropDownRows := StrToInt(NextValue);
            //  Expanded := (NextValue = 'Yes');
            //  Font.Charset := TFontCharSet(StrToInt(NextValue));
            //  Font.Color := TColor(StrToInt(NextValue));
            //  Font.Height := StrToInt(NextValue);
            //  Font.Name := NextValue;
            //  t := NextValue;
            //  if pos('b',t)> 0 then Font.Style := Font.Style + [fsBold];
            //  if pos('I',t)> 0 then Font.Style := Font.Style + [fsItalic];
            //  if pos('u',t)> 0 then Font.Style := Font.Style + [fsUnderline];
            //  if pos('S',t)> 0 then Font.Style := Font.Style + [fsStrikeOut];
            //  DelimitedStringToList(NextValue,'|',PickList);
            //  ReadOnly := (NextValue = 'Yes');
            Visible := (NextValue = 'Y');
            Width := StrToInt(NextValue);
            NextValue; // spare
            Index := ThisCol;
            Inc(ThisCol);
            Break;
          end;
    if not Found then
      for I := 0 to 19 do
        NextValue;
  end;
  Result := S;
end;

function TLKDBGridCollection.GetItem(Index: Integer): TLKDcDBGridControl;
begin
  Result := TLKDcDBGridControl(inherited GetItem(Index));
end;

procedure TLKDBGridCollection.RebulldControls;
var
  I: Integer;
begin
  Clear;
  if Assigned(Grid) then with Grid do
    for I := 0 to Columns.Count - 1 do
      with Add do
      begin
        Grid := Self.FGrid;
        ControlOwner := Self.ControlOwner;
        GridColumn := Columns.Items[I];
        if Visible then
          Visible := (FExcludeFieldList.IndexOf(FieldName) < 0);
      end;
end;

procedure TLKDBGridCollection.SetFocus;
begin
  Grid.SetFocus;
end;

procedure TLKDBGridCollection.SetItem(Index: Integer; Value:
    TLKDcDBGridControl);
begin
  inherited SetItem(Index, Value);
end;

procedure TLKDBGridCollection.SetReadOnly(const Value: Boolean);
begin
  if (Value = FReadOnly) then
    Exit;
  inherited SetReadOnly(Value);
  if Assigned(Grid) then
    with Grid do
    begin
      ParentColor := Value;
      if Value then
      begin
        Font.Color := FontReadOnlyColor;
        Options := Options - [dgEditing];
      end
      else
      begin
        Color := EditBackgroundColor;
        Font.Color := FontEditColor;
        Options := Options + [dgEditing];
      end;
    end;
end;

{
****************************** TLKDcDBEditControl ******************************
}
constructor TLKDcDBEditControl.Create(Collection: TCollection);
begin
  inherited;
  FOnEnter := nil;
  FOnExit := nil;
end;

function TLKDcDBEditControl.CreateControlsFromControlType: Boolean;
begin
  Result := True;
  if not Assigned(FField) then
    Exit;
  case ControlType of
    dbcUnAssigned: Exit;
    dbcCheckbox, dbcRadioGroup: ;
  else
    if not Assigned(FCaptionLabel) then
    begin
      FCaptionLabel := TStaticText.Create(FControlOwner);
      FCaptionLabel.AutoSize := False;
      FCaptionLabel.Height := 15;
    end;
  end;
  LinesHeight := 1;
  case ControlType of
    dbcDCLookup:
      begin
        FEditControl := TLKDBDcLookupEdit.Create(FControlOwner);
        TLKDBDcLookupEdit(FEditControl).Autosize := False;
        LinesHeight := 1;
      end;
    dbcDatePicker:
      begin
        FEditControl := TJvDBDatePickerEdit.Create(FControlOwner);
        with TJvDBDatePickerEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcDateTimePicker:
      begin
        FEditControl := TJvDBDateTimePicker.Create(FControlOwner);
        with TJvDBDateTimePicker(FEditControl) do
        begin
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcRichEdit:
      begin
        FEditControl := TJvDBRichEdit.Create(FControlOwner);
        with TJvDBRichEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
        end;
        LinesHeight := 3;
      end;
    dbcSpinEdit:
      begin
        FEditControl := TJvDBSpinEdit.Create(FControlOwner);
        with TJvDBSpinEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
        end;
        LinesHeight := 1.3;
      end;
    dbcLookUpListBox:
      begin
        FEditControl := TJvDBLookUpList.Create(FControlOwner);
        with TJvDBLookUpList(FEditControl) do
        begin
          BorderStyle := bsNone;
        end;
        LinesHeight := 3;
      end;
    dbcLookUpComboBox:
      begin
        FEditControl := TJvDBLookUpCombo.Create(FControlOwner);
        with TJvDBLookUpCombo(FEditControl) do
        begin
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcRadioPanel:
      begin
        FEditControl := TJvDBRadioPanel.Create(FControlOwner);
        LinesHeight := 2;
      end;
    dbcComboBox:
      begin
        FEditControl := TDBComboBox.Create(FControlOwner);
        with TDBComboBox(FEditControl) do
        begin
          BevelKind := bkFlat;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.5;
      end;
    dbcTreeView:
      begin
        FEditControl := TJvDBTreeView.Create(FControlOwner);
        with TJvDBTreeView(FEditControl) do
        begin
          BevelKind := bkFlat;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 3;
      end;
    dbcLookupTreeViewCombo:
      begin
        FEditControl := TJvDBLookupTreeViewCombo.Create(FControlOwner);
        with TJvDBLookupTreeViewCombo(FEditControl) do
        begin
          BevelKind := bkFlat;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcLookupTreeView:
      begin
        FEditControl := TJvDBLookupTreeView.Create(FControlOwner);
        with TJvDBLookupTreeView(FEditControl) do
        begin
          BevelKind := bkFlat;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcComboEdit:
      begin
        FEditControl := TJvDBComboEdit.Create(FControlOwner);
        with TJvDBComboEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
        end;
        LinesHeight := 1.3;
      end;
    dbcDateEdit:
      begin
        FEditControl := TJvDBDateEdit.Create(FControlOwner);
        with TJvDBDateEdit(FEditControl) do
        begin
          BorderStyle := bsNone; //bug prevents size below 21
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.5;
      end;
    dbcTimeEdit:
      begin
        FEditControl := TJvDBMaskEdit.Create(FControlOwner);
        with TJvDBMaskEdit(FEditControl) do
        begin
          FField.EditMask := '!90:00;1;_';
          BorderStyle := bsNone;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcCalcEdit:
      begin
        FEditControl := TJvDBCalcEdit.Create(FControlOwner);
        with TJvDBCalcEdit(FEditControl) do
        begin
          BorderStyle := bsNone; //bug prevents size below 21
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.4;
      end;
    dbcMaskEdit:
      begin
        FEditControl := TJvDBMaskEdit.Create(FControlOwner);
        with TJvDBMaskEdit(FEditControl) do
        begin
          BorderStyle := bsNone; //bug prevents size below 21
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.5;
      end;
    dbcLookupComboEdit:
      begin
        FEditControl := TJvDBLookUpComboEdit.Create(FControlOwner);
        with TJvDBLookUpComboEdit(FEditControl) do
        begin
          BevelKind := bkFlat;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.5;
      end;
    dbcSearchEdit:
      begin
        FEditControl := TJvDBSearchEdit.Create(FControlOwner);
        with TJvDBSearchEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcSearchComboBox:
      begin
        FEditControl := TJvDBSearchComboBox.Create(FControlOwner);
        with TJvDBSearchComboBox(FEditControl) do
        begin
          BevelKind := bkFlat;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcFindEdit:
      begin
        FEditControl := TJvDBFindEdit.Create(FControlOwner);
        with TJvDBFindEdit(FEditControl) do
        begin
          BorderStyle := bsNone;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
        LinesHeight := 1.3;
      end;
    dbcImage:
      begin
        FEditControl := TJvDBImage.Create(FControlOwner);
        LinesHeight := 3;
      end;
    dbcCheckbox:
      begin
        FEditControl := TDBCheckBox.Create(FControlOwner);
        with TDBCheckBox(FEditControl) do
        begin
          Alignment := taLeftJustify;
          if Assigned(FField) then
          begin
            if (FField.Datatype in [ftSmallInt, ftBoolean]) then
            begin
              ValueChecked := '1';
              ValueUnchecked := '0';
            end;
          end;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcText:
      begin
        FEditControl := TDBText.Create(FControlOwner);
        with TDBText(EditControl) do
        begin
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcEdit:
      begin
        FEditControl := TDBEdit.Create(FControlOwner);
        with TDBEdit(EditControl) do
        begin
          Borderstyle := bsNone;
          if (FField.tag = 1) then
            with Font do
              Style := Style + [fsBold];
        end;
      end;
    dbcMemo:
      begin
        FEditControl := TDBMemo.Create(FControlOwner);
        with TDBMemo(EditControl) do
        begin
          Borderstyle := bsNone;
        end;
        LinesHeight := 3;
      end;
    dbcListBox:
      begin
        FEditControl := TDBListBox.Create(FControlOwner);
        with TDBListBox(EditControl) do
        begin
          Borderstyle := bsNone;
        end;
        LinesHeight := 3;
      end;
    dbcRadioGroup:
      begin
        FEditControl := TDBRadioGroup.Create(FControlOwner);
        LinesHeight := 2;
      end;
    dbcButtonEdit:
      begin
        FEditControl := TLKDBButtonEdit.Create(FControlOwner);
      end;
    dbcImageFile:
      begin
        FEditControl := TLKDBImageEdit.Create(FControlOwner);
      end;
    dbcFilePath:
      begin
        FEditControl := TLKDBFilePathEdit.Create(FControlOwner);
      end;
    dbcURL:
      begin
        FEditControl := TLKDBUrlEdit.Create(FControlOwner);
      end;
    dbcEMail:
      begin
        FEditControl := TLKDBEMailEdit.Create(FControlOwner);
      end;
    else
      Result := inherited CreateControlsFromControlType;
    if Assigned(FCaptionLabel) then
    begin
      if (FEditControl is TWinControl) then
        FCaptionLabel.FocusControl := TWinControl(FEditControl);
    end;
  end;
end;

function TLKDcDBEditControl.GetLinesHeight: Double;
begin
  Result := 0;
  if Visible then
    Result := FLinesHeight;
end;

function TLKDcDBEditControl.GetWidth: Integer;
begin
  case ControlType of
    dbcUnAssigned: Result := 0;
    dbcCheckbox: Result := Trunc((EditControl.Width / 0.35) - 16);
    dbcRadioGroup: Result := EditControl.Width - 8;
  else
    begin
    Result := EditControl.Width;
    if Assigned(FCaptionLabel) then
      Result := FCaptionLabel.Width + Result;
    end;
  end;
end;

procedure TLKDcDBEditControl.SetCaption(const Value: string);
begin
  inherited SetCaption(Value);
  case ControlType of
    dbcUnAssigned: Exit;
    dbcCheckbox:
      begin
        TDBCheckBox(EditControl).Caption := Value;
        FCaption := Value;
      end;
    dbcRadioGroup: TDBRadioGroup(EditControl).Caption := Value;
  else if Assigned(FCaptionLabel) then
    FCaptionLabel.Caption := Value;
  end;
end;

procedure TLKDcDBEditControl.SetControlParent(AParent: TWinControl);
begin
  inherited SetControlParent(AParent);
  if Assigned(FCaptionLabel) then
    FCaptionLabel.Parent := AParent;
end;

procedure TLKDcDBEditControl.SetControlType(const Value: TLKDBControlType);
begin
  if (ControlType = Value) then
    Exit;
  FreeAndNil(FCaptionLabel);
  inherited SetControlType(Value);
  SetLinesHeight(FLinesHeight);
  SetTop(Top);
  SetLeft(Left);
  SetTabOrder(FTabOrder);
end;

procedure TLKDcDBEditControl.SetDisplayName(const Value: string);
begin
  inherited SetDisplayName(Value);
  if Assigned(EditControl) then
    case ControlType of
      dbcCheckbox: TDBCheckBox(EditControl).Caption := Value;
      dbcRadioGroup: TDBRadioGroup(EditControl).Caption := Value + '  ';
    end;
  if Assigned(FCaptionLabel) then
    FCaptionLabel.Caption := Value;
end;

procedure TLKDcDBEditControl.SetLeft(const Value: Integer);
begin
  FLeft := Value;
  if (Layout = aceHorizontal) then
  begin
    if Assigned(FCaptionLabel) then
      FCaptionLabel.Left := Value;
    if Assigned(EditControl) then
      EditControl.Left := Value;
  end
  else
  begin
    if Assigned(FCaptionLabel) then
    begin
      FCaptionLabel.Left := Value;
      if Assigned(EditControl) then
        EditControl.Left := Value + FCaptionLabel.Width;
    end
    else if Assigned(EditControl) then
    if (ControlType = dbcRadioGroup) then
      EditControl.Left := Value
    else
      EditControl.Left := Value - 8;
  end;
end;

procedure TLKDcDBEditControl.SetLinesHeight(const Value: Double);
begin
  FLinesHeight := Value;
  if Assigned(EditControl) then
    case ControlType of
      dbcRadioGroup: EditControl.Height := Round(Value * 16 - 4);
      dbcComboBox: ;
    else
      EditControl.Height := Round(Value * 16 - 1);
    end;
end;

procedure TLKDcDBEditControl.SetReadOnly(const Value: Boolean);
begin
  inherited SetReadOnly(Value);
  if Assigned(EditControl) then
    case ControlType of
    dbcDatePicker: with TJvDBDatePickerEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
        //ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcDateTimePicker: with TJvDBDateTimePicker(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
       // ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcRichEdit: with TJvDBRichEdit(EditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcSpinEdit: with TJvDBSpinEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcLookUpListBox: with TJvDBLookUpList(EditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcLookUpComboBox: with TJvDBLookUpCombo(EditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcRadioPanel: with TJvDBRadioPanel(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
        if FReadOnly then
          Font.Color := FontReadOnlyColor
        else
          Font.Color := FontEditColor;
      end;
    dbcCombobox: with TDBComboBox(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
        ParentColor := FReadOnly;
        if FReadOnly then
          Font.Color := FontReadOnlyColor
        else
          Font.Color := FontEditColor;
      end;
    dbcTreeView: with TJvDBTreeView(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcLookupTreeViewCombo: with TJvDBLookupTreeViewCombo(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcLookupTreeView: with TJvDBLookupTreeView(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcComboEdit: with TJvDBComboEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcDateEdit: with TJvDBDateEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcTimeEdit: with TJvDBMaskEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcCalcEdit: with TJvDBCalcEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcMaskEdit: with TJvDBMaskEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcLookupComboEdit: with TJvDBLookUpComboEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcSearchComboBox: with TJvDBSearchComboBox(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcFindEdit: with TJvDBFindEdit(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
  //      ParentColor := FReadOnly;
        if FReadOnly then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcImage: ;
    dbcCheckbox: with TDBCheckBox(FEditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
      end;
    dbcMemo: with TDBMemo(EditControl) do
      begin
        ReadOnly := Value;
        TabStop := not Value;
//        ParentColor := Value;
        if Value then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcListBox: with TDBListBox(EditControl) do
      begin
        ReadOnly := Value;
        TabStop := not Value;
//        ParentColor := Value;
        if Value then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
        else
        begin
          Color := EditBackgroundColor;
          Font.Color := FontEditColor;
        end;
      end;
    dbcRadioGroup: with TDBRadioGroup(EditControl) do
      begin
        ReadOnly := FReadOnly;
        TabStop := not FReadOnly;
        if FReadOnly then
          Font.Color := FontReadOnlyColor
        else
          Font.Color := FontEditColor;
      end;
    else if EditControl is TDBEdit then
    with TDBEdit(EditControl) do
    begin
      ReadOnly := Value;
      TabStop := not Value;
//      ParentColor := Value;
      if Value then
        begin
          Color := EditReadOnlyBackgroundColor;
          Font.Color := FontReadOnlyColor;
        end
      else
      begin
        Color := EditBackgroundColor;
        Font.Color := FontEditColor;
      end;
    end;
  end
end;

procedure TLKDcDBEditControl.SetTabOrder(const Value: Integer);
begin
  FTabOrder := Value;
  if Assigned(EditControl) then
    if EditControl is TWinControl then
      TWinControl(EditControl).TabOrder := Value;
end;

procedure TLKDcDBEditControl.SetTop(const Value: Integer);
begin
  FTop := Value;
  if (Layout = aceHorizontal) then
  begin
    if Assigned(FCaptionLabel) then
      FCaptionLabel.Top := Value;
    if Assigned(EditControl) then
      EditControl.Top := Value + 16;
  end
  else
  begin
    if Assigned(FCaptionLabel) then
      FCaptionLabel.Top := Value;
    if Assigned(EditControl) then
      EditControl.Top := Value;
  end;
end;

procedure TLKDcDBEditControl.SetVisible(const Value: Boolean);
begin
  inherited SetVisible(Value);
  if Assigned(FCaptionLabel) then
    FCaptionLabel.Visible := Visible;
  if Assigned(EditControl) then
    EditControl.Visible := Visible;
end;

procedure TLKDcDBEditControl.SetWidth(const Value: Integer);
begin
  case ControlType of
    dbcUnAssigned: ;
    dbcCheckbox: EditControl.Width := Trunc(0.35 * Value + 16);
    dbcRadioGroup: EditControl.Width := Value + 8;
  else
    begin
      FCaptionLabel.Width := Trunc(Value * 0.35);
      with EditControl do
      begin
        Left := FCaptionLabel.Left + FCaptionLabel.Width;
        Width := Trunc(Value * 0.65);
      end;
    end;
  end;
end;

{
***************************** TLKDBEditCollection ******************************
}
constructor TLKDBEditCollection.Create(ItemClass: TLKDcDBEditControlClass);
begin
  inherited Create(ItemClass);
end;

function TLKDBEditCollection.Add: TLKDcDBEditControl;
begin
  Result := TLKDcDBEditControl(inherited Add);
end;

procedure TLKDBEditCollection.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    FreeAndNil(Items[I].FCaptionLabel);
  inherited Clear;
end;

procedure TLKDBEditCollection.ConstructFromDatasource(const ExcludeFieldNames:
    string);
var
  I, X: Integer;
  Table: TDataSet;
begin
  inherited ConstructFromDatasource(ExcludeFieldNames);
  if not Assigned(FControlDataSource) then
    Exit;
  X := 0;
  Table := FControlDataSource.Dataset;
  if not Assigned(Table) then
    Exit;
  Table.FieldDefs.Update;
  for I := 0 to Table.Fields.Count - 1 do
    if (Table.Fields[I].Datatype <> ftBlob) and
      (FExcludeFieldList.IndexOf(Table.Fields[I].FieldName) < 0) then
      with Add do
      begin
        ControlOwner := Self.ControlOwner;
        ControlParent := Self.ControlParent;
        DataSource := ControlDataSource;
        Field := Table.Fields[I];
        TabOrder := X;
        Inc(X);
      end;
end;

function TLKDBEditCollection.FormatGet: string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Count - 1 do
  begin
    with Items[I] do
    begin
      Result := Result + Field.FieldName + ';';
      Result := Result + Caption + ';';
      if Visible then
        Result := Result + 'Yes;'
      else
        Result := Result + 'No;';
      Result := Result + FloatToStr(LinesHeight) + ';';
    end;
    if (Items[I].EditControl is TLKDBImageEdit) then
      with TLKDBImageEdit(Items[I].EditControl).ImgDlg do
      begin
        Result := IntToStr(Top)+ '|' + IntToStr(Left)+ '|'
                + IntToStr(Width)+ '|' + IntToStr(Height)+ ';';
      end
    else
      Result := Result + ';';
    Result := Result + ';;'; // 2 spares for future development
  end;
end;

function TLKDBEditCollection.FormatSet(const Value: string): string;
var
  MySpare: string;
  I: Integer;
  S: string;

  function NextValue: string;
  begin
    if (S = '') then
      Result := ''
    else
    begin
      Result := copy(S, 1, pos(';', S) - 1);
      System.Delete(S, 1, pos(';', S));
    end;
  end;

  procedure  SetImageEditDlg(ADBImageEdit: TLKDBImageEdit; const ANextValue: string);
  var
    S: string;

    function ExtractNext: Integer;
    var
      I: Integer;
    begin
      try
        I := Pos('|',S);
        if (I = 0) then
          Result := StrToInt(S)
        else
        begin
          Result := StrToInt(Copy(S,1,I-1));
          System.Delete(S,1,I);
        end;
      except
        Result := 200;
      end;
    end;

  begin
    if (ANextValue = '') then
      Exit;
    S:= ANextValue;
    with ADBImageEdit.ImgDlg do
    begin
      Top   := ExtractNext;
      Left  := ExtractNext;
      Width := ExtractNext;
      Height:= ExtractNext;
    end;
  end;

begin
  // A string list of field names is used so that format is retained even if fields or field order changes
  S := Value;
  if (S > '') then for I := 0 to Count - 1 do
  begin
    with Items[I] do
    begin
      if (FieldName <> NextValue) then
        Abort;
      Caption := NextValue;
      Visible := (NextValue = 'Yes');
      LinesHeight := StrToFloat(NextValue);
    end;
    if (Items[I].EditControl is TLKDBImageEdit) then
      SetImageEditDlg(TLKDBImageEdit(Items[I].EditControl), NextValue)
    else MySpare := NextValue;
    MySpare := NextValue;
    MySpare := NextValue;
  end;
  Result := S;
end;

function TLKDBEditCollection.GetItem(Index: Integer): TLKDcDBEditControl;
begin
  Result := TLKDcDBEditControl(inherited GetItem(Index));
end;

procedure TLKDBEditCollection.SetControlDataSource(ADataSource: TDataSource);
var
  I: Integer;
begin
  FControlDataSource := ADataSource;
  for I := 0 to Count - 1 do
    Items[I].DataSource := ADataSource;
end;

procedure TLKDBEditCollection.SetControlParent(AParent: TWinControl);
var
  I: Integer;
begin
  FControlParent := AParent;
  for I := 0 to Count-1 do
    Items[I].ControlParent := AParent;
end;

procedure TLKDBEditCollection.SetItem(Index: Integer; Value:
    TLKDcDBEditControl);
begin
  inherited SetItem(Index, Value);
end;

procedure TLKDBEditCollection.UpdateFields;
var
  I: Integer;
  Table: TDataset;
begin
  if not Assigned(FControlDataSource) then
    Exit;
  Table := ControlDataSource.Dataset;
  if Assigned(Table) then
    for I := 0 to Count - 1 do
      Items[I].Field := Table.FindField(Items[I].FieldName);
end;

{
****************************** TLKDcDBEditsPanel *******************************
}
constructor TLKDcDBEditsPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDisplayHeight := Height;
  FEditControls := TLKDBEditCollection.Create(TLKDcDBEditControl);
  FEditControls.ControlOwner := AOwner;
  FEditControls.ControlParent := Self;
end;

destructor TLKDcDBEditsPanel.Destroy;
begin
  FEditControls.Free;
  inherited Destroy;
end;

procedure TLKDcDBEditsPanel.ConstructFromDatasource(const ExcludeFieldNames:
    string);
begin
  FEditControls.ConstructFromDatasource(ExcludeFieldNames);
end;

function TLKDcDBEditsPanel.DlgEditListPropsExecute(AOwner: TComponent):
    TModalResult;
begin
  with TDlgLKUtilsEditListProps.Create(AOwner) do
  begin
    try
      Construct(Self);
      Result := ShowModal;
    finally Free;
    end;
  end;
end;

procedure TLKDcDBEditsPanel.DoOnFormClose;
begin
  FEditControls.DoOnFormClose;
end;

function TLKDcDBEditsPanel.GetCount: Integer;
begin
  try
    Result := FEditControls.Count;
  except
    Result := 0;
  end;
end;

function TLKDcDBEditsPanel.GetDataSource: TDataSource;
begin
  Result := FEditControls.ControlDataSource;
end;

function TLKDcDBEditsPanel.GetFormat: string;
begin
  Result := FEditControls.FormatGet
          + IntToStr(Ord(Layout)) + ';'
          + IntToStr(ColWidth0) + ';'
          + IntToStr(ColWidth1) + ';'
          + IntToStr(ColWidth2) + ';'
          + ';;;'; // 3 spares for future development
end;

function TLKDcDBEditsPanel.GetItem(Index: Integer): TLKDcDBEditControl;
begin
  Result := FEditControls.Items[Index];
end;

function TLKDcDBEditsPanel.GetReadOnly: Boolean;
begin
  Result := FEditControls.ReadOnly;
end;

function TLKDcDBEditsPanel.LoadFormat(const AComponent: TLKSystemComponent;
    ASection, AnItem: string): Boolean;
var
  S: string;
  S1: string;
  I: Integer;
begin
  S := '';
  I := 0;
  repeat
    S1 := AComponent.RecallString(ASection, AnItem + IntToStr(I), '', false);
    S := S + S1;
    Inc(I);
  until (S1 = '');
  Result := (S > '');
  if Result then
  try
    Format := S;
  except
    Result := False;
  end;
end;

procedure TLKDcDBEditsPanel.Resize;
var
  I: Integer;
  ClipWidth: Integer;
  NextLeft: Integer;
  NextTop: Integer;
  ColDown: Integer;
  Edits0Width: Integer;
  Edits1Width: Integer;
  Edits2Width: Integer;
  Edits0Left: Integer;
  Edits1Left: Integer;
  Edits2Left: Integer;
begin
  inherited Resize;
  if (FEditControls.Count = 0) then
    Exit;
  if (Width < 20) then
    Exit;
  for I := 0 to FEditControls.Count - 1 do
    FEditControls[I].Alignment := taLeftJustify;
  case FLayout of
    aceHorizontal:
      begin
        ClipWidth := Width - 16;
        NextLeft := 8;
        NextTop := 4;
        ColDown := 48;
        for I := 0 to FEditControls.Count - 1 do
        begin
          FEditControls[I].Width :=
            Min(FEditControls[I].Field.DisplayWidth + 2, ClipWidth);
          if (NextLeft + FEditControls[I].Width > ClipWidth) then
          begin
            NextLeft := 8;
            Inc(NextTop, ColDown);
          end;
          Left := NextLeft;
          Top := NextTop;
          Inc(NextLeft, Width);
        end;
        FDisplayHeight := NextTop + ColDown;
      end;
    aceOneCol:
      begin
        ClipWidth := Width - 16;
        for I := 0 to FEditControls.Count - 1 do
        begin
          FEditControls[I].Left := 8;
          FEditControls[I].Width := ClipWidth;
        end;
      end;
    aceTwoCol:
      begin
        ClipWidth := Width - 24;
        Edits0Width := Trunc(ClipWidth * ColWidth0 / 100);
        Edits1Width := Trunc(ClipWidth * ColWidth1 / 100);
        Edits0Left := 8;
        Edits1Left := Edits0Left + Edits0Width + 8;
        for I := 0 to FEditControls.Count - 1 do
          case FEditControls[I].Column of
            0:
              begin
                FEditControls[I].Left := Edits0Left;
                FEditControls[I].Width := Edits0Width;
              end;
          else
            FEditControls[I].Left := Edits1Left;
            FEditControls[I].Width := Edits1Width;
          end;
      end;
    aceThreeCol:
      begin
        ClipWidth := Width - 32;
        Edits0Width := Trunc(ClipWidth * ColWidth0 / 100);
        Edits1Width := Trunc(ClipWidth * ColWidth1 / 100);
        Edits2Width := Trunc(ClipWidth * ColWidth2 / 100);
        Edits0Left := 8;
        Edits1Left := Edits0Left + Edits0Width + 8;
        Edits2Left := Edits1Left + Edits1Width + 8;
        for I := 0 to FEditControls.Count - 1 do
          case FEditControls[I].Column of
            0:
              begin
                FEditControls[I].Left := Edits0Left;
                FEditControls[I].Width := Edits0Width;
              end;
            1:
              begin
                FEditControls[I].Left := Edits1Left;
                FEditControls[I].Width := Edits1Width;
              end;
          else
            FEditControls[I].Left := Edits2Left;
            FEditControls[I].Width := Edits2Width;
          end;
      end;
  end;
end;

procedure TLKDcDBEditsPanel.SaveFormat(const AComponent: TLKSystemComponent;
    ASection, AnItem: string);
var
  S: string;
  I: Integer;
  B: Integer;
  S1: string;
begin
  S := Format;
  I := 0;
  B := 1;
  while (B <= Length(S)) do
  begin
    S1 := Copy(S, B, 255);
    AComponent.RememberString(ASection, AnItem + IntToStr(I), S1, False);
    inc(I);
    B := 255 * I + 1;
  end;
end;

procedure TLKDcDBEditsPanel.SetDataSource(ADataSource: TDataSource);
begin
  FEditControls.ControlDataSource := ADataSource;
end;

procedure TLKDcDBEditsPanel.SetFormat(const Value: string);
var
  S, MySpare: string;

  function NextValue: string;
  begin
    if (S = '') then
      Result := ''
    else
    begin
      Result := copy(S, 1, pos(';', S) - 1);
      System.Delete(S, 1, pos(';', S));
    end;
  end;

begin
  S := FEditControls.FormatSet(Value);
  Layout := TLKEditsLayout(StrToInt(NextValue));
  ColWidth0 := StrToInt(NextValue);
  ColWidth1 := StrToInt(NextValue);
  ColWidth2 := StrToInt(NextValue);
  MySpare := NextValue;
  MySpare := NextValue;
  MySpare := NextValue;
  Resize;
end;

procedure TLKDcDBEditsPanel.SetLayout(const Value: TLKEditsLayout);
var
  I, NextTop, ColNo: Integer;
  TotalLines: Double;
  MaxEditTop: Double;
  LastColNumber: Integer;
begin
  FLayout := Value;
  TotalLines := 0;
  for I := 0 to EditControls.Count - 1 do
  begin
    EditControls[I].Layout := FLayout;
    if EditControls[I].Visible then
      TotalLines := TotalLines + EditControls[I].LinesHeight;
  end;
  MaxEditTop := 0;
  LastColNumber := 0;
  case FLayout of
    aceHorizontal:
      begin
        Resize;
        Exit;
      end;
    aceOneCol:
      begin
        MaxEditTop := 16 * TotalLines;
        ColWidth0 := 100;
        ColWidth1 := 0;
        ColWidth2 := 0;
      end;
    aceTwoCol:
      begin
        MaxEditTop := 16 * Trunc((TotalLines + 1.1) / 2);
        ColWidth0 := 55;
        ColWidth1 := 45;
        ColWidth2 := 0;
        LastColNumber := 1;
      end;
    aceThreeCol:
      begin
        MaxEditTop := 16 * Trunc((TotalLines + 2.1) / 3);
        ColWidth0 := 40;
        ColWidth1 := 30;
        ColWidth2 := 30;
        LastColNumber := 2;
      end;
  end;
  // set the DBControls Top and Column positions
  ColNo := 0;
  FDisplayHeight := 0;
  NextTop := 4; // 4 pixel space at top
  for I := 0 to FEditControls.Count - 1 do
  begin
    with FEditControls[I] do
    begin
      Column := ColNo;
      if Visible then
      begin
        Top := NextTop;
        NextTop := NextTop + Round(LinesHeight * 16);
        FDisplayHeight := Max(FDisplayHeight, NextTop);
      end;
    end;
    //check for new column
    if (NextTop >= MaxEditTop) and (ColNo < LastColNumber) then
    begin
      NextTop := 4;
      Inc(ColNo);
    end;
  end;
  Resize;
end;

procedure TLKDcDBEditsPanel.SetReadOnly(const Value: Boolean);
begin
  FEditControls.ReadOnly := Value;
end;

procedure TLKDcDBEditsPanel.UpdateFields;
begin
  FEditControls.UpdateFields;
end;

{
**************************** TLKDcDBGridController *****************************
}
constructor TLKDcDBGridController.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FGridColumnItems := TLKDBGridCollection.Create(TLKDcDBGridControl);
  FGridColumnItems.ControlOwner := AOwner;
end;

destructor TLKDcDBGridController.Destroy;
begin
  FGridColumnItems.Free;
  inherited Destroy;
end;

procedure TLKDcDBGridController.ClearFormat(const AComponent:
    TLKSystemComponent; ASection, AnItem: string);
var
  i: Integer;
begin
  i := 0;
  while (AComponent.RecallString(ASection, AnItem + IntToStr(i), '', false) > '')
    do
  begin
    AComponent.SystemSettings.TblMaster.Delete;
    inc(i);
  end;
  AComponent.SystemSettings.ApplyUpdates(0);
end;

procedure TLKDcDBGridController.ConstructFromDatasource(const
    ExcludeFieldNames: string);
begin
  FGridColumnItems.ConstructFromDatasource(ExcludeFieldNames);
end;

function TLKDcDBGridController.DlgEditListPropsExecute(AOwner: TComponent):
    TModalResult;
begin
  if Assigned(Grid) then
    begin
      Grid.ShowColumnsDialog;
      Result := mrOk;
    end
  else
    Result := mrCancel;
end;

procedure TLKDcDBGridController.DoOnFormClose;
begin
  FGridColumnItems.DoOnFormClose;
end;

function TLKDcDBGridController.GetCount: Integer;
begin
  Result := FGridColumnItems.Count;
end;

function TLKDcDBGridController.GetItem(Index: Integer): TLKDcDBGridControl;
begin
  Result := FGridColumnItems.Items[Index];
end;

function TLKDcDBGridController.GetReadOnly: Boolean;
begin
  Result := FGridColumnItems.ReadOnly;
end;

function TLKDcDBGridController.GetSelectedField: TField;
begin
  if Assigned(Grid) then
    Result := Grid.SelectedField
  else Result := nil;
end;

function TLKDcDBGridController.LoadFormat(const AComponent: TLKSystemComponent;
    ASection, AnItem: string): Boolean;
var
  S: string;
  S1: string;
  I: Integer;
begin
  S := '';
  I := 0;
  repeat
    S1 := AComponent.RecallString(ASection, AnItem + IntToStr(I), '', false);
    S := S + S1;
    inc(I);
  until (S1 = '');
  Result := (GridColumnItems.FormatSet(S) = '');
end;

procedure TLKDcDBGridController.SaveFormat(const AComponent: TLKSystemComponent;
    ASection, AnItem: string);
var
  S: string;
  I: Integer;
  B: Integer;
  S1: string;
begin
  S := GridColumnItems.FormatGet;
  I := 0;
  B := 1;
  while (B <= Length(S)) do
  begin
    S1 := Copy(S, B, 255);
    AComponent.RememberString(ASection, AnItem + IntToStr(I), S1, False);
    inc(I);
    B := 255 * I + 1;
  end;
end;

procedure TLKDcDBGridController.SetGrid(Value: TJvDBUltimGrid);
begin
  FGrid := Value;
  GridColumnItems.Grid := Value;
end;

procedure TLKDcDBGridController.SetReadOnly(const Value: Boolean);
begin
  FGridColumnItems.ReadOnly := Value;
end;

procedure TLKDcDBGridController.SetSelectedField(Value: TField);
begin
  if Assigned(Grid) then
    Grid.SelectedField := Value;
end;



end.
