{*****************************************************************************}
{                                                                             }
{   This file is part of the LKRecords components Library                     }
{                                                                             }
{   See the file COPYING.LGPL.txt included in this distribution,              }
{   for details about the Licence and Copyright.                              }
{                                                                             }
{   This code is distributed in the hope that it will be useful,              }
{   but WITHOUT ANY WARRANTY; without even the implied warranty of            }
{   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      }
{                                                                             }
{   Author: Leslie John Kaye :    www.leskaye.com                             }
{                                                                             }
{   Abstract:                                                                 }
{      various utility panels and dialogs used by TLKDataComponent etc.       }
{                                                                             }
{*****************************************************************************}

unit LKDlgs;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, System.UITypes,
  Graphics, Controls, Variants,
  Forms, StdCtrls, ExtCtrls, Buttons, Dialogs, DB, DBTables, ComCtrls,
  JvExStdCtrls, JvTextListBox, LKDefine, LKRecordsComponents;

type

  TLKWizPanel = class(TPanel)
  private
    FOnButtonClick: TNotifyEvent;
    FStatusText: string;
    FHelpStr: string;
    FHelpURI: string;
    function GetButtonCaption: string;
    procedure SetButtonCaption(const Value: string);
    procedure SetHelpStr(const Value: string);
  protected
    FBtnExecute: TButton;
    FBtnHelp: TButton;
    procedure CheckHelpButtonVisible;
    procedure DoButtonClick(Sender: TObject); virtual;
    procedure DoBtnHelpClick(Sender: TObject);
  public
    LabelItemNo: TLabel;
    procedure Resize; override;
    constructor Create(AOwner: TComponent); override;
    property StatusText: string read FStatusText write FStatusText;
  published
    property ButtonCaption: string read GetButtonCaption write SetButtonCaption;
    property HelpStr: string read FHelpStr write SetHelpStr;
    property HelpURI: string read FHelpURI write FHelpURI;
    property OnButtonClick: TNotifyEvent read FOnButtonClick write
        FOnButtonClick;
  end;

  TLKComponentWizPanel = class(TLKWizPanel)
  private
    FLKComponent: TLKSystemComponent;
  protected
    procedure SetLKComponent(Value: TLKSystemComponent); virtual;
  public
    property LKComponent: TLKSystemComponent read FLKComponent write
        SetLKComponent;
  end;

  TLKACSetupWizPanel = class(TLKComponentWizPanel)
  protected
    procedure DoButtonClick(Sender: TObject); override;
    procedure SetLKComponent(Value: TLKSystemComponent); override;
  end;

  TLKFilterPanel = class(TPanel)
  private
    cbFilter: TComboBox;
    ckbxFilter: TCheckbox;
    FDataComponent: TLKDataComponent;
    FFilterDataset: TLKACTable;
    Reloading: Boolean;
    sbFilter: TSpeedButton;
    procedure CbFilterChange(Sender: TObject);
    procedure CbFilterExit(Sender: TObject);
    procedure CkbxFilterClick(Sender: TObject);
    procedure DlgTableFilterExecute(Sender: TObject);
    function GetFilter: string;
    function GetFiltered: Boolean;
    procedure SetDataComponent(const Value: TLKDataComponent);
    procedure SetFilter(const Value: string);
    procedure SetFilterDataset(const Value: TLKACTable);
    procedure SetFiltered(Value: Boolean);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation);
        override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure FilterReload;
    property Filter: string read GetFilter write SetFilter;
    property Filtered: Boolean read GetFiltered write SetFiltered;
  published
    property DataComponent: TLKDataComponent read FDataComponent write
        SetDataComponent;
    property FilterDataset: TLKACTable read FFilterDataset write
        SetFilterDataset;
  end;

  TLKTableFilterDialog = class(TComponent)
  private
    FACTable: TLKACTable;
    FDataComponent: TLKDataComponent;
    FFieldName: string;
    FOnClose: TNotifyEvent;
    FOnShow: TNotifyEvent;
    FTable: TDataset;
    FVersion: string;
    myTable: TDataset;
  protected
    procedure doOnClose;
    procedure doOnShow;
    procedure SetACTable(const AACTable: TLKACTable);
    procedure SetDataComponent(const ADataComponent: TLKDataComponent);
    procedure SetTable(const ATable: TDataset);
    procedure SetVersion(const Value: string);
  public
    Folder: string;
    constructor Create(AOwner: TComponent); override;
    function Execute: Boolean;
  published
    property ACTable: TLKACTable read FACTable write SetACTable default
        tbMaster;
    property DataComponent: TLKDataComponent read FDataComponent write
        SetDataComponent;
    property FilterFieldName: string read FFieldName write FFieldName;
    property Table: TDataset read FTable write SetTable;
    property Version: string read FVersion write SetVersion;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
  end;

  TJvSearchListBox = class(TJvTextListBox)
  private
    FSearchItems: TStringList;
    FSearchText: string;
    procedure SetSearchItems(Value: TStringList);
    procedure SetSearchText(const Value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property SearchItems: TStringList read FSearchItems write SetSearchItems;
    property SearchText: string read FSearchText write SetSearchText;
  end;

procedure Register;

implementation

uses LKHeaderDlg, LKTblFilterDg, LKZipProgressDlg, LKWrZipDlg;

procedure Register;
begin
  // RegisterComponents('LKUtils', [TLKDBACLookupEdit]);
  RegisterComponents('LKUtils', [TJvSearchListBox]);
  RegisterComponents('LKUtils', [TLKFilterPanel]);
  RegisterComponents('LKUtils', [TLKTableFilterDialog]);
end;


{
***************************** TLKComponentWizPanel *****************************
}

procedure TLKComponentWizPanel.SetLKComponent(Value: TLKSystemComponent);
begin
  FLKComponent := Value;
  if Assigned(Value) then
  begin
    FBtnExecute.Caption := Value.NameFriendly;
    HelpContext := Value.HelpContext;
  end;
  CheckHelpButtonVisible;
end;


{
********************************* TLKWizPanel **********************************
}
procedure TLKWizPanel.CheckHelpButtonVisible;
begin
  FBtnHelp.Visible := (FHelpStr > '') or (HelpContext > -1);
end;

constructor TLKWizPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  BorderStyle := bsNone;
  Height := 32;
  Caption := ' ';
  LabelItemNo := TLabel.Create(AOwner);
  OnClick := DoButtonClick;
  with LabelItemNo do
  begin
    Parent := Self;
    Left := 8;
    Top := 8;
    Font.Color := clNavy;
    Font.Style := Font.Style + [fsBold];
  end;
  FBtnExecute := TButton.Create(AOwner);
  with FBtnExecute do
  begin
    Parent := Self;
    Left := 36;
    Top := 4;
    Height := 24;
    Font.Color := clNavy;
    Font.Style := Font.Style + [fsBold];
    OnClick := DoButtonClick;
  end;
  FBtnHelp := TButton.Create(AOwner);
  with FBtnHelp do
  begin
    Parent := Self;
    Left := 48;
    Top := 4;
    Anchors := [akTop] + [akRight];
    Caption := 'Help...';
    OnClick := DoBtnHelpClick;
  end;
end;

procedure TLKWizPanel.DoBtnHelpClick(Sender: TObject);
begin
  if (HelpStr = '') then
  begin
    if (HelpContext > -1) then
      Application.HelpCommand(HELP_CONTEXT, HelpContext);
  end
  else if (Owner is TFormLKHeaderDlg) then
  begin
    TFormLKHeaderDlg(Owner).LKHeaderPanel1.Comments.Text := HelpStr;
    TFormLKHeaderDlg(Owner).LKHeaderPanel1.HelpURI := HelpURI;
  end
  else ShowMessage(HelpStr);
end;

procedure TLKWizPanel.DoButtonClick(Sender: TObject);
begin
  if Assigned(FOnButtonClick) then
    FOnButtonClick(Sender);
end;

function TLKWizPanel.GetButtonCaption: string;
begin
  Result := FBtnExecute.Caption;
end;

procedure TLKWizPanel.Resize;
begin
  inherited Resize;
  FBtnExecute.Width := FBtnHelp.Left - 40;
end;


procedure TLKWizPanel.SetButtonCaption(const Value: string);
begin
  FBtnExecute.Caption := Value;
end;

procedure TLKWizPanel.SetHelpStr(const Value: string);
begin
  FHelpStr := Value;
  CheckHelpButtonVisible;
end;

{
****************************** TLKACSetupWizPanel ******************************
}
procedure TLKACSetupWizPanel.DoButtonClick(Sender: TObject);
begin
  inherited DoButtonClick(Sender);
  LKComponent.DlgSetupExecute;
end;

procedure TLKACSetupWizPanel.SetLKComponent(Value: TLKSystemComponent);
begin
  if Assigned(Value) then
  begin
    HelpContext := Value.HelpContextSetup;
  end;
  inherited SetLKComponent(Value);
end;


{
******************************** TLKFilterPanel ********************************
}
constructor TLKFilterPanel.Create(AOwner: TComponent);
begin
  inherited;
  Reloading := true;
  FFilterDataset := tbMaster;
  BevelOuter := bvNone;
  Height := 25;
  Width := 280;
  Caption := ' ';
  ShowCaption := False;
  ParentColor := true;
  CkbxFilter := TCheckbox.Create(Self);
  with CkbxFilter do
  begin
    Alignment := taRightJustify;
    Caption := 'Master filter';
    Height := 17;
    Width := 88;
    Top := 4;
    Left := 8;
    Parent := Self;
    OnClick := ckbxFilterClick;
  end;
  cbFilter := TComboBox.Create(Self);
  with cbFilter do
  begin
    Left := 104;
    Top := 2;
    Width := 149;
    Height := 21;
    ItemHeight := 13;
    Anchors := [akLeft, akTop, akRight];
    Text := '';
    OnChange := cbFilterChange;
    Parent := Self;
    OnExit := cbFilterExit;
  end;
  sbFilter := TSpeedButton.Create(Self);
  with sbFilter do
  begin
    Left := 254;
    Top := 0;
    Width := 25;
    Height := 25;
    Anchors := [akTop, akRight];
    Flat := True;
    try
      //Glyph.LoadFromResourceName(HInstance,'WIZARD');
      Glyph.LoadFromResourceName(MainInstance, 'WIZARD');
      //Glyph.LoadFromResourceID(Application.Handle,0);
    except
      //Temp fix for any problem with load resource
      Caption := '...';
    end;
    Hint := 'Filter construction wizard...';
    ShowHint := True;
    Parent := Self;
    OnClick := DlgTableFilterExecute;
  end;
  Reloading := false;
end;

procedure TLKFilterPanel.CbFilterChange(Sender: TObject);
begin
  if ckbxFilter.Checked then
    Filtered := false;
end;

procedure TLKFilterPanel.CbFilterExit(Sender: TObject);
begin
  Filter := cbFilter.Text;
end;

procedure TLKFilterPanel.CkbxFilterClick(Sender: TObject);
begin
  Filtered := ckbxFilter.Checked;
end;

procedure TLKFilterPanel.DlgTableFilterExecute(Sender: TObject);
begin
  if Assigned(FDataComponent) then
  begin
    DataComponent.DlgTableFilterExecute(FilterDataset);
    FilterReload;
  end;
end;

procedure TLKFilterPanel.FilterReload;
begin
  CbFilter.Items.Clear;
  if Assigned(FDataComponent) then
    with FDataComponent do
      case FFilterDataset of
        tbMaster:
          begin
            CbFilter.Items.Assign(Filters);
            CbFilter.Text := Filter;
            CkbxFilter.Checked := Filtered;
          end;
        tbDetail:
          begin
            CbFilter.Items.Assign(FiltersDetail);
            CbFilter.Text := TblDetail.Filter;
            CkbxFilter.Checked := TblDetail.Filtered;
          end;
      end
  else
  begin
    CbFilter.Text := '';
    CkbxFilter.Checked := False;
  end;
end;

function TLKFilterPanel.GetFilter: string;
begin
  if Assigned(FDataComponent) then
    with FDataComponent do
      case FFilterDataset of
        tbMaster: cbFilter.Text := Filter;
        tbDetail: cbFilter.Text := TblDetail.Filter;
      end
  else
    cbFilter.Text := '';
  Result := cbFilter.Text;
end;

function TLKFilterPanel.GetFiltered: Boolean;
begin
  if Assigned(FDataComponent) then
    with FDataComponent do
      case FFilterDataset of
        tbMaster: ckbxFilter.Checked := Filtered;
        tbDetail: ckbxFilter.Checked := TblDetail.Filtered;
      end
  else
    ckbxFilter.Checked := false;
  Result := ckbxFilter.Checked;
end;

procedure TLKFilterPanel.Notification(AComponent: TComponent; Operation:
    TOperation);
begin
  if (Operation = opRemove) then
  begin
    if (AComponent = FDataComponent) then
      FDataComponent := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TLKFilterPanel.SetDataComponent(const Value: TLKDataComponent);
begin
  FDataComponent := Value;
  //if csLoading in ComponentState then exit;
  if Assigned(Value) then
  begin
    Value.FreeNotification(Self);
    SetFilterDataset(FFilterDataset);
  end;
  FilterReload;
end;

procedure TLKFilterPanel.SetFilter(const Value: string);
begin
  cbFilter.Text := Value;
  if Assigned(DataComponent) then
    case FilterDataset of
      tbMaster: if (DataComponent.Filter <> Value) then
          DataComponent.Filter := Value;
      tbDetail: if (DataComponent.TblDetail.Filter <> Value) then
          DataComponent.TblDetail.Filter := Value;
    end;
end;

procedure TLKFilterPanel.SetFilterDataset(const Value: TLKACTable);
begin
  FFilterDataset := Value;
  CbFilter.Items.Clear;
  case FFilterDataset of
    tbMaster:
        CkbxFilter.Caption := 'Master filter';
    tbDetail:
        CkbxFilter.Caption := 'Detail filter';
  end;
  FilterReload;
end;

procedure TLKFilterPanel.SetFiltered(Value: Boolean);
begin
  ckbxFilter.Checked := Value;
  if Assigned(DataComponent) then
    case FilterDataset of
      tbMaster: if (DataComponent.Filtered <> Value) then
          DataComponent.Filtered := Value;
      tbDetail: if (DataComponent.TblDetail.Filtered <> Value) then
          DataComponent.TblDetail.Filtered := Value;
    end;
end;


{
***************************** TLKTableFilterDialog *****************************
}
constructor TLKTableFilterDialog.Create(AOwner: TComponent);
begin
  inherited create(AOwner);
  setVersion(' ');
  FFieldName := '';
end;

procedure TLKTableFilterDialog.doOnClose;
begin
  if Assigned(FOnClose) then
    FOnClose(self);
end;

procedure TLKTableFilterDialog.doOnShow;
begin
  if Assigned(FOnShow) then
    FOnShow(self);
end;

function TLKTableFilterDialog.Execute: Boolean;
var
  newFilter, thisFieldType: string;
  n: Integer;
begin
  Result := false;
  FormLKTableFilter := TFormLKTableFilter.create(TForm(Owner));
  try
    doOnShow;
    with FormLKTableFilter do
    begin
      with cbxFieldName do
      begin
        items.clear;
        if Assigned(myTable) then
          with myTable do
            for n := 0 to (FieldDefs.count - 1) do
            begin
              case FieldDefs.items[n].DataType of
                ftString: thisFieldType := 'Text';
                ftBoolean: thisFieldType := 'TRUE or FALSE';
                ftSmallint, ftInteger, ftWord, ftFloat, ftCurrency, ftBCD,
                  ftAutoInc: thisFieldType := 'Number';
                ftDate, ftTime, ftDateTime: thisFieldType := 'Date';
              else
                thisFieldType := 'unknown';
              end;
              if (thisFieldType <> 'unknown') then
              begin
                items.add(FieldDefs.items[n].name);
                stringFields.add(thisFieldType);
                if (items[n] = FFieldName) then
                  text := FFieldName;
              end;
            end;
        if (text = '') and (Items.Count > 0) then
          text := items[0];
      end;
      caption := 'Set table filter ' + myTable.Name;
      if (Showmodal = mrOK) then
      begin
        newFilter := '';
        for n := 0 to mLink.lines.Count - 1 do
          newFilter := NewFilter + ' ' + mLink.lines[n]
            + ' ' + mFieldName.lines[n] + ' ' + mTest.lines[n] + ' ' +
              mData.lines[n];
        case ACTable of
          tbMaster: DataComponent.Filter := newFilter;
          tbDetail: DataComponent.TblDetail.Filter := newFilter;
        end;
      end;
      doOnClose;
    end;
  finally FormLKTableFilter.free;
  end;
end;

procedure TLKTableFilterDialog.SetACTable(const AACTable: TLKACTable);
begin
  FACTable := AACTable;
  if Assigned(FDataComponent) then
    case FACTable of
      tbMaster: myTable := FDataComponent.TblMaster;
      tbDetail: myTable := FDataComponent.TblDetail;
    end;
end;

procedure TLKTableFilterDialog.SetDataComponent(const ADataComponent:
    TLKDataComponent);
begin
  FDataComponent := ADataComponent;
  FTable := nil;
  if Assigned(FDataComponent) then
    case ACTable of
      tbMaster: myTable := FDataComponent.TblMaster;
      tbDetail: myTable := FDataComponent.TblDetail;
    end;
end;

procedure TLKTableFilterDialog.SetTable(const ATable: TDataset);
begin
  FTable := ATable;
  myTable := FTable;
  FDataComponent := nil;
end;

procedure TLKTableFilterDialog.SetVersion(const Value: string);
begin
  FVersion := 'v.1.03';
end;


{
******************************* TJvSearchListBox *******************************
}
constructor TJvSearchListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSearchItems := TStringList.Create;
  FSearchText := '';
end;

destructor TJvSearchListBox.Destroy;
begin
  FSearchItems.Free;
  inherited Destroy;
end;

procedure TJvSearchListBox.SetSearchItems(Value: TStringList);
begin
  FSearchItems.Assign(Value);
end;

procedure TJvSearchListBox.SetSearchText(const Value: string);
var
  i: Integer;
begin
  FSearchText := Value;
  with Items do
  begin
    Clear;
    for i := 0 to FSearchItems.Count - 1 do
      if ((Value = '') or (Pos(Value, FSearchItems[i]) = 1)) then
        AddItem(FSearchItems[i], FSearchItems.Objects[i]);
    Height := ItemHeight * Count + 2;
  end;
end;

end.

