{*****************************************************************************}
{                                                                             }
{   This file is part of the LKRecords components Library                     }
{                                                                             }
{   See the file COPYING.LGPL.txt included in this distribution,              }
{   for details about the Licence and Copyright.                              }
{                                                                             }
{   This code is distributed in the hope that it will be useful,              }
{   but WITHOUT ANY WARRANTY; without even the implied warranty of            }
{   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      }
{                                                                             }
{   Author: Leslie John Kaye :    www.leskaye.com                             }
{                                                                             }
{   Abstract:                                                                 }
{      TLKRecords components property editors                                 }
{                                                                             }
{*****************************************************************************}

unit LKRecordsPropEdits;

interface

uses
{$IFNDEF LINUX}
  Windows, VCLEditors, Controls,
    ImgList, Graphics,
  ShellAPI,
{$ELSE}
  ShellAPI, QImgList, QGraphics,
{$ENDIF}
  Classes, DB, DesignIntf, DesignEditors,
  System.UITypes, Dialogs, SysUtils, DBClient, ActnList, WideStrings,
  LKUtils, LKRecordsComponents;

type
  TLKBaseComponentEditor = class(TComponentEditor)
  protected
    procedure MoveFieldsFromDataModule(AComponent: TComponent; ADataComponent:
        TLKDataComponent);
  end;

  TLKDataComponentEditor = class(TLKBaseComponentEditor)
  public
    constructor Create(AComponent: TComponent; ADesigner: IDesigner); override;
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure Register;

implementation

uses LKBaseComponent, LKImageIndexEditor;

type
  TLKComponentImageIndexProperty = class(TIntegerProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

{
************************ TLKComponentImageIndexProperty ************************
}
procedure TLKComponentImageIndexProperty.Edit;
var
  Item: TLKComponent;
  Images: TCustomImageList;
begin
  //Get Action of ImageIndex
  Item := TLKComponent(GetComponent(0));
  //Get ImageList of TActionlist of this TAction
  Images := Item.Images;
  if Images = nil then
    exit;
  with TDlgLKImageIndexEditor.Create(nil) do
  try
    ImageList1.AddImages(Images);
    FormCaptionName := Item.NameFriendly;
    ImageIndex := GetOrdValue;
    if ShowModal = mrOk then
      SetOrdValue(ImageIndex);
  finally Free;
  end;
end;

function TLKComponentImageIndexProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paMultiSelect]
end;

type
  TLKCloneProperty = class(TComponentProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{
******************************* TLKCloneProperty *******************************
}
function TLKCloneProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paMultiSelect];
end;

procedure TLKCloneProperty.GetValueList(List: TStrings);
var
  I: Integer;
  Comp: TComponent;
begin
  with TComponent(getComponent(0)).Owner.Owner do
    for I := 0 to ComponentCount - 1 do
    begin
      Comp := Components[I];
      if (Comp is TLKDataComponent) then
        with TLKDataComponent(Comp) do
        begin
          if (TblMaster.TableName > '') then
            list.Append(Name + '.' + TblMaster.name);
          if (TblDetail.TableName > '') then
            list.Append(Name + '.' + TblDetail.name);
        end;
    end;
end;

procedure TLKCloneProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do
      Proc(Values[I]);
  finally
    Values.Free;
  end;
end;

type
  TLKImageIndexProperty = class(TIntegerProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{
**************************** TLKImageIndexProperty *****************************
}
function TLKImageIndexProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paMultiSelect];
end;

procedure TLKImageIndexProperty.GetValueList(List: TStrings);
var
  i: Integer;

  //  b : TBitMap;

begin
  with TLKComponent(GetComponent(0)) do
    if assigned(Images) then
      with Images do
        for i := 0 to Count - 1 do
        begin
          List.Add(IntToStr(i));
          {
          b := TBitMap.Create;
          Images.GetBitMap(i,b);
          List.AddObject(IntToStr(i), b);
          //b.Free;
          }
        end
    else
      showmessage('No ImageList set');
end;

procedure TLKImageIndexProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do
      Proc(Values[I]);
  finally
    Values.Free;
  end;
end;

{ TStringsSelectProperty }
type
  TStringsSelectProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{
****************************** TMDBStringProperty ******************************
}
function TStringsSelectProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TStringsSelectProperty.GetValueList(List: TStrings);
begin
end;

procedure TStringsSelectProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do
      Proc(Values[I]);
  finally
    Values.Free;
  end;
end;


{ TWideStringsSelectProperty }
type
  TWideStringsSelectProperty = class(TWideStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TWideStrings); virtual;
    procedure GetValues(Proc: TGetWideStrProc); override;
  end;

{
****************************** TWideStringsSelectProperty ******************************
}
function TWideStringsSelectProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TWideStringsSelectProperty.GetValueList(List: TWideStrings);
begin
end;

procedure TWideStringsSelectProperty.GetValues(Proc: TGetWideStrProc);
var
  I: Integer;
  Values: TWideStringList;
begin
  Values := TWideStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do
      Proc(Values[I]);
  finally
    Values.Free;
  end;
end;



type
  TDBSortedFieldProperty = class(TStringsSelectProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

{
**************************** TDBSortedFieldProperty ****************************
}
procedure TDBSortedFieldProperty.GetValueList(List: TStrings);
var
  i: Integer;
begin
  with TDataSet(GetComponent(0)) do
    for i := 0 to FieldCount - 1 do
      if (Fields[i].Owner = Owner)
        and (Fields[i].FieldName <> '') then
        List.Add(Fields[i].FieldName);
end;

//*****************************************

type
  TIdxNameProperty = class(TWideStringsSelectProperty)
  public
    procedure GetValueList(List: TWideStrings); override;
  end;

{
******************************* TIdxNameProperty *******************************
}
procedure TIdxNameProperty.GetValueList(List: TWideStrings);
var
  SL: TStringList;
begin
  List.Clear;
  SL := TStringList.Create;
  try
    if GetComponent(0) is TLKDataComponent then
      TLKDataComponent(GetComponent(0)).TblMaster.GetIndexNames(SL)
    else if GetComponent(0) is TLKTable then
      TLKTable(GetComponent(0)).GetIndexNames(SL);
    List.AddStrings(SL);
  finally
    SL.Free;
  end;
end;

//*****************************************

type
  TSubClassProperty = class(TClassProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
  end;

{
****************************** TSubClassProperty *******************************
}
function TSubClassProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paSubProperties];
end;

type
  TDBStringProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{
****************************** TDBStringProperty *******************************
}
function TDBStringProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TDBStringProperty.GetValueList(List: TStrings);
begin
  //
end;

procedure TDBStringProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for I := 0 to Values.Count - 1 do
      Proc(Values[I]);
  finally
    Values.Free;
  end;
end;

type
  TLKGroupProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
    procedure SetValue(const Value: string); override;
  end;

{
******************************* TLKGroupProperty *******************************
}
procedure TLKGroupProperty.GetValueList(List: TStrings);
var
  i: Integer;
  Comp: TPersistent;
begin
  for i := 0 to propcount - 1 do
  begin
    Comp := getComponent(i);
    if Comp is TLKGroup then
      list.Append(TComponent(Comp).name);
  end;
end;

procedure TLKGroupProperty.SetValue(const Value: string);
var
  i: Integer;
  Comp: TPersistent;
begin
  for i := 0 to propcount - 1 do
  begin
    Comp := getComponent(i);
    if (TComponent(Comp).name = value) then
    begin
      TLKDataComponent(GetComponent(0)).ACGroup := TLKGroup(Comp);
      designer.modified;
      break;
    end;
  end;
end;

type
  TDBFileNameProperty = class(TStringsSelectProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

{
***************************** TDBFileNameProperty ******************************
}
procedure TDBFileNameProperty.GetValueList(List: TStrings);
begin
//  if GetComponent(0) is TLKTable then
    TLKTable(GetComponent(0)).SQLConnection.GetTableNames(List)
//  else if GetComponent(0) is TLKDataComponent then
//    TLKDataComponent(GetComponent(0)).SQLConnection.GetTableNames(List);
end;

{
**************************** TLKBaseComponentEditor ****************************
}
procedure TLKBaseComponentEditor.MoveFieldsFromDataModule(AComponent:
    TComponent; ADataComponent: TLKDataComponent);
var
  i: Integer;
  f: TField;
begin
  {
    move the fields from the datamodule (root) into the BaseComponent so that the
    TClientDatasetComponentEditor can find them
    this procedure cannot be reversed in the component editor because the editor is
    destroyed before edit is finished therefore it is done in CB.getChildren method
  }
  with ADataComponent do
  begin
    for i := 0 to TblMaster.FieldCount - 1 do
    begin
      f := TField(TblMaster.Fields[i]);
      if (f = TField(AComponent.Owner.FindComponent(f.name))) then
      begin
        AComponent.Owner.RemoveComponent(f);
        InsertComponent(f);
      end;
    end;
    // do the same for the DetailTable
    for i := 0 to TblDetail.FieldCount - 1 do
    begin
      f := TField(TblDetail.Fields[i]);
      if (f = TField(AComponent.Owner.FindComponent(f.name))) then
      begin
        AComponent.Owner.RemoveComponent(f);
        InsertComponent(f);
      end;
    end;
  end;
end;

type
  TLKHubRecordsEditor = class(TLKBaseComponentEditor)
  public
    constructor Create(AComponent: TComponent; ADesigner: IDesigner); override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

{
***************************** TLKHubAccountsEditor *****************************
}
constructor TLKHubRecordsEditor.Create(AComponent: TComponent; ADesigner:
    IDesigner);
begin
  inherited;
  MoveFieldsFromDataModule(AComponent, TLKHubRecords(AComponent).Diary);
end;

procedure TLKHubRecordsEditor.ExecuteVerb(Index: Integer);
var
  MTable, DTable: TLKTable;
  DiaryMTableEditor, DiaryDTableEditor: IComponentEditor;
  SystemSettingsMTableEditor: IComponentEditor;
begin
  inherited;
  MTable := TLKHubRecords(Component).Diary.TblMaster;
  DTable := TLKHubRecords(Component).Diary.TblDetail;
  DiaryMTableEditor := getComponentEditor(MTable, designer);
  DiaryDTableEditor := getComponentEditor(DTable, designer);
  MTable := TLKHubRecords(Component).SystemSettings.TblMaster;
  SystemSettingsMTableEditor := getComponentEditor(MTable, designer);
  case index of
    0: DiaryMTableEditor.Edit;
    1: DiaryDTableEditor.Edit;
    2: SystemSettingsMTableEditor.Edit;
  end;
end;

function TLKHubRecordsEditor.GetVerb(Index: Integer): string;
begin
  case index of
    0: result := 'Editor: Diary Master...';
    1: result := 'Editor: Diary Detail...';
    2: result := 'Editor: System settings Master...';
  end;
end;

function TLKHubRecordsEditor.GetVerbCount: Integer;
begin
  result := 3;
end;

{
**************************** TLKDataComponentEditor ****************************
}
constructor TLKDataComponentEditor.Create(AComponent: TComponent; ADesigner:
    IDesigner);
begin
  inherited;
  MoveFieldsFromDataModule(AComponent, TLKDataComponent(AComponent));
end;

procedure TLKDataComponentEditor.Edit;
begin
      //when component Double-clicked at designtime
  ExecuteVerb(2);
  ExecuteVerb(1);
  ExecuteVerb(0);
end;

procedure TLKDataComponentEditor.ExecuteVerb(Index: Integer);
var
  MTable, DTable: TCustomClientDataSet;
  MActionList: TActionList;
  MActionEditor, MTableEditor, DTableEditor: IComponentEditor;
begin
  MTable := TLKDataComponent(Component).TblMaster;
  DTable := TLKDataComponent(Component).TblDetail;
  MActionList := TLKDataComponent(Component).ActionList;
  MTableEditor := getComponentEditor(MTable, designer);
  DTableEditor := getComponentEditor(DTable, designer);
  MActionEditor := getComponentEditor(MActionList, designer);
  case index of
    0: MActionEditor.Edit;
    1: MTableEditor.Edit;
    2: DTableEditor.Edit;
  end;
end;

function TLKDataComponentEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := '&Actionlist editor...';
    1: Result := 'Editor: Table&Master...';
    2: Result := 'Editor: Table&Detail...';
  end;
end;

function TLKDataComponentEditor.GetVerbCount: Integer;
begin
  result := 3;
end;

type
  TLKGroupEditor = class(TComponentEditor)
  public
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

{
******************************** TLKGroupEditor ********************************
}
procedure TLKGroupEditor.Edit;
begin
  ExecuteVerb(0);
end;

procedure TLKGroupEditor.ExecuteVerb(Index: Integer);
begin
  TLKGroup(Component).GroupComponents.DlgChangeListOrderExecute;
end;

function TLKGroupEditor.GetVerb(Index: Integer): string;
begin
  case index of
    0: Result := '&Group editor...';
  end;
end;

function TLKGroupEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(System.UITypes.TImageIndex), TLKComponent, 'ImageIndex',
    TLKComponentImageIndexProperty);
  RegisterPropertyEditor(TypeInfo(TLKTable), TLKTable,
    'Clone', TLKCloneProperty);
  RegisterPropertyEditor(TypeInfo(WideString), TLKTable,
    'TableName', TDBFileNameProperty);
  RegisterPropertyEditor(TypeInfo(WideString), TLKTable, 'IndexName',
    TIdxNameProperty);
  RegisterPropertyEditor(TypeInfo(WideString), TLKDataComponent, 'IndexName',
    TIdxNameProperty);
  // RegisterPropertyEditor(TypeInfo(string), TLKDaybookNominalControl, 'LedgerIndexName', TLedgerIdxNameProperty);
  RegisterComponentEditor(TLKDataComponent, TLKDataComponentEditor);
  RegisterComponentEditor(TLKGroup, TLKGroupEditor);
  RegisterComponentEditor(TLKHubRecords, TLKHubRecordsEditor);
  RegisterPropertyEditor(TypeInfo(string), TLKDataComponent, 'ACGroup',
    TLKGroupProperty);
  //RegisterPropertyEditor(TypeInfo(TEnhTwain), TDocumentsImages, 'MyTwain', TSubClassProperty);
//  RegisterPropertyEditor(TypeInfo(TMemo), TLKACGroupBox, 'Memo', TSubClassProperty);
  RegisterPropertyEditor(TypeInfo(TActionList), TLKDataComponent, 'ActionList',
    TSubClassProperty);
end;

end.

