unit LKTableBase;

interface

uses
  Classes, SysUtils, SQLConst, DB, DBClient, SqlExpr, Provider,
  Variants, LKDefine;

type

  TInternalSQLDataSet = class(TCustomSQLDataSet)
  published
    property CommandText;
    property CommandType;
    property DataSource;
    property GetMetadata default false;
    property MaxBlobSize default 0;
    property ParamCheck;
    property Params;
    property SortFieldNames;
  end;

  TLKCustomClientDataset = class(TCustomClientDataSet)
  private
    FIsFirstOpen: Boolean;
    function GetCurrentIndexFieldNames: string;
    function GetCurrentIndexFieldValues: Variant;
  protected
    procedure AssignTo(Dest: TPersistent); override;
    procedure SetActive(Value: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure EmptyTable; virtual;
    function GetFieldValues(FieldNames: string): Variant;
    function HasValidData: Boolean;
    property Active;
    property AfterApplyUpdates;
    property AfterCancel;
    property AfterClose;
    property AfterDelete;
    property AfterEdit;
    property AfterExecute;
    property AfterGetParams;
    property AfterGetRecords;
    property AfterInsert;
    property AfterOpen;
    property AfterPost;
    property AfterRefresh;
    property AfterRowRequest;
    property AfterScroll;
    property Aggregates;
    property AggregatesActive;
    property AutoCalcFields;
    property BeforeApplyUpdates;
    property BeforeCancel;
    property BeforeClose;
    property BeforeDelete;
    property BeforeEdit;
    property BeforeExecute;
    property BeforeGetParams;
    property BeforeGetRecords;
    property BeforeInsert;
    property BeforeOpen;
    property BeforePost;
    property BeforeRefresh;
    property BeforeRowRequest;
    property BeforeScroll;
    property CommandText;
    property ConnectionBroker;
    property Constraints;
    property CurrentIndexFieldNames: string read GetCurrentIndexFieldNames;
    property CurrentIndexFieldValues: Variant read GetCurrentIndexFieldValues;
    property DataSetField;
    property DisableStringTrim;
    property FetchOnDemand;
    property FieldDefs;
    property FileName;
    property Filter;
    property Filtered;
    property FilterOptions;
    property IndexDefs;
    property IndexFieldNames;
    property IndexName;
    property MasterFields;
    property MasterSource;
    property ObjectView;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property OnReconcileError;
    property PacketRecords;
    property Params;
    property ProviderName;
    property ReadOnly;
    property RemoteServer;
    property StoreDefs;
  end;

  TLKTableBase = class(TLKCustomClientDataset)
  private
    FClone: TLKTableBase;
    FCursorCloned: Boolean;
    FInternalDataSet: TInternalSQLDataSet;
    FInternalProvider: TDataSetProvider;
    FOldFilter: string;
    FTableName: Widestring;
    FWasFiltered: Boolean;
    FWhereClause: string;
    function GetSQLConnection: TSQLConnection;
    procedure SetClone(Value: TLKTableBase);
    procedure SetSQLConnection(Value: TSQLConnection);
    procedure SetTableName(const Value: Widestring);
    procedure SetWhereClause(const Value: string);
  protected
    procedure AssignTo(Dest: TPersistent); override;
    function GetTableName: Widestring; virtual;
    procedure SetActive(Value: Boolean); override;
    procedure SetName(const NewName: TComponentName); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateTable;
    procedure DeleteTable;
    procedure EmptyTable; override;
    procedure RestoreFilter;
    procedure SaveFilter;
    property InternalDataSet: TInternalSQLDataSet read FInternalDataSet;
    property InternalProvider: TDataSetProvider read FInternalProvider;
    property SQLConnection: TSQLConnection read GetSQLConnection write
        SetSQLConnection;
  published
    property Active;
    property Aggregates;
    property AggregatesActive;
    property AutoCalcFields;
    property Clone: TLKTableBase read FClone write SetClone;
    property Constraints;
    property FieldDefs;
    property FileName;
    property FilterOptions;
    property Params;
    property ReadOnly;
    property StoreDefs;
    property TableName: Widestring read GetTableName write SetTableName;
    property WhereClause: string read FWhereClause write SetWhereClause;
  end;


implementation

uses LKUtils;

{
**************************** TLKCustomClientDataset ****************************
}
constructor TLKCustomClientDataset.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIsFirstOpen := True;
end;

procedure TLKCustomClientDataset.AssignTo(Dest: TPersistent);
var
  D: TDataSet;
  C: TLKCustomClientDataSet;
begin
  if not (Dest is TDataset) then
    Exit;
  D := TDataset(Dest);
  D.FieldDefs.Assign(FieldDefs);
  D.Filter := Filter;
  D.OnFilterRecord := OnFilterRecord;
  D.FilterOptions := FilterOptions;
  //D.MasterSource := MasterSource;
  if Dest is TLKCustomClientDataSet then
  begin
    C := TLKCustomClientDataSet(Dest);
    C.IndexDefs.Assign(IndexDefs);
    if (IndexName > '') then
      C.IndexName := IndexName
    else
      C.IndexFieldNames := IndexFieldNames;
    C.MasterFields := MasterFields;
  //  C.RemoteServer := RemoteServer;
  //  C.ProviderName := Source.ProviderName;
  //  if HasAppServer then
  //    C.AppServer := AppServer;
    C.ReadOnly := ReadOnly;
  end;
end;

procedure TLKCustomClientDataset.EmptyTable;
begin
  EmptyDataset;
  ApplyUpdates(-1);
end;

function TLKCustomClientDataset.GetCurrentIndexFieldNames: string;
var
  s: string;
  i: Integer;
begin
  s := IndexName;
  if (s > '') then
    with IndexDefs do
    begin
      Result := Items[0].Fields;
      for i := 0 to count - 1 do
        if (Items[i].Name = s) then
        begin
          Result := Items[i].Fields;
          break;
        end;
    end
  else
    Result := IndexFieldNames;
end;

function TLKCustomClientDataset.GetCurrentIndexFieldValues: Variant;
var
  myIndexFieldNames: TStringList;
  i: Integer;
begin
  Result := null;
  if not Active then
    exit;
  myIndexFieldNames := TStringList.Create;
  try
    DelimitedStringToList(CurrentIndexFieldNames, ';', myIndexFieldNames);
    with myIndexFieldNames do
      case Count of
        0: ;
        1: Result := FieldByName(Strings[0]).Value;
      else
        begin
          Result := VarArrayCreate([0, Count - 1], varVariant);
          for i := 0 to Count - 1 do
            Result[i] := FieldByName(Strings[i]).Value;
        end;
      end;
  finally myIndexFieldNames.Free;
  end;
end;

function TLKCustomClientDataset.GetFieldValues(FieldNames: string): Variant;
var
  I, x: Integer;
  myFieldList: TStringList;
begin
  myFieldList := TStringList.Create;
  try
    DelimitedStringToList(FieldNames, ';', myFieldList);
    x := myFieldList.Count - 1;
    case x of
      -1: Result := null;
      0: Result := FieldByName(myFieldList[0]).asString;
    else
      begin
        Result := VarArrayCreate([0, x], varVariant);
        for i := 0 to x do
          Result[i] := FieldByName(myFieldList[i]).value;
      end;
    end;
  finally myFieldList.Free;
  end;
end;

function TLKCustomClientDataset.HasValidData: Boolean;
var
  bm: TBookMark;
begin
  bm := GetBookMark;
  try
    Result := Assigned(bm);
    if Result then
      Result := BookMarkValid(bm);
  finally FreeBookMark(bm);
  end;
end;

procedure TLKCustomClientDataset.SetActive(Value: Boolean);
var
  I: Integer;
  S1, S2: string;
begin
  if (Value = Active) then
    Exit;
  inherited SetActive(Value);
  if Active and FIsFirstOpen then
  begin
    FIsFirstOpen := False;
    for I := 0 to FieldCount - 1 do
    begin
      with Fields[i] do
        if (DisplayName = FieldName) then
        begin
          S1 := Trim(StringReplace(DisplayName, '_', ' ', [rfReplaceAll]));
          S2 := Lowercase(Copy(S1, 2, Length(S1) - 1));
          DisplayLabel := S1[1] + S2;
        end;
      if Fields[I] is TFMTBCDField then
        with TFMTBCDField(Fields[I]) do
        begin
          Currency := (Size = 4);
          DisplayWidth := 12;
          Alignment := taRightJustify;
        end;
    end;
  end;
end;

{
********************************* TLKTableBase *********************************
}
constructor TLKTableBase.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FClone := nil;
  FCursorCloned := False;
  FInternalDataSet := TInternalSQLDataSet.Create(AOwner);
  FInternalDataSet.SetSubComponent(True);
  FInternalDataSet.CommandType := ctQuery;
  FInternalDataSet.GetMetadata := True;
  FInternalDataSet.MaxBlobSize := -1;
  FInternalDataSet.ParamCheck := True;
  FInternalProvider := TDataSetProvider.Create(AOwner);
  FInternalProvider.SetSubComponent(True);
  FInternalProvider.Exported := True;
  FInternalProvider.DataSet := FInternalDataSet;
  FInternalProvider.Options := [poUseQuoteChar];
  FInternalProvider.UpdateMode := upWhereAll;
  SetProvider(FInternalProvider);
  ObjectView := True;
  PacketRecords := -1;
end;

procedure TLKTableBase.AssignTo(Dest: TPersistent);
var
  D: TLKTableBase;
begin
  inherited AssignTo(Dest);
  if not (Dest is TLKTableBase) then
    Exit;
  D := TLKTableBase(Dest);
  D.TableName := TableName;
  D.WhereClause := WhereClause;
end;

procedure TLKTableBase.CreateTable;
var
  S: string;
  Fn: string;
  IndexFieldList: TStringList;
  I: Integer;
begin
  IndexFieldList := TStringList.Create;
  try
    if (IndexDefs.Count > 0) then
    begin
      S := IndexDefs.Items[0].Fields;
      repeat
        I := pos(';', S); //find out if more than one link field
        if (I > 0) then
        begin
          Fn := copy(S, 1, I - 1);
          System.delete(S, 1, I);
        end
        else
          Fn := S;
        IndexFieldList.Add(Fn);
      until (I = 0);
    end;
    S := 'CREATE TABLE `' + TableName + '` ( ';
    with FieldDefs do
      for I := 0 to count - 1 do
        with Items[I] do
        begin
          S := S + '`' + Name + '`' + ' ';
          case DataType of
            ftString, ftFixedChar, ftWideString: S := S + 'char(' +
              intToStr(Size) + ') ';
            ftFmtMemo, //  Formatted text memo field
            ftMemo: S := S + 'varchar'; //Text memo field
            ftWord, //  : S := S + 16-bit unsigned integer field
            ftSmallint: S := S + 'smallint(6)'; //16-bit integer field
            ftInteger: S := S + 'int(11)'; //32-bit integer field
            ftLargeint: S := S + 'bigint(16)'; //  Large integer field
            ftAutoInc: S := S + 'int(11)';
              //  Auto-incrementing 32-bit integer counter field
            ftBoolean: S := S + 'enum(''Y'',''N'')'; //  Boolean field
            ftFloat: S := S + 'float '; //Floating-point numeric field
            ftCurrency: S := S + 'double'; //  Money field
            ftBCD, //  Binary-Coded Decimal field that can be converted to Currency type without a loss of precision.
            ftFMTBcd: S := S + 'tinyblob';
              //Binary-Coded Decimal field that is too large for ftBCD.
            ftDateTime: S := S + 'datetime'; //    Date and time field
            ftDate: S := S + 'date'; //  Date field
            ftTime: S := S + 'time'; //  Time field
            ftTimeStamp: S := S + 'timestamp';
              //  Date and time field accessed through dbExpress
            ftBytes, //  Fixed number of bytes (binary storage)
            ftVarBytes, //  Variable number of bytes (binary storage)
            ftGraphic, //  Bitmap field
            ftParadoxOle, //  Paradox OLE field
            ftDBaseOle, //  dBASE OLE field
            ftTypedBinary, //  Typed binary field
            ftOraBlob, //  BLOB fields in Oracle 8 tables
            ftOraClob, //  CLOB fields in Oracle 8 tables
            ftVariant, //  Data of unknown or undetermined type
            ftInterface, //  References to interfaces (IUnknown)
            ftIDispatch, //  References to IDispatch interfaces
            ftGuid, //  globally unique identifier (GUID) values
            ftArray: S := S + 'blob'; //   Array field
            //ftCursor  Output cursor from an Oracle stored procedure (TParam only)
            //ftADT  Abstract Data Type field
            //ftReference  REF field
            ftBlob, //  Binary Large OBject field
            ftDataSet: S := S + 'longblob'; //DataSet field
          end; //case
          if (IndexFieldList.IndexOf(Name) > -1) then
            case DataType of
              ftString, ftFixedChar, ftWideString: S := S + ' NOT NULL default '
                + chr(39) + chr(39);
            else
              S := S + ' NOT NULL default ' + chr(39) + '0' + chr(39);
            end
          else
            S := S + ' default NULL';
          if (I < count - 1) then
            S := S + ',';
        end;
    if (IndexDefs.Count > 0) then
    begin
      S := S + ', PRIMARY KEY (';
      with IndexFieldList do
      begin
        S := S + '`' + Strings[0] + '`';
        for I := 1 to count - 1 do
          S := S + ',' + '`' + Strings[0] + '`';
      end;
      S := S + ')';
    end;
    S := S + ' ) TYPE=MyISAM';
    try
      SQLConnection.ExecuteDirect(S);
    except;
    end;
  finally IndexFieldList.Free;
  end;
end;

procedure TLKTableBase.DeleteTable;
begin
  Close;
  with SQLConnection do
  begin
    ExecuteDirect('DROP TABLE ' + TableName);
    ExecuteDirect('COMMIT');
  end;
end;

procedure TLKTableBase.EmptyTable;
begin
  if Assigned(SQLConnection) then
  begin
    with SQLConnection do
    begin
      ExecuteDirect('DELETE * FROM ' + TableName);
      ExecuteDirect('COMMIT');
    end;
    if Active then
      Refresh;
  end
  else
  inherited EmptyTable;
end;

function TLKTableBase.GetSQLConnection: TSQLConnection;
begin
  Result := FInternalDataSet.SQLConnection;
end;

function TLKTableBase.GetTableName: Widestring;
begin
  Result := FTableName;
end;

procedure TLKTableBase.RestoreFilter;
begin
  Filter := FOldFilter;
  Filtered := FWasFiltered;
end;

procedure TLKTableBase.SaveFilter;
begin
  FOldFilter := Filter;
  FWasFiltered := Filtered;
end;

procedure TLKTableBase.SetActive(Value: Boolean);
begin
  if csLoading in ComponentState then
    Exit;
  if (Value = Active) then
    Exit;
  if Value and Assigned(FClone) then
  begin
    if not FCursorCloned then
    begin
      // set properties before cloning the cursor
      FInternalProvider.DataSet := Clone.FInternalDataSet;

      MasterFields := FClone.MasterFields;
      //  MasterSource := Source.MasterSource;
      if (FClone.IndexName > '') then
        IndexName := FClone.IndexName
      else
        IndexFieldNames := FClone.IndexFieldNames;
      RemoteServer := FClone.RemoteServer;
      ProviderName := FClone.ProviderName;
      if FClone.HasAppServer then
        AppServer := FClone.AppServer;
      ReadOnly := FClone.ReadOnly;
      // All below settings not overwritten
      {
      Filter := FClone.Filter;
      OnFilterRecord := FClone.OnFilterRecord;
      FilterOptions := FClone.FilterOptions;
      MasterSource := Source.MasterSource;
      MasterFields := FClone.MasterFields;
      }
//      CloneCursor(FClone, False, True);    //crash
      FCursorCloned := True;
    end;
  end;
  inherited SetActive(Value);
end;

procedure TLKTableBase.SetClone(Value: TLKTableBase);
begin
  FClone := Value;
  if Assigned(Value) then
  begin
    TableName := Value.TableName;
    SetProvider(nil);
    FInternalProvider.DataSet := nil;
    // Value.FreeNotification(Self); done in LKComponent
    if (Value.IndexName > '') then
    begin
      IndexDefs.Assign(Value.IndexDefs);
      IndexName := Value.IndexName
    end else
      IndexFieldNames := Value.IndexFieldNames;
    RemoteServer := Value.RemoteServer;
    FInternalProvider.DataSet := Value.InternalDataSet;
    if Value.HasAppServer then
      AppServer := Value.AppServer;
    ReadOnly := Value.ReadOnly;
  end
  else
    SetProvider(FInternalProvider);
//    FInternalProvider.DataSet := FInternalDataSet;
end;

procedure TLKTableBase.SetName(const NewName: TComponentName);
begin
  inherited SetName(NewName);
  FInternalDataSet.Name := Name + 'InternalDataSet';
  FInternalProvider.Name := Name + 'InternalProvider';
  if Assigned(SQLConnection) then
    ProviderName := FInternalProvider.Name
  else
    ProviderName := '';
end;

procedure TLKTableBase.SetSQLConnection(Value: TSQLConnection);
begin
  FInternalDataSet.SQLConnection := Value;
  if Assigned(Value) then
    ProviderName := FInternalProvider.Name
  else
    ProviderName := '';
end;

procedure TLKTableBase.SetTableName(const Value: Widestring);
var
  S: string;
begin
  FTableName := Value;
  //InternalDataSet.TableName := Value;
  if (Value = '') then
    InternalDataset.CommandText := ''
  else
  begin
    S := Value;
    {
    if Assigned(AccountsComponent) and (S > '') then
     S := SQLObjectDelimited(S,AccountsComponent.DatabaseType);
     }
    if (WhereClause > ' ') then
      S := S + ' WHERE ' + WhereClause;
    InternalDataset.CommandText := 'SELECT * FROM ' + S;
  end;
end;

procedure TLKTableBase.SetWhereClause(const Value: string);
var
  I: Integer;
begin
  FWhereClause := Trim(Value);
  I := pos('WHERE ', Uppercase(FWhereClause));
  if (I > 0) then
    System.Delete(FWhereClause, I, 6);
  SetTableName(TableName);
end;



end.
