{*****************************************************************************}
{                                                                             }
{   This file is part of the LKAccounts components Library                    }
{                                                                             }
{   See the file COPYING.LGPL.txt included in this distribution,              }
{   for details about the Licence and Copyright.                              }
{                                                                             }
{   This code is distributed in the hope that it will be useful,              }
{   but WITHOUT ANY WARRANTY; without even the implied warranty of            }
{   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      }
{                                                                             }
{   Author: Leslie John Kaye :    www.leskaye.com                             }
{                                                                             }
{   Abstract:                                                                 }
{      Components to provide core ledger accounting together with daybooks    }
{      and data entry documents and records                                   }
{                                                                             }
{*****************************************************************************}

unit LKAccountsComponents;

{:
Prime entry for Nominal and Cost Ledgers
}
{:
A 'Journal' is an alternative name for 'Daybook' however has
in common usage become to mean a record of transfers between
ledger accounts.
}
// Importer Classes

{:
A Daybook for a ledger which is an analysis of a Nominal Ledger
Control Account. Eg. a Sales Daybook records invoices entered
into a Sales Ledger. If all of the accounts in the Sales Ledger
are added together then the total equals the balance shown
on the Sales Ledger Control Account of the Nominal Ledger.
}

{
Contact Ledgers will normally retain records from earlier
financial years, perhaps back to when a customer/supplier etc. first opened their account.
These therefore do not have an 'Opening balance' so that a Contacts Ledger current
balance is the 'Account Balance' with any opening balance brought forward and
included as a 'B/FWD' transaction entry.
}

{ TLKDaybookSalesTax }

{:
TLKDaybookSalesTax.CheckDocumentAnalysisTotals
(ABatchReference)
}

{:
A Daybook for a ledger which is an analysis of a Nominal Ledger
Control Account. Eg. a Sales Daybook records invoices entered
into a Sales Ledger. If all of the accounts in the Sales Ledger
are added together then the total equals the balance shown
on the Sales Ledger Control Account of the Nominal Ledger.
}
{:
A 'Journal' is an alternative name for 'Daybook' however has
in common usage become to mean a record of transfers between
ledger accounts.
}

{:
Records the entry of documents into an accounts system.
The Base component is aware of the AccountsSystemHub and
introduces the ExecuteDaybook method. The CreateDaybookForm
method is overwritten in dependent classes to show the
appropriate dialog.
}

interface

uses
  Windows, Messages, SysUtils, Classes, System.Types, System.UITypes,
  Graphics, Controls, Forms, Dialogs,
  ImgList, StdCtrls, ComCtrls, Contnrs, Variants, Menus, ShellApi,
  ActnList, DB, DBTables, DBClient, SqlExpr, DBXCommon, Provider, Registry,
  DBCtrls, DBGrids, SimpleDS,
  xmldom, XMLIntf, msxmldom, XMLDoc,
  Math, RecError, JvExControls, JvComponent, JvOutlookBar,
  LKBaseComponent, LKRecordsComponents, LKDefine,
  LKLedgerNominalEntryErrorsDlg, LKFormBaseDlg,  LKDaybookBatchErrorsDlg;

type
  TLKLedgerContact = class;
  TLKLedgerPurchase = class;
  TLKLedgerSales = class;
  TLKDocumentsSalesInvoices = class;
  TLKHubAccounts = class;
  TLKLedgerCost = class;
  TLKLedgerAnalysis = class;
  TLKLedgerNominal = class;
  TLKLedgerNominalBudget = class;
  TLKDaybookJournal = class;
  TLKJournalSkeletons = class;
  TLKSalesTaxReturns = class;

// Geberal accounts components
  TLKLedgerAccounts = class(TLKLedgerBase)
  private
    function GetAccountsHub: TLKHubAccounts;
    procedure SetAccountsHub(Value: TLKHubAccounts);
  protected
    procedure DeleteInputDocument; override;
    function EndOfMonthNext: TDateTime; override;
  public
    property AccountsHub: TLKHubAccounts read GetAccountsHub write
        SetAccountsHub;
  end;

  TLKDaybookAccounts = class(TLKDaybookBase)
  private
    function GetAccountsHub: TLKHubAccounts;
    function GetNominalLedger: TLKLedgerNominal;
    function GetTblDCostCentre: string;
    function GetTblDDescription: string;
    function GetTblDNominalCode: string;
    procedure SetAccountsHub(Value: TLKHubAccounts);
    procedure SetTblDCostCentre(const Value: string);
    procedure SetTblDDescription(const Value: string);
    procedure SetTblDNominalCode(const Value: string);
  protected
    procedure PbCheckBatchItemDetailIsValid; override;
    procedure PbCheckBatchItemMasterIsValid; override;
    procedure PbTablesApplyUpdates; override;
    procedure PbTablesCancelUpdates; override;
    procedure PbTablesRefresh; override;
  public
    property AccountsHub: TLKHubAccounts read GetAccountsHub write
        SetAccountsHub;
    property NominalLedger: TLKLedgerNominal read GetNominalLedger;
    property TblDCostCentre: string read GetTblDCostCentre write
        SetTblDCostCentre;
    property TblDDescription: string read GetTblDDescription write
        SetTblDDescription;
    property TblDNominalCode: string read GetTblDNominalCode write
        SetTblDNominalCode;
  end;

  TLKNominalPosting = class(TComponent)
  private
    FCostCentre: string;
    FDescription: string;
    FNominalCode: string;
    FValue: Currency;
  public
    property CostCentre: string read FCostCentre write FCostCentre;
    property Description: string read FDescription write FDescription;
    property NominalCode: string read FNominalCode write FNominalCode;
    property Value: Currency read FValue write FValue;
  end;

  TLKNominalPostingList = class(TComponentList)
  protected
    function GetItems(Index: Integer): TLKNominalPosting;
    procedure SetItems(Index: Integer; AComponent: TLKNominalPosting);
  public
    procedure Accumulate(const ANominalCode, ACostCentre, ADescription: string;
        AValue: currency);
    function Add(AComponent: TLKNominalPosting): Integer;
    function Extract(Item: TLKNominalPosting): TLKNominalPosting;
    function First: TLKNominalPosting;
    function IndexOf(AComponent: TLKNominalPosting): Integer;
    procedure Insert(Index: Integer; AComponent: TLKNominalPosting);
    function Last: TLKNominalPosting;
    property Items[Index: Integer]: TLKNominalPosting read GetItems write
        SetItems; default;
  end;

  TLKRecordsAccounts = class(TLKRecords)
  private
    function GetAccountsHub: TLKHubAccounts;
    procedure SetAccountsHub(Value: TLKHubAccounts);
  public
    property AccountsHub: TLKHubAccounts read GetAccountsHub write
        SetAccountsHub;
  end;

  TLKContainedRecordsAccounts = class(TLKContainedRecords)
  private
    function GetAccountsHub: TLKHubAccounts;
    procedure SetAccountsHub(Value: TLKHubAccounts);
  public
    property AccountsHub: TLKHubAccounts read GetAccountsHub write
        SetAccountsHub;
  end;

  TLKDocuments = class(TLKRecordsPosting)
  private
    function GetAccountsHub: TLKHubAccounts;
    function GetNominalLedger: TLKLedgerNominal;
    function GetTDCostCentre: string;
    function GetTDNominalCode: string;
    procedure SetAccountsHub(Value: TLKHubAccounts);
    procedure SetTDCostCentre(const Value: string);
    procedure SetTDNominalCode(const Value: string);
  protected
    procedure DoDetailAfterPost; override;
  public
    property AccountsHub: TLKHubAccounts read GetAccountsHub write
        SetAccountsHub;
    property NominalLedger: TLKLedgerNominal read GetNominalLedger;
    property TDCostCentre: string read GetTDCostCentre write SetTDCostCentre;
    property TDNominalCode: string read GetTDNominalCode write SetTDNominalCode;
  end;


// Ledger Analysis
  TLKLedgerAnalysisContainedDocuments = class(TLKContainedRecordsAccounts)
  private
    function GetContainerComponent: TLKLedgerAnalysis;
  public
    constructor Create(AOwner: TComponent); override;
    property ContainerComponent: TLKLedgerAnalysis read GetContainerComponent;
  end;

  TLKLedgerAnalysisAccountsPeriods = class(TLKLedgerAnalysisContainedDocuments)
  private
    function GettmPeriodEnd: TDate;
    procedure SettmPeriodEnd(Value: TDate);
  public
    property tmPeriodEnd: TDate read GettmPeriodEnd write SettmPeriodEnd;
  end;

  TLKLedgerAnalysisDeferredPostings = class(TLKLedgerAnalysisContainedDocuments)
  end;

  TLKLedgerAnalysisReportLayouts = class(TLKLedgerAnalysisContainedDocuments)
  private
    function GettmDescription: string;
    procedure SettmDescription(const Value: string);
  protected
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
  public
    constructor Create(AOwner: TComponent); override;
    property tmDescription: string read GettmDescription write SettmDescription;
  end;

  TLKLedgerAnalysisReportLines = class(TLKLedgerAnalysisContainedDocuments)
  private
    function GettmDescription: string;
    procedure SettmDescription(const Value: string);
  protected
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    procedure DlgImportExportExecute(AOwner: TComponent; const AsExportDlg,
        Autoexecute: boolean); override;
    function View(const KeyValues: Variant; ShowAsModal: Boolean; Buttons:
        TMsgDlgButtons): TModalResult; override;
    property tmDescription: string read GettmDescription write SettmDescription;
  end;

  TLKLedgerAnalysisManagementAccounts = class(
      TLKLedgerAnalysisContainedDocuments)
  private
    function GetTblDPeriodEnd: TDate;
    procedure SetTblDPeriodEnd(Value: TDate);
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    procedure DlgManagementAccountsPrintExecute; virtual;
    function EndOfMonth: TModalResult; override;
    procedure ManagementAccountsCalculate;
    property TblDPeriodEnd: TDate read GetTblDPeriodEnd write SetTblDPeriodEnd;
  end;

  TLKLedgerAnalysis = class abstract(TLKLedgerAccounts)
  private
    FAccountsPeriods: TLKLedgerAnalysisAccountsPeriods;
    FDeferredPostings: TLKLedgerAnalysisDeferredPostings;
    FReportLayouts: TLKLedgerAnalysisReportLayouts;
    FReportLines: TLKLedgerAnalysisReportLines;
  protected
    FManagementAccounts: TLKLedgerAnalysisManagementAccounts;
    procedure CreateManagementAccounts; virtual; abstract;
    procedure CreateSubComponents; override;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
    procedure RefreshAccountCodes; virtual; abstract;
    procedure SetBalanceFieldPairs; override;
    procedure SetContainedComponentNames; override;
    procedure TrialBalanceCalculate(const AccountsDate: TDateTime); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    function EndOfYear(ProgressText: string): TModalResult; override;
  published
    property AccountsPeriods: TLKLedgerAnalysisAccountsPeriods read
        FAccountsPeriods;
    property DeferredPostings: TLKLedgerAnalysisDeferredPostings read
        FDeferredPostings;
    property ManagementAccounts: TLKLedgerAnalysisManagementAccounts read
        FManagementAccounts;
    property ReportLayouts: TLKLedgerAnalysisReportLayouts read FReportLayouts;
    property ReportLines: TLKLedgerAnalysisReportLines read FReportLines;
  end;


// Ledger Nominal
  TLKLedgerNominalContainedDocuments = class(
      TLKLedgerAnalysisContainedDocuments)
  private
    function GetContainerComponent: TLKLedgerNominal;
  public
    constructor Create(AOwner: TComponent); override;
    property ContainerComponent: TLKLedgerNominal read GetContainerComponent;
  end;

  TLKLedgerNominalCodesLookup = class(TLKLedgerNominalContainedDocuments)
  protected
    procedure CreateViewFormRecord; override;
  end;

  TLKLedgerNominalFixedAssetCategories = class(TLKLedgerNominalCodesLookup)
  private
    function GettmAssetLife: Double;
    function GettmDepreciationNominalCodeBS: string;
    function GettmDepreciationNominalCodePandL: string;
    function GettmDepreciationPercent: Double;
    function GettmDepreciationStyle: TLKFixedAssetDepreciationStyle;
    function GettmDescription: string;
    procedure SettmAssetLife(Value: Double);
    procedure SettmDepreciationNominalCodeBS(const Value: string);
    procedure SettmDepreciationNominalCodePandL(const Value: string);
    procedure SettmDepreciationPercent(Value: Double);
    procedure SettmDepreciationStyle(Value: TLKFixedAssetDepreciationStyle);
    procedure SettmDescription(const Value: string);
  protected
    procedure CreateViewFormRecord; override;
  public
    property tmAssetLife: Double read GettmAssetLife write SettmAssetLife;
    property tmDepreciationNominalCodeBS: string read
        GettmDepreciationNominalCodeBS write SettmDepreciationNominalCodeBS;
    property tmDepreciationNominalCodePandL: string read
        GettmDepreciationNominalCodePandL write
        SettmDepreciationNominalCodePandL;
    property tmDepreciationPercent: Double read GettmDepreciationPercent write
        SettmDepreciationPercent;
    property tmDepreciationStyle: TLKFixedAssetDepreciationStyle read
        GettmDepreciationStyle write SettmDepreciationStyle;
    property tmDescription: string read GettmDescription write SettmDescription;
  end;

  TLKLedgerNominalNominalCodes = class(TLKLedgerNominalContainedDocuments)
  private
    function GettmDefaultSalesTaxCode: Integer;
    function GettmDescription: string;
    function GettmTurnoverNominalCode: string;
    function GettmVariableCostPercent: Double;
    procedure SettmDefaultSalesTaxCode(const Value: Integer);
    procedure SettmDescription(const Value: string);
    procedure SettmTurnoverNominalCode(const Value: string);
    procedure SettmVariableCostPercent(const Value: Double);
  protected
    procedure DoMasterBeforePost; override;
  public
    property tmDefaultSalesTaxCode: Integer read GettmDefaultSalesTaxCode write
        SettmDefaultSalesTaxCode;
    property tmDescription: string read GettmDescription write SettmDescription;
    property tmTurnoverNominalCode: string read GettmTurnoverNominalCode write
        SettmTurnoverNominalCode;
    property tmVariableCostPercent: Double read GettmVariableCostPercent write
        SettmVariableCostPercent;
  end;

  TLKLedgerNominalReservesBase = class(TLKLedgerNominalCodesLookup)
  private
    function GettmDescription: string;
    procedure SettmDescription(const Value: string);
  protected
    procedure CreateViewFormRecord; override;
  public
    procedure CalculateAccountsReserves; virtual;
    procedure CalculateBalances; virtual;
    property tmDescription: string read GettmDescription write SettmDescription;
  end;

  TLKLedgerNominalFixedAssets = class(TLKLedgerNominalReservesBase)
  private
    function GettdBalance: Currency;
    function GettmAmount: Currency;
    function GettmCostCentre: string;
    function GettmDateFrom: TDateTime;
    function GettmDateTo: TDateTime;
    function GettmDisposalValue: Currency;
    procedure SettdBalance(Value: Currency);
    procedure SettmAmount(Value: Currency);
    procedure SettmCostCentre(const Value: string);
    procedure SettmDateFrom(Value: TDateTime);
    procedure SettmDateTo(Value: TDateTime);
    procedure SettmDisposalValue(Value: Currency);
  protected
    procedure CreateViewFormRecord; override;
    procedure DoMasterBeforePost; override;
  public
    procedure CalculateAccountsReserves; override;
    procedure CalculateBalances; override;
    property tdBalance: Currency read GettdBalance write SettdBalance;
    property tmAmount: Currency read GettmAmount write SettmAmount;
    property tmCostCentre: string read GettmCostCentre write SettmCostCentre;
    property tmDateFrom: TDateTime read GettmDateFrom write SettmDateFrom;
    property tmDateTo: TDateTime read GettmDateTo write SettmDateTo;
    property tmDisposalValue: Currency read GettmDisposalValue write
        SettmDisposalValue;
  end;

  TLKLedgerNominalReserves = class(TLKLedgerNominalReservesBase)
  private
    function GettdBalance: Currency;
    function GettmAmount: Currency;
    function GettmCostCentre: string;
    function GettmDateFrom: TDateTime;
    function GettmDateTo: TDateTime;
    function GettmItemNo: Integer;
    function GettmNominalCode: string;
    procedure SettdBalance(Value: Currency);
    procedure SettmAmount(Value: Currency);
    procedure SettmCostCentre(const Value: string);
    procedure SettmDateFrom(Value: TDateTime);
    procedure SettmDateTo(Value: TDateTime);
    procedure SettmItemNo(Value: Integer);
    procedure SettmNominalCode(const Value: string);
  public
    property tdBalance: Currency read GettdBalance write SettdBalance;
    property tmAmount: Currency read GettmAmount write SettmAmount;
    property tmCostCentre: string read GettmCostCentre write SettmCostCentre;
    property tmDateFrom: TDateTime read GettmDateFrom write SettmDateFrom;
    property tmDateTo: TDateTime read GettmDateTo write SettmDateTo;
    property tmItemNo: Integer read GettmItemNo write SettmItemNo;
    property tmNominalCode: string read GettmNominalCode write SettmNominalCode;
  end;

  TLKLedgerNominalPrepayments = class(TLKLedgerNominalReserves)
  public
    procedure CalculateAccountsReserves; override;
    procedure CalculateBalances; override;
  end;

  TLKLedgerNominalAccruals = class(TLKLedgerNominalReserves)
  public
    procedure CalculateAccountsReserves; override;
    procedure CalculateBalances; override;
  end;

  TLKLedgerNominalEnterBalances = class(TLKLedgerNominalCodesLookup)
  private
    function GettdCostCentre: string;
    function GettdCredit: Currency;
    function GettdDebit: Currency;
    function GettdDescription: string;
    function GettdItemNo: Integer;
    function GettdNominalCode: string;
    function GettmCostCentre: string;
    function GettmNominalCode: string;
    function GettmReserveRefererence: string;
    procedure SettdCostCentre(const Value: string);
    procedure SettdCredit(Value: Currency);
    procedure SettdDebit(Value: Currency);
    procedure SettdDescription(const Value: string);
    procedure SettdItemNo(Value: Integer);
    procedure SettdNominalCode(const Value: string);
    procedure SettmCostCentre(const Value: string);
    procedure SettmNominalCode(const Value: string);
    procedure SettmReserveRefererence(const Value: string);
  protected
    DlgErrors: TLKDlgLedgerNominalEntryErrors;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
    function PlCheckBatchIsValid: TModalResult; virtual;
    procedure PlCheckBatchItemDetailIsValid; virtual;
    procedure PlCheckBatchItemMasterIsValid; virtual;
    procedure PlCheckNominal(const NominalCode, CostCentre: string);
    procedure PlClearReserves; virtual;
    function PlDlgErrorsCorrectError(ErrorInfo: TStrings): TModalResult;
        virtual;
    procedure PlDlgErrorsCorrectWarnings; virtual;
    procedure PlPostReserve; virtual;
    procedure PlProcessedComplete; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    procedure EntriesCalculateWizardExecute(AOwner: TComponent); virtual;
    function PlDlgErrorsCorrectErrors: TModalResult; virtual;
    procedure PostToLedger(AOwner: TComponent);
    property tdCostCentre: string read GettdCostCentre write SettdCostCentre;
    property tdCredit: Currency read GettdCredit write SettdCredit;
    property tdDebit: Currency read GettdDebit write SettdDebit;
    property tdDescription: string read GettdDescription write SettdDescription;
    property tdItemNo: Integer read GettdItemNo write SettdItemNo;
    property tdNominalCode: string read GettdNominalCode write SettdNominalCode;
    property tmCostCentre: string read GettmCostCentre write SettmCostCentre;
    property tmNominalCode: string read GettmNominalCode write SettmNominalCode;
    property tmReserveRefererence: string read GettmReserveRefererence write
        SettmReserveRefererence;
  end;

  TLKLedgerNominalEnterOpeningBalances = class(TLKLedgerNominalEnterBalances)
  protected
    procedure PlClearReserves; override;
    procedure PlPostReserve; override;
  end;

  TLKLedgerNominalEnterReserves = class(TLKLedgerNominalEnterBalances)
  private
    function GettmReserveDate: TDateTime;
    procedure SettmReserveDate(Value: TDateTime);
  protected
    procedure PlClearReserves; override;
    procedure PlPostReserve; override;
  public
    procedure CreateViewForm; override;
    procedure DrillDownRecord; override;
    function EndOfMonth: TModalResult; override;
    procedure EntriesCalculateWizardExecute(AOwner: TComponent); override;
    property tmReserveDate: TDateTime read GettmReserveDate write
        SettmReserveDate;
  end;

  TLKLedgerNominalManagementAccounts = class(
      TLKLedgerAnalysisManagementAccounts)
  public
    procedure DlgManagementAccountsPrintExecute; override;
  end;


// Ledger Cost
  TLKLedgerNominalBudget = class(TLKLedgerNominalContainedDocuments)
  protected
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    procedure DlgBudgetPrintExecute;
    procedure DlgImportExportExecute(AOwner: TComponent; const AsExportDlg,
        Autoexecute: boolean); override;
  end;

  TLKLedgerNominal = class(TLKLedgerAnalysis)
  private
    FAccruals: TLKLedgerNominalAccruals;
    FBudget: TLKLedgerNominalBudget;
    FFixedAssetCategories: TLKLedgerNominalFixedAssetCategories;
    FFixedAssets: TLKLedgerNominalFixedAssets;
    FNominalCodes: TLKLedgerNominalNominalCodes;
    FOpeningBalances: TLKLedgerNominalEnterOpeningBalances;
    FPrepayments: TLKLedgerNominalPrepayments;
    FReserves: TLKLedgerNominalEnterReserves;
    function GetCodeAccruals: string;
    function GetCodePrepayments: string;
    function GetCodeProfitAndLoss: string;
    function GetCostLedger: TLKLedgerCost;
    function GetManagementAccounts: TLKLedgerNominalManagementAccounts;
    procedure SetCodeAccruals(const Value: string);
    procedure SetCodePrepayments(const Value: string);
    procedure SetCodeProfitAndLoss(const Value: string);
  protected
    procedure CreateManagementAccounts; override;
    procedure CreateSubComponents; override;
    procedure DoAfterSetHub; override;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
    procedure Loaded; override;
    procedure RefreshAccountCodes; override;
    procedure SetContainedComponentNames; override;
  public
    constructor Create(AOwner: TComponent); override;
    function AccountIsValid(const KeyValues: Variant; RaiseError: Boolean =
        False): Boolean; override;
    procedure AccountPost(const DetailRecord: array of const); override;
    procedure CreateViewForm; override;
    function DlgNominalCodeSelectExecute(const ACaption, EncodedNomCost,
        HelpTxt: string; const getCostcentre: boolean): string;
    function EndOfMonth: TModalResult; override;
    function EndOfYear(ProgressText: string): TModalResult; override;
    function RecordNew: TModalResult; override;
    procedure TrialBalanceCalculate(const AccountsDate: TDateTime); override;
    property CodeAccruals: string read GetCodeAccruals write SetCodeAccruals;
    property CodePrepayments: string read GetCodePrepayments write
        SetCodePrepayments;
    property CodeProfitAndLoss: string read GetCodeProfitAndLoss write
        SetCodeProfitAndLoss;
    property CostLedger: TLKLedgerCost read GetCostLedger;
  published
    property Accruals: TLKLedgerNominalAccruals read FAccruals;
    property Budget: TLKLedgerNominalBudget read FBudget;
    property FixedAssetCategories: TLKLedgerNominalFixedAssetCategories read
        FFixedAssetCategories;
    property FixedAssets: TLKLedgerNominalFixedAssets read FFixedAssets;
    property ManagementAccounts: TLKLedgerNominalManagementAccounts read
        GetManagementAccounts;
    property NominalCodes: TLKLedgerNominalNominalCodes read FNominalCodes;
    property OpeningBalances: TLKLedgerNominalEnterOpeningBalances read
        FOpeningBalances;
    property Prepayments: TLKLedgerNominalPrepayments read FPrepayments;
    property Reserves: TLKLedgerNominalEnterReserves read FReserves;
  end;

  TLKLedgerCostContainedDocuments = class(TLKLedgerAnalysisContainedDocuments)
  private
    function GetContainerComponent: TLKLedgerCost;
  public
    constructor Create(AOwner: TComponent); override;
    property ContainerComponent: TLKLedgerCost read GetContainerComponent;
  end;


// SalesTax
  TLKLedgerCostCostCentres = class(TLKLedgerCostContainedDocuments)
  private
    function GettmDescription: string;
    procedure SettmDescription(const Value: string);
  public
    property tmDescription: string read GettmDescription write SettmDescription;
  end;

  TLKLedgerCostManagementAccounts = class(TLKLedgerAnalysisManagementAccounts)
  public
    procedure DlgManagementAccountsPrintExecute; override;
  end;


// Nominal Ledger Control account Ledgers
  TLKLedgerCost = class(TLKLedgerAnalysis)
  private
    FCostCentreCurrent: string;
    FCostCentres: TLKLedgerCostCostCentres;
    function GetManagementAccounts: TLKLedgerCostManagementAccounts;
    function GetNominalLedger: TLKLedgerNominal;
    procedure SetCostCentreCurrent(const Value: string);
  protected
    procedure CreateManagementAccounts; override;
    procedure CreateSubComponents; override;
    procedure DoAfterSetHub; override;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
    procedure RefreshAccountCodes; override;
    procedure SetContainedComponentNames; override;
  public
    constructor Create(AOwner: TComponent); override;
    function AccountIsValid(const KeyValues: Variant; RaiseError: Boolean =
        False): Boolean; override;
    procedure AccountPost(const DetailRecord: array of const); override;
    procedure AddValueToAccount(CostCentre, NominalCode: string; Value:
        currency);
    procedure CreateViewForm; override;
    function DlgSelectCostCentreExecute(AOwner: TComponent): TModalResult;
    function EndOfMonth: TModalResult; override;
    function EndOfYear(ProgressText: string): TModalResult; override;
    function RecordNew: TModalResult; override;
    procedure TrialBalanceCalculate(const AccountsDate: TDateTime); override;
    property CostCentreCurrent: string read FCostCentreCurrent write
        SetCostCentreCurrent;
    property NominalLedger: TLKLedgerNominal read GetNominalLedger;
  published
    property CostCentres: TLKLedgerCostCostCentres read FCostCentres;
    property ManagementAccounts: TLKLedgerCostManagementAccounts read
        GetManagementAccounts;
  end;

  TLKSalesTaxCodes = class(TLKContainedRecordsAccounts)
  private
    function GetContainerComponent: TLKSalesTaxReturns;
  protected
    procedure DoMasterBeforePost; override;
  public
    constructor Create(AOwner: TComponent); override;
    property ContainerComponent: TLKSalesTaxReturns read GetContainerComponent;
  end;

  TLKSalesTaxReturns = class(TLKRecordsAccounts)
  private
    FSalesTaxCodes: TLKSalesTaxCodes;
    function GetDefaultStandardSalesTaxCode: Integer;
    function GetNominalControlCode: string;
    function GetNominalLedger: TLKLedgerNominal;
    function GetSalesTaxRegistration: string;
    procedure SetDefaultStandardSalesTaxCode(Value: Integer);
    procedure SetNominalControlCode(const Value: string);
    procedure SetSalesTaxRegistration(const Value: string);
  protected
    procedure CreateSubComponents; override;
    procedure CreateViewFormRecord; override;
    procedure DoAfterSetHub; override;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
    procedure SetContainedComponentNames; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CalculateSalesTaxReturnValues;
    procedure CreateViewForm; override;
    procedure DrillDownRecord; override;
    function RecordNew: TModalResult; override;
    function SalesTax(const GoodsValue: Currency; SalesTaxCode: Integer):
        Currency;
    property DefaultStandardSalesTaxCode: Integer read
        GetDefaultStandardSalesTaxCode write SetDefaultStandardSalesTaxCode;
    property NominalControlCode: string read GetNominalControlCode write
        SetNominalControlCode;
    property NominalLedger: TLKLedgerNominal read GetNominalLedger;
    property SalesTaxRegistration: string read GetSalesTaxRegistration write
        SetSalesTaxRegistration;
  published
    property SalesTaxCodes: TLKSalesTaxCodes read FSalesTaxCodes;
  end;

  TLKLedgerNominalControl = class(TLKLedgerAccounts)
  private
    function GetNominalControlCode: string;
    function GetNominalLedger: TLKLedgerNominal;
    procedure SetNominalControlCode(const Value: string);
  protected
    procedure AssignTo(Dest: TPersistent); override;
    procedure CreateViewFormRecord; override;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
    function GetSettings: TModalResult; override;
    procedure SaveSettings; override;
    procedure SetBalanceFieldPairs; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CheckAccountBalancesToNominalControl(showOK: boolean); virtual;
    procedure CreateViewForm; override;
    procedure DlgDiaryExecute; virtual;
    procedure DlgJnlAccountToAccountExecute; virtual;
    procedure DlgJnlAccountToNominalExecute; virtual;
    procedure DlgNominalCodeSetupExecute; virtual;
    function EndOfMonth: TModalResult; override;
    function EndOfYear(ProgressText: string): TModalResult; override;
    function NominalControlBalance: Currency;
    property NominalControlCode: string read GetNominalControlCode write
        SetNominalControlCode;
    property NominalLedger: TLKLedgerNominal read GetNominalLedger;
  end;

  TLKDaybookNominal = class abstract(TLKDaybookAccounts)
  private
    FEntryType: TLKACDaybookEntryType;
    FNominalPostings: TLKNominalPostingList;
    function GetLedger: TLKLedgerNominalControl;
    procedure SetLedger(Value: TLKLedgerNominalControl);
    procedure SetPostNominalTransactionDetail(Value:
        TLKPostNominalTransactionDetail);
  protected
    FPostNominalTransactionDetail: TLKPostNominalTransactionDetail;
    NominalBatchTotal: Currency;
    procedure AssignTo(Dest: TPersistent); override;
    function GetBatchDate: TDateTime; override;
    function GetSettings: TModalResult; override;
    procedure PbBeforeProcessBatch; override;
    procedure PbCheckBatchItemMasterIsValid; override;
    procedure PbFinalisePostings; override;
    procedure PbPostCostAndNominalAnalysis; virtual; abstract;
    procedure PbPostNominalControl; virtual;
    procedure PbProcessDocument; override;
    function PbSetFilter: Boolean; override;
    property NominalPostings: TLKNominalPostingList read FNominalPostings;
  public
    procedure CreateViewForm; override;
    property PostNominalTransactionDetail: TLKPostNominalTransactionDetail read
        FPostNominalTransactionDetail write SetPostNominalTransactionDetail;
  published
    property EntryType: TLKACDaybookEntryType read FEntryType write FEntryType
        default detDebits;
    property Ledger: TLKLedgerNominalControl read GetLedger write SetLedger;
  end;

  TLKDaybookSalesTax = class(TLKDaybookNominal)
  private
    function GetSalesTaxReturns: TLKSalesTaxReturns;
    function GetTblDGoods: Currency;
    function GetTblDSalesTax: Currency;
    function GetTblDSalesTaxCode: Integer;
    function GetTblDTotal: Currency;
    function GetTblMGoods: Currency;
    function GetTblMSalesTax: Currency;
    function GetTblMTotal: Currency;
    procedure SetTblDGoods(Value: Currency);
    procedure SetTblDSalesTax(Value: Currency);
    procedure SetTblDSalesTaxCode(Value: Integer);
    procedure SetTblDTotal(Value: Currency);
    procedure SetTblMGoods(Value: Currency);
    procedure SetTblMSalesTax(Value: Currency);
    procedure SetTblMTotal(Value: Currency);
  protected
    DocsHaveSalesTax: Boolean;
    TotalSalesTax: Currency;
    procedure PbBeforeProcessBatch; override;
    procedure PbCheckBatchItemDetailIsValid; override;
    procedure PbCheckBatchItemMasterIsValid; override;
    procedure PbFinalisePostings; override;
    procedure PbPostCostAndNominalAnalysis; override;
    procedure PbPostSalesTaxControl;
    procedure PbTablesApplyUpdates; override;
    procedure PbTablesCancelUpdates; override;
    procedure PbTablesRefresh; override;
    procedure SetActive(Value: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CalculateSalesTaxPeriodValues(SalesTaxSummary: TDataset; const
        SalesTaxPeriod: string; SalestaxPeriodEnd: TDateTime);
    property SalesTaxReturns: TLKSalesTaxReturns read GetSalesTaxReturns;
    property TblDGoods: Currency read GetTblDGoods write SetTblDGoods;
    property TblDSalesTax: Currency read GetTblDSalesTax write SetTblDSalesTax;
    property TblDSalesTaxCode: Integer read GetTblDSalesTaxCode write
        SetTblDSalesTaxCode;
    property TblDTotal: Currency read GetTblDTotal write SetTblDTotal;
    property TblMGoods: Currency read GetTblMGoods write SetTblMGoods;
    property TblMSalesTax: Currency read GetTblMSalesTax write SetTblMSalesTax;
    property TblMTotal: Currency read GetTblMTotal write SetTblMTotal;
  end;


// Cash and Bank
  TLKDocumentsNominal = class(TLKDocuments)
  private
    function GetDaybook: TLKDaybookNominal;
    procedure SetDaybook(Value: TLKDaybookNominal);
  protected
    procedure CreateViewFormRecord; override;
  published
    property Daybook: TLKDaybookNominal read GetDaybook write SetDaybook;
  end;

  TLKJournal = class(TLKDocuments)
  private
    FSkeletonJournals: TLKJournalSkeletons;
    function GetDaybook: TLKDaybookJournal;
    procedure SetDaybook(Value: TLKDaybookJournal);
  protected
    procedure CreateSubComponents; override;
    procedure DoAfterSetHub; override;
    procedure DoDetailAfterInsert; override;
    procedure SetContainedComponentNames; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure ContactLedgerAccountTransferExecute(ALedger: TLKLedgerContact);
    procedure CreateViewForm; override;
    procedure NomControlLedgerAccountTransferExecute(ALedger:
        TLKLedgerNominalControl);
    procedure NomControlLedgerNominalJournalExecute(ALedger:
        TLKLedgerNominalControl);
    function RecordNew: TModalResult; override;
  published
    property Daybook: TLKDaybookJournal read GetDaybook write SetDaybook;
    property SkeletonJournals: TLKJournalSkeletons read FSkeletonJournals;
  end;

  TLKJournalSkeletons = class(TLKContainedRecordsAccounts)
  private
    Description: string;
    function GetContainerComponent: TLKJournal;
  protected
    procedure CreateViewFormRecord; override;
    procedure DoDetailAfterInsert; override;
    procedure DoDetailAfterPost; override;
    procedure DoMasterAfterPost; override;
  public
    constructor Create(AOwner: TComponent); override;
    property ContainerComponent: TLKJournal read GetContainerComponent;
  end;

  TLKDocumentsSalesTax = class(TLKDocumentsNominal)
  private
    function GetSalesTaxReturns: TLKSalesTaxReturns;
    function GetTDGoods: Currency;
    function GetTDSalesTax: Currency;
    function GetTDSalesTaxCode: Integer;
    function GetTDTotal: Currency;
    function GetTMGoods: Currency;
    function GetTMSalesTax: Currency;
    function GetTMTotal: Currency;
    procedure SetTDGoods(Value: Currency);
    procedure SetTDSalesTax(Value: Currency);
    procedure SetTDSalesTaxCode(Value: Integer);
    procedure SetTDTotal(Value: Currency);
    procedure SetTMGoods(Value: Currency);
    procedure SetTMSalesTax(Value: Currency);
    procedure SetTMTotal(Value: Currency);
  protected
    procedure CreateViewFormRecord; override;
    function DaybookAdjusted(const ALedgerMatchValue: currency): Currency;
    procedure DoDetailBeforePost; override;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    property SalesTaxReturns: TLKSalesTaxReturns read GetSalesTaxReturns;
    property TDGoods: Currency read GetTDGoods write SetTDGoods;
    property TDSalesTax: Currency read GetTDSalesTax write SetTDSalesTax;
    property TDSalesTaxCode: Integer read GetTDSalesTaxCode write
        SetTDSalesTaxCode;
    property TDTotal: Currency read GetTDTotal write SetTDTotal;
    property TMGoods: Currency read GetTMGoods write SetTMGoods;
    property TMSalesTax: Currency read GetTMSalesTax write SetTMSalesTax;
    property TMTotal: Currency read GetTMTotal write SetTMTotal;
  end;

  TLKLedgerCash = class(TLKLedgerNominalControl)
  private
    PostingDate: TDateTime;
    ValueDeleted: Currency;
  protected
    procedure DeleteInputDocument; override;
    procedure DoDetailBeforePost; override;
  public
    function AccountCheckTotals(const showOK, correctErrors, raiseError:
        boolean): Boolean; override;
    procedure CreateViewForm; override;
    function EndOfYear(ProgressText: string): TModalResult; override;
    function RecordNew: TModalResult; override;
    function RefreshBalances: Currency; virtual;
  end;

  TLKLedgerBank = class(TLKLedgerCash)
  protected
    procedure CreateViewFormRecord; override;
  public
    procedure CreateViewForm; override;
    function DlgBankAgreementExecute: TModalResult; virtual;
    function EndOfMonth: TModalResult; override;
  end;

  TLKDaybookCash = class(TLKDaybookSalesTax)
  private
    PbLedgers: TLKLedgerList;
    function GetLedger: TLKLedgerCash;
    procedure SetLedger(Value: TLKLedgerCash);
  protected
    procedure PbPostDocumentToLedger; override;
    procedure PbProcessedComplete; override;
    procedure PbTablesApplyUpdates; override;
    procedure PbTablesCancelUpdates; override;
    procedure PbTablesRefresh; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Ledger: TLKLedgerCash read GetLedger write SetLedger;
  end;

  TLKDaybookPettyCash = class(TLKDaybookSalesTax)
  private
    function GetLedger: TLKLedgerCash;
    procedure SetLedger(Value: TLKLedgerCash);
  protected
    procedure PbPostDocumentToLedger; override;
  published
    property Ledger: TLKLedgerCash read GetLedger write SetLedger;
  end;

  TLKDaybookBank = class(TLKDaybookCash)
  private
    function GetLedger: TLKLedgerBank;
    procedure SetLedger(Value: TLKLedgerBank);
  published
    property Ledger: TLKLedgerBank read GetLedger write SetLedger;
  end;

  TLKDocumentsCashBook = class(TLKDocumentsSalesTax)
  private
    FDefaultPaymentLedger: TLKLedgerContact;
    function GetDaybook: TLKDaybookCash;
    function GettmAccount: Variant;
    procedure SetDaybook(Value: TLKDaybookCash);
    procedure SetDefaultPaymentLedger(Value: TLKLedgerContact);
  protected
    procedure CreateViewFormRecord; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
        override;
  public
    procedure ContactLedgerPayment(AOwner: TComponent; ALedger:
        TLKLedgerContact; const APaymentDate: TDateTime);
    procedure CreateViewForm; override;
    function DlgContactLedgerPaymentExecute(AOwner: TComponent; ALedger:
        TLKLedgerContact): Boolean; virtual;
    property tmAccount: Variant read GettmAccount;
  published
    property Daybook: TLKDaybookCash read GetDaybook write SetDaybook;
    property DefaultPaymentLedger: TLKLedgerContact read FDefaultPaymentLedger
        write SetDefaultPaymentLedger;
  end;

  TLKDocumentsBankBook = class(TLKDocumentsCashBook)
  private
    function GetDaybook: TLKDaybookBank;
    procedure SetDaybook(Value: TLKDaybookBank);
  published
    property Daybook: TLKDaybookBank read GetDaybook write SetDaybook;
  end;

  TLKDocumentsPettyCash = class(TLKDocumentsSalesTax)
  private
    function GetDaybook: TLKDaybookPettyCash;
    procedure SetDaybook(Value: TLKDaybookPettyCash);
  published
    property Daybook: TLKDaybookPettyCash read GetDaybook write SetDaybook;
  end;


// Invoicng, Sales and Purchase Ledgers
  TLKLedgerContact = class(TLKLedgerNominalControl)
  private
    function GetTDCurrent: Boolean;
    function GetTDUnpaid: Currency;
    function GetTMTermsDays: Integer;
    procedure SetTDCurrent(Value: Boolean);
    procedure SetTDUnpaid(Value: Currency);
    procedure SetTMTermsDays(Value: Integer);
  protected
    procedure CreateViewFormRecord; override;
    procedure DlgPrintMonthEndBalancesExecute; virtual;
    procedure DoDetailBeforeDelete; override;
    function EndOfYearFilter(TransactionMonthsToRetain: Integer): string;
        override;
    procedure SetBalanceFieldPairs; override;
  public
    constructor Create(AOwner: TComponent); override;
    function AccountCheckTotals(const showOK, correctErrors, raiseError:
        boolean): Boolean; override;
    procedure AccountPost(const DetailRecord: array of const); override;
    procedure CreateViewForm; override;
    procedure DlgJnlAccountToAccountExecute; override;
    procedure DlgLedgerAllocateExecute; override;
    function EndOfMonth: TModalResult; override;
    procedure RecalculateAgeing;
    procedure UpdateAgeing(const TransactionDate: TDateTime; SignedValue:
        Currency); virtual;
    property TDCurrent: Boolean read GetTDCurrent write SetTDCurrent;
    property TDUnpaid: Currency read GetTDUnpaid write SetTDUnpaid;
    property TMTermsDays: Integer read GetTMTermsDays write SetTMTermsDays;
  end;

  TLKBankTransferPaymentsList = class(TLKContainedRecordsAccounts)
  private
    function GetBankPaymentsBook: TLKDocumentsBankBook;
    function GetContainerComponent: TLKLedgerPurchase;
  protected
    property BankPaymentsBook: TLKDocumentsBankBook read GetBankPaymentsBook;
  public
    constructor Create(AOwner: TComponent); override;
    procedure MakePaymentListPayment; virtual;
    property ContainerComponent: TLKLedgerPurchase read GetContainerComponent;
  end;

  TLKLedgerPurchase = class(TLKLedgerContact)
  private
    FBankPaymentsBook: TLKDocumentsBankBook;
    FBankTransferPaymentsList: TLKBankTransferPaymentsList;
    FChequeBook: TLKDocumentsBankBook;
    procedure SetBankPaymentsBook(Value: TLKDocumentsBankBook);
    procedure SetChequeBook(Value: TLKDocumentsBankBook);
  protected
    procedure CreateSubComponents; override;
    procedure CreateViewFormRecord; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
        override;
    procedure SetContainedComponentNames; override;
  public
    procedure AddPaymentToPaymentList;
    procedure CreateViewForm; override;
    procedure DrawCheque(AOwner: TComponent);
  published
    property BankPaymentsBook: TLKDocumentsBankBook read FBankPaymentsBook
        write SetBankPaymentsBook;
    property BankTransferPaymentsList: TLKBankTransferPaymentsList read
        FBankTransferPaymentsList;
    property ChequeBook: TLKDocumentsBankBook read FChequeBook write
        SetChequeBook;
  end;

  TLKSalesManagers = class(TLKRecordsAccounts)
  private
    FSalesLedger: TLKLedgerSales;
    procedure SetSalesLedger(Value: TLKLedgerSales);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation);
        override;
  public
    property SalesLedger: TLKLedgerSales read FSalesLedger write SetSalesLedger;
  end;

  TLKLedgerSales = class(TLKLedgerContact)
  private
    FCustomerStatements: TLKRecordsAccounts;
    FSalesManagers: TLKSalesManagers;
    procedure SetCustomerStatements(Value: TLKRecordsAccounts);
    procedure SetSalesManagers(Value: TLKSalesManagers);
  protected
    procedure CreateViewFormRecord; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
        override;
  public
    procedure CreateViewForm; override;
    function EndOfMonth: TModalResult; override;
    procedure PrintCustomerStatements;
  published
    property CustomerStatements: TLKRecordsAccounts read FCustomerStatements
        write SetCustomerStatements;
    property SalesManagers: TLKSalesManagers read FSalesManagers write
        SetSalesManagers;
  end;

  TLKDaybookContact = class(TLKDaybookSalesTax)
  private
    function GetLedger: TLKLedgerContact;
    function GetTblMDueDate: TDateTime;
    procedure SetLedger(Value: TLKLedgerContact);
    procedure SetTblMDueDate(Value: TDateTime);
  protected
    procedure PbPostDocumentToLedger; override;
  public
    property TblMDueDate: TDateTime read GetTblMDueDate write SetTblMDueDate;
  published
    property Ledger: TLKLedgerContact read GetLedger write SetLedger;
  end;

  TLKDaybookPurchase = class(TLKDaybookContact)
  private
    function GetLedger: TLKLedgerPurchase;
    procedure SetLedger(Value: TLKLedgerPurchase);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Ledger: TLKLedgerPurchase read GetLedger write SetLedger;
  end;

  TLKDaybookSales = class(TLKDaybookContact)
  private
    function GetLedger: TLKLedgerSales;
    procedure SetLedger(Value: TLKLedgerSales);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Ledger: TLKLedgerSales read GetLedger write SetLedger;
  end;

  TLKDocumentsContact = class(TLKDocumentsSalesTax)
  private
    FContactNameField: Integer;
    FContactNameFieldLedger: Integer;
    function GetMasterComponent: TLKLedgerContact;
    function GetTDQuantity: Double;
    function GetTDRate: Currency;
    function GettmDueDate: TDateTime;
    procedure SetMasterComponent(Value: TLKLedgerContact);
    procedure SetTDQuantity(Value: Double);
    procedure SetTDRate(Value: Currency);
    procedure SettmDueDate(Value: TDateTime);
  protected
    procedure CreateViewFormRecord; override;
    procedure DoDetailBeforePost; override;
    procedure DoMasterAfterInsert; override;
    procedure DoMasterBeforePost; override;
    property ContactNameField: Integer read FContactNameField write
        FContactNameField;
    property ContactNameFieldLedger: Integer read FContactNameFieldLedger write
        FContactNameFieldLedger;
  public
    constructor Create(AOwner: TComponent); override;
    property TDQuantity: Double read GetTDQuantity write SetTDQuantity;
    property TDRate: Currency read GetTDRate write SetTDRate;
    property tmDueDate: TDateTime read GettmDueDate write SettmDueDate;
  published
    property MasterComponent: TLKLedgerContact read GetMasterComponent write
        SetMasterComponent;
  end;

  TLKDocumentsPurchaseInvoices = class(TLKDocumentsContact)
  private
    function GetDaybook: TLKDaybookPurchase;
    procedure SetDaybook(Value: TLKDaybookPurchase);
  published
    property Daybook: TLKDaybookPurchase read GetDaybook write SetDaybook;
  end;

  TLKDocumentsSalesInvoices = class(TLKDocumentsContact)
  private
    FDefaultSalesNominalCode: string;
    function GetDaybook: TLKDaybookSales;
    function GetDefaultSalesNominalCode: string;
    procedure SetDaybook(Value: TLKDaybookSales);
    procedure SetDefaultSalesNominalCode(const Value: string);
  protected
    NextSalesInvoiceDate: TDateTime;
    procedure DoMasterAfterInsert; override;
    procedure DoMasterAfterPost; override;
    function DoSetup(const ForceIfSettingsExist: Boolean): TModalResult;
        override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateViewForm; override;
    property DefaultSalesNominalCode: string read GetDefaultSalesNominalCode
        write SetDefaultSalesNominalCode;
  published
    property Daybook: TLKDaybookSales read GetDaybook write SetDaybook;
  end;

  TLKDaybookJournal = class(TLKDaybookAccounts)
  private
    PbUpdatedLedgers: TLKLedgerList;
    function GetLedger: TLKLedgerNominal;
    procedure SetLedger(Value: TLKLedgerNominal);
  protected
    function GetBatchDate: TDateTime; override;
    procedure PbBeforeProcessBatch; override;
    procedure PbCheckBatchItemDetailIsValid; override;
    procedure PbPostDocumentToLedger; override;
    procedure PbProcessedComplete; override;
    procedure PbTablesApplyUpdates; override;
    procedure PbTablesCancelUpdates; override;
    procedure PbTablesRefresh; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Ledger: TLKLedgerNominal read GetLedger write SetLedger;
  end;

  TLKHubAccountsDiary = class(TLKHubRecordsDiary)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TLKHubAccounts = class(TLKHubRecords)
  private
    FAccountsInterval: TLKAccountsInterval;
    FAccountsIntervalDays: Integer;
    FCostLedger: TLKLedgerCost;
    FJournal: TLKJournal;
    FNominalLedger: TLKLedgerNominal;
    FP11D: TLKDataComponent;
    FSalesTaxReturns: TLKSalesTaxReturns;
    function GetAccountsInterval: TLKAccountsInterval;
    function GetAccountsIntervalDays: Integer;
    function GetAccountsPeriodEnd: TDateTime;
    function GetAccountsPeriodStart: TDateTime;
    function GetAccountsYear: string;
    function GetDiary: TLKHubAccountsDiary;
    function GetSalesTaxRegistration: string;
    procedure SetAccountsInterval(Value: TLKAccountsInterval);
    procedure SetAccountsIntervalDays(Value: Integer);
    procedure SetAccountsPeriodEnd(Value: TDateTime);
    procedure SetAccountsPeriodStart(Value: TDateTime);
    procedure SetCostLedger(Value: TLKLedgerCost);
    procedure SetJournal(Value: TLKJournal);
    procedure SetNominalLedger(Value: TLKLedgerNominal);
    procedure SetP11D(Value: TLKDataComponent);
    procedure SetSalesTaxRegistration(const Value: string);
    procedure SetSalesTaxReturns(Value: TLKSalesTaxReturns);
  protected
    procedure CreateDiary; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
        override;
  public
    constructor Create(AOwner: TComponent); override;
    function DlgEnterpriseExecute: TModalResult; override;
    procedure DlgP11DExecute(Sender: TLKDataComponent);
    procedure DrillDownDiary; override;
    procedure DrillDownSalesTaxReturn(const DaybookID, SalesTaxPeriod: string);
    function EndOfYear(ProgressText: string): TModalResult; override;
    function PeriodEndThis(ADate: TDateTime): TDateTime;
    property AccountsInterval: TLKAccountsInterval read GetAccountsInterval
        write SetAccountsInterval;
    property AccountsIntervalDays: Integer read GetAccountsIntervalDays write
        SetAccountsIntervalDays;
    property AccountsPeriodEnd: TDateTime read GetAccountsPeriodEnd write
        SetAccountsPeriodEnd;
    property AccountsPeriodStart: TDateTime read GetAccountsPeriodStart write
        SetAccountsPeriodStart;
    property AccountsYear: string read GetAccountsYear;
    property CostLedger: TLKLedgerCost read FCostLedger write SetCostLedger;
    property Journal: TLKJournal read FJournal write SetJournal;
    property NominalLedger: TLKLedgerNominal read FNominalLedger write
        SetNominalLedger;
    property P11D: TLKDataComponent read FP11D write SetP11D;
    property SalesTaxRegistration: string read GetSalesTaxRegistration write
        SetSalesTaxRegistration;
    property SalesTaxReturns: TLKSalesTaxReturns read FSalesTaxReturns write
        SetSalesTaxReturns;
  published
    property Diary: TLKHubAccountsDiary read GetDiary;
  end;

procedure Register;

implementation

uses
  LKUtils, LKUtilsDB, LKUtilComponents, LKSystemComponentSQLWizardDlg,
  LKdlgs,
  LKRecordsBaseDlg, LKRecordsBaseRecordDlg,
  LKRecordsInputRecordDlg,
  LKRecordsSequencedRecordDlg,
  prLKTableBasePrintRecordDlg, prLKRecordDlg,
  prLKEnvelopeDlg, LKDaybookBaseDlg,
  LKGroupDlg,
  LKHubAccountsEnterpriseDlg,
  LKLedgerAnalysisSetupWizDlg, LKLedgerAnalysisReportLinesDlg,
  LKLedgerNominalEnterBalancesDlg, LKLedgerNominalEnterReservesDlg,
  LKLedgerNominalFixedAssteCategoriesRecordDlg,
  LKLedgerNominalReservesBaseInputRecordDlg,
  LKLedgerNominalFixedAssetsRecordDlg, LKLedgerNominalBudgetDlg,
  LKDocumentsNominalRecordDlg,LKDocumentsInputNominalRecordDlg,
  LKLedgerNominalControlDlg, LKLedgerNomAnalysisDiaryDlg,
  LKLedgerContactDlg, LKLedgerPurchaseDlg, LKLedgerPurchaseRecordDlg,
  LKLedgerContactRecordDlg, LKLedgerSalesDlg, LKLedgerSalesRecordDlg,
  LKLedgerAnalysisDlg, LKLedgerNominalDlg, LKLedgerCostDlg,
  LKLedgerAnalysisManagementAccountsDlg, prLKLedgerNominalBudget,
  prLKLedgerNominalManagementAcs, prLKLedgerCostManagementAcs,
  LKLedgerNominalCodeSelectDlg,
  LKDaybookNominalDlg,
  LKRecordsSalesTaxReturnsDlg, LKRecordsSalesTaxReturnsRecordDlg,
  LKDocumentsPIIDDlg,
  LKDocumentsSalesTaxDlg, LKDocumentsSalesTaxRecordDlg,
  LKDocumentsContactRecordDlg,
  LKLedgerCashDlg, LKLedgerBankDlg, LKLedgerBankRecordDlg,
  LKDocumentsCashBookDlg, LKDocumentsCashBookRecordDlg,
  LKLedgerBankUpOsDlg,
  LKLedgerContactAllocateDlg,
  LKLedgerContactJnlDlg,
  LKJournalDlg, LKJournalNomCostDlg, LKLedgerNomControlRecordDlg,
  LKLedgerNomControlJnlDlg, LKLedgerNomControlNomJnlDlg,
  LKDocumentsSalesInvoicesDlg;

procedure Register;
begin
  // RegisterComponents('LKAccounts', [TLKHubAccounts]);
  RegisterComponents('LKAccounts', [TLKHubAccounts]);
  RegisterComponents('LKAccounts', [TLKRecordsAccounts]);
end;


{
****************************** TLKLedgerAccounts *******************************
}
procedure TLKLedgerAccounts.DeleteInputDocument;
var
  DbkID: string;
  DocRef: string;
  D: TLKDaybookBase;
begin
  with TblDetail do
  begin
    DbkID := FieldByName('DAYBOOK_ID').asString;
    DocRef := FieldByName('DOCUMENT_NUMBER').asString;
    D := Daybooks.ComponentByID(DbkID);
    if Assigned(D) then
      if D.Locate(D.TblMaster.Fields[0].FieldName, DocRef, []) then
        D.Delete;
  end;
end;

function TLKLedgerAccounts.EndOfMonthNext: TDateTime;
begin
if not Assigned(AccountsHub) then
  Result := inherited
else
  case AccountsHub.AccountsInterval of
    aciWeekly: Result := LedgerDate + 7;
    aciFortnightly: Result := LedgerDate + 14;
    aciFourWeeks: Result := LedgerDate + 28;
    aciAnnual: Result := MonthEndThis(LedgerDate + 360);
    aciDays: Result := LedgerDate + AccountsHub.AccountsIntervalDays;
    else Result := MonthEndNext(LedgerDate);
  end;
end;

function TLKLedgerAccounts.GetAccountsHub: TLKHubAccounts;
begin
  Result := TLKHubAccounts(inherited AccountsHub);
end;

procedure TLKLedgerAccounts.SetAccountsHub(Value: TLKHubAccounts);
begin
  inherited AccountsHub := Value;
end;


{
****************************** TLKDaybookAccounts ******************************
}
function TLKDaybookAccounts.GetAccountsHub: TLKHubAccounts;
begin
  Result := TLKHubAccounts(inherited AccountsHub);
end;

function TLKDaybookAccounts.GetNominalLedger: TLKLedgerNominal;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.NominalLedger
  else
    Result := nil;
end;

function TLKDaybookAccounts.GetTblDCostCentre: string;
begin
  Result := TblDetail.FieldByName('COST_CENTRE').AsString;
  if (Result = ' ') then
    Result := '';
end;

function TLKDaybookAccounts.GetTblDDescription: string;
begin
  Result := TblDetail.FieldByName('DESCRIPTION').AsString;
  if (Result = ' ') then
    Result := '';
end;

function TLKDaybookAccounts.GetTblDNominalCode: string;
begin
  Result := TblDetail.FieldByName('NOMINAL_CODE').AsString;
end;

procedure TLKDaybookAccounts.PbCheckBatchItemDetailIsValid;
var
  I: Integer;
  Item: string;
  DocRef: string;
  NomCode: string;
  CostCentre: string;
begin
  inherited PbCheckBatchItemDetailIsValid;
  DocRef := TblMDocumentReference;
  Item := IntToStr(TblDDocumentItemNumber);
  NomCode := TblDNominalCode;
  CostCentre := TblDCostCentre;
  with TblDetail, DlgDaybookErrors do
  begin
    if (NomCode = '') then
    begin
      for I := 2 to FieldCount - 1 do
        if not Fields[I].isNull then
        begin
          AddError(['201', DocRef, Item, 'Nominal code is blank',
            'Display & correct']);
          exit;
        end;
      AddError(['200', DocRef, Item, 'Blank Item', 'Delete']);
      exit;
    end;
    if not NominalLedger.AccountIsValid(VarArrayOf([NomCode, CostCentre]), False) then
      AddError(['201', DocRef, Item, 'Incorrect code ' + NomCode + ' ' + CostCentre,
        'Display & correct']);
  end;
end;

procedure TLKDaybookAccounts.PbCheckBatchItemMasterIsValid;
begin
  if not TblDetail.FindFirst then
    FDlgDaybookErrors.AddError(['100', TblMaster.Fields[0].asString, '',
      'has no detail records', 'delete record']);
end;

procedure TLKDaybookAccounts.PbTablesApplyUpdates;
begin
  if Assigned(NominalLedger) then
    with NominalLedger do
    begin
      ApplyUpdates(0);
      DeferredPostings.ApplyUpdates(0);
      if Assigned(CostLedger) then
        CostLedger.ApplyUpdates(0);
    end;
  inherited PbTablesApplyUpdates;
end;

procedure TLKDaybookAccounts.PbTablesCancelUpdates;
begin
  if Assigned(NominalLedger) then
    with NominalLedger do
    begin
      CancelUpdates;
      DeferredPostings.CancelUpdates;
      if Assigned(CostLedger) then
        CostLedger.CancelUpdates;
    end;
  inherited PbTablesCancelUpdates;
end;

procedure TLKDaybookAccounts.PbTablesRefresh;
begin
  if Assigned(NominalLedger) then
    with NominalLedger do
    begin
      TblDetail.Refresh;
      DeferredPostings.Refresh;
      if Assigned(CostLedger) then
        CostLedger.TblDetail.Refresh;
    end;
  inherited PbTablesRefresh;
end;

procedure TLKDaybookAccounts.SetAccountsHub(Value: TLKHubAccounts);
begin
  inherited AccountsHub := Value;
end;

procedure TLKDaybookAccounts.SetTblDCostCentre(const Value: string);
begin
  TblDetail.FieldByName('COST_CENTRE').AsString := Value;
end;

procedure TLKDaybookAccounts.SetTblDDescription(const Value: string);
begin
  TblDetail.FieldByName('DESCRIPTION').AsString := Value;
end;

procedure TLKDaybookAccounts.SetTblDNominalCode(const Value: string);
begin
  TblDetail.FieldByName('NOMINAL_CODE').AsString := Value;
end;



{
**************************** TLKNominalPostingList *****************************
}
procedure TLKNominalPostingList.Accumulate(const ANominalCode, ACostCentre,
    ADescription: string; AValue: currency);
var
  I: Integer;
  RecFound: Boolean;
  MyPosting: TLKNominalPosting;
begin
  RecFound := False;
  for I := 0 to Count - 1 do
    with Items[I] do
      if (NominalCode = ANominalCode) and (CostCentre = ACostCentre) then
      begin
        if (length(Description) < 33) then
          Description := Description + ',' + ADescription;
        Value := Value + AValue;
        RecFound := True;
        break;
      end;
  if not RecFound then
  begin
    MyPosting := TLKNominalPosting.Create(nil);
    Add(MyPosting);
    with MyPosting do
    begin
      NominalCode := ANominalCode;
      CostCentre := ACostCentre;
      Description := ADescription;
      Value := AValue;
    end;
  end;
end;

function TLKNominalPostingList.Add(AComponent: TLKNominalPosting): Integer;
begin
  Result := inherited Add(AComponent);
end;

function TLKNominalPostingList.Extract(Item: TLKNominalPosting):
    TLKNominalPosting;
begin
  Result := TLKNominalPosting(inherited Extract(Item));
end;

function TLKNominalPostingList.First: TLKNominalPosting;
begin
  Result := TLKNominalPosting(inherited First);
end;

function TLKNominalPostingList.GetItems(Index: Integer): TLKNominalPosting;
begin
  Result := TLKNominalPosting(inherited Items[Index]);
end;

function TLKNominalPostingList.IndexOf(AComponent: TLKNominalPosting): Integer;
begin
  Result := inherited IndexOf(AComponent);
end;

procedure TLKNominalPostingList.Insert(Index: Integer; AComponent:
    TLKNominalPosting);
begin
  inherited Insert(Index, AComponent);
end;

function TLKNominalPostingList.Last: TLKNominalPosting;
begin
  Result := TLKNominalPosting(inherited Last);
end;

procedure TLKNominalPostingList.SetItems(Index: Integer; AComponent:
    TLKNominalPosting);
begin
  inherited Items[Index] := AComponent;
end;


{
****************************** TLKRecordsAccounts ******************************
}
function TLKRecordsAccounts.GetAccountsHub: TLKHubAccounts;
begin
  Result := TLKHubAccounts(inherited AccountsHub);
end;

procedure TLKRecordsAccounts.SetAccountsHub(Value: TLKHubAccounts);
begin
  inherited AccountsHub := Value;
end;


{
************************* TLKContainedRecordsAccounts **************************
}
function TLKContainedRecordsAccounts.GetAccountsHub: TLKHubAccounts;
begin
  Result := TLKHubAccounts(inherited AccountsHub);
end;

procedure TLKContainedRecordsAccounts.SetAccountsHub(Value: TLKHubAccounts);
begin
  inherited AccountsHub := Value;
end;


{
********************************* TLKDocuments *********************************
}
procedure TLKDocuments.DoDetailAfterPost;
begin
  NominalLedger.CostLedger.CostCentreCurrent := TDCostCentre;
  inherited DoDetailAfterPost;
end;

function TLKDocuments.GetAccountsHub: TLKHubAccounts;
begin
  Result := TLKHubAccounts(inherited AccountsHub);
end;

function TLKDocuments.GetNominalLedger: TLKLedgerNominal;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.NominalLedger
  else
    Result := nil;
end;

function TLKDocuments.GetTDCostCentre: string;
begin
  Result := TblDetail.FieldByName('COST_CENTRE').AsString;
  if (Result = ' ') then
    Result := '';
end;

function TLKDocuments.GetTDNominalCode: string;
begin
  Result := TblDetail.FieldByName('NOMINAL_CODE').AsString;
end;

procedure TLKDocuments.SetAccountsHub(Value: TLKHubAccounts);
begin
  inherited AccountsHub := Value;
end;

procedure TLKDocuments.SetTDCostCentre(const Value: string);
begin
  TblDetail.FieldByName('COST_CENTRE').AsString := Value;
end;

procedure TLKDocuments.SetTDNominalCode(const Value: string);
begin
  TblDetail.FieldByName('NOMINAL_CODE').AsString := Value;
end;


{
********************* TLKLedgerAnalysisContainedDocuments **********************
}
constructor TLKLedgerAnalysisContainedDocuments.Create(AOwner: TComponent);
begin
  if not (AOwner is TLKLedgerAnalysis) then
    raise LKError.Create('Program error: '
      + 'LKLedgerAnalysisContainedDocuments.Owner should be a TLKLedgerAnalysis');
  inherited Create(AOwner);
end;

function TLKLedgerAnalysisContainedDocuments.GetContainerComponent:
    TLKLedgerAnalysis;
begin
  Result := TLKLedgerAnalysis(inherited ContainerComponent);
end;


{
*********************** TLKLedgerAnalysisAccountsPeriods ***********************
}
function TLKLedgerAnalysisAccountsPeriods.GettmPeriodEnd: TDate;
begin
  Result := GetDate(TblMaster.FieldByName('PERIOD_END'));
end;

procedure TLKLedgerAnalysisAccountsPeriods.SettmPeriodEnd(Value: TDate);
begin
  TblMaster.FieldByName('PERIOD_END').AsDateTime := Value;
end;



{
************************ TLKLedgerAnalysisReportLayouts ************************
}
constructor TLKLedgerAnalysisReportLayouts.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
end;

function TLKLedgerAnalysisReportLayouts.DoSetup(const ForceIfSettingsExist:
    Boolean): TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  case Result of
    mrAbort: exit;
  end;
  if not ForceIfSettingsExist then
    exit;
  Result := View(Null, True, [mbOK, mbAbort]);

  {if (Result <> mrOK) then exit;
  showmessage('You should enter the main Management Accounts report heads of your '
   + NameFriendly);
  showmessage('Layout head codes may be up to 8 characters and are sorted alphabetically.'
   + ' You may just like to use single letters A to Z');
  Result := ReportLayouts.ViewRecord(Null,True,[mbOK,mbAbort]);
  case Result of mrAbort : exit; end;}
end;

function TLKLedgerAnalysisReportLayouts.GettmDescription: string;
begin
  Result := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

procedure TLKLedgerAnalysisReportLayouts.SettmDescription(const Value: string);
begin
  TblMaster.FieldByName('DESCRIPTION').AsString := Value;
end;


{
************************* TLKLedgerAnalysisReportLines *************************
}
constructor TLKLedgerAnalysisReportLines.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
end;

procedure TLKLedgerAnalysisReportLines.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerAnalysisReportLines, FFormView);
  TFormLKLedgerAnalysisReportLines(FormView).Construct(Self);
end;

procedure TLKLedgerAnalysisReportLines.DlgImportExportExecute(AOwner:
    TComponent; const AsExportDlg, Autoexecute: boolean);
var
  Okay: Boolean;
  S: string;
  ReportType: string;
  LayoutHead: string;
  LineNumber: string;
  CSVFileName: string;
  CSVFile: TextFile;

  function BlankFields(const NoOfFields: Integer): string;
  var
    I: Integer;
  begin
    Result := '"';
    for I := 1 to NoOfFields do
      Result := Result + ',""';
  end;

  function GetNextField: string;
  var
    I: Integer;
  begin
    if (Pos('"', S) = 1) then
    begin
      Result := Copy(S, 2, Pos('",', S) - 2);
      System.Delete(S, 1, Length(Result) + 3);
    end
    else
    begin
      I := Pos(',', S);
      if I = 0 then
        Result := S
      else
        Result := Copy(S, 1, I - 1);
      System.Delete(S, 1, Length(Result) + 1);
    end;
    Result := Trim(Result);
  end;

  function IntFieldToYesNo(F: TField): string;
  begin
    if (GetInteger(F) = 0) then
      Result := 'No'
    else Result := 'Yes';
  end;

  function YesNoToInt(const S: string): Integer;
  begin
    if (S = 'Yes') then
      Result := 1
    else Result := 0;
  end;

begin
  if AsExportDlg then
  begin
    CSVFileName := 'AccountsLayout';
    if AutoExecute then
    begin
      S := IntToStr(Trunc(Now * 1000));
      System.Delete(S, 1, Length(S) - 7);
      CSVFileName := CSVFileName + S;
    end
    else
    begin
      with TLKViewDialog.Create(Application) do
      try
        DataType := vString;
        Title := NameFriendly + 'Report layout Export';
        HeaderCaption := 'Enter the export file name';
        HeaderComments := 'The export CSV file can be opened with your '
          + 'spreadsheet application (e.g. Excel).' + Chr(10)
          + 'The file will be saved to ' + InstallFolder;
        Value := CSVFileName;
        Okay := Execute;
        if (Value > '') then
          CSVFileName := Value;
      finally Free;
      end;
      if not Okay then
        Exit;
    end;
    CSVFileName := InstallFolder + PathDelim + CSVFileName + '.csv';
    AssignFile(CSVFile, CSVFileName);
    Rewrite(CSVFile);
    Screen.Cursor := crHourglass;
    MasterComponent.ExternalDisableControls;
    ExternalDisableControls;
    try
      WriteLn(CSVFile,'"The records in this file are defined by a letter '
      + 'A to D in the first (A) column as follows'+BlankFields(6));
      WriteLn(CSVFile,'" '+ BlankFields(6));
      WriteLn(CSVFile,'"A -'+'","'+'The Report reference'+'","'
      + 'The Report Title'+'","'+'Is it a Profit & Loss type report (Yes or '
      + 'No; No = Balance sheet Type)'
      + BlankFields(3));
      WriteLn(CSVFile,'"     - In the Nominal Ledger Report reference ''A'' is '
      + 'defined as a full Profit and Loss Account and type ''N'' is a full '
      + 'Balance Sheet' + BlankFields(6));
      WriteLn(CSVFile,'"     - Other report references can be used to re-order '
      + 'the A or N reports or to print subsets' + BlankFields(6));
      WriteLn(CSVFile,'"B -'+'","'+'Subsection reference (Alphabetical)'+'","'
      + 'Subsection Title'+'","'+'Normally a Debit (Yes or No)'+'","'
      + 'Zero running total (Yes or No)'+'","'+'Subtotal Title'+'","'
      +'Subtotal normally a Debit (Yes or No)"');
      WriteLn(CSVFile,'"C -'+'","'+'Summary line reference (Alphabetical)'+'","'
      + 'Summary line Title' + BlankFields(4));
      WriteLn(CSVFile,'"D -'+'","'+'Nominal Code (Alphabetical)'+'","'+'Detail '
      + 'line Title (Nominal Account name)'+'","'+'default Sales Tax code'+'","'
      +'Variable percent for Budgeting' + BlankFields(2));
      WriteLn(CSVFile,'" '+ BlankFields(6));
      WriteLn(CSVFile,'"Anything other than A to D in the first column '
      + 'signifies a comment' + BlankFields(6));
      WriteLn(CSVFile,'"There should be only one A type record in the '
      + ' file which is the first record.' + BlankFields(6));
      WriteLn(CSVFile,'"A type A record must be followed by a B; a B '
      + 'by a C and a C by a D type record.' + BlankFields(6));
      WriteLn(CSVFile,'"A D type record can be followed by a D C or B type '
      + 'record. The last record in the file must be a type D'+ BlankFields(6));
      WriteLn(CSVFile,'" '+ BlankFields(6));
      WriteLn(CSVFile,'" '+ BlankFields(6));
      with MasterComponent.TblMaster do if FindFirst then
      repeat
        WriteLn(CSVFile,'"A","'+Fields[0].AsString+'","'+Fields[1].AsString
        +'","' + IntFieldToYesNo(Fields[2]) + BlankFields(3));
        with MasterComponent.TblDetail do if FindFirst then
        repeat
          WriteLn(CSVFile,'" '+ BlankFields(6));
          WriteLn(CSVFile, '"B","'+ Fields[1].AsString+'","' + Fields[2].AsString
          + '","' + IntFieldToYesNo(Fields[3]) + '","'+ IntFieldToYesNo(Fields[4])
          + '","' + Fields[5].AsString + '","' + IntFieldToYesNo(Fields[6]) +'"');
          with Self.TblMaster do if FindFirst then
          repeat
            WriteLn(CSVFile,'" '+ BlankFields(6));
            WriteLn(CSVFile, '"C","'+ Fields[2].AsString+'","'
            + Fields[3].AsString + BlankFields(4));
            with Self.TblDetail do if FindFirst then
            repeat
              WriteLn(CSVFile,'"D","'+ Fields[0].AsString+'","'
              + Fields[2].AsString + '","' + Fields[3].AsString+'","'
              + Fields[4].AsString + BlankFields(2));
            until not FindNext;
          until not FindNext;
        until not FindNext;
      until not FindNext;
    finally CloseFile(CSVFile);
      Screen.Cursor := crDefault;
      MasterComponent.ExternalEnableControls;
      ExternalEnableControls;
      MasterComponent.TblMaster.First;
    end;
    if (MessageDlg('Export success! - Open file now?', mtInformation,
      [mbYes, mbNo],0) = mrYes) then
      ShellExecute(Application.Handle, 'open', PChar(CSVFileName), nil, nil, 1);
    end
  else
  begin
    with TOpenDialog.Create(Application) do
    try
      Title := 'Select Report layout Import file';
      DefaultExt := '.csv';
      Filter := 'Comma Separated Variable (CSV)|*.csv';
      InitialDir := InstallFolder;
      Options := Options + [ofFileMustExist];
      Okay := Execute;
      CSVFileName := FileName;
    finally Free;
    end;
    if not Okay then
      Exit;
    AssignFile(CSVFile, CSVFileName);
    Reset(CSVFile);
    Screen.Cursor := crHourglass;
    MasterComponent.ExternalDisableControls;
    ExternalDisableControls;
    try
      repeat
        ReadLn(CSVFile, S);
        ReportType := GetNextField;
        if ReportType = 'A' then
        begin
          ReportType := GetNextField;
          if MasterComponent.Locate('LAYOUT_TYPE',ReportType,[]) then
          begin
            if (MessageDlg('Report type ' + ReportType
              + ' already exists: confirm overwrite?', mtInformation,
              [mbYes,mbNo],0) <> mrYes) then
                Abort;
            DeleteDetailWithMasterRecord :=
              ((ReportType = 'A') or (ReportType = 'N'));
            with MasterComponent.TblDetail do if FindLast then
            repeat
              with Self.TblMaster do if FindLast then
              repeat
                Delete;
              until not FindLast;
              Delete;
            until not FindLast;
            MasterComponent.TblMaster.Delete;
            Self.ApplyUpdates(0);
            MasterComponent.ApplyUpdates(0);
          end;
          MasterComponent.TblMaster.InsertRecord(
            [ReportType, GetNextField, YesNoToInt(GetNextField)]);
        end else
        if ReportType = 'B' then
        begin
          LayoutHead := GetNextField;
          MasterComponent.TblDetail.InsertRecord(
            [ReportType, LayoutHead, GetNextField, YesNoToInt(GetNextField),
              YesNoToInt(GetNextField), GetNextField, YesNoToInt(GetNextField)]);
        end else
        if ReportType = 'C' then
        begin
          LineNumber := GetNextField;
          TblMaster.InsertRecord(
            [ReportType, LayoutHead, LineNumber, GetNextField]);
        end else
        if ReportType = 'D' then
        begin
          TblDetail.InsertRecord(
            [GetNextField, LineNumber, GetNextField,
              GetNextField, GetNextField]);
        end;
      until Eof(CSVFile);
      Self.ApplyUpdates(0);
      MasterComponent.ApplyUpdates(0);
    finally
      CloseFile(CSVFile);
      Screen.Cursor := crDefault;
      MasterComponent.ExternalEnableControls;
      MasterComponent.TblDetail.First;
      ExternalEnableControls;
    end;
  end;
end;

function TLKLedgerAnalysisReportLines.DoSetup(const ForceIfSettingsExist:
    Boolean): TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  case Result of
    mrAbort: exit;
  end;
  if not ForceIfSettingsExist then
    exit;
  Result := View(Null, True, [mbOK, mbAbort]);

  {if (Result <> mrOK) then exit;
  showmessage('You should now enter the line descriptions on your Management Accounts');
  //  + ' the Nominal Codes which make them up e.g with "Printing and Stationery" you'
  //  + ' could enter Nominal Codes for "Printing" and for "Stationery"');
  //showmessage('Line numbers could be the Layout head code + "10" to "90" allowing a 10 spacing'
  //  + ' the Nominal Codes could be the Line number + "00" to "90" allowing a 10 spacing,'
  //  + ' e.g. Layout head "J", Line number "J30", Nominal code "J3040"');
  Result := ReportLines.View(Null,True,[mbOK,mbAbort]);}
end;

function TLKLedgerAnalysisReportLines.GettmDescription: string;
begin
  Result := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

procedure TLKLedgerAnalysisReportLines.SettmDescription(const Value: string);
begin
  TblMaster.FieldByName('DESCRIPTION').AsString := Value;
end;

function TLKLedgerAnalysisReportLines.View(const KeyValues: Variant;
    ShowAsModal: Boolean; Buttons: TMsgDlgButtons): TModalResult;
begin
  Result := inherited View(KeyValues, ShowAsModal, Buttons);
  ContainerComponent.RefreshAccountCodes;
end;


{
********************* TLKLedgerAnalysisManagementAccounts **********************
}
constructor TLKLedgerAnalysisManagementAccounts.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

procedure TLKLedgerAnalysisManagementAccounts.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerAnalysisManagementAccounts, FFormView);
  TFormLKLedgerAnalysisManagementAccounts(FormView).Construct(Self);
end;

procedure TLKLedgerAnalysisManagementAccounts.DlgManagementAccountsPrintExecute;
begin
  // override as required
end;

function TLKLedgerAnalysisManagementAccounts.EndOfMonth: TModalResult;
begin
  Result := inherited EndOfMonth;
  if Result = mrOK then
    Result := View(Null, True, [mbOK]);
end;

function TLKLedgerAnalysisManagementAccounts.GetTblDPeriodEnd: TDate;
begin
  Result := GetDate(TblDetail.FieldByName('PERIOD_END'));
end;

procedure TLKLedgerAnalysisManagementAccounts.ManagementAccountsCalculate;
var
  AccountsDate: TDateTime;
  ThisPeriod: Currency;
  YearToDate: Currency;

  procedure CalculateManagementAccounts;
  var
    AccountsStart: TDateTime;
    PeriodEnd: TDateTime;
  begin
    AccountsStart := AccountsHub.AccountsPeriodStart;
//    TblDetail.Filter := 'PERIOD_END > '''
//          + DateToStr(AccountsHub.AccountsPeriodStart)
//          + ''' AND PERIOD_END <= ''' + DateToStr(AccountsDate) + '''';
    TblDetail.Filtered := False;
    Screen.Cursor := crHourGlass;
    try
    if ContainerComponent.ReportLayouts.TblDetail.FindFirst then
      repeat
        if ContainerComponent.ReportLines.TblMaster.FindFirst then
          repeat
            if ContainerComponent.ReportLines.TblDetail.FindFirst then
              repeat
                  if TblMaster.FindFirst then
                  repeat
                    with TblMaster do
                    begin
                      // P&L + BS does not work for P&L code on Balance Sheet
                      YearToDate := GetCurrency(FieldByName('BALANCE_SHEET'));
                      if CurrencyIsZero(YearToDate) then // P&L item
                        YearToDate :=
                          GetCurrency(FieldByName('PROFIT_AND_LOSS'));
                      // movement this year in total
                      ThisPeriod := YearToDate -
                        GetCurrency(FieldByName('OPENING_BALANCE'));
                      // this month is this period less previous current periods
                    end;
                    with TblDetail do
                    begin
                      PeriodEnd := AccountsStart;
                      if FindFirst then
                      repeat
                        PeriodEnd := GetDate(FieldByName('PERIOD_END'));
                        if (PeriodEnd >= AccountsDate) then
                          Break;
                        if (PeriodEnd > AccountsStart) then
                          ThisPeriod := ThisPeriod -
                            GetCurrency(FieldByName('THIS_MONTH'));
                      until not FindNext
                        or (GetDate(FieldByName('PERIOD_END')) > AccountsDate);
                      if (PeriodEnd = AccountsDate) then
                        Edit
                      else
                      begin
                        Insert;
                        TblDPeriodEnd := AccountsDate;
                      end;
                      FieldByName('THIS_MONTH').AsCurrency := ThisPeriod;
                      FieldByName('YEAR_TO_DATE').AsCurrency := YearToDate;
                      Post;
                    end;
                  until not TblMaster.FindNext;
              until not ContainerComponent.ReportLines.TblDetail.FindNext;
          until not ContainerComponent.ReportLines.TblMaster.FindNext;
      until not ContainerComponent.ReportLayouts.TblDetail.FindNext;
    ApplyUpdates(0);
    finally
      Screen.Cursor := crDefault;
    end;
  end;

begin
  AccountsDate := ContainerComponent.LedgerDate;
  with TLKViewDialog.Create(Self) do
  begin
    try
      DataType := vDate;
      Value := AccountsDate;
      Title := 'Calculate ' + NameFriendly;
      HeaderCaption := 'Confirm Accounts Date to calculate';
      HeaderComments := 'If this is not the current accounting period then you'
        + ' should recalculate later periods once you are satisfied that the'
        + ' Accounts are correct';
      if not Execute then
        SysUtils.Abort;
      AccountsDate := Value;
    finally Free;
    end;
  end;
  Screen.Cursor := crHourglass;
  try
    // Insert the accounts date in AccountsPeriods
    with ContainerComponent.AccountsPeriods do
      if not Locate('PERIOD_END',AccountsDate,[]) then
      begin
        InsertRecord([AccountsDate]);
        ApplyUpdates(0);
      end;

    // Calculate P&L and Balance Sheet year to date figures
    ContainerComponent.TrialBalanceCalculate(AccountsDate);

    with ContainerComponent.ReportLayouts do
    begin
      if Locate('LAYOUT_TYPE','A',[]) then
        CalculateManagementAccounts
      else
        raise LKError.Create('Report Layout "A" not found. It must contain all'
          + ' Profit and Loss Nominal Codes');
      if Locate('LAYOUT_TYPE','N',[]) then
        CalculateManagementAccounts
      else
        raise LKError.Create('Report Layout "N" not found. It must contain all'
          + ' Balance Sheet Nominal Codes');
    end;
  finally
    Screen.cursor := crDefault;
  end;
  Showmessage(NameFriendly + ' Management accounts calculated for ' +
    DateToStr(AccountsDate));
end;

procedure TLKLedgerAnalysisManagementAccounts.SetTblDPeriodEnd(Value: TDate);
begin
  TblDetail.FieldByName('PERIOD_END').AsDateTime := Value;
end;


{
****************************** TLKLedgerAnalysis *******************************
}
constructor TLKLedgerAnalysis.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
  BalanceFieldName := 'ACCOUNT_BALANCE';
  SetBalanceFieldPairs;
  with ManagementAccounts do
  begin
    MasterComponent := ReportLines;
    MasterTable := tbDetail;
  end;
  with ReportLines do
  begin
    MasterComponent := ReportLayouts;
  end;
end;

procedure TLKLedgerAnalysis.CreateSubComponents;
begin
  inherited CreateSubComponents;
  FAccountsPeriods := TLKLedgerAnalysisAccountsPeriods.Create(Self);
  FDeferredPostings := TLKLedgerAnalysisDeferredPostings.Create(Self);
  FReportLayouts := TLKLedgerAnalysisReportLayouts.Create(Self);
  FReportLines := TLKLedgerAnalysisReportLines.Create(Self);
  CreateManagementAccounts;
end;

procedure TLKLedgerAnalysis.CreateViewForm;
begin
  Application.CreateForm(TFormLKAnalysisLedger, FFormView);
  TFormLKAnalysisLedger(FormView).Construct(Self);
end;

function TLKLedgerAnalysis.DoSetup(const ForceIfSettingsExist: Boolean):
    TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  case Result of
    mrAbort: exit;
  end;
  if not (ForceIfSettingsExist or (Status = acsNotSetUp)) then
    exit;
  Open;
  with TFormLKLedgerAnalysisSetupWiz.Create(nil) do
  try
    Construct(Self);
    Result := ShowModal;
  finally Free;
  end;
end;

function TLKLedgerAnalysis.EndOfYear(ProgressText: string): TModalResult;
var
  D: Currency;
begin
  Result := inherited EndOfYear(ProgressText);
  if (Result <> mrOK) then
    Exit;
  Screen.Cursor := crHourglass;
  ProgressText := 'Clearing Trial Balance to Previous Year';
  Application.ProcessMessages;
  try
    with TblMaster do
    begin
      if FindFirst then
        repeat
          D := GetCurrency(FieldByName('BALANCE_SHEET')) +
            GetCurrency(FieldByName('PROFIT_AND_LOSS'));
          Edit;
          FieldByName('OPENING_RESERVES').AsCurrency := 0;
          FieldByName('OPENING_BALANCE').AsCurrency := 0;
          FieldByName('ACCOUNT_BALANCE').AsCurrency := 0;
          FieldByName('RESERVES').AsCurrency := 0;
          FieldByName('PROFIT_AND_LOSS').AsCurrency := 0;
          FieldByName('BALANCE_SHEET').AsCurrency := 0;
          FieldByName('PREVIOUS_PERIOD').AsCurrency := D;
          Post;
        until not FindNext;
      ApplyUpdates(0);
      ProgressText := 'Removing unused Accounts from ' + NameFriendly;
      Application.ProcessMessages;
      if FindLast then
        repeat
          if CurrencyIsZero(GetCurrency(FieldByName('PREVIOUS_PERIOD'))) then
            Delete;
        until not FindPrior;
      ApplyUpdates(0);
    end;
    ManagementAccounts.Refresh;
  finally Screen.Cursor := crDefault;
  end;
end;

procedure TLKLedgerAnalysis.SetBalanceFieldPairs;
begin
  with FMSumDPairFieldNames do
  begin
    Clear;
    if (BalanceFieldName = '') then
      exit;
    add(BalanceFieldName);
    add('+DEBIT');
    add(BalanceFieldName);
    add('-CREDIT');
  end;
end;

procedure TLKLedgerAnalysis.SetContainedComponentNames;
begin
  inherited SetContainedComponentNames;
  FAccountsPeriods.Name := Name + 'AccountsPeriods';
  FDeferredPostings.Name := Name + 'DeferredPostings';
  FReportLayouts.Name := Name + 'ReportLayouts';
  FReportLines.Name := Name + 'ReportLines';
  FManagementAccounts.Name := Name + 'ManagementAccounts';
end;

procedure TLKLedgerAnalysis.TrialBalanceCalculate(const AccountsDate:
    TDateTime);
begin
  if (AccountsDate < LedgerDate - 0.5) then
  begin
    TblDetail.Filter := 'DATE_ <= ''' + DateToStr(AccountsDate) + '''';
    TblDetail.Filtered := True;
    if TblMaster.FindFirst then
      repeat
        AccountCheckTotals(False, True, False);
      until not TblMaster.FindNext;
  end;
end;


{
********************** TLKLedgerNominalContainedDocuments **********************
}
constructor TLKLedgerNominalContainedDocuments.Create(AOwner: TComponent);
begin
  if not (AOwner is TLKLedgerNominal) then
    raise LKError.Create('Program error: '
      + 'LKLedgerNominalContainedDocuments.Owner should be a TLKLedgerNominal');
  inherited Create(AOwner);
end;

function TLKLedgerNominalContainedDocuments.GetContainerComponent:
    TLKLedgerNominal;
begin
  Result := TLKLedgerNominal(inherited ContainerComponent);
end;


{
************************* TLKLedgerNominalCodesLookup **************************
}
procedure TLKLedgerNominalCodesLookup.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKDocumentsInputNominalRecord, FFormViewRecord);
  TFormLKDocumentsInputNominalRecord(FormViewRecord).Construct(Self,
    ContainerComponent);
end;


{
********************* TLKLedgerNominalFixedAssetCategories *********************
}
procedure TLKLedgerNominalFixedAssetCategories.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerNominalFixedAssteCategoriesRecord,
    FFormViewRecord);
  TFormLKLedgerNominalFixedAssteCategoriesRecord(FormViewRecord).Construct(Self);
end;

function TLKLedgerNominalFixedAssetCategories.GettmAssetLife: Double;
begin
  Result := GetFloat(TblMaster.FieldByName('ASSET_LIFE_IN_YEARS'));
end;

function TLKLedgerNominalFixedAssetCategories.GettmDepreciationNominalCodeBS:
    string;
begin
  Result := TblMaster.FieldByName('DEPRECIATION_NOMINAL_CODE_BS').AsString;
end;

function
    TLKLedgerNominalFixedAssetCategories.GettmDepreciationNominalCodePandL:
    string;
begin
  Result := TblMaster.FieldByName('DEPRECIATION_NOMINAL_CODE_PL').AsString;
end;

function TLKLedgerNominalFixedAssetCategories.GettmDepreciationPercent: Double;
begin
  Result := GetFloat(TblMaster.FieldByName('DEPRECIATION_RATE_PERCENT_PA'));
end;

function TLKLedgerNominalFixedAssetCategories.GettmDepreciationStyle:
    TLKFixedAssetDepreciationStyle;
begin
  Result := TLKFixedAssetDepreciationStyle(
    GetInteger(TblMaster.FieldByName('DEPRECIATION_METHOD')));
end;

function TLKLedgerNominalFixedAssetCategories.GettmDescription: string;
begin
  Result := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

procedure TLKLedgerNominalFixedAssetCategories.SettmAssetLife(Value: Double);
begin
  TblMaster.FieldByName('ASSET_LIFE_IN_YEARS').AsFloat := Value;
end;

procedure TLKLedgerNominalFixedAssetCategories.SettmDepreciationNominalCodeBS(
    const Value: string);
begin
  TblMaster.FieldByName('DEPRECIATION_NOMINAL_CODE_BS').AsString := Value;
end;

procedure
    TLKLedgerNominalFixedAssetCategories.SettmDepreciationNominalCodePandL(
    const Value: string);
begin
  TblMaster.FieldByName('DEPRECIATION_NOMINAL_CODE_PL').AsString := Value;
end;

procedure TLKLedgerNominalFixedAssetCategories.SettmDepreciationPercent(Value:
    Double);
begin
  TblMaster.FieldByName('DEPRECIATION_RATE_PERCENT_PA').AsFloat := Value;
end;

procedure TLKLedgerNominalFixedAssetCategories.SettmDepreciationStyle(Value:
    TLKFixedAssetDepreciationStyle);
begin
  TblMaster.FieldByName('DEPRECIATION_METHOD').AsInteger := Ord(Value);
end;

procedure TLKLedgerNominalFixedAssetCategories.SettmDescription(const Value:
    string);
begin
  TblMaster.FieldByName('DESCRIPTION').AsString := Value;
end;


{
************************* TLKLedgerNominalNominalCodes *************************
}
procedure TLKLedgerNominalNominalCodes.DoMasterBeforePost;
begin
  if (tmVariableCostPercent > 100) or (tmVariableCostPercent < 0) then
    raise LKError.Create('Percent value must be in the range 0 to 100');
  inherited DoMasterBeforePost;
end;

function TLKLedgerNominalNominalCodes.GettmDefaultSalesTaxCode: Integer;
begin
  Result := GetInteger(TblMaster.FieldByName('DEFAULT_VAT_CODE'));
end;

function TLKLedgerNominalNominalCodes.GettmDescription: string;
begin
  Result := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

function TLKLedgerNominalNominalCodes.GettmTurnoverNominalCode: string;
begin
  Result := TblMaster.FieldByName('TURNOVER_NOMINAL_CODE').AsString;
end;

function TLKLedgerNominalNominalCodes.GettmVariableCostPercent: Double;
begin
  Result := GetFloat(TblMaster.FieldByName('VARIABLE_COST_PERCENT'));
end;

procedure TLKLedgerNominalNominalCodes.SettmDefaultSalesTaxCode(const Value:
    Integer);
begin
  TblMaster.FieldByName('DEFAULT_VAT_CODE').AsInteger := Value;
end;

procedure TLKLedgerNominalNominalCodes.SettmDescription(const Value: string);
begin
  TblMaster.FieldByName('DESCRIPTION').AsString := Value;
end;

procedure TLKLedgerNominalNominalCodes.SettmTurnoverNominalCode(const Value:
    string);
begin
  TblMaster.FieldByName('TURNOVER_NOMINAL_CODE').AsString := Value;
end;

procedure TLKLedgerNominalNominalCodes.SettmVariableCostPercent(const Value:
    Double);
begin
  TblMaster.FieldByName('VARIABLE_COST_PERCENT').AsFloat := Value;
end;


{
************************* TLKLedgerNominalReservesBase *************************
}
procedure TLKLedgerNominalReservesBase.CalculateAccountsReserves;
begin
end;

procedure TLKLedgerNominalReservesBase.CalculateBalances;
begin
  if (TblMaster.State in [dsInsert, dsEdit]) then
    TblMaster.Post;
  //clear any existing records
  while TblDetail.FindLast do
    TblDetail.Delete;
  ApplyUpdates(0);
  Refresh;
  // calculate in decendants
end;

procedure TLKLedgerNominalReservesBase.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerNominalReservesBaseInputRecord,
    FFormViewRecord);
  TFormLKLedgerNominalReservesBaseInputRecord(FormViewRecord).Construct(Self);
end;

function TLKLedgerNominalReservesBase.GettmDescription: string;
begin
  Result := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

procedure TLKLedgerNominalReservesBase.SettmDescription(const Value: string);
begin
  TblMaster.FieldByName('DESCRIPTION').AsString := Value;
end;


{
************************* TLKLedgerNominalFixedAssets **************************
}
procedure TLKLedgerNominalFixedAssets.CalculateAccountsReserves;
var
  ED: TDateTime;
  I: Integer;
  C: string;
  A: Currency;
  TotDepn: Currency;
  NPL: string;
  NBS: string;
  Dc: string;
  Da: string;

const SectionName = 'FADepn';

begin
  if ContainerComponent.FixedAssetCategories.TblMaster.FindFirst then
  begin
    ED := ContainerComponent.LedgerDate;
    with ContainerComponent.Reserves do
    begin
      RefreshMSumDPairs := False;
      if not TblMaster.Locate('RESERVE_REFERENCE', SectionName, []) then
      begin
        DecodeNominalAccount(ContainerComponent.SuspensePostingsCode, NBS, C);
        TblMaster.InsertRecord(
          [SectionName, ED, NBS, C, 'Fixed Asset Depreciation']);
      end;
      while TblDetail.FindLast do
        TblDetail.Delete;
      ApplyUpdates(0);
    end;
    I := 0;
    repeat
      TotDepn := 0;
      with ContainerComponent.FixedAssetCategories do
      begin
        NPL := tmDepreciationNominalCodePandL;
        NBS := tmDepreciationNominalCodeBS;
        Dc := tmDescription;
      end;
      if TblMaster.FindFirst then
        repeat
          Da := tmDescription;
          C := tmCostCentre;
          if TblDetail.Locate('DATE_', ED, []) or TblDetail.FindLast then
          begin
            A := tdBalance;
            ContainerComponent.Reserves.TblDetail.InsertRecord(
              [SectionName, I, NPL, C, Da, DebitOrNull(A), CreditOrNull(A)]);
            Inc(I);
            TotDepn := TotDepn - A;
          end;
        until not TblMaster.FindNext;
      ContainerComponent.Reserves.TblDetail.InsertRecord([SectionName, I, NBS,
        '', Dc, DebitOrNull(TotDepn), CreditOrNull(TotDepn)]);
      Inc(I);
    until not ContainerComponent.FixedAssetCategories.TblMaster.FindNext;
    ContainerComponent.Reserves.RefreshMSumDPairs := True;
    ContainerComponent.Reserves.ApplyUpdates(0);
  end;
end;

procedure TLKLedgerNominalFixedAssets.CalculateBalances;
var
  TotalDays: Double;
  Rate: Double;
  Cost: Currency;
  CurrDate: TDateTime;
  DateFrom: TDateTime;
  DateTo: TDateTime;
  DepreciationMethod: TLKFixedAssetDepreciationStyle;
  TotalDepreciation: Currency;
  Residual: Currency;
  CurrDepBalance: Currency;
  WDV: Currency;
begin
  inherited CalculateBalances;
  with ContainerComponent.FixedAssetCategories do
  begin
    DepreciationMethod := tmDepreciationStyle;
    case DepreciationMethod of
      fadNone:
        begin
          ShowMessage('No depreciation is charged on this category of asset');
          exit;
        end;
    end;
    Rate := tmDepreciationPercent / 100;
  end;
  Cost := tmAmount;
  Residual := tmDisposalValue;
  TotalDepreciation := Cost - Residual;
  WDV := Cost;
  CurrDepBalance := 0;
  DateFrom := tmDateFrom;
  DateTo := tmDateTo;
  if (DateTo = 0) then
  begin
    Showmessage('Date to not entered!');
    Exit;
  end;
  if (DateFrom = 0) then
  begin
    Showmessage('Date from not entered!');
    Exit;
  end;
  TotalDays := 1 + DateTo - DateFrom;
  if (TotalDays < 1) then
  begin
    Showmessage('Date to is before Date from!');
    Exit;
  end;
  CurrDate := AccountsHub.PeriodEndThis(DateFrom);
  while (CurrDate < DateTo) do
  begin
    case DepreciationMethod of
      fadStraightLine:
        begin
          CurrDepBalance := RoundCents(TotalDepreciation * (CurrDate - DateFrom)
            / TotalDays);
          WDV := Cost - CurrDepBalance;
        end;
      fadReducingBalance:
        begin
          CurrDepBalance := CurrDepBalance + RoundCents(WDV * Rate * (CurrDate -
            DateFrom) / 365);
          WDV := Cost - CurrDepBalance;
          DateFrom := CurrDate;
        end;
    end;
    // period end balance is fraction of period expired
    TblDetail.InsertRecord([nil, CurrDate, CurrDepBalance, WDV]);
    // next period
    CurrDate := CurrDate + 1;
    CurrDate := AccountsHub.PeriodEndThis(CurrDate);
  end;
  with TblDetail do
  begin
    InsertRecord([nil, CurrDate, TotalDepreciation, Residual]);
    ApplyUpdates(0);
    Refresh;
  end;
end;

procedure TLKLedgerNominalFixedAssets.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerNominalFixedAssetsRecord,
    FFormViewRecord);
  TFormLKLedgerNominalFixedAssetsRecord(FormViewRecord).Construct(Self);
end;

procedure TLKLedgerNominalFixedAssets.DoMasterBeforePost;
var
  D: TDateTime;
begin
  inherited DoMasterBeforePost;
  if TblMaster.FieldByName('ANTICIPATED_DISPOSAL_DATE').IsNull then
  begin
    D := tmDateFrom;
    if (D > 30000) then
      tmDateTo := Round(D + ContainerComponent.FixedAssetCategories.tmAssetLife
        * 365.25);
  end;
end;

function TLKLedgerNominalFixedAssets.GettdBalance: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('DEPRECIATION_TO_DATE'));
end;

function TLKLedgerNominalFixedAssets.GettmAmount: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('COST_OR_VALUATION'));
end;

function TLKLedgerNominalFixedAssets.GettmCostCentre: string;
begin
  Result := TblMaster.FieldByName('COST_CENTRE').AsString;
end;

function TLKLedgerNominalFixedAssets.GettmDateFrom: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('DATE_ACQUIRED'));
end;

function TLKLedgerNominalFixedAssets.GettmDateTo: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('ANTICIPATED_DISPOSAL_DATE'));
end;

function TLKLedgerNominalFixedAssets.GettmDisposalValue: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('ANTICIPATED_DISPOSAL_VALUE'));
end;

procedure TLKLedgerNominalFixedAssets.SettdBalance(Value: Currency);
begin
  TblDetail.FieldByName('DEPRECIATION_TO_DATE').AsCurrency := Value;
end;

procedure TLKLedgerNominalFixedAssets.SettmAmount(Value: Currency);
begin
  TblMaster.FieldByName('COST_OR_VALUATION').AsCurrency := Value;
end;

procedure TLKLedgerNominalFixedAssets.SettmCostCentre(const Value: string);
begin
  TblMaster.FieldByName('COST_CENTRE').AsString := Value;
end;

procedure TLKLedgerNominalFixedAssets.SettmDateFrom(Value: TDateTime);
begin
  TblMaster.FieldByName('DATE_ACQUIRED').AsDateTime := Value;
end;

procedure TLKLedgerNominalFixedAssets.SettmDateTo(Value: TDateTime);
begin
  TblMaster.FieldByName('ANTICIPATED_DISPOSAL_DATE').AsDateTime := Value;
end;

procedure TLKLedgerNominalFixedAssets.SettmDisposalValue(Value: Currency);
begin
  TblMaster.FieldByName('ANTICIPATED_DISPOSAL_VALUE').AsCurrency := Value;
end;


{
*************************** TLKLedgerNominalReserves ***************************
}
function TLKLedgerNominalReserves.GettdBalance: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('BALANCE'));
end;

function TLKLedgerNominalReserves.GettmAmount: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('AMOUNT'));
end;

function TLKLedgerNominalReserves.GettmCostCentre: string;
begin
  Result := TblMaster.FieldByName('COST_CENTRE').AsString;
end;

function TLKLedgerNominalReserves.GettmDateFrom: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('DATE_FROM'));
end;

function TLKLedgerNominalReserves.GettmDateTo: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('DATE_TO'));
end;

function TLKLedgerNominalReserves.GettmItemNo: Integer;
begin
  Result := GetInteger(TblMaster.FieldByName('ITEM_NUMBER'));
end;

function TLKLedgerNominalReserves.GettmNominalCode: string;
begin
  Result := TblMaster.FieldByName('NOMINAL_CODE').AsString;
end;

procedure TLKLedgerNominalReserves.SettdBalance(Value: Currency);
begin
  TblDetail.FieldByName('BALANCE').AsCurrency := Value;
end;

procedure TLKLedgerNominalReserves.SettmAmount(Value: Currency);
begin
  TblMaster.FieldByName('AMOUNT').AsCurrency := Value;
end;

procedure TLKLedgerNominalReserves.SettmCostCentre(const Value: string);
begin
  TblMaster.FieldByName('COST_CENTRE').AsString := Value;
end;

procedure TLKLedgerNominalReserves.SettmDateFrom(Value: TDateTime);
begin
  TblMaster.FieldByName('DATE_FROM').AsDateTime := Value;
end;

procedure TLKLedgerNominalReserves.SettmDateTo(Value: TDateTime);
begin
  TblMaster.FieldByName('DATE_TO').AsDateTime := Value;
end;

procedure TLKLedgerNominalReserves.SettmItemNo(Value: Integer);
begin
  TblMaster.FieldByName('ITEM_NUMBER').AsInteger := Value;
end;

procedure TLKLedgerNominalReserves.SettmNominalCode(const Value: string);
begin
  TblMaster.FieldByName('NOMINAL_CODE').AsString := Value;
end;


{
************************* TLKLedgerNominalPrepayments **************************
}
procedure TLKLedgerNominalPrepayments.CalculateAccountsReserves;
var
  ED: TDateTime;
  I: Integer;
  N, C, D: string;
  A: Currency;

const SectionName = 'Prepaid';

begin
  ED := ContainerComponent.LedgerDate;
  if TblMaster.FindFirst then
  begin
    with ContainerComponent.Reserves do
    begin
      RefreshMSumDPairs := False;
      if not TblMaster.Locate('RESERVE_REFERENCE', SectionName, []) then
      begin
        DecodeNominalAccount(ContainerComponent.CodePrepayments, N, C);
        TblMaster.InsertRecord([SectionName, ED, N, C, 'Prepayments']);
      end;
      while TblDetail.FindLast do
        TblDetail.Delete;
      ApplyUpdates(0);
    end;
    repeat
      if Self.TblDetail.Locate('DATE_', ED, []) or Self.TblDetail.FindLast then
      begin
        I := tmItemNo;
        N := tmNominalCode;
        C := tmCostCentre;
        D := tmDescription;
        A := -tdBalance;
        ContainerComponent.Reserves.TblDetail.InsertRecord(
          [SectionName, I, N, C, D, DebitOrNull(A), CreditOrNull(A)]);
      end;
    until not TblMaster.FindNext;
    ContainerComponent.Reserves.RefreshMSumDPairs := True;
    ContainerComponent.Reserves.ApplyUpdates(0);
  end;
end;

procedure TLKLedgerNominalPrepayments.CalculateBalances;
var
  TotalDays: Double;
  L, B: Currency;
  CurrDate: TDateTime;
  DateFrom: TDateTime;
  DateTo: TDateTime;
begin
  inherited CalculateBalances;
  L := tmAmount;
  DateFrom := tmDateFrom;
  DateTo := tmDateTo;
  if (DateTo = 0) then
  begin
    Showmessage('Date to not entered!');
    Exit;
  end;
  if (DateFrom = 0) then
  begin
    Showmessage('Date from not entered!');
    Exit;
  end;
  TotalDays := 1 + DateTo - DateFrom;
  if (TotalDays < 1) then
  begin
    Showmessage('Date to is before Date from!');
    Exit;
  end;
  if (TotalDays > 1830) then
    if (MessageDlg('Date to is more than 5 years after Date from',
      mtConfirmation, [mbOK, mbCancel], 0) = mrCancel) then
      Exit;
  CurrDate := AccountsHub.PeriodEndThis(DateFrom);
  while (CurrDate < DateTo) do
  begin
    // period end balance is fraction of period left
    B := RoundCents(L * (DateTo - CurrDate + 1) / TotalDays);
    TblDetail.InsertRecord([nil, CurrDate, B]);
    // next period
    CurrDate := CurrDate + 1;
    CurrDate := AccountsHub.PeriodEndThis(CurrDate);
  end;
  // final balance must be zero
  TblDetail.InsertRecord([nil, DateTo, 0]);
  ApplyUpdates(0);
//  Refresh;
end;


{
*************************** TLKLedgerNominalAccruals ***************************
}
procedure TLKLedgerNominalAccruals.CalculateAccountsReserves;
var
  ED: TDateTime;
  I: Integer;
  N, C, D: string;
  A: Currency;

const SectionName = 'Accrual';

begin
  ED := ContainerComponent.LedgerDate;
  if FindFirst then
  begin
    with ContainerComponent.Reserves do
    begin
      RefreshMSumDPairs := False;
      if not TblMaster.Locate('RESERVE_REFERENCE', SectionName, []) then
      begin
        DecodeNominalAccount(ContainerComponent.CodePrepayments, N, C);
        TblMaster.InsertRecord([SectionName, ED, N, C, 'Accruals']);
      end;
      while TblDetail.FindLast do
        TblDetail.Delete;
      ApplyUpdates(0);
    end;
    repeat
      if TblDetail.Locate('DATE_', ED, []) or TblDetail.FindLast then
      begin
        I := tmItemNo;
        N := tmNominalCode;
        C := tmCostCentre;
        D := tmDescription;
        A := tdBalance;
        ContainerComponent.Reserves.TblDetail.InsertRecord(
          [SectionName, I, N, C, D, DebitOrNull(A), CreditOrNull(A)]);
      end;
    until not FindNext;
    ContainerComponent.Reserves.RefreshMSumDPairs := True;
    ContainerComponent.Reserves.ApplyUpdates(0);
  end;
end;

procedure TLKLedgerNominalAccruals.CalculateBalances;
var
  TotalDays: Double;
  L, B: Currency;
  CurrDate: TDateTime;
  DateFrom: TDateTime;
  DateTo: TDateTime;
begin
  inherited CalculateBalances;
  L := tmAmount;
  DateFrom := tmDateFrom;
  DateTo := tmDateTo;
  if (DateTo = 0) then
  begin
    Showmessage('Date to not entered!');
    Exit;
  end;
  if (DateFrom = 0) then
  begin
    Showmessage('Date from not entered!');
    Exit;
  end;
  TotalDays := 1 + DateTo - DateFrom;
  if (TotalDays < 1) then
  begin
    Showmessage('Date to is before Date from!');
    Exit;
  end;
  if (TotalDays > 1830) then
    if (MessageDlg('Date to is more than 5 years after Date from',
      mtConfirmation, [mbOK, mbCancel], 0) = mrCancel) then
      Exit;
  CurrDate := AccountsHub.PeriodEndThis(DateFrom);
  while (CurrDate <= DateTo) do
  begin
    // period end balance is fraction of period expired
    B := RoundCents(L * (CurrDate - DateFrom + 1) / TotalDays);
    TblDetail.InsertRecord([nil, CurrDate, B]);
    // next period
    CurrDate := AccountsHub.PeriodEndThis(CurrDate+1);
  end;
  ApplyUpdates(0);
//  Refresh;
end;


{
************************ TLKLedgerNominalEnterBalances *************************
}
constructor TLKLedgerNominalEnterBalances.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
  with FMSumDPairFieldNames do
  begin // Total needs to be the double entry posting of the difference
    Add('TOTAL');
    Add('-DEBIT');
    Add('TOTAL');
    Add('+CREDIT');
  end;
end;

procedure TLKLedgerNominalEnterBalances.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerNominalEnterBalances, FFormView);
  TFormLKLedgerNominalEnterBalances(FormView).Construct(Self);
end;

function TLKLedgerNominalEnterBalances.DoSetup(const ForceIfSettingsExist:
    Boolean): TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  case Result of
    mrAbort: Exit;
  end;
  if not ForceIfSettingsExist then
    Exit;
  Result := View(Null, True, [mbOK, mbAbort]);
end;

procedure TLKLedgerNominalEnterBalances.EntriesCalculateWizardExecute(AOwner:
    TComponent);
begin
  // override as required
end;

function TLKLedgerNominalEnterBalances.GettdCostCentre: string;
begin
  Result := TblDetail.FieldByName('COST_CENTRE').AsString;
end;

function TLKLedgerNominalEnterBalances.GettdCredit: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('CREDIT'));
end;

function TLKLedgerNominalEnterBalances.GettdDebit: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('DEBIT'));
end;

function TLKLedgerNominalEnterBalances.GettdDescription: string;
begin
  Result := TblDetail.FieldByName('DESCRIPTION').AsString;
end;

function TLKLedgerNominalEnterBalances.GettdItemNo: Integer;
begin
  Result := GetInteger(TblDetail.FieldByName('ITEM_NUMBER'));
end;

function TLKLedgerNominalEnterBalances.GettdNominalCode: string;
begin
  Result := TblDetail.FieldByName('NOMINAL_CODE').AsString;
end;

function TLKLedgerNominalEnterBalances.GettmCostCentre: string;
begin
  Result := TblMaster.FieldByName('COST_CENTRE').AsString;
end;

function TLKLedgerNominalEnterBalances.GettmNominalCode: string;
begin
  Result := TblMaster.FieldByName('NOMINAL_CODE').AsString;
end;

function TLKLedgerNominalEnterBalances.GettmReserveRefererence: string;
begin
  Result := TblMaster.FieldByName('RESERVE_REFERENCE').AsString;
end;

function TLKLedgerNominalEnterBalances.PlCheckBatchIsValid: TModalResult;
begin
  // ErrCodeInteger,DocRef,DocItem,DESCRIPTION,CorrectionAction
  DlgErrors.ProblemDaybook := Self;
  Result := mrOK;
  if TblMaster.FindFirst then
  begin
    repeat
      PlCheckBatchItemMasterIsValid;
      if TblDetail.FindFirst then
        repeat PlCheckBatchItemDetailIsValid;
        until not TblDetail.findNext
      else
        DlgErrors.AddWarning(['N', TblMaster.Fields[0].asString, '', '', '', '',
          'has no detail records', 'delete record']);
    until not TblMaster.FindNext;
  end
  else
    DlgErrors.AddError(['999', '', '', '', '', '', 'No entries found',
      'finish']);
  with DlgErrors do
    if areErrors then
      // if correct errors is actioned then re-check the batch
      if (ShowModal = mrOK) then
        Result := mrRetry;
end;

procedure TLKLedgerNominalEnterBalances.PlCheckBatchItemDetailIsValid;
var
  doc, item, nomCode, cost: string;
  I: Integer;
  isBlank: Boolean;
begin
  nomCode := tdNominalCode;
  cost := tdCostCentre;
  with TblDetail, DlgErrors do
  begin
    doc := Fields[0].asString;
    item := Fields[1].asString;
    if (nomCode = '') then
    begin
      isBlank := True;
      for i := 2 to FieldCount - 1 do
        if not fields[i].isNull then
          isBlank := False;
      if isBlank then
        AddError(['200', doc, item, nomCode, cost, 'Blank item', 'Delete'])
      else
        AddError(['201', doc, item, nomCode, cost, 'Nominal code is blank',
          'Display & correct']);
      exit;
    end;
    if not ContainerComponent.AccountIsValid(VarArrayOf([nomCode, cost]), False)
      then
      AddError(['201', doc, item, nomCode, cost, 'Incorrect code ' + nomCode +
        ' ' + cost,
          'Display & correct']);
  end;
end;

procedure TLKLedgerNominalEnterBalances.PlCheckBatchItemMasterIsValid;
var
  doc, item, nomCode, cost: string;
  IsBlank: Boolean;
  I: Integer;
begin
  nomCode := tmNominalCode;
  cost := tmCostCentre;
  with TblMaster, DlgErrors do
  begin
    doc := Fields[0].AsString;
    item := FieldByName('DESCRIPTION').AsString;
    if (nomCode = '') then
    begin
      IsBlank := True;
      for I := 2 to FieldCount - 1 do
        if not fields[I].isNull then
          IsBlank := False;
      if IsBlank then
        AddError(['200', doc, item, nomCode, cost, 'Blank item', 'Delete'])
      else
        AddError(['201', doc, item, nomCode, cost, 'Nominal code is blank',
          'Display & correct']);
      Exit;
    end;
    if not ContainerComponent.AccountIsValid(VarArrayOf([nomCode, cost]), False)
      then
      AddError(['201', doc, item, nomCode, cost, 'Incorrect code ' + nomCode +
        ' ' + cost,
          'Display & correct']);
  end;
end;

procedure TLKLedgerNominalEnterBalances.PlCheckNominal(const NominalCode,
    CostCentre: string);
var
  Desc: string;
begin
  with ContainerComponent do
  begin
    if not AccountIsValid(VarArrayOf([NominalCode, CostCentre]), True) then
      Exit;
    if not Locate('NOMINAL_CODE', NominalCode, []) then
    begin
      with NominalCodes.TblMaster do
      begin
        Locate('NOMINAL_CODE', NominalCode, []);
        Desc := FieldByName('DESCRIPTION').AsString;
      end;
      InsertRecord([NominalCode, Desc, 0, 0, 0]);
    end;
    if (CostCentre > ' ') then
      with CostLedger do
        if not Locate('COST_CENTRE;NOMINAL_CODE', VarArrayOf([CostCentre,
          NominalCode]), []) then
        begin
          with NominalCodes.TblMaster do
          begin
            Locate('NOMINAL_CODE', NominalCode, []);
            Desc := FieldByName('DESCRIPTION').asString;
          end;
          InsertRecord([CostCentre, NominalCode, Desc, 0, 0, 0]);
        end;
  end;
end;

procedure TLKLedgerNominalEnterBalances.PlClearReserves;
begin
  // override
end;

function TLKLedgerNominalEnterBalances.PlDlgErrorsCorrectError(ErrorInfo:
    TStrings): TModalResult;
var
  ErrorCode, DocItem: Integer;
  DocRef: string;
  DetailFound: Boolean;
begin
  Result := mrOK;
  ErrorCode := StrToInt(errorInfo[0]);
  DocRef := ErrorInfo[1];
  DocItem := StrToInt(errorInfo[2]);
  if Locate(TblMaster.Fields[0].FieldName, DocRef, []) then
  begin
    with TblDetail do
    try
      DetailFound := FindKey([DocRef, DocItem]);
    except
      DetailFound := Locate(IndexFieldNames,VarArrayOf([DocRef, DocItem]),[]);
    end;
    case ErrorCode of
      999:
        begin
          Result := mrAbort;
          Exit;
        end;
      100:
        if (MessageDlg(ErrorInfo.CommaText + ' - confirm Delete?', mtWarning,
          [mbYes, mbNo], 0) = mrYes) then
          RecordDelete
        else
          Result := ViewRecord(Null, True, [mbOK, mbAbort]);
      101:
        begin
          Showmessage(ErrorInfo.CommaText);
          Result := ViewRecord(Null, True, [mbOK, mbAbort]);
        end;
      200: if DetailFound then
          TblDetail.Delete;
      201: if DetailFound then
        begin
          Showmessage(ErrorInfo.CommaText);
          Result := ViewRecord(Null, True, [mbOK, mbAbort]);
        end;
    end;
  end;
end;

function TLKLedgerNominalEnterBalances.PlDlgErrorsCorrectErrors: TModalResult;
var
  i: Integer;
begin
  Result := mrOK;
  with DlgErrors do
  begin
    if areErrors then
      with sgErrors do
        for i := 1 to RowCount - 1 do
        begin
          Result := PlDlgErrorsCorrectError(Rows[i]);
          if (Result <> mrOK) then
            break;
        end;
    if areWarnings then
      if MessageDlg('There are warnings - do you want to edit the entries?',
        mtInformation, [mbYes, mbNo], 0) = mrYes then
        PlDlgErrorsCorrectWarnings;
  end;
end;

procedure TLKLedgerNominalEnterBalances.PlDlgErrorsCorrectWarnings;
begin
  View(DlgErrors.sgWarnings.Cells[1, 2], True, [mbOK]);
end;

procedure TLKLedgerNominalEnterBalances.PlPostReserve;
begin
  //override
end;

procedure TLKLedgerNominalEnterBalances.PlProcessedComplete;
begin
  ShowMessage('Processing complete!');
end;

procedure TLKLedgerNominalEnterBalances.PostToLedger(AOwner: TComponent);
var
  TX: TDBXTransaction;
  Mr: TModalResult;

  label
    again;

begin
  TblMaster.Filtered := False;
  TblDetail.Filtered := False;
  DlgErrors := TLKDlgLedgerNominalEntryErrors.Create(AOwner);
  try
    again:
    DlgErrors.ClearMessages;
    case PlCheckBatchIsValid of
      mrAbort, mrCancel: sysutils.Abort;
      mrRetry: goto again;
    end;
  finally DlgErrors.Free;
  end;
  //apply any ccorrections prior to processing
  ApplyUpdates(0);
  PlClearReserves;
  ContainerComponent.ApplyUpdates(0);
  DlgPrintExecute(NameFriendly, 'Default', Null, False);
  Screen.Cursor := crHourglass;
  try
    with TblMaster do
      if FindFirst then
        repeat
          PlPostReserve;
        until not FindNext;
    Mr := mrRetry;
    if SQLConnection.InTransaction then
      while Mr = mrRetry do
      begin
        if SQLConnection.InTransaction then
          Mr :=
            MessageDlg('Cannot continue : a database transaction is in progress',
            mtInformation, [mbRetry, mbAbort], 0)
        else
          Mr := mrOK;
        if Mr = mrAbort then
        begin
          ContainerComponent.CancelUpdates;
          SysUtils.Abort;
        end;
      end;
    try
      SQLConnection.BeginTransaction;
      ContainerComponent.ApplyUpdates(0);
      SQLConnection.CommitFreeAndNil(TX); {on success, commit the changes}
    except
      SQLConnection.RollbackFreeAndNil(TX); {on failure, undo the changes}
      raise;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
  PlProcessedComplete;
end;

procedure TLKLedgerNominalEnterBalances.SettdCostCentre(const Value: string);
begin
  TblDetail.FieldByName('COST_CENTRE').AsString := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettdCredit(Value: Currency);
begin
  TblDetail.FieldByName('CREDIT').AsCurrency := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettdDebit(Value: Currency);
begin
  TblDetail.FieldByName('DEBIT').AsCurrency := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettdDescription(const Value: string);
begin
  TblDetail.FieldByName('DESCRIPTION').AsString := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettdItemNo(Value: Integer);
begin
  TblDetail.FieldByName('ITEM_NUMBER').AsInteger := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettdNominalCode(const Value: string);
begin
  TblDetail.FieldByName('NOMINAL_CODE').AsString := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettmCostCentre(const Value: string);
begin
  TblMaster.FieldByName('COST_CENTRE').AsString := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettmNominalCode(const Value: string);
begin
  TblMaster.FieldByName('NOMINAL_CODE').AsString := Value;
end;

procedure TLKLedgerNominalEnterBalances.SettmReserveRefererence(const Value:
    string);
begin
  TblMaster.FieldByName('RESERVE_REFERENCE').AsString := Value;
end;


{
********************* TLKLedgerNominalEnterOpeningBalances *********************
}
procedure TLKLedgerNominalEnterOpeningBalances.PlClearReserves;
begin
  with ContainerComponent do
  begin
    OKToEdit := True;
    try
      with TblMaster do
        if FindFirst then
          repeat
            Edit;
            FieldByName('OPENING_RESERVES').Value := Null;
            FieldByName('OPENING_BALANCE').Value := Null;
            Post;
          until not FindNext;
    finally OKToEdit := False;
    end;
    with CostLedger do
    begin
      OKToEdit := True;
      try
        with TblMaster do
          if FindFirst then
            repeat
              Edit;
              FieldByName('OPENING_RESERVES').Value := Null;
              FieldByName('OPENING_BALANCE').Value := Null;
              Post;
            until not FindNext;
      finally OKToEdit := False;
      end;
    end;
  end;
end;

procedure TLKLedgerNominalEnterOpeningBalances.PlPostReserve;
var
  L: Currency;
  s: string;
  n: string;
  c: string;
begin
  if (TblMaster.Fields[0].asString = '~~BSItems') then
    s := 'OPENING_BALANCE'
  else
    s := 'OPENING_RESERVES';
  with TblDetail do
    if FindFirst then
      repeat
        n := tdNominalCode;
        c := tdCostCentre;
        L := tdDebit - tdCredit;
        PlCheckNominal(n, c);
        with ContainerComponent do
        begin
          OKToEdit := True;
          try
            with TblMaster do
            begin
              Edit;
              AddCurrencyToField(L, FieldByName(s));
              Post;
            end;
          finally OKToEdit := False;
          end;
          if (c > ' ') then
            with CostLedger do
            begin
              OKToEdit := True;
              try
                with TblMaster do
                begin
                  Edit;
                  AddCurrencyToField(L, FieldByName(s));
                  Post;
                end;
              finally OKToEdit := False;
              end;
            end;
        end;
      until not FindNext;
  // post the control account
  n := tmNominalCode;
  c := tmCostCentre;
  PlCheckNominal(n, c);
  L := GetCurrency(TblMaster.FieldByName('TOTAL'));
  with ContainerComponent do
  begin
    OKToEdit := True;
    try
      with TblMaster do
      begin
        Edit;
        AddCurrencyToField(L, FieldByName(s));
        Post;
      end;
    finally OKToEdit := False;
    end;
    if (c > ' ') then
      with CostLedger do
      begin
        OKToEdit := True;
        try
          with TblMaster do
          begin
            Edit;
            AddCurrencyToField(L, FieldByName(s));
            Post;
          end;
        finally OKToEdit := False;
        end;
      end;
  end;
end;


{
************************ TLKLedgerNominalEnterReserves *************************
}
procedure TLKLedgerNominalEnterReserves.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerNominalEnterReserves, FFormView);
  TFormLKLedgerNominalEnterReserves(FormView).Construct(Self);
end;

procedure TLKLedgerNominalEnterReserves.DrillDownRecord;
var
  i: Integer;
  r, n, c, d: string;
begin
  if assigned(OnDetailDrillDown) then
    inherited
  else
  begin
    r := tmReserveRefererence;
    i := tdItemNo;
    n := tdNominalCode;
    c := tdCostCentre;
    d := tdDescription;
    if (r = 'Accrual') then
      with ContainerComponent.Accruals do
      begin
        if not Locate('ITEM_NUMBER', i, []) then
          InsertRecord([i, n, c, d]);
        ViewRecord(Null, True, [mbOK]);
      end;
    if (r = 'Prepaid') then
      with ContainerComponent.Prepayments do
      begin
        if not Locate('ITEM_NUMBER', i, []) then
          InsertRecord([i, n, c, d]);
        ViewRecord(Null, True, [mbOK]);
      end;
  end;
end;

function TLKLedgerNominalEnterReserves.EndOfMonth: TModalResult;
begin
  Result := inherited EndOfMonth;
  if Result = mrOK then
    Result := View(Null, True, [mbOK]);
end;

procedure TLKLedgerNominalEnterReserves.EntriesCalculateWizardExecute(AOwner:
    TComponent);
var
  ED: TDateTime;
begin
  with ContainerComponent do // Nominal Ledger
  begin
    FixedAssets.CalculateAccountsReserves;
    Accruals.CalculateAccountsReserves;
    Prepayments.CalculateAccountsReserves;
    ED := LedgerDate;
  end;
  with TblMaster do
  begin
    FindFirst;
    repeat Edit;
      tmReserveDate := ED;
      Post;
    until not FindNext;
  end;
  ApplyUpdates(0);
end;

function TLKLedgerNominalEnterReserves.GettmReserveDate: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('DATE_'));
end;

procedure TLKLedgerNominalEnterReserves.PlClearReserves;
begin
  with ContainerComponent do
  begin
    OKToEdit := True;
    try
      with TblMaster do
        if FindFirst then
          repeat
            Edit;
            FieldByName('RESERVES').Value := Null;
            Post;
          until not FindNext;
      ApplyUpdates(0);
    finally OKToEdit := False;
    end;
    with CostLedger do
    begin
      OKToEdit := True;
      try
        with TblMaster do
          if FindFirst then
            repeat
              Edit;
              FieldByName('RESERVES').Value := Null;
              Post;
            until not FindNext;
        ApplyUpdates(0);
      finally OKToEdit := False;
      end;
    end;
  end;
end;

procedure TLKLedgerNominalEnterReserves.PlPostReserve;
var
  L: Currency;
  N: string;
  C: string;
begin
  //post the detail reserve items
  with TblDetail do
    if FindFirst then
      repeat
        N := tdNominalCode;
        C := tdCostCentre;
        L := tdDebit - tdCredit;
        PlCheckNominal(N, C);
        with ContainerComponent do
        begin
          OKToEdit := True;
          try
            with TblMaster do
            begin
              Edit;
              AddCurrencyToField(L, FieldByName('RESERVES'));
              Post;
            end;
          finally OKToEdit := False;
          end;
          if (C > ' ') then
            with CostLedger do
            begin
              OKToEdit := True;
              try
                with TblMaster do
                begin
                  Edit;
                  AddCurrencyToField(L, FieldByName('RESERVES'));
                  Post;
                end;
              finally OKToEdit := False;
              end;
            end;
        end;
      until not FindNext;
  // post the control account
  N := tmNominalCode;
  C := tmCostCentre;
  PlCheckNominal(N, C);
  L := GetCurrency(TblMaster.FieldByName('TOTAL'));
  with ContainerComponent do
  begin
    OKToEdit := True;
    try
      with TblMaster do
      begin
        Edit;
        AddCurrencyToField(L, FieldByName('RESERVES'));
        Post;
      end;
    finally OKToEdit := False;
    end;
    if (C > ' ') then
      with CostLedger do
      begin
        OKToEdit := True;
        try
          with TblMaster do
          begin
            Edit;
            AddCurrencyToField(L, FieldByName('RESERVES'));
            Post;
          end;
        finally OKToEdit := False;
        end;
      end;
  end;
end;

procedure TLKLedgerNominalEnterReserves.SettmReserveDate(Value: TDateTime);
begin
  TblMaster.FieldByName('DATE_').AsDateTime := Value;
end;


{
********************** TLKLedgerNominalManagementAccounts **********************
}
procedure TLKLedgerNominalManagementAccounts.DlgManagementAccountsPrintExecute;
begin
  with TDlgLKLedgerNominalManagementAcsPrint.Create(Owner) do
  begin
    try
      Construct(Self, '');
      PopUpParent := FormView;
      ShowModal;
    finally free;
    end;
  end;
end;


{
**************************** TLKLedgerNominalBudget ****************************
}
constructor TLKLedgerNominalBudget.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
  NameFriendly := 'Nominal Ledger Budget';
end;

procedure TLKLedgerNominalBudget.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerNominalBudget, FFormView);
  TFormLKLedgerNominalBudget(FormView).Construct(Self);
end;

procedure TLKLedgerNominalBudget.DlgBudgetPrintExecute;
begin
  with TDlgLKLedgerNominalBudgetPrint.Create(Owner) do
  begin
    try
      Construct(Self, '');
      PopUpParent := FormView;
      ShowModal;
    finally free;
    end;
  end;
end;

procedure TLKLedgerNominalBudget.DlgImportExportExecute(AOwner: TComponent;
    const AsExportDlg, Autoexecute: boolean);
var
  Okay: Boolean;
  NormallyACredit: Boolean;
  S: string;
  I: Integer;
  NominalCode: string;
  BudgetDates: array[0..500] of TDateTime;
  PeriodEndDate: TDateTime;
  PeriodEndDateCount: Integer;
  Budget: Currency;
  CSVFileName: string;
  CSVFile: TextFile;
  Numerals: string;
  TmpDataSet: TSimpleDataSet;
  AskYesNo: Char;

  function GetNextField: string;
  var
    I: Integer;
  begin
    if (Pos('"', S) = 1) then
    begin
      Result := Copy(S, 2, Pos('",', S) - 2);
      System.Delete(S, 1, Length(Result) + 3);
    end
    else
    begin
      I := Pos(',', S);
      if I = 0 then
        Result := S
      else
        Result := Copy(S, 1, I - 1);
      System.Delete(S, 1, Length(Result) + 1);
    end;
    Result := Trim(Result);
  end;

  function GetNextFieldAsNumber: string;
  var
    I: Integer;
    N: string;
  begin
    N := GetNextField;
    Result := '';
    for I := 1 to Length(N) do
      if (Pos(N[I], Numerals) > 0) then
        Result := Result + N[I];
  end;

  function TryNextFieldAsPeriodEnd: Boolean;
  var
    CalcDate: TDateTime;
  begin
    case AccountsHub.AccountsInterval of
      aciWeekly: CalcDate := PeriodEndDate + 7;
      aciFortnightly: CalcDate := PeriodEndDate + 14;
      aciFourWeeks: CalcDate := PeriodEndDate + 28;
      aciAnnual: CalcDate := MonthEndThis(PeriodEndDate + 360);
      aciDays: CalcDate := PeriodEndDate + AccountsHub.AccountsIntervalDays;
      else CalcDate := MonthEndNext(PeriodEndDate);
    end;
    Result := TryStrToDate(GetNextField, PeriodEndDate);
    if Result then
      Result := (PeriodEndDate = CalcDate);
    if not Result then
      case AskYesNo of
        'Y': PeriodEndDate := CalcDate;
        'N':;
        else case MessageDlg('The next period end date expected is '
          + DateToStr(CalcDate) + ', however '+ DateToStr(PeriodEndDate)
          + ' is found. Correct?', mtWarning,
          [mbYes, mbNo, mbAbort, mbYesToAll, mbNoToAll],0) of
          mrYes: PeriodEndDate := CalcDate;
          mrNo:;
          mrAbort: Exit;
          mrYesToAll:
          begin
            PeriodEndDate := CalcDate;
            AskYesNo := 'Y';
          end;
          mrNoToAll: AskYesNo := 'N';
        end;
      end;
    Result := True;
  end;

begin
  if AsExportDlg then
  begin
    ShowMessage('To Export, please Print as Spreadsheet');
    DlgBudgetPrintExecute;
  end
  else
  begin
    with TOpenDialog.Create(Application) do
    try
      Title := 'Select Report layout Import file';
      DefaultExt := '.csv';
      Filter := 'Comma Separated Variable (CSV)|*.csv';
      InitialDir := InstallFolder;
      Options := Options + [ofFileMustExist];
      Okay := Execute;
      CSVFileName := FileName;
    finally Free;
    end;
    if not Okay then
      Exit;
    AskYesNo := 'A';
    Numerals := '-0123456789' + FormatSettings.DecimalSeparator;
    NormallyACredit := False;
    AssignFile(CSVFile, CSVFileName);
    Reset(CSVFile);
    Screen.Cursor := crHourglass;
    MasterComponent.ExternalDisableControls;
    ExternalDisableControls;
    PeriodEndDateCount := 0;
    TmpDataSet := TSimpleDataSet.Create(ContainerComponent.Owner);
    TmpDataSet.DataSet.SQLConnection := SQLConnection;
    TmpDataSet.DataSet.CommandText := 'SELECT * FROM ' + TblDetail.TableName;
    try
      TmpDataSet.Open;
      repeat
        ReadLn(CSVFile, S);
        NominalCode := GetNextField;
        if (PeriodEndDateCount = 0) then
        begin
          GetNextField; // discard the description if any
          if TryStrToDate(GetNextField, PeriodEndDate) then
          repeat
            with ContainerComponent.AccountsPeriods do
             if not TblMaster.Locate('PERIOD_END', PeriodEndDate, []) then
               InsertRecord([PeriodEndDate]);
            BudgetDates[PeriodEndDateCount] := PeriodEndDate;
            Inc(PeriodEndDateCount);
          until not TryNextFieldAsPeriodEnd;
          ContainerComponent.AccountsPeriods.TblMaster.ApplyUpdates(-1);
        end
        else if ContainerComponent.NominalCodes.TblMaster.Locate(
          'NOMINAL_CODE', NominalCode, []) then
        begin
          GetNextField; // discard the description if any
          for I := 0 to PeriodEndDateCount -1 do
            if TryStrToCurr(GetNextFieldAsNumber, Budget) then
            begin
              if NormallyACredit then
                Budget := -Budget;
              if TmpDataSet.Locate('NOMINAL_CODE;PERIOD_END',
                VarArrayOf([NominalCode, BudgetDates[I]]),[]) then
              begin
                TmpDataSet.Edit;
                TmpDataSet.FieldByName('BUDGET').AsCurrency := Budget;
                TmpDataSet.Post;
              end
              else
              begin
                TmpDataSet.InsertRecord([NominalCode, BudgetDates[I], Budget]);
              end;
            end;
           TmpDataSet.ApplyUpdates(-1);
        end
        else
        begin
          S := Uppercase(GetNextField);
          if (Pos('CREDIT',S) > 0) then
            NormallyACredit := True
          else if (Pos('DEBIT',S) > 0) then
            NormallyACredit := False;
        end;
      until Eof(CSVFile);
    finally
      CloseFile(CSVFile);
      Screen.Cursor := crDefault;
      TmpDataSet.Close;
      TmpDataSet.Free;
      TblDetail.Refresh;
      MasterComponent.ExternalEnableControls;
      ExternalEnableControls;
      MasterComponent.TblDetail.First;
    end;
  end;
end;

function TLKLedgerNominalBudget.DoSetup(const ForceIfSettingsExist: Boolean):
    TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  case Result of
    mrAbort: Exit;
  end;
  if not ForceIfSettingsExist then
    Exit;
  Result := View(Null, True, [mbOK, mbAbort]);
end;


{
******************************* TLKLedgerNominal *******************************
}
constructor TLKLedgerNominal.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  NameFriendly := 'Nominal Ledger';
  with ManagementAccounts do
  begin
    IndexName := 'DEFAULT_ORDER';
    MasterFields := 'NOMINAL_CODE';
    TblDetail.IndexName := 'PRIMARY_KEY';
    TblDetail.MasterFields := 'NOMINAL_CODE';
    TblDetail.TableName := 'NOMINAL_ACCOUNTS';
    TblMaster.TableName := 'NOMINAL_LEDGER';
    //TblMaster.Clone := Self.TblMaster;
  end;
  with NominalCodes do
  begin
    IndexName := 'DEFAULT_ORDER';
    NameFriendly := 'Nominal codes';
    TblMaster.TableName := 'NOMINAL_CODES';
  end;
  //TblMaster.TableName := 'NominalLedger';
  //TblDetail.TableName := 'AnalysisLedgerDtl';
  //TblDetail.MasterFields := 'NOMINAL_CODE';
  with Budget do
  begin
    IndexName := 'DEFAULT_ORDER';
    MasterComponent := ReportLines;
    MasterFields := 'NOMINAL_CODE';
    MasterTable := tbDetail;
    NameFriendly := 'Budget';
    TblDetail.IndexName := 'PRIMARY_KEY';
    TblDetail.MasterFields := 'NOMINAL_CODE';
    TblDetail.TableName := 'NOMINAL_BUDGETS';
    TblMaster.TableName := 'NOMINAL_LEDGER';
    //TblMaster.Clone := Self.TblMaster;
  end;

  {
  with NominalCategories do
   begin
   TblMaster.TableName := 'NominalCategories';
   IndexFieldNames := 'LINE_NUMBER';
   TblDetail.Clone := NominalCodes.TblMaster;
   //TblDetail.TableName := 'NominalCodes';
   //TblDetail.IndexName := 'idxNominalCategory';
   TblDetail.IndexFieldNames := 'LINE_NUMBER;NOMINAL_CODE';
   TblDetail.MasterFields := 'LINE_NUMBER';
   NameFriendly := 'Nominal analysis Categories';
   end;
  with TLKLedgerNominalEnterOpeningBalances(OpeningBalances) do
   begin
   NameFriendly := 'Opening balances';
   TblMaster.TableName := 'NOMINALOPENINGBALANCES';
   IndexFieldNames := 'RESERVE_REFERENCE';
   TblDetail.TableName := 'NOMINALOPENINGBALANCESDTL';
   TblDetail.IndexFieldNames := 'RESERVE_REFERENCE;NOMINAL_CODE;COST_CENTRE';
   TblDetail.MasterFields := 'RESERVE_REFERENCE';
   end;
  with TLKLedgerNominalEnterReserves(Reserves) do
   begin
   NameFriendly := 'Accounts reserves';
   TblMaster.TableName := 'NominalReserves';
   IndexFieldNames := 'RESERVE_REFERENCE';
   TblDetail.TableName := 'NominalReservesDtl';
   TblDetail.IndexFieldNames := 'RESERVE_REFERENCE;NOMINAL_CODE;COST_CENTRE';
   TblDetail.MasterFields := 'RESERVE_REFERENCE';
   end;
  // }
  //showmessage('Create finish');
end;

function TLKLedgerNominal.AccountIsValid(const KeyValues: Variant; RaiseError:
    Boolean = False): Boolean;
var
  Code: string;
begin
  if not Active then
    Open;
  Code := Trim(KeyValues[0]);
  Result := (Code = '');
  if Result then
    Exit
  else
    Result := NominalCodes.Locate('NOMINAL_CODE', Code, []);
  if not Result and RaiseError then
    raise LKError.Create(NameFriendly + ' reference ' + Code +
      ' not found');
  if Result and Assigned(CostLedger) then
    Result := CostLedger.AccountIsValid(KeyValues[1], RaiseError);
end;

procedure TLKLedgerNominal.AccountPost(const DetailRecord: array of const);
var
  Code, Cost, Desc: string;
  I: Integer;
  Value: Currency;
  D: Extended;
  NewRecord: array[0..9] of TVarRec;

  procedure ChangeAccountCode(const newCode: array of const);
  begin
    NewRecord[1] := newCode[0];
    NewRecord[2] := newCode[1];
  end;

begin
  D := DetailRecord[3].vExtended^;
  if (D > LedgerDate) then
  begin
    DeferredPostings.InsertRecord(DetailRecord);
    Exit;
  end;
  Code := string(DetailRecord[1].vAnsiString);
  Cost := Trim(string(DetailRecord[2].vAnsiString));
  if AccountIsValid(VarArrayOf([Code, Cost]), False) then
  begin
    if not Locate('NOMINAL_CODE', Code, []) then
    begin
      with NominalCodes.TblMaster do
      begin
        Locate('NOMINAL_CODE', Code, []);
        Desc := FieldByName('DESCRIPTION').AsString;
      end;
      InsertRecord([Code, Desc]);
    end;
    OKToEdit := True;
    try
      with TblDetail do
      begin
        InsertRecord(DetailRecord); //updates account totals
        Value := GetCurrency(FieldByName('DEBIT')) -
          GetCurrency(FieldByName('CREDIT'));
      end;
    finally OKToEdit := False;
    end;
    if Assigned(CostLedger) then
      CostLedger.AddValueToAccount(Cost, Code, value);
    Exit;
  end
  else
  begin
    //ensure posting to suspence even with invalid code
    for I := 0 to 9 do
      NewRecord[I] := DetailRecord[I];
    DecodeNominalAccount(SuspensePostingsCode, Code, Cost);
    ChangeAccountCode([Code, Cost]);
    AccountPost(NewRecord);
  end;
end;

procedure TLKLedgerNominal.CreateManagementAccounts;
begin
  FManagementAccounts := TLKLedgerNominalManagementAccounts.Create(Self);
end;

procedure TLKLedgerNominal.CreateSubComponents;
begin
  inherited CreateSubComponents;
  FAccruals := TLKLedgerNominalAccruals.Create(Self);
  FBudget := TLKLedgerNominalBudget.Create(Self);
  FFixedAssetCategories := TLKLedgerNominalFixedAssetCategories.Create(Self);
  FFixedAssets := TLKLedgerNominalFixedAssets.Create(Self);
  FFixedAssets.MasterComponent := FFixedAssetCategories;
  FNominalCodes := TLKLedgerNominalNominalCodes.Create(Self);
  FOpeningBalances := TLKLedgerNominalEnterOpeningBalances.Create(Self);
  FPrepayments := TLKLedgerNominalPrepayments.Create(Self);
  FReserves := TLKLedgerNominalEnterReserves.Create(Self);
end;

procedure TLKLedgerNominal.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerNominal, FFormView);
  TFormLKLedgerNominal(FormView).Construct(Self);
end;

function TLKLedgerNominal.DlgNominalCodeSelectExecute(const ACaption,
    EncodedNomCost, HelpTxt: string; const getCostcentre: boolean): string;
var
  n, c: string;
begin
  Result := '';
  with TDlgLedgerNominalCodeSelect.Create(Owner) do
  begin
    try
      Caption := ACaption;
      NominalLedger := Self;
      PanelCostCentre.Visible :=
        (getCostCentre and (Status > acsBasic));
      if PanelCostCentre.Visible then
        PanelCostCentre.Visible := assigned(CostLedger);
      if (EncodedNomCost > '') then
      begin
        DecodeNominalAccount(EncodedNomCost, n, c);
        eNomCode.Text := n;
        eCostCentre.Text := c;
        CheckForFinish; //does OKBtn.Enabled if valid code
      end;
      HelpText := HelpTxt;
      btnHelp.Visible := (HelpText > '');
      if (ShowModal = mrOK) then
        Result := EncodeNominalAccount(eNomCode.Text, eCostCentre.Text);
    finally Free;
    end;
  end;
end;

procedure TLKLedgerNominal.DoAfterSetHub;
begin
  inherited DoAfterSetHub;
  if assigned(AccountsHub) then
    AccountsHub.NominalLedger := Self;
end;

function TLKLedgerNominal.DoSetup(const ForceIfSettingsExist: Boolean):
    TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  {showmessage('You should enter the main heads of your Management Accounts '
   + 'Profit and Loss Account.');
  Result := ReportLayouts.ViewRecord('A',True,[mbOK,mbAbort]);
  case Result of mrAbort : exit; end;
  showmessage('You should enter the main heads of your Management Accounts '
   + 'Balance Sheet.');
  Result := ReportLayouts.ViewRecord('N',True,[mbOK,mbAbort]);
  case Result of mrAbort : exit; end;
  showmessage('You should now enter the line descriptions on your Management Accounts and '
    + ' the Nominal Codes which make them up e.g with "Printing and Stationery" you'
    + ' could enter Nominal Codes for "Printing" and for "Stationery"');
  showmessage('Line numbers could be the Layout head code + "10" to "90" allowing a 10 spacing'
    + ' the Nominal Codes could be the Line number + "00" to "90" allowing a 10 spacing,'
    + ' e.g. Layout head "J", Line number "J30", Nominal code "J3040"');
  showmessage('You should remember to setup a Suspense Postings Nominal code,'
    + 'on Balance Sheet where postings will be entered if an account is not found.');
  Result := ReportLines.View(Null,True,[mbOK,mbAbort]);
  SuspensePostingsCode := DlgNominalCodeSelectExecute(
   'Suspense postings account',SuspensePostingsCode,
   'Enter the account for Suspense Postings where transactions will be entered if a valid Nominal account is not found.',
   False);
  showmessage('You should now enter your Fixed Asset Categories - Property, Equipment,'
    +'Vehicles etc..');
  showmessage('Category codes may be up to 8 characters and are sorted alphabetically.');
  Result := FixedAssetCategories.View(Null,True,[mbOK,mbAbort]);
  case Result of mrAbort : exit; end;
  if (FixedAssetCategories.TblMaster.RecordCount > 0) then
   begin
   showmessage('You should now enter your Fixed Assets against the Categories');
   showmessage('You may wish to track the costs of individual assets in which case'
    + ' Cost Centres may be the same as Asset references.');
   Result := FixedAssetCategories.View(Null,True,[mbOK,mbAbort]);
   case Result of mrAbort : exit; end;
   end;
  CodeAccruals := DlgNominalCodeSelectExecute(
  'Accruals (reserves for future costs) account on Balance Sheet',CodeAccruals,
  '',False);
  CodePrepayments := DlgNominalCodeSelectExecute(
  'Prepayments (costs paid in advance) account on Balance Sheet',CodePrepayments,
  '',False);
  CodeProfitAndLoss := DlgNominalCodeSelectExecute(
  'Profit and Loss account on Balance Sheet',CodeProfitAndLoss,
  '',False);
  Status := acsBasic;
  Result := mrOK;// save status}
end;

function TLKLedgerNominal.EndOfMonth: TModalResult;
var
  NomCode, CostCentre, DaybookID, Doc, LedgerAc: string;
  Value: Currency;
  BatchDate: TDateTime;
  LedgerTempURN: Integer;
  I: Integer;
  Desc: string;
begin
  Result := inherited EndOfMonth;
  with DeferredPostings.TblMaster do
  begin
    Filter := 'DATE_ <= ' + chr(39) + dateTostr(LedgerDate) + chr(39);
    Filtered := True;
  end;
  LedgerTempURN := -1;
  try
    while DeferredPostings.TblMaster.FindFirst do
    begin
      with DeferredPostings.TblMaster do
      begin
        NomCode := FieldByName('NOMINAL_CODE').asString;
        CostCentre := FieldByName('COST_CENTRE').asString;
        BatchDate := FieldByName('DATE_').asDateTime;
        DaybookID := FieldByName('DAYBOOK_ID').asString;
        LedgerAC := FieldByName('LEDGER_ACCOUNT').asString;
        Doc := FieldByName('DOCUMENT_NUMBER').asString;
        Desc := FieldByName('DESCRIPTION').asString;
        Value := GetCurrency(FieldByName('DEBIT')) -
          GetCurrency(FieldByName('CREDIT'));
        Delete;
      end;
      AccountPost([LedgerTempURN, NomCode, CostCentre, BatchDate, DaybookID,
        LedgerAC, Doc, Desc, DebitOrNull(Value), CreditOrNull(Value)]);
      Dec(LedgerTempURN);
    end;
    ApplyUpdates(0);
    Refresh;
  finally
    with DeferredPostings.TblMaster do
    begin
      Filtered := False;
      Filter := '';
    end;
  end; // finallyif assigned(CostLedger) then with CostLedger do
  if Assigned(CostLedger) then
    CostLedger.EndOfMonth;
  BatchDate := Ledgerdate;
  if assigned(AccountsHub) then
    with AccountsHub do
    begin
      for I := 0 to Ledgers.count - 1 do
        with TLKLedgerNominalControl(Ledgers[I]) do
          if (LedgerDate < BatchDate) then
            EndOfMonth;
      //OnEndOfMonth(Sender);
    end;
end;

function TLKLedgerNominal.EndOfYear(ProgressText: string): TModalResult;
var
  D, C: Currency;
  Code, Desc: string;
  Item: Integer;
  S: string;
begin
  if not Active then
    if Assigned(ACGroup) then
      ACGroup.Open
    else
      Open;
  Result := MessageDlg(NameFriendly +
    ' Year End. The Nominal Ledger will now be cleared for the new year. Do you'
    + ' want a full printout?',
    mtInformation, [mbYes, mbNo, mbAbort], 0);
  case Result of
    mrYes: DlgPrintExecute('', 'Default', Null, False);
    mrAbort: Exit;
  end;
  Screen.Cursor := crHourglass;
  ProgressText := 'Writing Opening Balances for ' + NameFriendly;
  Application.ProcessMessages;
  OKToEdit := True;
  try
    OpeningBalances.EmptyTables;
    OpeningBalances.CopyRecords(Reserves, '', tbsBoth, False, batAppend);
    with OpeningBalances do
      if FindFirst then
        repeat
          with TblDetail do
            if FindFirst then
              repeat
                D := GetCurrency(FieldByName('DEBIT'));
                C := GetCurrency(FieldByName('CREDIT'));
                Edit;
                FieldByName('DEBIT').Value := CurrencyOrNull(C);
                FieldByName('CREDIT').Value := CurrencyOrNull(D);
                Post;
              until not FindNext;
        until not FindNext;
    DecodeNominalAccount(CodeProfitAndLoss, Code, S);
    OpeningBalances.AppendRecord(['~~BSItems', LedgerDate, Code, Null,
      'Balance sheet values']);
    Item := 1;
    TblDetail.EmptyTable;
    with TblMaster do
    begin
      if FindFirst then
        repeat
          D := GetCurrency(FieldByName('BALANCE_SHEET'));
          if not CurrencyIsZero(D) then
          begin
            Code := FieldByName('NOMINAL_CODE').AsString;
            Desc := FieldByName('DESCRIPTION').AsString;
            // get balance sheet value
            OpeningBalances.TblDetail.AppendRecord(['~~BSItems', Item, Code,
              Null, Desc, DebitOrNull(D), CreditOrNull(D)]);
            Inc(Item);
          end;
        until not FindNext;
    end;
    OpeningBalances.ApplyUpdates(0);
    ProgressText := 'Writing Fixed Asset Depreciation Opening Balances for '
      + NameFriendly;
    Application.ProcessMessages;
    try
      if FixedAssetCategories.FindFirst then
      repeat
        with FixedAssets do if FindFirst then
        repeat
          if TblDetail.Locate('DATE_', LedgerDate,[]) then
          begin
            TblMaster.Edit;
            TblMaster.FieldByName('DEPRECIATION_LAST_PERIOD_END').AsCurrency :=
              GetCurrency(TblDetail.FieldByName('DEPRECIATION_TO_DATE'));
            TblMaster.Post;
            Break;
          end;
        until not FindNext;
      until not FixedAssetCategories.FindNext;
    finally
      FixedAssets.ApplyUpdates(0);
    end;
    ProgressText := 'Clearing accounts for ' + NameFriendly;
    Application.ProcessMessages;
      // clear the NL balances in inherited
    Result := inherited EndOfYear(ProgressText);
    with CostLedger do
    begin
      Refresh;
      OKToEdit := True;
      try
        EndOfYear(ProgressText);
      finally OKToEdit := False;
      end;
    end;
    ProgressText := 'Posting Opening balances for ' + NameFriendly;
    Application.ProcessMessages;
    OpeningBalances.PostToLedger(Owner);
    Reserves.ApplyUpdates(0);
    ApplyUpdates(0);
    ManagementAccounts.Refresh;
  finally Screen.Cursor := crDefault;
    OKToEdit := False;
  end;
  ProgressText := 'Year end complete for ' + NameFriendly;
  Application.ProcessMessages;
end;

function TLKLedgerNominal.GetCodeAccruals: string;
begin
  Result := RecallString('Settings', 'AccrualsCode', '', True);
end;

function TLKLedgerNominal.GetCodePrepayments: string;
begin
  Result := RecallString('Settings', 'PrepaymentsCode', '', True);
end;

function TLKLedgerNominal.GetCodeProfitAndLoss: string;
begin
  Result := RecallString('Settings', 'ProfitAndLossCode', '', True);
end;

function TLKLedgerNominal.GetCostLedger: TLKLedgerCost;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.CostLedger
  else
    Result := nil;
end;

function TLKLedgerNominal.GetManagementAccounts:
    TLKLedgerNominalManagementAccounts;
begin
  Result := TLKLedgerNominalManagementAccounts(inherited ManagementAccounts);
end;

procedure TLKLedgerNominal.Loaded;
begin
  inherited Loaded;
  try
    {
    with ManagementAccounts do
     begin
     NameFriendly := 'Management Accounts';
     TblMaster.Clone := Self.TblMaster;
     //TblMaster.TableName := 'NominalLedger';
     IndexFieldNames := 'NOMINAL_CODE';
     TblDetail.TableName := 'NominalAccounts';
     TblDetail.IndexFieldNames := 'NOMINAL_CODE;PERIOD_END';
     TblDetail.MasterFields := 'NOMINAL_CODE';
     end;
    with NominalCodes do
     begin
     Editing := False;
     TblMaster.TableName := 'NominalCodes';
     IndexFieldNames := 'NOMINAL_CODE';
     NameFriendly := 'Nominal codes';
     end;
    with NominalCategories do
     begin
     TblMaster.TableName := 'NominalCategories';
     IndexFieldNames := 'LINE_NUMBER';
     TblDetail.Clone := NominalCodes.TblMaster;
     //TblDetail.TableName := 'NominalCodes';
     //TblDetail.IndexName := 'idxNominalCategory';
     TblDetail.IndexFieldNames := 'LINE_NUMBER;NOMINAL_CODE';
     TblDetail.MasterFields := 'LINE_NUMBER';
     NameFriendly := 'Nominal analysis Categories';
     end;
    with OpeningBalances do
     begin
     NameFriendly := 'Opening balances';
     TblMaster.TableName := 'NominalOpeningBalances';
     IndexFieldNames := 'RESERVE_REFERENCE';
     TblDetail.TableName := 'NominalOpeningBalancesDtl';
     TblDetail.IndexFieldNames := 'RESERVE_REFERENCE;NOMINAL_CODE;COST_CENTRE';
     TblDetail.MasterFields := 'RESERVE_REFERENCE';
     end;
    with Reserves do
     begin
     NameFriendly := 'Accounts reserves';
     TblMaster.TableName := 'NominalReserves';
     IndexFieldNames := 'RESERVE_REFERENCE';
     TblDetail.TableName := 'NominalReservesDtl';
     TblDetail.IndexFieldNames := 'RESERVE_REFERENCE;NOMINAL_CODE;COST_CENTRE';
     TblDetail.MasterFields := 'RESERVE_REFERENCE';
     end;
    // }
  except
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
  //showmessage('Nominal Loaded finish');
end;

function TLKLedgerNominal.RecordNew: TModalResult;
begin
  Result := MessageDlg(
    'You cannot enter a new Account directly into the '+ NameFriendly
    + '.  You should create a new Nominal Code. Do you want to do this now?',
    mtInformation,[mbYes, mbNo],0);
  if (Result = mrYes) then
    Result := ReportLines.DlgViewExecute;
end;

procedure TLKLedgerNominal.RefreshAccountCodes;
begin
  NominalCodes.Refresh;
end;

procedure TLKLedgerNominal.SetCodeAccruals(const Value: string);
begin
  RememberString('Settings', 'AccrualsCode', value, True);
end;

procedure TLKLedgerNominal.SetCodePrepayments(const Value: string);
begin
  RememberString('Settings', 'PrepaymentsCode', value, True);
end;

procedure TLKLedgerNominal.SetCodeProfitAndLoss(const Value: string);
begin
  RememberString('Settings', 'ProfitAndLossCode', value, True);
end;

procedure TLKLedgerNominal.SetContainedComponentNames;
begin
  inherited SetContainedComponentNames;
  FFixedAssetCategories.Name := Name + 'FixedAssetCategories';
  FFixedAssets.Name := Name + 'FixedAssets';
  FNominalCodes.Name := Name + 'Codes';
  FOpeningBalances.Name := Name + 'OpeningBalances';
  FReserves.Name := Name + 'Reserves';
  FAccruals.Name := Name + 'Accruals';
  FPrepayments.Name := Name + 'Prepayments';
  FBudget.Name := Name + 'Budget';
end;

procedure TLKLedgerNominal.TrialBalanceCalculate(const AccountsDate: TDateTime);
var
  TotalProfit: Currency;
  ProfitAndLoss: Boolean;
  N, C: string;

  procedure CalculateLedgerAccounts;
  var
    b, p, l: Currency;
  begin
    if ReportLayouts.TblDetail.FindFirst then
      repeat
        if ReportLines.TblMaster.FindFirst then
          repeat
            if ReportLines.TblDetail.FindFirst then
              repeat
                p := 0;
                b := 0;
                with Self.TblMaster do
                  if FindFirst then
                  begin
                    l := GetCurrency(FieldByName('OPENING_RESERVES')) +
                      GetCurrency(FieldByName('OPENING_BALANCE')) +
                      GetCurrency(FieldByName('ACCOUNT_BALANCE')) +
                      GetCurrency(FieldByName('RESERVES'));
                    if ProfitAndLoss then
                    begin
                      p := l;
                      TotalProfit := TotalProfit + p;
                    end
                    else
                      b := l;
                    OKToEdit := True;
                    try
                      Edit;
                      FieldByName('PROFIT_AND_LOSS').AsCurrency := p;
                      FieldByName('BALANCE_SHEET').AsCurrency := b;
                      Post;
                    finally OKToEdit := False;
                    end;
                  end;
              until not ReportLines.TblDetail.FindNext;
          until not ReportLines.TblMaster.FindNext;
      until not ReportLayouts.TblDetail.FindNext;
    ApplyUpdates(0);
  end;

begin
  inherited;
  // Attatch to ReportLayouts in order to identify P&L and Balance Sheet codesClose
  {with ManagementAccounts do
  begin
    Close;
    MasterComponent := nil;
    MasterFields := '';
    TblMaster.Params.Clear;
    Open;
  end;
  }
  //Close;
  MasterComponent := ReportLines;
  MasterTable := tbDetail;
  MasterFields := 'NOMINAL_CODE';
  //Open;
  // Calculate P&L or BS values for all codes
  try
    TotalProfit := 0;
    with ReportLayouts do
    begin
      if Locate('LAYOUT_TYPE','A',[]) then
      begin
        ProfitAndLoss := True;
        CalculateLedgerAccounts;
      end
      else
        raise LKError.Create(
          'Report Layout "A" not found. It must contain all Profit and Loss Nominal Codes');
      if Locate('LAYOUT_TYPE','N',[]) then
      begin
        ProfitAndLoss := False;
        CalculateLedgerAccounts;
      end
      else
        raise LKError.Create(
          'Report Layout "N" not found. It must contain all Balance Sheet Nominal Codes');
    end;
  finally
    //Close;
    MasterComponent := nil;
    MasterFields := '';
    TblMaster.Params.Clear;
    //Open;
    {
    with ManagementAccounts do
    begin
      Close;
      MasterComponent := ReportLines;
      MasterFields := 'NOMINAL_CODE';
      Open;
    end;
    }
  end;
  // Apply the total to the Balance Shet Profit and Loss code
  DecodeNominalAccount(CodeProfitAndLoss, N, C);
  if not TblMaster.Locate('NOMINAL_CODE', N, []) then
    raise
      LKError.Create('Profit and loss account balance sheet code not found ' +
      CodeProfitAndLoss);
  OKToEdit := True;
  try
    with TblMaster do
    begin
      Edit;
      //AddCurrencyToField(-TotalProfit,FieldByName('PROFIT_AND_LOSS'));
      AddCurrencyToField(TotalProfit, FieldByName('BALANCE_SHEET'));
      Post;
      //b := FieldByName('BALANCE_SHEET').AsCurrency;
    end;
    ApplyUpdates(0);
    try
      ManagementAccounts.Refresh;
      //ReportLines.Refresh;
    except on E: Exception do
      Showmessage(E.Message);
    end;
  finally OKToEdit := False;
  end;
end;


{
*********************** TLKLedgerCostContainedDocuments ************************
}
constructor TLKLedgerCostContainedDocuments.Create(AOwner: TComponent);
begin
  if not (AOwner is TLKLedgerCost) then
    raise LKError.Create('Program error: '
      + 'LKLedgerCostContainedDocuments.Owner should be a TLKLedgerCost');
  inherited Create(AOwner);
end;

function TLKLedgerCostContainedDocuments.GetContainerComponent: TLKLedgerCost;
begin
  Result := TLKLedgerCost(inherited ContainerComponent);
end;


{
*************************** TLKLedgerCostCostCentres ***************************
}
function TLKLedgerCostCostCentres.GettmDescription: string;
begin
  Result := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

procedure TLKLedgerCostCostCentres.SettmDescription(const Value: string);
begin
  TblMaster.FieldByName('DESCRIPTION').AsString := Value;
end;


{
*********************** TLKLedgerCostManagementAccounts ************************
}
procedure TLKLedgerCostManagementAccounts.DlgManagementAccountsPrintExecute;
begin
  with TDlgLKLedgerCostManagementAcsPrint.Create(Owner) do
  begin
    try
      Construct(Self, '');
      PopUpParent := FormView;
      ShowModal;
    finally
      Free;
    end;
  end;
end;


{
******************************** TLKLedgerCost *********************************
}
constructor TLKLedgerCost.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCostCentreCurrent := '';
  FHasPeriodEndProcedure := False; // done in Nominal ledger
end;

function TLKLedgerCost.AccountIsValid(const KeyValues: Variant; RaiseError:
    Boolean = False): Boolean;
var
  Cost: string;
begin
  Cost := Trim(KeyValues);
  Result := (Cost = '');
  if Result then
    Exit
  else
    Result := CostCentres.Locate('COST_CENTRE', Cost, []);
  if RaiseError and not Result then
    raise LKError.Create(NameFriendly + ' reference ' + Cost + ' not found');
end;

procedure TLKLedgerCost.AccountPost(const DetailRecord: array of const);
begin
  LKError.Create('Direct postings to ' + nameFriendly
    + ' are not allowed. All postings should be through Nominal Ledger.');
end;

procedure TLKLedgerCost.AddValueToAccount(CostCentre, NominalCode: string;
    Value: currency);
var
  Desc: string;
  WasOKToEdit: Boolean;
begin
  if (costCentre = '') then
    exit;
  if Locate(TblDetail.MasterFields,VarArrayOf([CostCentre,NominalCode]),[]) then
  begin
    WasOKToEdit := OKToEdit;
    OKToEdit := True;
    try
      with TblMaster do
      begin
        Edit;
        AddCurrencyToField(Value, FieldByName('ACCOUNT_BALANCE'));
        Post;
      end
    finally OKToEdit := WasOKToEdit;
    end;
  end
  else if CostCentres.Locate('COST_CENTRE', CostCentre, []) then
  begin
    with NominalLedger.NominalCodes.TblMaster do
    begin
      Locate('NOMINAL_CODE', NominalCode, []);
      Desc := FieldByName('DESCRIPTION').asString;
    end;
    InsertRecord([CostCentre, NominalCode, Desc, 0, 0, value]);
  end;
end;

procedure TLKLedgerCost.CreateManagementAccounts;
begin
  FManagementAccounts := TLKLedgerCostManagementAccounts.Create(Self);
end;

procedure TLKLedgerCost.CreateSubComponents;
begin
  inherited CreateSubComponents;
  FCostCentres := TLKLedgerCostCostCentres.Create(Self);
end;

procedure TLKLedgerCost.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerCost, FFormView);
  TFormLKLedgerCost(FormView).Construct(Self);
end;

function TLKLedgerCost.DlgSelectCostCentreExecute(AOwner: TComponent):
    TModalResult;
begin
  with CostCentres do
  begin
    if (FormView = nil) then
      CreateViewForm;
    FormView.Caption := 'Select Cost Centre';
    try
      Result := View(CostCentreCurrent, True, [mbOK, mbCancel]);
    finally FormView.Caption := NameFriendly;
    end;
  end;
end;

procedure TLKLedgerCost.DoAfterSetHub;
begin
  inherited DoAfterSetHub;
  if Assigned(AccountsHub) then
    AccountsHub.CostLedger := Self;
end;

function TLKLedgerCost.DoSetup(const ForceIfSettingsExist: Boolean):
    TModalResult;
begin
  Result := mrOK;
  if not (ForceIfSettingsExist or (Status = acsNotSetUp)) then
    exit;
  Open;
  Showmessage('You should enter the main categories of your '
    + NameFriendly);
  Result := ReportLayouts.View(Null, True, [mbOK, mbAbort]);
  case Result of
    mrAbort: exit;
  end;
  showmessage('You should now enter the assets which you want to track in your '
    + NameFriendly);
  Result := ReportLines.View(Null, True, [mbOK, mbAbort]);
  case Result of
    mrAbort: exit;
  else
    Status := acsBasic;
  end;
end;

function TLKLedgerCost.EndOfMonth: TModalResult;
begin
  Result := inherited EndOfMonth;
end;

function TLKLedgerCost.EndOfYear(ProgressText: string): TModalResult;
begin
  if not OKToEdit then
  begin
    Showmessage(NameFriendly +
      ' Year End must be done using the NominalLedger');
    Result := mrCancel;
    Exit;
  end;
  // clear the NL balances in inherited
  Result := inherited EndOfYear(ProgressText);
  ProgressText := 'Year end complete for ' + NameFriendly;
  Application.ProcessMessages;
end;

function TLKLedgerCost.GetManagementAccounts: TLKLedgerCostManagementAccounts;
begin
  Result := TLKLedgerCostManagementAccounts(inherited ManagementAccounts);
end;

function TLKLedgerCost.GetNominalLedger: TLKLedgerNominal;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.NominalLedger
  else
    Result := nil;
end;

function TLKLedgerCost.RecordNew: TModalResult;
begin
  Result := MessageDlg(
    'You cannot enter a new Account directly into the '+ NameFriendly
    + '.  You should create a new Cost Centre. Do you want to do this now?',
    mtInformation,[mbYes, mbNo],0);
  if (Result = mrYes) then
    Result := CostCentres.DlgViewExecute;
end;

procedure TLKLedgerCost.RefreshAccountCodes;
begin
  CostCentres.Refresh;
end;

procedure TLKLedgerCost.SetContainedComponentNames;
begin
  inherited SetContainedComponentNames;
  FCostCentres.Name := Name + 'CostCentres';
end;

procedure TLKLedgerCost.SetCostCentreCurrent(const Value: string);
begin
  if (Value > '') and not CostCentres.TblMaster.Locate('COST_CENTRE',Value,[]) then
    raise LKError.Create('Cost centre not found: ' + Value);
  FCostCentreCurrent := Value;
end;

procedure TLKLedgerCost.TrialBalanceCalculate(const AccountsDate: TDateTime);
var
  n: Currency;
begin
  inherited;
  with TblMaster do
    if FindFirst then
      repeat
        n := GetCurrency(FieldByName('OPENING_RESERVES')) +
          GetCurrency(FieldByName('OPENING_BALANCE')) +
          GetCurrency(FieldByName('ACCOUNT_BALANCE')) +
          GetCurrency(FieldByName('RESERVES'));
        OKToEdit := True;
        try
          Edit;
          FieldByName('PROFIT_AND_LOSS').AsCurrency := n;
          Post;
        finally OKToEdit := False;
        end;
      until not FindNext;
  ApplyUpdates(0);
  ManagementAccounts.Refresh;
end;


{
******************************* TLKSalesTaxCodes *******************************
}
constructor TLKSalesTaxCodes.Create(AOwner: TComponent);
begin
  if not (AOwner is TLKSalesTaxReturns) then
    raise LKError.Create('Program error: '
     + 'TLKSalesTaxReturnsContainedDocuments.Owner should be a TLKSalesTaxReturns');
  inherited Create(AOwner);
end;

procedure TLKSalesTaxCodes.DoMasterBeforePost;
begin
  inherited DoMasterBeforePost;
  if FieldByName('RATE_PERCENT').IsNull then
    Exit;
  if (FieldByName('RATE_PERCENT').AsInteger > 0) then
    Exit;
  Showmessage('Code "0" indicates a non ' +
    'SalesTaxable transaction so you should start at code "1"');
end;

function TLKSalesTaxCodes.GetContainerComponent: TLKSalesTaxReturns;
begin
  Result := TLKSalesTaxReturns(inherited ContainerComponent);
end;


{
****************************** TLKSalesTaxReturns ******************************
}
constructor TLKSalesTaxReturns.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
end;

procedure TLKSalesTaxReturns.CalculateSalesTaxReturnValues;
var
  SalesTaxPeriod: string;
  SalesTaxPeriodEnd: TDateTime;
  OutputGoods, OutputSalesTax, InputGoods, InputSalesTax: Currency;
  I: Integer;
begin
  Screen.Cursor := crHourglass;
  try
    with TblMaster do
    begin
      if State in [dsEdit, dsInsert] then
        Post;
      SalesTaxPeriod := FieldByName('VAT_PERIOD').AsString;
      SalesTaxPeriodEnd := GetDate(FieldByName('VAT_PERIOD_END'));
    end;
    if (Length(SalesTaxPeriod) <> 4) then
    begin
      Showmessage('Sales Tax period should be 4 numeric characters');
      Exit;
    end;
    with TblDetail do
      while FindLast do
        Delete;
    ApplyUpdates(0);
    with AccountsHub.Daybooks do
      for I := 0 to Count - 1 do
        if (Items[I] is TLKDaybookSalesTax) then
          TLKDaybookSalesTax(Items[I]).CalculateSalesTaxPeriodValues(
            TblDetail, SalesTaxPeriod, SalesTaxPeriodEnd);
    OutputGoods := 0;
    OutputSalesTax := 0;
    InputGoods := 0;
    InputSalesTax := 0;
    with TblDetail do
      if FindFirst then
        repeat
          if (GetInteger(FieldByName('IS_INPUT')) = 1) then
          begin
            InputGoods := InputGoods + GetCurrency(FieldByName('TOTAL_GOODS'));
            InputSalesTax := InputSalesTax + GetCurrency(FieldByName('TOTAL_VAT'));
          end
          else
          begin
            OutputGoods := OutputGoods + GetCurrency(FieldByName('TOTAL_GOODS'));
            OutputSalesTax := OutputSalesTax + GetCurrency(FieldByName('TOTAL_VAT'));
          end
        until not FindNext;
    with TblMaster do
    begin
      Edit;
      FieldByName('INPUT_GOODS').AsCurrency := InputGoods;
      FieldByName('INPUT_VAT').AsCurrency := InputSalesTax;
      FieldByName('OUTPUT_GOODS').AsCurrency := OutputGoods;
      FieldByName('OUTPUT_VAT').AsCurrency := OutputSalesTax;
      Post;
    end;
    ApplyUpdates(0);
  finally Screen.Cursor := crDefault;
  end;
end;

procedure TLKSalesTaxReturns.CreateSubComponents;
begin
  inherited CreateSubComponents;
  FSalesTaxCodes := TLKSalesTaxCodes.Create(Self);
end;

procedure TLKSalesTaxReturns.CreateViewForm;
begin
  Application.CreateForm(TFormLKRecordsSalesTaxReturns, FFormView);
  TFormLKRecordsSalesTaxReturns(FormView).Construct(Self);
end;

procedure TLKSalesTaxReturns.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKRecordsSalesTaxReturnsRecord, FFormViewRecord);
  TFormLKRecordsSalesTaxReturnsRecord(FormViewRecord).Construct(Self);
end;

procedure TLKSalesTaxReturns.DoAfterSetHub;
begin
  inherited DoAfterSetHub;
  if Assigned(AccountsHub) then
    AccountsHub.SalesTaxReturns := Self;
end;

function TLKSalesTaxReturns.DoSetup(const ForceIfSettingsExist: Boolean):
    TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  if (Result <> mrOK) or not ForceIfSettingsExist then
    Exit;
  Open;
  Showmessage('Enter VAT codes and percentages. Code "0" indicates a non ' +
    'VATable transaction so you should start at code "1" (eg Standard Rate VAT)');
  Result := SalesTaxCodes.View(Null, True, [mbOK, mbAbort]);
  if (Result <> mrOK) then
    exit;
  with TLKViewDialog.Create(Owner) do
  begin
    try
      Title := nameFriendly + ' setup';
      HeaderCaption := 'Enter your VAT Registration number';
      HeaderComments := 'Enterprises must register for Value Added Tax when turnover'
        + ' is expected to exceed the Registration limit. Time limits apply';
      DataType := vString;
      Value := SalesTaxRegistration;
      if Execute then
        SalesTaxRegistration := Value;
      Title := 'Enter the VAT Standard Rate code';
      DataType := vInteger;
      Value := DefaultStandardSalesTaxCode;
      if Execute then
        DefaultStandardSalesTaxCode := Value;
    finally Free;
    end;
  end;
  if assigned(NominalLedger) then
    NominalControlCode := NominalLedger.DlgNominalCodeSelectExecute(
      NameFriendly + ' Control account', NominalControlCode,
      'Enter the Balance Sheet Nominal Control account for VAT',
      False);
end;

procedure TLKSalesTaxReturns.DrillDownRecord;
var
  DbkID: string;
  DocNo: string;
begin
  if assigned(AccountsHub) then
  begin
    with TblDetail do
    begin
      DbkID := FieldByName('DAYBOOK_ID').asString;
      DocNo := FieldByName('VAT_PERIOD').asString;
    end;
    AccountsHub.DrillDownSalesTaxReturn(DbkID, DocNo);
  end;
end;

function TLKSalesTaxReturns.GetDefaultStandardSalesTaxCode: Integer;
begin
  Result := RecallInteger('Settings', 'DefaultStandardSalesTaxCode', 0, True);
end;

function TLKSalesTaxReturns.GetNominalControlCode: string;
begin
  Result := RecallString('Settings', 'NominalControlCode', '', True);
  if (Result = '') and assigned(NominalLedger) then
    Result := NominalLedger.SuspensePostingsCode;
end;

function TLKSalesTaxReturns.GetNominalLedger: TLKLedgerNominal;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.NominalLedger
  else
    Result := nil;
end;

function TLKSalesTaxReturns.GetSalesTaxRegistration: string;
begin
  Result := RecallString('Settings', 'SalesTaxRegistration', '', True);
end;

function TLKSalesTaxReturns.RecordNew: TModalResult;
var
  s, ms, ys: string;
  e: TDateTime;
  y, m, d: Word;
begin
  Result := mrAbort;
  with TblMaster do
    if FindLast then
      e := MonthEndThis(GetDate(FieldByName('VAT_PERIOD_END')) + 80)
    else
      e := MonthEndThis(Date);
  with TLKViewDialog.Create(Self) do
  try
    DataType := vDate;
    Title := 'New VAT Period';
    HeaderCaption := 'Enter the VAT Period end date';
    HeaderComments := 'After entry, use the [Calculate] button to calculate the '
      + ' VAT and Goods totals for the period.';
    Value := e;
    if not Execute then
      exit;
    e := Value;
  finally Free;
  end;
  if (FormViewRecord = nil) then
  begin
    CreateViewFormRecord;
    TFormLKRecordsSalesTaxReturnsRecord(FormViewRecord).Construct(Self);
  end;
  DecodeDate(e, y, m, d);
  ms := IntToStr(m);
  if Length(ms) = 1 then
    ms := '0' + ms;
  ys := IntToStr(y);
  if Length(ys) = 1 then
    ys := '0' + ys;
  if Length(ys) = 4 then
    ys := Copy(ys, 3, 2);
  s := ms + ys;
  InsertRecord([s, e]);
  Result := ViewRecord(Null, True, [mbOK, mbCancel]);
  if (Result <> mrOK) then
  begin
    s := NameFriendly;
    if (s[length(s)] = 's') then
      System.Delete(s, length(s), 1);
    if (MessageDlg('Do you want to cancel (delete) this ' + s, mtWarning,
      [mbYes, mbNo], 0)
      = mrYes) then
      if TblMaster.State = dsInsert then
        TblMaster.Cancel
      else
        RecordDelete;
  end;
  ApplyUpdates(0);
end;

function TLKSalesTaxReturns.SalesTax(const GoodsValue: Currency; SalesTaxCode:
    Integer): Currency;
begin
  Result := 0.00;
  //if (Status in [acsNotSetUp,acsPartSetUp,acsNotUsed]) then exit;
  if (SalesTaxCode = 0) then
    Exit;
  with SalesTaxCodes.TblMaster do
    if Locate('VAT_CODE',SalesTaxCode,[]) then
      Result := RoundCents(GoodsValue
        * GetCurrency(FieldByName('RATE_PERCENT')) / 100)
    else
      Showmessage('VAT code ' + IntToStr(SalesTaxCode) +
        ' not found in VAT Returns VAT Codes');
end;

procedure TLKSalesTaxReturns.SetContainedComponentNames;
begin
  inherited SetContainedComponentNames;
  FSalesTaxCodes.Name := Name + 'SalesTaxCodes';
end;

procedure TLKSalesTaxReturns.SetDefaultStandardSalesTaxCode(Value: Integer);
begin
  RememberInteger('Settings', 'DefaultStandardSalesTaxCode', Value, True);
end;

procedure TLKSalesTaxReturns.SetNominalControlCode(const Value: string);
begin
  RememberString('Settings', 'NominalControlCode', Value, True);
end;

procedure TLKSalesTaxReturns.SetSalesTaxRegistration(const Value: string);
begin
  if (Value > '') then
    RememberString('Settings', 'SalesTaxRegistration', Value, True);
end;


{
*************************** TLKLedgerNominalControl ****************************
}
constructor TLKLedgerNominalControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDlgSetupIsImplemented := True;
  BalanceFieldName := 'ACCOUNT_BALANCE';
  SetBalanceFieldPairs;
end;

procedure TLKLedgerNominalControl.AssignTo(Dest: TPersistent);
var
  DestNCL: TLKLedgerNominalControl;
begin
  inherited AssignTo(Dest);
  if not (Dest is TLKLedgerNominalControl) then
    exit;
  DestNCL := TLKLedgerNominalControl(Dest);
  DestNCL.NominalControlCode := NominalControlCode;
end;

procedure TLKLedgerNominalControl.CheckAccountBalancesToNominalControl(showOK:
    boolean);
var
  myBalance, NLBalance: Currency;
begin
  NLBalance := NominalControlBalance;
  myBalance := SumLedgerBalances;
  if not EqualCurrency(myBalance, NLBalance) then
    Showmessage(Format(NameFriendly +
      ' Balances %m do not equal Nominal Control balance %m.' +
      #10 + 'Difference is ', [myBalance, NLBalance])
      + FloatToStrF(myBalance - NLBalance, ffFixed, 15, 4))
  else if ShowOK then
    Showmessage(Format(NameFriendly +
      ' Balances %m agree with Nominal Control balance.', [myBalance]));
end;

procedure TLKLedgerNominalControl.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerNominalControl, FFormView);
  TFormLKLedgerNominalControl(FormView).Construct(Self);
end;

procedure TLKLedgerNominalControl.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerNominalControlRecord, FFormViewRecord);
  TFormLKLedgerNominalControlRecord(FormViewRecord).Construct(Self);
end;

procedure TLKLedgerNominalControl.DlgDiaryExecute;
begin
  if assigned(AccountsHub) then
    if assigned(AccountsHub.Diary) then
      with TLKLedgerNominalAnalysisDiaryDlg.Create(Owner) do
      begin
        try
          Construct(Self);
          ShowModal;
        finally Free;
        end;
      end;
end;

procedure TLKLedgerNominalControl.DlgJnlAccountToAccountExecute;
begin
  AccountsHub.Journal.NomControlLedgerAccountTransferExecute(Self);
end;

procedure TLKLedgerNominalControl.DlgJnlAccountToNominalExecute;
begin
  //TLKDaybookJournal(AccountsHub.Journal).NomControlLedgerNominalJournalExecute(Owner,Self);
  AccountsHub.Journal.NomControlLedgerNominalJournalExecute(Self);
end;

procedure TLKLedgerNominalControl.DlgNominalCodeSetupExecute;
begin
  if Assigned(NominalLedger) then
    NominalControlCode := NominalLedger.DlgNominalCodeSelectExecute(
      NameFriendly + ' Control account', NominalControlCode,
      'Enter the Balance Sheet Nominal account which will hold the Control '
      + 'running total for the ' + NameFriendly,
      False);
end;

function TLKLedgerNominalControl.DoSetup(const ForceIfSettingsExist: Boolean):
    TModalResult;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  case Result of
    mrAbort: Exit;
  end;
  DlgNominalCodeSetupExecute;
  Status := acsBasic;
end;

function TLKLedgerNominalControl.EndOfMonth: TModalResult;
begin
  Result := inherited EndOfMonth;
  CheckAccountBalancesToNominalControl(False);
end;

function TLKLedgerNominalControl.EndOfYear(ProgressText: string): TModalResult;
var
  I: Integer;
  MonthsToRetain: Integer;
begin
  Result := inherited EndOfYear(ProgressText);
  if (Result <> mrOK) then
    Exit;
  // get the number of months transactions to retain
  MonthsToRetain := RecallInteger('EndOfYear', 'MonthsToRetain', 24, True);
  if (MonthsToRetain < 12) or (MonthsToRetain > 240) then
    MonthsToRetain := 24;
  with TLKViewDialog.Create(Owner) do
  try
    DataType := vInteger;
    Title := nameFriendly + ' Year End';
    HeaderCaption := 'Delete audit trail for completed items in the ' + nameFriendly;
    HeaderComments := 'Enter the number of months prior to which matched'
      + ' documents and transactions will be removed.';
    Value := MonthsToRetain;
    Result := mrRetry;
    repeat
      if Execute then
      begin
        if (MonthsToRetain < 12) or (MonthsToRetain > 240) then
          ShowMessage('Months should be between 12 and 240')
        else
        begin
          Result := mrOK;
          if (Value <> MonthsToRetain) then
          begin
            RememberInteger('EndOfYear', 'MonthsToRetain', Value, True);
            MonthsToRetain := Value;
          end;
        end;
      end
      else
      begin
        Result := mrCancel;
        Showmessage('Old transactions for ' + nameFriendly + ' retained.');
      end;
    until not (Result = mrRetry);
  finally Free;
  end;
  if (Result <> mrOK) then
    exit;
  // ensure all transactions and documents are visible
  with Daybooks do
    for I := 0 to Count - 1 do
      with Items[I] do
      begin
        Close;
        TblMaster.WhereClause := '';
        if not ACGroup.Active then
          ACGroup.Open;
        Open;
        with InputDocuments do
        begin
          Filter := '';
          Filtered := False;
          TblDetail.Filter := '';
          TblDetail.Filtered := False;
        end;
        Filter := '';
        Filtered := False;
        TblDetail.Filter := '';
        TblDetail.Filtered := False;
      end;
  ProgressText := 'Clearing old transactions from ' + NameFriendly;
  Application.ProcessMessages;
  Filter := '';
  Filtered := False;
  // Delete the transactions and documents
  TblDetail.Filter := EndOfYearFilter(MonthsToRetain);
  TblDetail.Filtered := True;
  OKToEdit := True;
  try
    if FindFirst then
      repeat
        while TblDetail.FindFirst do
        begin
          DeleteInputDocument;
          TblDetail.Delete;
        end;
      until not FindNext;
    // Apply the updates
    with Daybooks do
      for I := 0 to Count - 1 do
        with Items[I] do
        begin
          ApplyUpdates(0);
          InputDocuments.ApplyUpdates(0);
        end;
    ApplyUpdates(0);
  finally
    OKToEdit := False;
    TblDetail.Filter := '';
    TblDetail.Filtered := False;
  end;
  ProgressText := 'Old transactions removed okay from ' + NameFriendly;
  Application.ProcessMessages;
end;

function TLKLedgerNominalControl.GetNominalControlCode: string;
begin
  Result := RecallString('Settings', 'NominalControlCode', '', True);
  if (Result = '') and assigned(NominalLedger) then
    Result := NominalLedger.SuspensePostingsCode;
end;

function TLKLedgerNominalControl.GetNominalLedger: TLKLedgerNominal;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.NominalLedger
  else
    Result := nil;
end;

function TLKLedgerNominalControl.GetSettings: TModalResult;
begin
  Result := inherited GetSettings;
  NominalControlCode := RecallString('Settings', 'NominalControlCode', ',,',
    True);
end;

function TLKLedgerNominalControl.NominalControlBalance: Currency;
var
  n, c: string;
begin
  DecodeNominalAccount(NominalControlCode, n, c);
  Result := 0;
  with NominalLedger.TblMaster do
    if Locate('NOMINAL_CODE', n, []) then
      Result := GetCurrency(FieldByName('OPENING_BALANCE'))
        + GetCurrency(FieldByName('ACCOUNT_BALANCE'));
end;

procedure TLKLedgerNominalControl.SaveSettings;
begin
  inherited SaveSettings;
  RememberString('Settings', 'NominalControlCode', NominalControlCode, True);
end;

procedure TLKLedgerNominalControl.SetBalanceFieldPairs;
begin
  with FMSumDPairFieldNames do
  begin
    Clear;
    if (BalanceFieldName = '') then
      exit;
    add(BalanceFieldName);
    add('+DEBIT');
    add(BalanceFieldName);
    add('-CREDIT');
  end;
end;

procedure TLKLedgerNominalControl.SetNominalControlCode(const Value: string);
begin
  RememberString('Settings', 'NominalControlCode', Value, True);
end;


{
****************************** TLKDaybookNominal *******************************
}
procedure TLKDaybookNominal.AssignTo(Dest: TPersistent);
var
  DestDB: TLKDaybookNominal;
begin
  inherited AssignTo(Dest);
  DestDB := TLKDaybookNominal(Dest);
  DestDB.EntryType := FEntryType;
end;

procedure TLKDaybookNominal.CreateViewForm;
begin
  Application.CreateForm(TFormLKDaybookNominal, FFormView);
  TFormLKDaybookNominal(FormView).Construct(Self);
end;

function TLKDaybookNominal.GetBatchDate: TDateTime;
begin
  if Assigned(Ledger) and (FBatchDate > Ledger.LedgerDate) then
    BatchDate := Ledger.LedgerDate;
  Result := inherited GetBatchDate;
end;

function TLKDaybookNominal.GetLedger: TLKLedgerNominalControl;
begin
  Result := TLKLedgerNominalControl(inherited Ledger);
end;

function TLKDaybookNominal.GetSettings: TModalResult;
begin
  Result := inherited GetSettings;
  FPostNominalTransactionDetail := TLKPostNominalTransactionDetail(
    RecallInteger('Settings', 'PostNominalAsBatch', 1, True));
end;

procedure TLKDaybookNominal.PbBeforeProcessBatch;
begin
  inherited PbBeforeProcessBatch;
  NominalBatchTotal := 0;
  NominalPostings.Clear;
end;

procedure TLKDaybookNominal.PbCheckBatchItemMasterIsValid;
var
  Myidxvalues: array[0..20] of variant;
  IdxFieldNames: TStringList;
  I: Integer;
  KeyValues: Variant;
begin
  inherited PbCheckBatchItemMasterIsValid;
  // Check that there is a Ledger Account to post to
  IdxFieldNames := TStringList.Create;
  try
    DelimitedStringToList(LedgerAccountNoFieldName, ';', IdxFieldNames);
    for I := 0 to IdxFieldNames.Count - 1 do
      Myidxvalues[I] := TblMaster.FieldByName(IdxFieldNames[I]).Value;

   // with Ledger.TblMaster do
    case IdxFieldNames.Count of
      1: KeyValues := Myidxvalues[0];
    else
      begin
        KeyValues := VarArrayCreate([0, IdxFieldNames.Count -1], varVariant);
        for I := 0 to IdxFieldNames.Count -1 do
          KeyValues[I] := Myidxvalues[I];
      end;
    end;
  {
    if not Ledger.Locate(LedgerAccountNoFieldName, KeyValues, []) then
      DlgDaybookErrors.AddError(['101', TblMDocumentReference, '',
        Ledger.NameFriendly + ' Account number ' + KeyValues + ' not found',
          'Display & correct']);
  }
    if not Ledger.TblMaster.Locate(LedgerAccountNoFieldName, KeyValues, []) then
      DlgDaybookErrors.AddError(['101', TblMDocumentReference, '',
        Ledger.NameFriendly + ' Account number ' + KeyValues + ' not found',
          'Display & correct']);
  finally IdxFieldNames.Free;
  end;
end;

procedure TLKDaybookNominal.PbFinalisePostings;
begin
  inherited PbFinalisePostings;
  PbPostNominalControl;
end;

procedure TLKDaybookNominal.PbPostNominalControl;
var
  NomCode, CostCentre: string;
  i: Integer;
begin
  case PostNominalTransactionDetail of
    ptdBatch: for i := 0 to NominalPostings.Count - 1 do
        with NominalPostings[i] do
        begin
          NominalLedger.AccountPost([LedgerTempURN, NominalCode, CostCentre,
            BatchDate, DaybookID,
              Null, BatchDescription, NameFriendly, DebitOrNull(Value),
              CreditOrNull(Value)]);
          Dec(LedgerTempURN);
        end;
  end; // case
  NominalPostings.Clear;
  // posting to nominal control
  DecodeNominalAccount(TLKLedgerNominalControl(Ledger).NominalControlCode,
    NomCode, CostCentre);
  if CurrencyIsZero(NominalBatchTotal) then
    exit;
  NominalLedger.AccountPost([LedgerTempURN, NomCode, CostCentre, BatchDate,
    DaybookID, Null,
      BatchDescription, NameFriendly, DebitOrNull(NominalBatchTotal),
      CreditOrNull(NominalBatchTotal)]);
  Dec(LedgerTempURN);
end;

procedure TLKDaybookNominal.PbProcessDocument;
begin
  inherited PbProcessDocument;
  PbPostCostAndNominalAnalysis;
end;

function TLKDaybookNominal.PbSetFilter: Boolean;
begin
  with TLKViewDialog.Create(Owner) do
  begin
    try
      Title := 'Batch date';
      HeaderCaption := 'Enter the Batch date';
      HeaderComments := 'All transactios up to this date will be posted to the Ledger'
        + ' provided that the date is not beyond the Ledger posting date.';
      DataType := vDate;
      Value := BatchDate;
      Result := Execute;
      if Result then
        BatchDate := Value;
      Result := inherited;
    finally Free;
    end;
  end;
end;

procedure TLKDaybookNominal.SetLedger(Value: TLKLedgerNominalControl);
begin
  if (Ledger = Value) then
    Exit;
  if Assigned(Ledger) then
    Ledger.DaybookRemove(Self);
  inherited Ledger := Value;
  if Assigned(Value) then
    Value.DaybookAdd(Self);
end;

procedure TLKDaybookNominal.SetPostNominalTransactionDetail(Value:
    TLKPostNominalTransactionDetail);
begin
  if (FPostNominalTransactionDetail <> Value) then
    RememberInteger('Settings', 'PostNominalAsBatch', Ord(Value), True);
  FPostNominalTransactionDetail := Value;
end;


{
****************************** TLKDaybookSalesTax ******************************
}
constructor TLKDaybookSalesTax.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FNominalPostings := TLKNominalPostingList.Create(True);
end;

destructor TLKDaybookSalesTax.Destroy;
begin
  FNominalPostings.Free;
  inherited Destroy;
end;

procedure TLKDaybookSalesTax.CalculateSalesTaxPeriodValues(SalesTaxSummary:
    TDataset; const SalesTaxPeriod: string; SalestaxPeriodEnd: TDateTime);
var
  GoodsTotals: array[0..100] of currency;
  SalesTaxTotals: array[0..100] of currency;
  I: Integer;
  MaxSalesTaxCodes: Integer;
  TotGoods, TotSalesTax: Currency;
  IsInput: Integer;
  HoldWhere: string;
  D,M,Y: Word;
begin
  with AccountsHub.SalesTaxReturns.SalesTaxCodes.TblMaster do
  begin
    if not FindLast then
      LKError.Create('No records found in the VAT Returns VAT Codes');
    MaxSalesTaxCodes := GetInteger(FieldByName('VAT_CODE'));
  end;
  if Active then
  begin
//    TblMaster.EmptyDataSet;
//    TblMaster.Data := Null;
    Close;
  end;
  HoldWhere := TblMaster.WhereClause;
  DecodeDate(SalestaxPeriodEnd, Y, M, D);
  TblMaster.WhereClause :=
    '( DATE_ <= '+Chr(39) + IntToStr(M)+'/'+IntToStr(D)+'/'+IntToStr(Y) + Chr(39)
    + ' ) AND ( ( VAT_PERIOD <= ' + Chr(39) + '    ' + Chr(39) + ' ) OR '
    + ' ( VAT_PERIOD IS NULL ) OR ( VAT_PERIOD = '+Chr(39)
    + SalesTaxPeriod + Chr(39) + ' ) )';
  Open;
  if not DocsHaveSalesTax then
  begin
    Close;
    TblMaster.WhereClause := HoldWhere;
    Exit;
  end;
  Screen.Cursor := crHourGlass;
  try
    if (EntryType = detDebits) then
      IsInput := 0
    else
      IsInput := 1;
      Filter := '';
      Filtered := False;
      for I := 0 to MaxSalesTaxCodes do
      begin
        GoodsTotals[I] := 0;
        SalesTaxTotals[I] := 0;
      end;
      if TblMaster.FindFirst then
      repeat
        with TblMaster do
        begin
          Edit;
          FieldByName('VAT_PERIOD').AsString := SalesTaxPeriod;
          Post;
        end;
        with TblDetail do
          if FindFirst then
            repeat
              I := GetInteger(FieldByName('VAT_CODE'));
              if I > MaxSalesTaxCodes then
              begin
                ShowMessage('Incorrect VAT Code (too large)');
                ViewRecord(Null, True, [mbAbort]);
                SysUtils.Abort;
              end
              else
              begin
                GoodsTotals[I] := GoodsTotals[I] + TblDGoods;
                SalesTaxTotals[I] := SalesTaxTotals[I] + TblDSalesTax;
              end;
            until not FindNext;
      until not FindNext;
      ApplyUpdates(0);
      with SalesTaxSummary do
      begin
        InsertRecord([SalesTaxPeriod, IsInput, DaybookID]);
        Edit;
        FieldByName('GOODS_0').AsCurrency := GoodsTotals[0];
        FieldByName('VAT_0').AsCurrency := SalesTaxTotals[0];
        TotGoods := 0;
        TotSalesTax := 0;
        for I := 1 to MaxSalesTaxCodes do
          if not ((GoodsTotals[I] = 0) and (SalesTaxTotals[I] = 0)) then
          begin
            FieldByName('GOODS_' + IntToStr(I)).AsCurrency := GoodsTotals[I];
            TotGoods := TotGoods + GoodsTotals[I];
            FieldByName('VAT_' + IntToStr(I)).AsCurrency := SalesTaxTotals[I];
            TotSalesTax := TotSalesTax + SalesTaxTotals[I];
          end;
        FieldByName('TOTAL_GOODS').AsCurrency := TotGoods;
        FieldByName('TOTAL_VAT').AsCurrency := TotSalesTax;
        Post;
      end;
  finally
    Screen.Cursor := crDefault;
    Close;
    TblMaster.WhereClause := HoldWhere;
  end;
end;

function TLKDaybookSalesTax.GetSalesTaxReturns: TLKSalesTaxReturns;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.SalesTaxReturns
  else
    Result := nil;
end;

function TLKDaybookSalesTax.GetTblDGoods: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('GOODS'));
end;

function TLKDaybookSalesTax.GetTblDSalesTax: Currency;
begin
  if DocsHaveSalesTax then
    Result := GetCurrency(TblDetail.FieldByName('VAT'))
  else
    Result := 0;
end;

function TLKDaybookSalesTax.GetTblDSalesTaxCode: Integer;
begin
  if DocsHaveSalesTax then
    Result := GetInteger(TblDetail.FieldByName('VAT_CODE'))
  else
    Result := 0;
end;

function TLKDaybookSalesTax.GetTblDTotal: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('TOTAL'));
end;

function TLKDaybookSalesTax.GetTblMGoods: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('GOODS'));
end;

function TLKDaybookSalesTax.GetTblMSalesTax: Currency;
begin
  if DocsHaveSalesTax then
    Result := GetCurrency(TblMaster.FieldByName('VAT'))
  else
    Result := 0;
end;

function TLKDaybookSalesTax.GetTblMTotal: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('TOTAL'));
end;

procedure TLKDaybookSalesTax.PbBeforeProcessBatch;
begin
  inherited PbBeforeProcessBatch;
  TotalSalesTax := 0;
end;

procedure TLKDaybookSalesTax.PbCheckBatchItemDetailIsValid;
var
  I: Integer;
  DocRef: string;
  Item: string;
  Goods, SalesTax, L: Currency;
begin
  inherited PbCheckBatchItemDetailIsValid;
  if not DocsHaveSalesTax then
    Exit;
  DocRef := TblMDocumentReference;
  Item := IntToStr(TblDDocumentItemNumber);
  Goods := TblDGoods;
  SalesTax := TblDSalesTax;
  I := TblDSalesTaxCode;
  if (I = 0) then
  begin
    if not CurrencyIsZero(SalesTax) then
      DlgDaybookErrors.AddError(['102', DocRef, Item,
        'Item has '+CurrToStr(SalesTax)+' VAT but the VAT code is zero',
          'Display & correct']);
  end
  else
  begin
    if Assigned(SalesTaxReturns) then
    begin
      L := SalesTax - SalesTaxReturns.SalesTax(Goods, I);
      if (L < -0.02) or (L > 0.02) then
        DlgDaybookErrors.AddWarning(['Y', DocRef, Item,
          'Item VAT amount not correct for VAT code.']);
    end
    else if not CurrencyIsZero(SalesTax) then
      DlgDaybookErrors.AddWarning(['Y', DocRef, Item,
        'VAT accounting is not set up but VAT code is ' + IntToStr(I)]);
  end;
end;

procedure TLKDaybookSalesTax.PbCheckBatchItemMasterIsValid;
var
  DocRef: string;
  Goods, SalesTax, Total, DtlGoods, DtlSalesTax: Currency;
begin
  inherited PbCheckBatchItemMasterIsValid;
  if not DocsHaveSalesTax then
    Exit;
  //TotalSalesTax := 0;
  Goods := TblMGoods;
  SalesTax := TblMSalesTax;
  Total := TblMTotal;
  if not EqualAmounts(Goods + SalesTax, Total) then
    DlgDaybookErrors.AddError(['102', DocRef, '',
      'Goods ' + FloatToStrF(Goods, ffCurrency, 15, 2)
        + ' plus VAT ' + FloatToStrF(SalesTax, ffCurrency, 4, 2)
        + ' do not equal total ' + FloatToStrF(Total, ffCurrency, 4, 2),
        'Display & correct']);
  // get the detail totals
  DtlGoods := 0;
  DtlSalesTax := 0;
  with TblDetail do
    if FindFirst then
      repeat
        DtlGoods := DtlGoods + TblDGoods;
        DtlSalesTax := DtlSalesTax + TblDSalesTax;
      until not Findnext;
  if not EqualAmounts(Goods, DtlGoods) then
    DlgDaybookErrors.AddError(['102', DocRef, '',
      'goods ' + FloatToStrF(Goods, ffCurrency, 15, 2) +
        ' does not equal goods analysis '
        + FloatToStrF(DtlGoods, ffCurrency, 15, 2),
        'Display & correct']);
  if not CurrencyIsZero(SalesTax) then
    if assigned(SalesTaxReturns) then
    begin
      if not equalAmounts(SalesTax, DtlSalesTax) then
        DlgDaybookErrors.AddError(['102', DocRef, '',
          'VAT ' + FloatToStrF(SalesTax, ffCurrency, 15, 2) +
            ' does not equal VAT analysis '
            + FloatToStrF(DtlSalesTax, ffCurrency, 15, 2),
            'Display & correct']);
    end
    else
      DlgDaybookErrors.AddError([DocRef, '',
        'VAT accounting is not set up but VAT is ' + FloatToStrF(SalesTax,
          ffCurrency, 15, 2),
          'Finish & correct', '999']);
end;

procedure TLKDaybookSalesTax.PbFinalisePostings;
begin
  inherited PbFinalisePostings;
  if DocsHaveSalesTax then
    PbPostSalesTaxControl;
end;

procedure TLKDaybookSalesTax.PbPostCostAndNominalAnalysis;
var
  NomCode, CostCent, LedgerAc: string;
  DocRef: string;
  Goods: Currency;
  I: Integer;
  TaxPoint: TDateTime;
  Desc: string;
  SalesTax: Currency;
begin
  case PostNominalTransactionDetail of
    ptdDocument: NominalPostings.Clear;
  end;
  if (LedgerAccountNoFieldName = '') then
    LedgerAc := ''
  else
    LedgerAc := TblMaster.FieldByName(LedgerAccountNoFieldName).AsString;
  TaxPoint := TblMDocumentDate;
  DocRef := TblMDocumentReference;
  with TblDetail do
    if FindFirst then
      repeat
        NomCode := TblDNominalCode;
        CostCent := TblDCostCentre;
        Desc := TblDDescription;
        Goods := TblDGoods;
        SalesTax := TblDSalesTax;
        // debit to control account therefore credit to nominal expense and income
        if (EntryType = detDebits) then
        begin
          Goods := -Goods;
          SalesTax := -SalesTax;
        end;
        if (Goods <> 0) then
          case PostNominalTransactionDetail of
            ptdDocument, ptdBatch: NominalPostings.Accumulate(NomCode, CostCent,
                Desc, Goods);
            ptdDocumentItems:
              begin
                NominalLedger.AccountPost([LedgerTempURN, NomCode, CostCent,
                  TaxPoint, DaybookID,
                    LedgerAc, DocRef, Desc, DebitOrNull(Goods),
                    CreditOrNull(Goods)]);
                Dec(LedgerTempURN);
              end;
          end;
        NominalBatchTotal := NominalBatchTotal - Goods - SalesTax;
        TotalSalesTax := TotalSalesTax + SalesTax;
      until not FindNext;
  case PostNominalTransactionDetail of
    ptdDocument: for I := 0 to NominalPostings.Count - 1 do
        with NominalPostings[I] do
        begin
          NominalLedger.AccountPost([LedgerTempURN, NominalCode, CostCentre,
            TaxPoint, DaybookID,
              LedgerAc, DocRef, Description, DebitOrNull(Value),
              CreditOrNull(Value)]);
          Dec(LedgerTempURN);
        end;
  end;
end;

procedure TLKDaybookSalesTax.PbPostSalesTaxControl;
var
  NomCode, CostCentre, Desc: string;
begin
  // Posting to SalesTax control
  if CurrencyIsZero(TotalSalesTax) or (SalesTaxReturns = nil) then
    Exit;
  Desc := 'Batch ' + intToStr(BatchNumber);
  DecodeNominalAccount(SalesTaxReturns.NominalControlCode, NomCode, CostCentre);
  NominalLedger.AccountPost([LedgerTempURN, NomCode, CostCentre, BatchDate,
    DaybookID,
      Null, Desc, NameFriendly, DebitOrNull(TotalSalesTax), CreditOrNull(TotalSalesTax)]);
  Dec(LedgerTempURN);
end;

procedure TLKDaybookSalesTax.PbTablesApplyUpdates;
begin
  if assigned(SalesTaxReturns) then
    SalesTaxReturns.ApplyUpdates(0);
  inherited PbTablesApplyUpdates;
end;

procedure TLKDaybookSalesTax.PbTablesCancelUpdates;
begin
  if assigned(SalesTaxReturns) then
    SalesTaxReturns.CancelUpdates;
  inherited PbTablesCancelUpdates;
end;

procedure TLKDaybookSalesTax.PbTablesRefresh;
begin
// TODO:
//if assigned(SalesTaxReturns) then
//  SalesTaxReturns.Refresh;
  inherited PbTablesRefresh;
end;

procedure TLKDaybookSalesTax.SetActive(Value: Boolean);
begin
  inherited SetActive(Value);
  if (csLoading in ComponentState) then
    Exit;
  if Value then
    DocsHaveSalesTax := Assigned(TblDetail.FindField('VAT'));
end;

procedure TLKDaybookSalesTax.SetTblDGoods(Value: Currency);
begin
  TblDetail.FieldByName('GOODS').AsCurrency := Value;
end;

procedure TLKDaybookSalesTax.SetTblDSalesTax(Value: Currency);
begin
  if DocsHaveSalesTax then
    TblDetail.FieldByName('VAT').AsCurrency := Value;
end;

procedure TLKDaybookSalesTax.SetTblDSalesTaxCode(Value: Integer);
begin
  if DocsHaveSalesTax then
    TblDetail.FieldByName('VAT_CODE').AsInteger := Value;
end;

procedure TLKDaybookSalesTax.SetTblDTotal(Value: Currency);
begin
  TblDetail.FieldByName('TOTAL').AsCurrency := Value;
end;

procedure TLKDaybookSalesTax.SetTblMGoods(Value: Currency);
begin
  TblMaster.FieldByName('GOODS').AsCurrency := Value;
end;

procedure TLKDaybookSalesTax.SetTblMSalesTax(Value: Currency);
begin
  if DocsHaveSalesTax then
    TblMaster.FieldByName('VAT').AsCurrency := Value;
end;

procedure TLKDaybookSalesTax.SetTblMTotal(Value: Currency);
begin
  TblMaster.FieldByName('TOTAL').AsCurrency := Value;
end;


{
***************************** TLKDocumentsNominal ******************************
}
procedure TLKDocumentsNominal.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKDocumentsNominalRecord, FFormViewRecord);
  TFormLKDocumentsNominalRecord(FormViewRecord).Construct(Self);
end;

function TLKDocumentsNominal.GetDaybook: TLKDaybookNominal;
begin
  Result := TLKDaybookNominal(inherited Daybook);
end;

procedure TLKDocumentsNominal.SetDaybook(Value: TLKDaybookNominal);
begin
  inherited Daybook := Value;
  MasterComponent := Value.Ledger;
end;


{
********************************** TLKJournal **********************************
}
constructor TLKJournal.Create(AOwner: TComponent);
begin
  inherited;
  DocumentRefLength := 8;
end;

procedure TLKJournal.ContactLedgerAccountTransferExecute(ALedger:
    TLKLedgerContact);
begin
  with TDlgContactLedgerAcToAc.Create(Self.Owner) do
  begin
    try
      Construct(Self, ALedger);
      ShowModal;
    finally
      Free;
    end;
  end;
end;

procedure TLKJournal.CreateSubComponents;
begin
  inherited CreateSubComponents;
  FSkeletonJournals := TLKJournalSkeletons.Create(Self);
end;

procedure TLKJournal.CreateViewForm;
begin
  Application.CreateForm(TFormLKJournal, FFormView);
  TFormLKJournal(FormView).Construct(Self);
end;

procedure TLKJournal.DoAfterSetHub;
begin
  inherited DoAfterSetHub;
  if Assigned(AccountsHub) then
    AccountsHub.Journal := Self;
end;

procedure TLKJournal.DoDetailAfterInsert;
begin
  TdCostCentre := NominalLedger.CostLedger.CostCentreCurrent;
  inherited DoDetailAfterInsert;
end;

function TLKJournal.GetDaybook: TLKDaybookJournal;
begin
  Result := TLKDaybookJournal(inherited Daybook);
end;

procedure TLKJournal.NomControlLedgerAccountTransferExecute(ALedger:
    TLKLedgerNominalControl);
begin
  with TDlgNomControlLedgerAcToAc.Create(Self.Owner) do
  begin
    try
      Construct(Self, ALedger);
      ShowModal;
    finally
      Free;
    end;
  end;
end;

procedure TLKJournal.NomControlLedgerNominalJournalExecute(ALedger:
    TLKLedgerNominalControl);
begin
  with TDlgNomControlLedgerAcToNom.Create(Self.Owner) do
  begin
    try
      Construct(Self, ALedger);
      ShowModal;
    finally
      Free;
    end;
  end;
end;

function TLKJournal.RecordNew: TModalResult;
begin
  with TDlgNominalAndCostJournal.Create(Owner) do
  begin
    try
      Construct(Self);
      Result := ShowModal;
    finally
      Free;
    end;
  end;
end;

procedure TLKJournal.SetContainedComponentNames;
begin
  inherited SetContainedComponentNames;
  FSkeletonJournals.Name := Name + 'SkeletonJournals';
end;

procedure TLKJournal.SetDaybook(Value: TLKDaybookJournal);
begin
  inherited Daybook := Value;
end;


{
***************************** TLKJournalSkeletons ******************************
}
constructor TLKJournalSkeletons.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Description := '';
end;

procedure TLKJournalSkeletons.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKDocumentsInputNominalRecord, FFormViewRecord);
  TFormLKDocumentsInputNominalRecord(FormViewRecord).Construct(Self,
    ContainerComponent.NominalLedger);
end;

procedure TLKJournalSkeletons.DoDetailAfterInsert;
begin
  inherited DoDetailAfterInsert;
  TblDetail.FieldByName('DESCRIPTION').AsString := Description;
end;

procedure TLKJournalSkeletons.DoDetailAfterPost;
begin
  inherited DoDetailAfterPost;
  Description := TblDetail.FieldByName('DESCRIPTION').AsString;
end;

procedure TLKJournalSkeletons.DoMasterAfterPost;
begin
  inherited DoMasterAfterPost;
  Description := TblMaster.FieldByName('DESCRIPTION').AsString;
end;

function TLKJournalSkeletons.GetContainerComponent: TLKJournal;
begin
  Result := TLKJournal(inherited ContainerComponent);
end;


{
***************************** TLKDocumentsSalesTax *****************************
}
constructor TLKDocumentsSalesTax.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  with MSumDPairFieldNames do
  begin
    Add('VAT');
    Add('VAT');
    Add('GOODS');
    Add('GOODS');
    Add('TOTAL');
    Add('TOTAL');
  end;
  FCopyFromExcludeFieldsMaster.Add('VAT_PERIOD');
  FSequenceActive := True;
end;

procedure TLKDocumentsSalesTax.CreateViewForm;
begin
  Application.CreateForm(TFormLKDocumentsSalesTax, FFormView);
  TFormLKDocumentsSalesTax(FormView).Construct(Self);
end;

procedure TLKDocumentsSalesTax.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKDocumentsSalesTaxRecord, FFormViewRecord);
  TFormLKDocumentsSalesTaxRecord(FormViewRecord).Construct(Self);
end;

function TLKDocumentsSalesTax.DaybookAdjusted(const ALedgerMatchValue:
    currency): Currency;
begin
  Result := ALedgerMatchValue;
  if (Daybook.EntryType = detCredits) then
    Result := -Result;
end;

procedure TLKDocumentsSalesTax.DoDetailBeforePost;
var
  Goods, SalesTax: Currency;
begin
  Goods := tdGoods;
  SalesTax := tdSalesTax;
  if (SalesTax = 0) and Assigned(SalesTaxReturns) then
  begin
    SalesTax := SalesTaxReturns.SalesTax(Goods, tdSalesTaxCode);
    tdSalesTax := SalesTax;
  end;
  tdTotal := Goods + SalesTax;
  inherited DoDetailBeforePost;
end;

function TLKDocumentsSalesTax.GetSalesTaxReturns: TLKSalesTaxReturns;
begin
  if Assigned(AccountsHub) then
    Result := AccountsHub.SalesTaxReturns
  else
    Result := nil;
end;

function TLKDocumentsSalesTax.GetTDGoods: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('GOODS'));
end;

function TLKDocumentsSalesTax.GetTDSalesTax: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('VAT'));
end;

function TLKDocumentsSalesTax.GetTDSalesTaxCode: Integer;
begin
  Result := GetInteger(TblDetail.FieldByName('VAT_CODE'));
end;

function TLKDocumentsSalesTax.GetTDTotal: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('TOTAL'));
end;

function TLKDocumentsSalesTax.GetTMGoods: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('GOODS'));
end;

function TLKDocumentsSalesTax.GetTMSalesTax: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('SalesTax'));
end;

function TLKDocumentsSalesTax.GetTMTotal: Currency;
begin
  Result := GetCurrency(TblMaster.FieldByName('TOTAL'));
end;

procedure TLKDocumentsSalesTax.Loaded;
begin
  inherited Loaded;
   if Assigned(Daybook) then
      Daybook.InputDocuments := Self;
end;

procedure TLKDocumentsSalesTax.SetTDGoods(Value: Currency);
begin
  TblDetail.FieldByName('GOODS').AsCurrency := Value;
end;

procedure TLKDocumentsSalesTax.SetTDSalesTax(Value: Currency);
begin
  TblDetail.FieldByName('VAT').AsCurrency := Value;
end;

procedure TLKDocumentsSalesTax.SetTDSalesTaxCode(Value: Integer);
begin
  TblDetail.FieldByName('VAT_CODE').AsInteger := Value;
end;

procedure TLKDocumentsSalesTax.SetTDTotal(Value: Currency);
begin
  TblDetail.FieldByName('TOTAL').AsCurrency := Value;
end;

procedure TLKDocumentsSalesTax.SetTMGoods(Value: Currency);
begin
  TblMaster.FieldByName('GOODS').AsCurrency := Value;
end;

procedure TLKDocumentsSalesTax.SetTMSalesTax(Value: Currency);
begin
  TblMaster.FieldByName('VAT').AsCurrency := Value;
end;

procedure TLKDocumentsSalesTax.SetTMTotal(Value: Currency);
begin
  TblMaster.FieldByName('TOTAL').AsCurrency := Value;
end;


{
******************************** TLKLedgerCash *********************************
}
function TLKLedgerCash.AccountCheckTotals(const showOK, correctErrors,
    raiseError: boolean): Boolean;
begin
  Result := inherited AccountCheckTotals(showOK, correctErrors, raiseError);
  RefreshBalances;
end;

procedure TLKLedgerCash.CreateViewForm;
begin
  Application.CreateForm(TFormCashLedger, FFormView);
  TFormCashLedger(FormView).Construct(Self);
end;

procedure TLKLedgerCash.DeleteInputDocument;
var
  D: TDateTime;
begin
  with TblDetail do
  begin
    ValueDeleted := ValueDeleted +
      GetCurrency(FieldByName('DEBIT')) - GetCurrency(FieldByName('CREDIT'));
    D := GetDate(FieldByName('DATE_'));
    if (D > PostingDate) then
      PostingDate := D;
  end;
  inherited DeleteInputDocument;
end;

procedure TLKLedgerCash.DoDetailBeforePost;
begin
  inherited doDetailBeforePost;
  if FNewDetailRecord then
    with TblDetail do
      FieldByName('BALANCE').AsCurrency :=
        +GetCurrency(FieldByName('DEBIT')) - GetCurrency(FieldByName('CREDIT'))
        + GetCurrency(TblMaster.FieldByName('ACCOUNT_BALANCE'));
end;

function TLKLedgerCash.EndOfYear(ProgressText: string): TModalResult;
begin
  ValueDeleted :=0;
  PostingDate := 0;
  Result := inherited EndOfYear(ProgressText);
  if not CurrencyIsZero(ValueDeleted) then
  begin
    AccountPost([-1, '', PostingDate, 'OPN', 'OPN', 'OPENING BALANCE',
      DebitOrNull(ValueDeleted), CreditOrNull(ValueDeleted), Null, Null, 1]);
    ApplyUpdates(0);
  end;
  ProgressText := 'Checking account balances for ' + NameFriendly;
  Application.ProcessMessages;
  RefreshBalances;
  ProgressText := 'Year end complete for ' + NameFriendly;
  Application.ProcessMessages;
end;

function TLKLedgerCash.RecordNew: TModalResult;
var
  transDate: TDateTime;
  acRef: string;
  OBal: Currency;
begin
  Result := inherited RecordNew;
  if (Result <> mrOK) then
    exit;
  acRef := TblMaster.Fields[0].asString;
  if (MessageDlg('Do you want to enter an Opening Balance on this account?',
    mtInformation, [mbYes, mbNo], 0) <> mrYes) then
    exit;
  transDate := MonthEndLast(LedgerDate);
  OBal := 0.00;
  with TLKViewDialog.Create(Owner) do
  begin
    try
      dataType := vCurrency;
      Title := 'Opening Balance';
      HeaderCaption := 'Enter the Opening Balance for the new account';
      HeaderComments := '(enter the amount as minus if overdrawn)';
      Value := OBal;
      execute;
    finally free;
    end;
  end;
  if not CurrencyIsZero(OBal) then
  begin
    AccountPost([-1, acRef, transDate, '', '', 'Opening balance',
      DebitOrNull(OBal), CreditOrNull(OBal)]);
    ApplyUpdates(0);
    Refresh;
  end;
end;

function TLKLedgerCash.RefreshBalances: Currency;
var
  OldFilter: string;
  WasFiltered: Boolean;
  wasOKToEdit: Boolean;
begin
  Result := 0;
  with TblDetail do
  begin
    // DisableControls;
    OldFilter := Filter;
    WasFiltered := Filtered;
    Filter := '';
    Filtered := False;
    screen.cursor := crHourglass;
    wasOKToEdit := OKToEdit;
    OKToEdit := True;
    RefreshMSumDPairs := False;
    try
      if FindFirst then
        repeat
          Result := Result +
            GetCurrency(FieldByName('DEBIT')) -
            GetCurrency(FieldByName('CREDIT'));
          Edit;
          FieldByName('BALANCE').AsCurrency := Result;
          Post;
          Application.ProcessMessages;
        until not Findnext;
    finally
      RefreshMSumDPairs := True;
      OldFilter := Filter;
      Filtered := WasFiltered;
      Screen.cursor := crDefault;
      ApplyUpdates(-1);
      OKToEdit := wasOKToEdit;
      //  EnableControls
    end;
  end;
end;


{
******************************** TLKLedgerBank *********************************
}
procedure TLKLedgerBank.CreateViewForm;
begin
  Application.CreateForm(TFormBankLedger, FFormView);
  TFormBankLedger(FormView).Construct(Self);
end;

procedure TLKLedgerBank.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerBankRecord, FFormViewRecord);
  TFormLKLedgerBankRecord(FormViewRecord).Construct(Self);
end;

function TLKLedgerBank.DlgBankAgreementExecute: TModalResult;
begin
  DlgBankUnpresentedOutstanding := TDlgBankUnpresentedOutstanding.Create(Self);
  with DlgBankUnpresentedOutstanding do
  try
    Construct(Self);
    Result := ShowModal;
  finally Free;
  end;
end;

function TLKLedgerBank.EndOfMonth: TModalResult;
begin
  if not Active then
    Open;
  with TblMaster do
    if findFirst then
      repeat DlgBankAgreementExecute;
      until not FindNext;
  Result := inherited EndOfMonth;
end;


{
******************************** TLKDaybookCash ********************************
}
constructor TLKDaybookCash.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  // EntryType := detCredits; //Payments cash book to bank/cash account
  PbLedgers := TLKLedgerList.Create(False);
end;

destructor TLKDaybookCash.Destroy;
begin
  PbLedgers.Free;
  inherited Destroy;
end;

function TLKDaybookCash.GetLedger: TLKLedgerCash;
begin
  Result := TLKLedgerCash(inherited Ledger);
end;

procedure TLKDaybookCash.PbPostDocumentToLedger;
var
  PaymentRef: string;
  LedgerRef, PayLedgerID, PayLedgerRef, PayLedgerDoc: string;
  InvDate: TDateTime;
  TransDate: TDateTime;
  Goods: Currency;
  ContactLedger: TLKLedgerContact;
  Desc: string;
  Total: Currency;
begin
  with TblMaster do
  begin
    //Get invoice information
    PaymentRef := Fields[0].asString;
    LedgerRef := Fields[3].asString;
    InvDate := GetDate(FieldByName('DATE_'));
    //Discount := GetCurrency(FieldByName('Discount'));
    //Goods := GetCurrency(FieldByName('Goods'));
    Total := GetCurrency(FieldByName('TOTAL'));
    if (EntryType = detCredits) then
      Total := -Total;
  end; // with TblMaster
  //Post Invoice to Cash (Bank) Ledger
  Ledger.AccountPost([LedgerTempURN, LedgerRef, InvDate, DaybookID, PaymentRef,
    DocumentDescription, DebitOrNull(Total), CreditOrNull(Total), Null,Null,1]);
  Dec(LedgerTempURN);
  //post payment and matching to contact ledgers
  if TblDetail.FindFirst then
  repeat
    with TblDetail do
    begin
      PayLedgerID := FieldByName('LEDGER_ID').asString;
      PayLedgerRef := FieldByName('LEDGER_REFERENCE').asString;
      PayLedgerDoc := FieldByName('LEDGER_DOCUMENT').asString;
      Goods := GetCurrency(FieldByName('GOODS'));
    end;
    Desc := tdDescription;
    if (EntryType = detDebits) then
      Goods := -Goods;
    ContactLedger :=
      TLKLedgerContact(AccountsHub.FindLedger(PayLedgerID));
    if Assigned(ContactLedger) then
    begin
      with PbLedgers do
        if (IndexOf(ContactLedger) < 0) then
        begin
          Add(ContactLedger);
          if not ContactLedger.Active then
            ContactLedger.Open;
        end;
//  post the bank payment.
//    Add to any existing items for this payment or post a fresh item
      with ContactLedger do
      begin
        if not Locate(TblDetail.MasterFields, PayLedgerRef, []) then
          raise LKError.Create('Problem locating Account ' + PayLedgerRef +
            ' in ' + NameFriendly);
        // should never happen as checked in PbCheckMaster
        // Match the invoice and payment
        try
          OKToEdit := True;
          if TblDetail.Locate('DOCUMENT_NUMBER', PayLedgerDoc, []) then
          begin
            // Allocated payment to invoice
            with TblDetail do
            begin
              TransDate := GetDate(FieldByName('DATE_'));
              Edit;
              AddCurrencyToField(Goods, FieldByName('UNPAID'));
              if CurrencyIsZero(FieldByName('UNPAID').AsCurrency) then
                FieldByName('CLEARED').AsDateTime := InvDate;
              Post; // updates Account Balance
            end;
            UpdateAgeing(TransDate, Goods);
            // allocate to the bank payment
            with TblDetail do
            begin
              if Locate('DOCUMENT_NUMBER', PaymentRef, []) then
                // thr Pament record is already entered on this account
              begin
                Edit;
                if (Goods > 0) then
                  AddCurrencyToField(Goods, FieldByName('DEBIT'))
                else
                  AddCurrencyToField(-Goods, FieldByName('CREDIT'));
                Post;
              end
              else  // post the Payment record
              begin
                AccountPost([LedgerTempURN, PayLedgerRef, InvDate, DaybookID,
                  PaymentRef, 'Payment', DebitOrNull(Goods),
                  CreditOrNull(Goods), 0, InvDate, Null, Null, 1]);
                Dec(LedgerTempURN);
              end;
            end;
          end
          else // unallocated payment
          begin
            AccountPost([LedgerTempURN, PayLedgerRef, InvDate, DaybookID,
              PaymentRef, Desc, DebitOrNull(Goods), CreditOrNull(Goods),
              Goods, InvDate, Null, Null, 1]);
            Dec(LedgerTempURN);
          end;
        finally OKToEdit := False;
        end;
      end;  //with ContactLedger
    end; // if assigned ContactLedger
  until not TblDetail.FindNext;
end;

procedure TLKDaybookCash.PbProcessedComplete;
begin
  PbLedgers.Clear;
  inherited PbProcessedComplete;
end;

procedure TLKDaybookCash.PbTablesApplyUpdates;
var
  I: Integer;
begin
  with PbLedgers do
    for I := 0 to Count - 1 do
      Items[I].ApplyUpdates(0);
  inherited PbTablesApplyUpdates;
end;

procedure TLKDaybookCash.PbTablesCancelUpdates;
var
  I: Integer;
begin
  with PbLedgers do
    for i := 0 to Count - 1 do
      Items[i].CancelUpdates;
  inherited PbTablesCancelUpdates;
end;

procedure TLKDaybookCash.PbTablesRefresh;
var I: Integer;
begin
  with PbLedgers do
    for I := 0 to Count - 1 do
      Items[I].TblDetail.Refresh;
  inherited PbTablesRefresh;
end;

procedure TLKDaybookCash.SetLedger(Value: TLKLedgerCash);
begin
  inherited Ledger := Value;
end;


{
***************************** TLKDaybookPettyCash ******************************
}
function TLKDaybookPettyCash.GetLedger: TLKLedgerCash;
begin
  Result := TLKLedgerCash(inherited Ledger);
end;

procedure TLKDaybookPettyCash.PbPostDocumentToLedger;
var
  DocRef, LedgerRef: string;
  InvDate: TDateTime;
  Goods, Total: Currency;
begin
  //Get invoice information
  LedgerRef := TblMaster.Fields[3].asString;

  InvDate := TblMDocumentDate;
  DocRef := TblMDocumentReference;
  Goods := TblMGoods;
  Total := TblMTotal;

  if (EntryType = detCredits) then
    Total := -Total
  else
    Goods := 0;
  //Post Invoice to Contact Ledger Cr else Dr
  with Ledger do
  begin
    AccountPost([LedgerTempURN, LedgerRef, InvDate, DaybookID, DocRef,
      DocumentDescription,
        DebitOrNull(Total), CreditOrNull(Total), Null, 1]);
    Dec(LedgerTempURN);
    OKToEdit := True;
    try
      with TblMaster do
      begin
        Edit;
        AddCurrencyToField(Goods, FieldByName('TURNOVER_THIS_YEAR'));
        Post;
      end;
    finally OKToEdit := False;
    end;
  end;
end;

procedure TLKDaybookPettyCash.SetLedger(Value: TLKLedgerCash);
begin
  inherited Ledger := Value;
end;


{
******************************** TLKDaybookBank ********************************
}
function TLKDaybookBank.GetLedger: TLKLedgerBank;
begin
  Result := TLKLedgerBank(inherited Ledger);
end;

procedure TLKDaybookBank.SetLedger(Value: TLKLedgerBank);
begin
  inherited Ledger := Value;
end;


{
***************************** TLKDocumentsCashBook *****************************
}
procedure TLKDocumentsCashBook.ContactLedgerPayment(AOwner: TComponent;
    ALedger: TLKLedgerContact; const APaymentDate: TDateTime);
var
  PayRef, NomCode, CostCentre, AccountRef, AccountName, desc: string;
  pay, Payment, Unallocated: Currency;
  ItemNo: Integer;
  docNo: string;
  myAccount: Variant;
  LedgerAllocateDlg: TLKContactLedgerAllocateDlg;

  label
    selectAgain;

begin
  payRef := tmDocumentReference;
  Payment := 0;
  Unallocated := 0;
  if not ALedger.Active then
    ALedger.ACGroup.Open;
  myAccount := tmAccount;
  selectAgain:
  with ALedger do
  begin
    case DlgSelectRecordExecute('Select account', myAccount) of
      mrCancel, mrRetry: abort;
    end;
    AccountRef := tmAccountReference;
    AccountName := tmAccountName;
    DecodeNominalAccount(NominalControlCode, NomCode, CostCentre);
  end;
  LedgerAllocateDlg := TLKContactLedgerAllocateDlg.Create(AOwner);
  with LedgerAllocateDlg do
  begin
    Construct(ALedger, False, (Daybook.EntryType = detDebits), APaymentDate);
    case showModal of
      mrCancel: abort;
      mrRetry: goto selectAgain;
    else
      begin
        Payment := DaybookAdjusted(-TotalPaid);
        Unallocated := DaybookAdjusted(-ValueUnallocated);
      end;
    end;
    if not CurrencyIsZero(Payment) then
    begin
      with TblDetail do
      begin
        ItemNo := GetInteger(FieldByName('ITEM_NUMBER'));
        Delete; // remove blank entry
      end;
      TempMatchList.First;
      while not TempMatchList.EOF do
      begin
        with TempMatchList do
        begin
          docNo := FieldByName('DOCUMENT_NUMBER').asString;
          desc := FieldByName('DESCRIPTION').asString;
          pay := DaybookAdjusted(GetCurrency(FieldByName('MATCH')));
          Next;
        end;
        if not CurrencyIsZero(pay) then
        begin
          TblDetail.InsertRecord([payRef, itemNo, NomCode, CostCentre,
            ALedger.LedgerID, AccountRef, docNo, Desc, 0, pay]);
          inc(itemNo);
        end;
      end;
      if not CurrencyIsZero(Unallocated) then
        TblDetail.InsertRecord([payRef, itemNo, NomCode, CostCentre,
          ALedger.LedgerID, AccountRef, '', UnallocatedDescription, 0,
            Unallocated]);
    end;
    // trigger update of TblMaster totals
    with TblDetail do
    begin
      Edit;
      Post;
    end;
  end;
end;

procedure TLKDocumentsCashBook.CreateViewForm;
begin
  Application.CreateForm(TFormLKDocumentsCashBook, FFormView);
  TFormLKDocumentsCashBook(FormView).Construct(Self);
end;

procedure TLKDocumentsCashBook.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKDocumentsCashBookRecord, FFormViewRecord);
  TFormLKDocumentsCashBookRecord(FormViewRecord).Construct(Self);
end;

function TLKDocumentsCashBook.DlgContactLedgerPaymentExecute(AOwner: TComponent;
    ALedger: TLKLedgerContact): Boolean;
var
  PayRef, NomCode, CostCentre, AccountName, a1, a2, a3, a4, docNo, desc, s: string;
  pay, Payment, Unallocated: Currency;
  PaymentDate: TDateTime;
  ItemNo: Integer;
  AccountRef: string;
  LedgerAllocateDlg: TLKContactLedgerAllocateDlg;
begin
  if not Active then
    ACGroup.Open;
  ItemNo := 1;
  Payment := 0;
  Unallocated := 0;
  with TLKViewDialog.Create(AOwner) do
  begin
    try
      Datatype := vDate;
      Title := ALedger.NameFriendly + ' Payment';
      HeaderCaption := 'Enter date of the payment';
      HeaderComments := 'Payment to an account in the ' + ALedger.NameFriendly;
      value := date;
      Result := execute;
      PaymentDate := value;
    finally Free;
    end;
  end;
  if not Result then
    exit;
  with ALedger do
  begin
    if not Active then
      ACGroup.Open;
    AccountRef := tmAccountReference;
    with TblMaster do
    begin
      // docRef := fields[0].asString;
      AccountName := fields[1].asString;
      a1 := FieldByName('INVOICE_ADDRESS1').asString;
      a2 := FieldByName('INVOICE_ADDRESS2').asString;
      a3 := FieldByName('INVOICE_ADDRESS3').asString;
      a4 := FieldByName('POSTCODE').asString;
    end;
  end;
  DecodeNominalAccount(ALedger.NominalControlCode, NomCode, CostCentre);
  LedgerAllocateDlg := TLKContactLedgerAllocateDlg.Create(AOwner);
  with TLKContactLedgerAllocateDlg(LedgerAllocateDlg) do
  begin
    Construct(ALedger, False, (Daybook.EntryType = detDebits), PaymentDate);
    btnBack.Visible := False;
    case showModal of
      mrCancel: abort;
    else
      begin
        Payment := DaybookAdjusted(-TotalPaid);
        Unallocated := DaybookAdjusted(-ValueUnallocated);
      end;
    end;
    if not CurrencyIsZero(Payment) then
    begin
      InsertRecord([nil, Null, Null, nil, PaymentDate, AccountName]);
      with TblMaster do
      begin
        Edit;
        if assigned(FindField('INVOICE_ADDRESS1')) then
        begin
          FieldByName('INVOICE_ADDRESS1').asString := a1;
          FieldByName('INVOICE_ADDRESS2').asString := a2;
          FieldByName('INVOICE_ADDRESS3').asString := a3;
          FieldByName('POSTCODE').asString := a4;
        end;
        FieldByName('GOODS').AsCurrency := Payment;
        FieldByName('VAT').AsCurrency := 0;
        FieldByName('TOTAL').AsCurrency := Payment;
        Post;
        payRef := fields[0].asString;
      end;
      TempMatchList.First;
      while not TempMatchList.EOF do
      begin
        with TempMatchList do
        begin
          docNo := FieldByName('DOCUMENT_NUMBER').asString;
          desc := FieldByName('DESCRIPTION').asString;
          pay := DaybookAdjusted(GetCurrency(FieldByName('MATCH')));
          Next;
        end;
        if not CurrencyIsZero(pay) then
        begin
          TblDetail.InsertRecord([payRef, ItemNo, NomCode, CostCentre,
            ALedger.LedgerID, AccountRef, docNo, Desc, 0, pay]);
          inc(ItemNo);
        end;
      end;
      if not CurrencyIsZero(Unallocated) then
        TblDetail.InsertRecord([payRef, ItemNo, NomCode, CostCentre,
          ALedger.LedgerID, AccountRef, payRef, UnallocatedDescription, 0,
            Unallocated]);
    end;
    // trigger update of TblMaster totals
    with TblDetail do
    begin
      Edit;
      Post;
    end;
  end;
  Result := (ViewRecord(Null, True, [mbOK, mbCancel]) = mrOK);
  if Result then
    DlgPrintExecute('', 'Default', payref, False)
  else
  begin
    s := NameFriendly;
    if (s[length(s)] = 's') then
      system.Delete(s, length(s), 1);
    Result := not (MessageDlg('Do you want to cancel (delete) this ' + s + '?',
      mtWarning, [mbYes, mbNo], 0) = mrYes);
    if not Result then
      TblMaster.Delete;
  end;
  ApplyUpdates(0);
end;

function TLKDocumentsCashBook.GetDaybook: TLKDaybookCash;
begin
  Result := TLKDaybookCash(inherited Daybook);
end;

function TLKDocumentsCashBook.GettmAccount: Variant;
begin
  try
    Result := TblMaster.Fields[5].Value;
  except Result := Null;
  end;
end;

procedure TLKDocumentsCashBook.Notification(AComponent: TComponent; Operation:
    TOperation);
begin
  if (Operation = opRemove) then
  begin
    if (AComponent = FDefaultPaymentLedger) then
      FDefaultPaymentLedger := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TLKDocumentsCashBook.SetDaybook(Value: TLKDaybookCash);
begin
  inherited Daybook := Value;
end;

procedure TLKDocumentsCashBook.SetDefaultPaymentLedger(Value: TLKLedgerContact);
begin
  FDefaultPaymentLedger := Value;
end;


{
***************************** TLKDocumentsBankBook *****************************
}
function TLKDocumentsBankBook.GetDaybook: TLKDaybookBank;
begin
  Result := TLKDaybookBank(inherited Daybook);
end;

procedure TLKDocumentsBankBook.SetDaybook(Value: TLKDaybookBank);
begin
  inherited Daybook := Value;
end;


{
**************************** TLKDocumentsPettyCash *****************************
}
function TLKDocumentsPettyCash.GetDaybook: TLKDaybookPettyCash;
begin
  Result := TLKDaybookPettyCash(inherited Daybook);
end;

procedure TLKDocumentsPettyCash.SetDaybook(Value: TLKDaybookPettyCash);
begin
  inherited Daybook := Value;
end;


{
******************************* TLKLedgerContact *******************************
}
constructor TLKLedgerContact.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FiltersDetail.Text := '(CURRENT_ > 0) OR (UNPAID <> 0)'
end;

function TLKLedgerContact.AccountCheckTotals(const showOK, correctErrors,
    raiseError: boolean): Boolean;
begin
  Result := inherited AccountCheckTotals(showOK, correctErrors, raiseError);
  RecalculateAgeing;
end;

procedure TLKLedgerContact.AccountPost(const DetailRecord: array of const);
var
  TransactionDate: TDateTime;
  Value: Currency;
  WasOKToEdit: Boolean;
begin
  inherited AccountPost(DetailRecord);
  WasOKToEdit := OKToEdit;
  OKToEdit := True;
  try
    with TblDetail do
    begin
      TransactionDate := GetDate(FieldByName('DATE_'));
      Value := GetCurrency(FieldByName('UNPAID'));
    end;
    UpdateAgeing(TransactionDate, Value);
  finally OKToEdit := WasOKToEdit;
  end;
end;

procedure TLKLedgerContact.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerContact, FFormView);
  TFormLKLedgerContact(FormView).Construct(Self);
end;

procedure TLKLedgerContact.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerContactRecord, FFormViewRecord);
  TFormLKLedgerContactRecord(FormViewRecord).Construct(Self);
end;

procedure TLKLedgerContact.DlgJnlAccountToAccountExecute;
begin
  AccountsHub.Journal.ContactLedgerAccountTransferExecute(Self);
end;

procedure TLKLedgerContact.DlgLedgerAllocateExecute;
begin
  with TLKContactLedgerAllocateDlg.Create(Owner) do
  begin
    try
      Construct(Self, True, True, LedgerDate);
      ShowModal;
    finally free;
    end;
  end;
  RecalculateAgeing;
end;

procedure TLKLedgerContact.DlgPrintMonthEndBalancesExecute;
begin
  with TblMaster do
  begin
    Filter := 'ACCOUNT_BALANCE < -0.005 OR ACCOUNT_BALANCE > 0.005';
    Filtered := True;
  end;
  with TblDetail do
  begin
    Filter := 'CURRENT_ = True';
    Filtered := True;
  end;
  try
    DlgPrintExecute(NameFriendly + ' Balances as at ' + dateToStr(LedgerDate),
      'Default', Null, False);
  finally
    with TblMaster do
    begin
      Filter := '';
      Filtered := False;
    end;
    with TblDetail do
    begin
      Filter := '';
      Filtered := False;
    end;
  end;
end;

procedure TLKLedgerContact.DoDetailBeforeDelete;
begin
  with TblDetail do
    UpdateAgeing(GetDate(FieldByName('DATE_')), -tdUnpaid);
  inherited DoDetailBeforeDelete;
end;

function TLKLedgerContact.EndOfMonth: TModalResult;
begin
  if not Active then
    ACGroup.Open;
  if (MessageDlg('Confirm that you want to print a list of the ' + NameFriendly
    + ' month end balances?', mtInformation, [mbYes, mbCancel], 0) = mrYes) then
      DlgPrintMonthEndBalancesExecute;
  Result := inherited EndOfMonth;
    // move cleared transactions to history
  Filter := '';
  Filtered := False;
  OKToEdit := True;
  RefreshMSumDPairs := False;
  Screen.Cursor := crHourglass;
  try
    {
    TblDetail.Filter := 'CURRENT_ = True AND UNPAID = 0';
    TblDetail.Filtered := True;
    if TblMaster.FindFirst then
    repeat
      with TblDetail do while FindLast do
      begin
        Edit;
        FieldByName('CURRENT_').Value := False;
        Post;
      end;
    until not TblMaster.FindNext;
    TblDetail.Filter := '';
    TblDetail.Filtered := False;
    ApplyUpdates(-1);
    if TblMaster.FindFirst then
    repeat
      RecalculateAgeing;
    until not TblMaster.FindNext;
    }

    if TblMaster.FindFirst then
    repeat
      if TblDetail.FindFirst then
      repeat
        if TDCurrent and CurrencyIsZero(TDUnpaid) then
        begin
          TblDetail.Edit;
          TDCurrent := False;
          TblDetail.Post;
        end;
      until not TblDetail.FindNext;
      RecalculateAgeing;
      ApplyUpdates(-1);
    until not TblMaster.FindNext;
  finally
    OKToEdit := False;
    Screen.Cursor := crDefault;
    RefreshMSumDPairs := True;
  end;
end;

function TLKLedgerContact.EndOfYearFilter(TransactionMonthsToRetain: Integer):
    string;
begin
  Result := 'DATE_ <= ' + chr(39)
    + DateToStr(MonthEndThis(LedgerDate - 15 - TransactionMonthsToRetain * 30.5))
    + chr(39) + ' AND UNPAID = 0';
end;

function TLKLedgerContact.GetTDCurrent: Boolean;
begin
  Result := (TblDetail.FieldByName('CURRENT_').Value = True);
end;

function TLKLedgerContact.GetTDUnpaid: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('UNPAID'));
end;

function TLKLedgerContact.GetTMTermsDays: Integer;
begin
  Result := GetInteger(TblMaster.FieldByName('CREDIT_TERMS_DAYS'));
end;

procedure TLKLedgerContact.RecalculateAgeing;
var
  value, cur, m1, m2, m3, o3: Currency;
  wMonth, wYear, tYear, tMonth, day: Word;
  years, months: Integer;
  wasOKToEdit: Boolean;
begin
  DecodeDate(LedgerDate, wYear, wMonth, day);
  cur := 0;
  m1 := 0;
  m2 := 0;
  m3 := 0;
  o3 := 0;
  wasOKToEdit := OKToEdit;
  OKToEdit := True;
  try
    with TblDetail do
    begin
      Filtered := False;
      Filter := '';
      if FindFirst then
      begin
        repeat
          Value := GetCurrency(FieldByName('UNPAID'));
          DecodeDate(getDate(FieldByName('DATE_')), tYear, tMonth, day);
          years := wYear - tYear;
          months := (years * 12) + wMonth - tMonth;
          if (months < 0) then
            months := 0;
          case months of
            0: cur := cur + Value;
            1: m1 := m1 + Value;
            2: m2 := m2 + Value;
            3: m3 := m3 + Value;
          else
            o3 := o3 + Value;
          end;
        until not FindNext;
      end;
    end;
    with TblMaster do
    begin
      Edit;
      FieldByName('CURRENT_').AsCurrency := cur;
      FieldByName('ONE_MONTH').AsCurrency := m1;
      FieldByName('TWO_MONTHS').AsCurrency := m2;
      FieldByName('THREE_MONTHS').AsCurrency := m3;
      FieldByName('OVER_3_MONTHS').AsCurrency := o3;
      Post;
    end;
    ApplyUpdates(-1);
  finally OKToEdit := wasOKToEdit;
  end;
end;

procedure TLKLedgerContact.SetBalanceFieldPairs;
begin
  with FMSumDPairFieldNames do
  begin
    Clear;
    if (BalanceFieldName = '') then
      Exit;
    Add(BalanceFieldName);
    Add('+UNPAID');
  end;
end;

procedure TLKLedgerContact.SetTDCurrent(Value: Boolean);
begin
  TblDetail.FieldByName('CURRENT_').Value := Value;
end;

procedure TLKLedgerContact.SetTDUnpaid(Value: Currency);
begin
  TblDetail.FieldByName('UNPAID').AsCurrency := Value;
end;

procedure TLKLedgerContact.SetTMTermsDays(Value: Integer);
begin
  TblMaster.FieldByName('CREDIT_TERMS_DAYS').AsInteger := Value;
end;

procedure TLKLedgerContact.UpdateAgeing(const TransactionDate: TDateTime;
    SignedValue: Currency);
var
  day, wMonth, wYear, tMonth, tYear: Word;
  months: Integer;
  wasOKToEdit: Boolean;
begin
  DecodeDate(LedgerDate, wYear, wMonth, day);
  DecodeDate(TransactionDate, tYear, tMonth, day);
  months := ((wYear - tYear) * 12) + wMonth - tMonth;
  if months < 0 then
    months := 0;
  wasOKToEdit := OKToEdit;
  OKToEdit := True;
  try
    with TblMaster do
    begin
      Edit;
      case months of
        0: AddCurrencyToField(SignedValue, FieldByName('CURRENT_'));
        1: AddCurrencyToField(SignedValue, FieldByName('ONE_MONTH'));
        2: AddCurrencyToField(SignedValue, FieldByName('TWO_MONTHS'));
        3: AddCurrencyToField(SignedValue, FieldByName('THREE_MONTHS'));
      else
        AddCurrencyToField(SignedValue, FieldByName('OVER_3_MONTHS'));
      end;
      Post;
    end;
  finally OKToEdit := wasOKToEdit;
  end;
end;


{
************************* TLKBankTransferPaymentsList **************************
}
constructor TLKBankTransferPaymentsList.Create(AOwner: TComponent);
begin
  if not (AOwner is TLKLedgerPurchase) then
    raise LKError.Create(
      'Program error TLKBankTransferPaymentsList.Owner should be a TLKLedgerPurchase');
  inherited Create(AOwner);
end;

function TLKBankTransferPaymentsList.GetBankPaymentsBook: TLKDocumentsBankBook;
begin
  if Assigned(ContainerComponent) then
    Result := ContainerComponent.BankPaymentsBook
  else
    Result := nil;
end;

function TLKBankTransferPaymentsList.GetContainerComponent: TLKLedgerPurchase;
begin
  Result := TLKLedgerPurchase(inherited ContainerComponent);
end;

procedure TLKBankTransferPaymentsList.MakePaymentListPayment;
var
//  docRef, PayRef, NomCode, CostCentre, AccountName, a1, a2, a3, a4, docNo, desc: string;
//  pay, Payment, Unallocated: Currency;
//  PaymentDate: TDateTime;
//  ItemNo: Integer;
  Okay: Boolean;
begin
  if not Assigned(BankPaymentsBook) then
    raise LKError.Create('The Bank Payments Book is not defined in this software');
  BankPaymentsBook.ACGroup.Open;
//  Payment := 0;
//  Unallocated := 0;
  with TLKViewDialog.Create(nil) do
  begin
    try
      Datatype := vDate;
      Title := NameFriendly + ' payment';
      HeaderCaption := 'Enter date of the payment';
      HeaderComments := 'Payment to an account in the ' + NameFriendly;
      Value := Date;
      Okay := Execute;
 //     PaymentDate := Value;
    finally Free;
    end;
  end;
  if not Okay then
    Exit;
  Showmessage('Bank payment not yet implemented');
    {
  with TblMaster do
  begin
    docRef := fields[0].asString;
    AccountName := fields[1].asString;
    a1 := FieldByName('INVOICE_ADDRESS1').asString;
    a2 := FieldByName('INVOICE_ADDRESS2').asString;
    a3 := FieldByName('INVOICE_ADDRESS3').asString;
    a4 := FieldByName('POSTCODE').asString;
  end;
  DecodeNominalAccount(NominalControlCode, NomCode, CostCentre);
  LKContactLedgerAllocateDlg := TLKContactLedgerAllocateDlg.Create(AOwner);
  try
    with LKContactLedgerAllocateDlg do
    begin
      btnBack.Visible := False;
      construct(Self, False, False, PaymentDate);
      case showModal of
        mrCancel: abort;
      else
        begin
          Payment := -TotalPaid;
          Unallocated := ValueUnallocated;
        end;
      end;
      if not CurrencyIsZero(Payment) then
      begin
        with Chequebook do
        begin
          InsertRecord([nil, Null, Null, nil, PaymentDate, AccountName, a1, a2,
            a3, a4]);
          payRef := TblMaster.fields[0].asString;
        end;
        TempMatchList.First;
        while not TempMatchList.EOF do
        begin
          with TempMatchList do
          begin
            docNo := FieldByName('DOCUMENT_NUMBER').asString;
            desc := FieldByName('DESCRIPTION').asString;
            pay := -GetCurrency(FieldByName('MATCH'));
            Next;
          end;
          if not CurrencyIsZero(pay) then
          begin
            Chequebook.TblDetail.InsertRecord([payRef, ItemNo, NomCode,
              CostCentre,
                LedgerID, docRef, docNo, Desc, 0, pay]);
            inc(ItemNo);
          end;
        end;
        if not CurrencyIsZero(Unallocated) then
        begin
          Chequebook.TblDetail.InsertRecord([payRef, ItemNo, NomCode,
            CostCentre,
              LedgerID, docRef, payRef, 'Unallocated', 0, Unallocated]);
        end;
      end;
    end;
    with ChequeBook do
      if (ViewRecord(Null, True, [mbOK, mbCancel]) = mrOK) then
        DlgPrintExecute('', 'Default', payref, False)
      else if (MessageDlg('Do you want to cancel (delete) this cheque?',
        mtWarning, [mbYes, mbNo], 0)
        = mrYes) then
        TblMaster.Delete;
  finally
    LKContactLedgerAllocateDlg.Free;
  end;
        }
end;


{
****************************** TLKLedgerPurchase *******************************
}
procedure TLKLedgerPurchase.AddPaymentToPaymentList;
var
  DocRef, PayRef, NomCode, CostCentre, AccountName, a1, a2, a3, a4, DocNo, Desc: string;
  Pay, Payment, Unallocated: Currency;
  ItemNo: Integer;
begin
  ItemNo := 1;
  Payment := 0;
  Unallocated := 0;
  with TblMaster do
  begin
    DocRef := fields[0].asString;
    AccountName := fields[1].asString;
    a1 := FieldByName('INVOICE_ADDRESS1').asString;
    a2 := FieldByName('INVOICE_ADDRESS2').asString;
    a3 := FieldByName('INVOICE_ADDRESS3').asString;
    a4 := FieldByName('POSTCODE').asString;
  end;
  DecodeNominalAccount(NominalControlCode, NomCode, CostCentre);
  LKContactLedgerAllocateDlg := TLKContactLedgerAllocateDlg.Create(nil);
  try
    with LKContactLedgerAllocateDlg do
    begin
      btnBack.Visible := False;
      Construct(Self, False, False, Date);
      case ShowModal of
        mrCancel: abort;
      else
        begin
          Payment := -TotalPaid;
          Unallocated := ValueUnallocated;
        end;
      end;
      if not CurrencyIsZero(Payment) then
      begin
        with BankTransferPaymentsList do
        begin
          InsertRecord([nil, Null, Null, nil, Null, AccountName, a1, a2, a3, a4]);
          PayRef := TblMaster.Fields[0].asString;
        end;
        TempMatchList.First;
        while not TempMatchList.EOF do
        begin
          with TempMatchList do
          begin
            DocNo := FieldByName('DOCUMENT_NUMBER').asString;
            Desc := FieldByName('DESCRIPTION').asString;
            Pay := -GetCurrency(FieldByName('MATCH'));
            Next;
          end;
          if not CurrencyIsZero(Pay) then
          begin
            FBankTransferPaymentsList.TblDetail.InsertRecord([PayRef, ItemNo,
             NomCode, CostCentre, LedgerID, DocRef, DocNo, Desc, 0, pay]);
            Inc(ItemNo);
          end;
        end;
        if not CurrencyIsZero(Unallocated) then
        begin
          BankTransferPaymentsList.TblDetail.InsertRecord([PayRef, ItemNo, NomCode,
            CostCentre, LedgerID, DocRef, payRef, 'Unallocated', 0, Unallocated]);
        end;
      end;
    end;
  finally
    LKContactLedgerAllocateDlg.Free;
  end;
end;

procedure TLKLedgerPurchase.CreateSubComponents;
begin
  inherited CreateSubComponents;
  FBankTransferPaymentsList := TLKBankTransferPaymentsList.Create(Self);
end;

procedure TLKLedgerPurchase.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerPurchase, FFormView);
  TFormLKLedgerPurchase(FormView).Construct(Self);
end;

procedure TLKLedgerPurchase.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerPurchaseRecord, FFormViewRecord);
  TFormLKLedgerPurchaseRecord(FormViewRecord).Construct(Self);
end;

procedure TLKLedgerPurchase.DrawCheque(AOwner: TComponent);
var
  docRef, PayRef, NomCode, CostCentre, AccountName, a1, a2, a3, a4, docNo, desc: string;
  pay, Payment, Unallocated: Currency;
  paymentDate: TDateTime;
  itemNo: Integer;
  okay: Boolean;
begin
  itemNo := 1;
  Chequebook.ACGroup.Open;
  Payment := 0;
  Unallocated := 0;
  with TLKViewDialog.Create(AOwner) do
  begin
    try
      Datatype := vDate;
      Title := NameFriendly + ' payment';
      HeaderCaption := 'Enter date of the payment';
      HeaderComments := 'Payment to an account in the ' + NameFriendly;
      value := date;
      okay := execute;
      PaymentDate := value;
    finally Free;
    end;
  end;
  if not okay then
    exit;
  with TblMaster do
  begin
    docRef := fields[0].asString;
    AccountName := fields[1].asString;
    a1 := FieldByName('INVOICE_ADDRESS1').asString;
    a2 := FieldByName('INVOICE_ADDRESS2').asString;
    a3 := FieldByName('INVOICE_ADDRESS3').asString;
    a4 := FieldByName('POSTCODE').asString;
  end;
  DecodeNominalAccount(NominalControlCode, NomCode, CostCentre);
  LKContactLedgerAllocateDlg := TLKContactLedgerAllocateDlg.Create(AOwner);
  try
    with LKContactLedgerAllocateDlg do
    begin
      btnBack.Visible := False;
      construct(Self, False, False, PaymentDate);
      case showModal of
        mrCancel: abort;
      else
        begin
          Payment := -TotalPaid;
          Unallocated := ValueUnallocated;
        end;
      end;
      if not CurrencyIsZero(Payment) then
      begin
        with Chequebook do
        begin
          InsertRecord([nil, Null, Null, nil, PaymentDate, AccountName, a1, a2,
            a3, a4]);
          payRef := TblMaster.fields[0].asString;
        end;
        TempMatchList.First;
        while not TempMatchList.EOF do
        begin
          with TempMatchList do
          begin
            docNo := FieldByName('DOCUMENT_NUMBER').asString;
            desc := FieldByName('DESCRIPTION').asString;
            pay := -GetCurrency(FieldByName('MATCH'));
            Next;
          end;
          if not CurrencyIsZero(pay) then
          begin
            Chequebook.TblDetail.InsertRecord([payRef, itemNo, NomCode,
              CostCentre,
                LedgerID, docRef, docNo, Desc, 0, pay]);
            inc(itemNo);
          end;
        end;
        if not CurrencyIsZero(Unallocated) then
        begin
          Chequebook.TblDetail.InsertRecord([payRef, itemNo, NomCode,
            CostCentre,
              LedgerID, docRef, payRef, 'Unallocated', 0, Unallocated]);
        end;
      end;
    end;
    with ChequeBook do
      if (ViewRecord(Null, True, [mbOK, mbCancel]) = mrOK) then
        DlgPrintExecute('', 'Default', payref, False)
      else if (MessageDlg('Do you want to cancel (delete) this cheque?',
        mtWarning, [mbYes, mbNo], 0)
        = mrYes) then
        TblMaster.Delete;
  finally
    LKContactLedgerAllocateDlg.Free;
  end;
end;

procedure TLKLedgerPurchase.Notification(AComponent: TComponent; Operation:
    TOperation);
begin
  if (Operation = opRemove) then
  begin
    if (AComponent = FBankPaymentsBook) then
      FBankPaymentsBook := nil;
    if (AComponent = FChequeBook) then
      FChequeBook := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TLKLedgerPurchase.SetBankPaymentsBook(Value: TLKDocumentsBankBook);
begin
  FBankPaymentsBook := Value;
  if assigned(Value) then
    Value.FreeNotification(Self);
end;

procedure TLKLedgerPurchase.SetChequeBook(Value: TLKDocumentsBankBook);
begin
  FChequeBook := Value;
  if assigned(Value) then
    Value.FreeNotification(Self);
end;

procedure TLKLedgerPurchase.SetContainedComponentNames;
begin
  inherited SetContainedComponentNames;
  FBankTransferPaymentsList.Name := Name + 'BankTransferPaymentsList';
end;


{
******************************* TLKSalesManagers *******************************
}
procedure TLKSalesManagers.Notification(AComponent: TComponent; Operation:
    TOperation);
begin
  if (Operation = opRemove) then
    if (AComponent = FSalesLedger) then
      FSalesLedger := nil;
  inherited Notification(AComponent, Operation);
end;

procedure TLKSalesManagers.SetSalesLedger(Value: TLKLedgerSales);
begin
  if assigned(Value) then
    Value.FreeNotification(Self);
  FSalesLedger := Value;
end;


{
******************************** TLKLedgerSales ********************************
}
procedure TLKLedgerSales.CreateViewForm;
begin
  Application.CreateForm(TFormLKLedgerSales, FFormView);
  TFormLKLedgerSales(FormView).Construct(Self);
end;

procedure TLKLedgerSales.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKLedgerSalesRecord, FFormViewRecord);
  TFormLKLedgerSalesRecord(FormViewRecord).Construct(Self);
end;

function TLKLedgerSales.EndOfMonth: TModalResult;
begin
  if Assigned(CustomerStatements) then
    if (messageDlg('Confirm that you want to print customer Statements?',
      mtInformation, [mbYes, mbCancel], 0) = mrYes) then
      PrintCustomerStatements;
  Result := inherited EndOfMonth;
end;

procedure TLKLedgerSales.Notification(AComponent: TComponent; Operation:
    TOperation);
begin
  if (Operation = opRemove) then
  begin
    if (AComponent = FCustomerStatements) then
      FCustomerStatements := nil;
    if (AComponent = FSalesManagers) then
      FSalesManagers := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TLKLedgerSales.PrintCustomerStatements;
begin
  {with TLKLedgerSales.create(Owner) do
   begin
   try
   Assign(Self);
   MasterComponent := Self.MasterComponent;
   MasterFields := Self.MasterFields;
  // TblMaster.TableName := Self.TblMaster.TableName;//Clone := Self.TblMaster;
  // TblDetail.Clone := Self.TblDetail;
   TblDetail.IndexDefs := Self.TblDetail.IndexDefs;
   Name := Self.Name + 'Statements';
   NameFriendly := 'Customer Statements';
   Filter := 'ACCOUNT_BALANCE > 5';
   Filtered := True;
   with TblDetail do
    begin
    Filter := 'CURRENT_ = True';
    Filtered := True;
    end;
   Open;
   // with TblDetail.IndexDefs.Add do
  // TblDetail.IndexName := Self.TblDetail.IndexName;
  // TblDetail.MasterFields := Self.TblDetail.MasterFields;
   View(Null,True,[mbOK])
   finally Close; free; end;
   end;
  }

  if assigned(FCustomerStatements) then
    with FCustomerStatements do
    begin
      ACGroup := Self.ACGroup;
      FolderLocalSettings := ACGroup.FolderLocalSettings;
      FolderSharedSettings := ACGroup.FolderSharedSettings;
      Open;
      // View(Null,True,[mbOK]);
      FCustomerStatements.DlgPrintExecute('Statement of Account', 'Default',
        Null, False);
    end;
end;

procedure TLKLedgerSales.SetCustomerStatements(Value: TLKRecordsAccounts);
begin
  if FCustomerStatements <> Value then
  begin
    FCustomerStatements := Value;
  end;
  if assigned(Value) then
    Value.FreeNotification(Self);
end;

procedure TLKLedgerSales.SetSalesManagers(Value: TLKSalesManagers);
begin
  if assigned(Value) then
    Value.FreeNotification(Self);
  FSalesManagers := Value;
end;


{
****************************** TLKDaybookContact *******************************
}
function TLKDaybookContact.GetLedger: TLKLedgerContact;
begin
  Result := TLKLedgerContact(inherited Ledger);
end;

function TLKDaybookContact.GetTblMDueDate: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('DUE_DATE'));
end;

procedure TLKDaybookContact.PbPostDocumentToLedger;
var
  DocRef, LedgerRef: string;
  InvDate, DueDate: TDateTime;
  Goods, Total: Currency;
  Query: Variant;
begin
  InvDate := TblMDocumentDate;
  DueDate := TblMDueDate;

  with TblMaster do
  begin
    //Get invoice information
    LedgerRef := FieldByName(LedgerAccountNoFieldName).asString;
    {
    Description := FieldByName('ORDER_NUMBER').AsString;
    if assigned(FindField('QUOTATION')) then
     Description := FieldByName('QUOTATION').AsString+ ' ' + Description;
    if assigned(FindField('SUPPLIER_INVOICE_NUMBER'))
     then Description := FieldByName('SUPPLIER_INVOICE_NUMBER').AsString+ ' ' + Description;
     }
   // Discount := GetCurrency(FieldByName('DISCOUNT'));
    if Assigned(FindField('QUERY')) then
      Query := FieldByName('QUERY').Value
    else Query := Null;
  end; // with TblMaster

  DocRef := TblMDocumentReference;
  Goods := TblMGoods;
  Total := TblMTotal;
  if (EntryType = detCredits) then
    Total := -Total
  else
    Goods := 0;
  //Post Invoice to Contact Ledger Cr else Dr
  with Ledger do
  begin
    if (DueDate = 0) then
    begin
      Duedate := GetFloat(TblMaster.FieldByname('CREDIT_TERMS_DAYS'));
      if DueDate = 0 then
        Duedate := 30;
      DueDate := DueDate + InvDate;
    end;
    AccountPost([LedgerTempURN, LedgerRef, InvDate, DaybookID, docRef,
      DocumentDescription, DebitOrNull(Total), CreditOrNull(Total),
        Total, DueDate, Query, Null, 1]);
    Dec(LedgerTempURN);
    OKToEdit := True;
    try
      with TblMaster do
      begin
        Edit;
        AddCurrencyToField(Goods, FieldByName('TURNOVER_THIS_YEAR'));
        Post;
      end;
    finally OKToEdit := False;
    end;
  end;
  //  adjustAgeing(theWorkDate,InvDate,Total); done in account post

   // loop back to next invoice until finish
end;

procedure TLKDaybookContact.SetLedger(Value: TLKLedgerContact);
begin
  inherited Ledger := Value;
end;

procedure TLKDaybookContact.SetTblMDueDate(Value: TDateTime);
begin
  TblMaster.FieldByName('DUE_DATE').AsDateTime := Value;
end;


{
****************************** TLKDaybookPurchase ******************************
}
constructor TLKDaybookPurchase.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  EntryType := detCredits;
end;

function TLKDaybookPurchase.GetLedger: TLKLedgerPurchase;
begin
  Result := TLKLedgerPurchase(inherited Ledger);
end;

procedure TLKDaybookPurchase.SetLedger(Value: TLKLedgerPurchase);
begin
  inherited Ledger := Value;
end;


{
******************************* TLKDaybookSales ********************************
}
constructor TLKDaybookSales.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  EntryType := detDebits;
end;

function TLKDaybookSales.GetLedger: TLKLedgerSales;
begin
  Result := TLKLedgerSales(inherited Ledger);
end;

procedure TLKDaybookSales.SetLedger(Value: TLKLedgerSales);
begin
  inherited Ledger := Value;
end;


{
***************************** TLKDocumentsContact ******************************
}
constructor TLKDocumentsContact.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ContactNameField := 4;
  ContactNameFieldLedger := 1;
end;

procedure TLKDocumentsContact.CreateViewFormRecord;
begin
  Application.CreateForm(TFormLKDocumentsContactRecord, FFormViewRecord);
  TFormLKDocumentsContactRecord(FormViewRecord).Construct(Self);
end;

procedure TLKDocumentsContact.DoDetailBeforePost;
begin
  tdGoods := RoundCents(tdRate * tdQuantity);
  inherited DoDetailBeforePost;
end;

procedure TLKDocumentsContact.DoMasterAfterInsert;
begin
  if Assigned(MasterComponent) then
    TblMaster.Fields[ContactNameField].AsString
      := MasterComponent.TblMaster.Fields[ContactNameFieldLedger].AsString;
  inherited DoMasterAfterInsert;
end;

procedure TLKDocumentsContact.DoMasterBeforePost;
begin
  if assigned(MasterComponent) and (tmDueDate = 0) then
    tmDueDate := tmDocumentDate + MasterComponent.tmTermsDays;
  inherited DoMasterBeforePost;
end;

function TLKDocumentsContact.GetMasterComponent: TLKLedgerContact;
begin
  Result := TLKLedgerContact(inherited MasterComponent);
end;

function TLKDocumentsContact.GetTDQuantity: Double;
begin
  Result := GetFloat(TblDetail.FieldByName('QUANTITY'));
end;

function TLKDocumentsContact.GetTDRate: Currency;
begin
  Result := GetCurrency(TblDetail.FieldByName('RATE'));
end;

function TLKDocumentsContact.GettmDueDate: TDateTime;
begin
  Result := GetDate(TblMaster.FieldByName('DUE_DATE'));
end;

procedure TLKDocumentsContact.SetMasterComponent(Value: TLKLedgerContact);
begin
  inherited MasterComponent := Value;
end;

procedure TLKDocumentsContact.SetTDQuantity(Value: Double);
begin
  TblDetail.FieldByName('QUANTITY').AsFloat := Value;
end;

procedure TLKDocumentsContact.SetTDRate(Value: Currency);
begin
  TblDetail.FieldByName('RATE').AsCurrency := Value;
end;

procedure TLKDocumentsContact.SettmDueDate(Value: TDateTime);
begin
  TblMaster.FieldByName('DUE_DATE').AsDateTime := Value;
end;


{
************************* TLKDocumentsPurchaseInvoices *************************
}
function TLKDocumentsPurchaseInvoices.GetDaybook: TLKDaybookPurchase;
begin
  Result := TLKDaybookPurchase(inherited Daybook);
end;

procedure TLKDocumentsPurchaseInvoices.SetDaybook(Value: TLKDaybookPurchase);
begin
  inherited Daybook := Value;
end;


{
************************** TLKDocumentsSalesInvoices ***************************
}
constructor TLKDocumentsSalesInvoices.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  NextSalesInvoiceDate := date;
  FDefaultSalesNominalCode := '';
end;

procedure TLKDocumentsSalesInvoices.CreateViewForm;
begin
  Application.CreateForm(TFormLKDocumentsSalesInvoices, FFormView);
  TFormLKDocumentsSalesInvoices(FormView).Construct(Self);
end;

procedure TLKDocumentsSalesInvoices.DoMasterAfterInsert;
begin
  TblMaster.FieldByName('DATE_').AsDateTime := NextSalesInvoiceDate;
  inherited DoMasterAfterInsert;
end;

procedure TLKDocumentsSalesInvoices.DoMasterAfterPost;
begin
  inherited DoMasterAfterPost;
  NextSalesInvoiceDate := GetDate(TblMaster.FieldByName('DATE_'));
end;

function TLKDocumentsSalesInvoices.DoSetup(const ForceIfSettingsExist:
    Boolean): TModalResult;
var
  s: string;
begin
  Result := inherited DoSetup(ForceIfSettingsExist);
  if ForceIfSettingsExist or (DefaultSalesNominalCode = '') then
  begin
    s := NominalLedger.DlgNominalCodeSelectExecute(
      'Default ' + NameFriendly
      + ' Sales Profit & Loss account Nominal code', DefaultSalesNominalCode +
      '||', '', False);
    if (s = '') then
      Result := mrAbort
    else
    begin
      if (pos('|', s) > 0) then
        s := System.Copy(s, 1, pos('|', s) - 1);
      DefaultSalesNominalCode := s;
    end;
  end;
end;

function TLKDocumentsSalesInvoices.GetDaybook: TLKDaybookSales;
begin
  Result := TLKDaybookSales(inherited Daybook);
end;

function TLKDocumentsSalesInvoices.GetDefaultSalesNominalCode: string;
begin
  if (FDefaultSalesNominalCode = '') and not (csLoading in ComponentState) then
    FDefaultSalesNominalCode := RecallString('Settings',
      'DefaultSalesNominalCode', '', True);
  Result := FDefaultSalesNominalCode;
end;

procedure TLKDocumentsSalesInvoices.SetDaybook(Value: TLKDaybookSales);
begin
  inherited Daybook := Value;
end;

procedure TLKDocumentsSalesInvoices.SetDefaultSalesNominalCode(const Value:
    string);
begin
  if (Value > '') and (Value <> FDefaultSalesNominalCode) then
    RememberString('Settings', 'DefaultSalesNominalCode', Value, True);
  FDefaultSalesNominalCode := Value;
end;


{
****************************** TLKDaybookJournal *******************************
}
constructor TLKDaybookJournal.Create(AOwner: TComponent);
begin
  inherited;
  PbUpdatedLedgers := TLKLedgerList.Create(False);
end;

destructor TLKDaybookJournal.Destroy;
begin
  PbUpdatedLedgers.Free;
  inherited Destroy;
end;

function TLKDaybookJournal.GetBatchDate: TDateTime;
begin
  if Assigned(NominalLedger) and (FBatchDate > NominalLedger.LedgerDate) then
    BatchDate := NominalLedger.LedgerDate;
  Result := inherited GetBatchDate;
end;

function TLKDaybookJournal.GetLedger: TLKLedgerNominal;
begin
  Result := TLKLedgerNominal(inherited Ledger);
end;

procedure TLKDaybookJournal.PbBeforeProcessBatch;
begin
  Ledger := NominalLedger;
  inherited PbBeforeProcessBatch;
end;

procedure TLKDaybookJournal.PbCheckBatchItemDetailIsValid;
var
  CLedgerID, CRef: string;
  ErrMsg: string;
begin
  inherited PbCheckBatchItemDetailIsValid;
  with TblDetail do
  begin
    CLedgerID := FieldByName('LEDGER_ID').asString;
    CRef := FieldByName('LEDGER_REFERENCE').asString;
  end;
  if not ((CLedgerID > ' ') or (CRef > ' ')) then
    Exit; // nominal posting only
  ErrMsg := '';
  if (CLedgerID <= ' ') then
    ErrMsg := 'Ledger identification not entered.'
  else if not Assigned(AccountsHub.Ledgers.ComponentByID(CLedgerID)) then
    ErrMsg := 'Ledger identification not correct '+CLedgerID;
  if (CRef <= ' ') then
    ErrMsg := ErrMsg + 'Ledger account number not entered. ';
  if (ErrMsg > '') then
    DlgDaybookErrors.AddError(['101', TblMDocumentReference, '', ErrMsg,
      'display and correct']);
end;

procedure TLKDaybookJournal.PbPostDocumentToLedger;
var
  JournalDate: TDateTime;
  Dr, Cr, Value: Currency;
  CLedgerID, CRef, NomCode, CostCentre, Description: string;
  NomControlLedger: TLKLedgerBase;
  JnlNo: string;
begin
  JnlNo := TblMDocumentReference;
  JournalDate := GetDate(TblMaster.FieldByName('DATE_'));
  with TblDetail do
    if FindFirst then
      repeat
        // Nominal posting
        NomCode := TblDNominalCode;
        CostCentre := TblDCostCentre;
        CRef := FieldByName('LEDGER_REFERENCE').AsString;
        Description := TblDDescription;
        Dr := GetCurrency(FieldByName('DEBIT'));
        Cr := GetCurrency(FieldByName('CREDIT'));
        if (Dr <> 0) or (Cr <> 0) then
          Ledger.AccountPost([LedgerTempURN, NomCode, CostCentre, JournalDate,
            DaybookID,
              CRef, JnlNo, Description, CurrencyOrNull(Dr),
                CurrencyOrNull(Cr)]);
        // Nominal analysis posting
        if (CRef > ' ') then
        begin
          CLedgerID := FieldByName('LEDGER_ID').AsString;
          with AccountsHub do
            NomControlLedger := Ledgers.ComponentByID(CLedgerID);
          with PbUpdatedLedgers do
            if (IndexOf(NomControlLedger) < 0) then
              Add(NomControlLedger);
          Value := Dr - Cr; // credit is -
          with NomControlLedger do
          begin
            if not Active then
              Open;
            if NomControlLedger is TLKLedgerContact then
              AccountPost([LedgerTempURN, CRef, JournalDate, DaybookID, JnlNo,
                NomCode + ' ' + Description,
                  CurrencyOrNull(Dr), CurrencyOrNull(Cr), value]);
            if NomControlLedger is TLKLedgerCash then
              AccountPost([LedgerTempURN, CRef, JournalDate, DaybookID, JnlNo,
                NomCode + ' ' + Description,
                  CurrencyOrNull(Dr), CurrencyOrNull(Cr)]);
          end;
        end;
        Dec(LedgerTempURN);
      until not FindNext;
end;

procedure TLKDaybookJournal.PbProcessedComplete;
begin
  PbUpdatedLedgers.Clear;
  inherited PbProcessedComplete;
end;

procedure TLKDaybookJournal.PbTablesApplyUpdates;
var
  I: Integer;
begin
  inherited PbTablesApplyUpdates;
  with PbUpdatedLedgers do
    for I := 0 to Count - 1 do
      Items[I].ApplyUpdates(0);
end;

procedure TLKDaybookJournal.PbTablesCancelUpdates;
var
  I: Integer;
begin
  with PbUpdatedLedgers do
    for I := 0 to Count - 1 do
      Items[I].CancelUpdates;
  inherited PbTablesCancelUpdates;
end;

procedure TLKDaybookJournal.PbTablesRefresh;
var
  I: Integer;
begin
  with PbUpdatedLedgers do
    for I := 0 to Count - 1 do
      Items[I].TblDetail.Refresh;
  inherited PbTablesRefresh;
end;

procedure TLKDaybookJournal.SetLedger(Value: TLKLedgerNominal);
begin
  if (Value = NominalLedger) then
    inherited Ledger := Value;
end;


{
***************************** TLKHubAccountsDiary ******************************
}
constructor TLKHubAccountsDiary.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  TblDetail.TableName := 'DIARY_DTL';
  TblDetail.MasterFields := 'DIARY_DATE';
end;


{
******************************** TLKHubAccounts ********************************
}
constructor TLKHubAccounts.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAccountsInterval := aciUnknown;
  FAccountsIntervalDays := 0;
  FDlgSetupIsImplemented := True;
end;

procedure TLKHubAccounts.CreateDiary;
begin
  FDiary := TLKHubAccountsDiary.Create(Self);
end;

function TLKHubAccounts.DlgEnterpriseExecute: TModalResult;
begin
  with TDlgLKHubAccountsEnterprise.Create(Owner) do
  try
    Construct(Self);
    Result := Showmodal;
  finally Free;
  end;
end;

procedure TLKHubAccounts.DlgP11DExecute(Sender: TLKDataComponent);
begin
  if Assigned(P11D) then
    with TLKDocumentP11DDlg.Create(nil) do
    try
      MyP11D := P11D;
      Execute(Sender);
    finally
      Free;
    end;
end;

procedure TLKHubAccounts.DrillDownDiary;
var
  LedgerRef: string;
  L: TLKLedgerBase;
begin
  with Diary.TblDetail do
  begin
    L := FLedgers.ComponentByID(FieldByName('LEDGER_ID').asString);
    LedgerRef := FieldByName('LEDGER_REFERENCE').asString;
  end;
  if not Assigned(L) or (LedgerRef = '') then
    Exit;
  L.View(LedgerRef, False, []);
  L.ViewRecord(LedgerRef, True, [mbOK]);
end;

procedure TLKHubAccounts.DrillDownSalesTaxReturn(const DaybookID,
    SalesTaxPeriod: string);
var
  OldFilter: string;
  WasFiltered: Boolean;
  D: TLKDaybookBase;
begin
  D := Daybooks.ComponentByID(DaybookID);
  if Assigned(D) then
  with D do
  begin
    if not ACGroup.Active then
      ACGroup.Open;
    if Assigned(InputDocuments) then
    begin
      with InputDocuments do
      begin
        OldFilter := Filter;
        WasFiltered := Filtered;
        Filter := '';
        Filtered := False;
      end;
      Open;
    end
    else
    begin
      OldFilter := '';
      WasFiltered := False;
    end;
    Filter := 'VAT_PERIOD = ' + Chr(39) + SalesTaxPeriod + Chr(39);
    Filtered := True;
    try
      View(Null, True, [mbOK]);
    finally
      if Assigned(InputDocuments) then
        with InputDocuments do
        begin
          Filter := Oldfilter;
          Filtered := Wasfiltered;
        end;
    end;
  end;
end;

function TLKHubAccounts.EndOfYear(ProgressText: string): TModalResult;
begin
  Result := inherited EndOfYear(ProgressText);
  if (Result = mrOK) then
    with TLKViewDialog.Create(Owner) do
    begin
      try
        DataType := vDate;
        Title := 'New Accounts Year';
        HeaderCaption := 'Confirm the new year end date';
        HeaderComments := 'the last date in the new accounting period';
        Value := MonthEndThis(AccountsPeriodEnd + 360);
        if not Execute then
        begin
          Showmessage('Year end date not changed from ' +
            DateToStr(AccountsPeriodEnd));
          Result := mrCancel;
        end;
        if (AccountsPeriodEnd = Value) then
        begin
          Showmessage('Year end date not changed from ' +
            DateToStr(AccountsPeriodEnd));
          Result := mrCancel;
        end;
        if (Result = mrOK) then
        begin
          AccountsPeriodStart := AccountsPeriodEnd + 1;
          AccountsPeriodEnd := Value;
        end;
      finally Free;
      end;
    end;
end;

function TLKHubAccounts.GetAccountsInterval: TLKAccountsInterval;
begin
  if (FAccountsInterval = aciUnknown) and not (csLoading in ComponentState) then
    FAccountsInterval := TLKAccountsInterval(
      RecallInteger(EnterpriseCurrent, 'AccountsInterval', Ord(aciMonthly),
      True));
  Result := FAccountsInterval;
end;

function TLKHubAccounts.GetAccountsIntervalDays: Integer;
begin
  if (FAccountsIntervalDays < 1) and not (csLoading in ComponentState) then
    FAccountsIntervalDays := RecallInteger(
      EnterpriseCurrent, 'AccountsIntervalDays', 30, True);
  Result := FAccountsIntervalDays;
end;

function TLKHubAccounts.GetAccountsPeriodEnd: TDateTime;
begin
  Result := 0;
  if not (csLoading in ComponentState) then
    Result := RecallDate(EnterpriseCurrent, 'AccountsPeriodEnd',
      MonthEndThis(date + 335.0), True);
end;

function TLKHubAccounts.GetAccountsPeriodStart: TDateTime;
begin
  Result := 0;
  if not (csLoading in ComponentState) then
    Result := RecallDate(EnterpriseCurrent, 'AccountsPeriodStart',
      MonthStartThis(date), True);
end;

function TLKHubAccounts.GetAccountsYear: string;
var
  d, m, y: Word;
begin
  Result := '';
  if csLoading in ComponentState then
    exit;
  DecodeDate(AccountsPeriodEnd, y, m, d);
  Result := intToStr(y);
end;

function TLKHubAccounts.GetDiary: TLKHubAccountsDiary;
begin
  Result := TLKHubAccountsDiary(inherited Diary);
end;

function TLKHubAccounts.GetSalesTaxRegistration: string;
begin
  if Assigned(SalesTaxReturns) then
    Result := SalesTaxReturns.SalesTaxRegistration
  else
    Result := '';
end;

procedure TLKHubAccounts.Notification(AComponent: TComponent; Operation:
    TOperation);
begin
  if (Operation = opRemove) then
  begin
    if (AComponent = FJournal) then
      FJournal := nil;
    if (AComponent = FNominalLedger) then
      FNominalLedger := nil;
    if (AComponent = FCostLedger) then
      FCostLedger := nil;
    if (AComponent = FSalesTaxReturns) then
      FSalesTaxReturns := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

function TLKHubAccounts.PeriodEndThis(ADate: TDateTime): TDateTime;
var
  ay, y, m, d: Word;
  n: Double;
begin
  case AccountsInterval of
    aciWeekly:
      begin
        n := (AccountsPeriodEnd - Date) / 7;
        if (n < 0) then
          n := -n;
        Result := Round(ADate + Frac(n) * 7);
      end;
    aciMonthly: Result := MonthEndThis(ADate);
    aciAnnual:
      begin
        DecodeDate(ADate, ay, m, d);
        DecodeDate(AccountsPeriodEnd, y, m, d);
        Result := EncodeDate(ay, m, d);
      end;
    //aciDays
  else
    Result := ADate;
  end;
end;

procedure TLKHubAccounts.SetAccountsInterval(Value: TLKAccountsInterval);
begin
  if FAccountsInterval <> Value then
  begin
    RememberInteger(EnterpriseCurrent, 'AccountsInterval', Ord(Value), True);
    FAccountsInterval := Value;
  end;
end;

procedure TLKHubAccounts.SetAccountsIntervalDays(Value: Integer);
begin
  if FAccountsIntervalDays <> Value then
  begin
    RememberInteger(EnterpriseCurrent, 'AccountsIntervalDays', Value, True);
    FAccountsIntervalDays := Value;
  end;
end;

procedure TLKHubAccounts.SetAccountsPeriodEnd(Value: TDateTime);
begin
  if not (csLoading in ComponentState) then
    RememberDate(EnterpriseCurrent, 'AccountsPeriodEnd', Value, True);
end;

procedure TLKHubAccounts.SetAccountsPeriodStart(Value: TDateTime);
begin
  if not (csLoading in ComponentState) then
    RememberDate(EnterpriseCurrent, 'AccountsPeriodStart', Value, True);
end;

procedure TLKHubAccounts.SetCostLedger(Value: TLKLedgerCost);
begin
  if assigned(Value) then
    Value.FreeNotification(Self);
  //use components.AccountsHub to refresh NominalLedger
  //with Daybooks do for i := 0 to Count-1 do TLKDaybookSalesTax(Items[i]).AccountsHub := Self;
  //with Documents do for i := 0 to Count-1 do TLKRecordsAccounts(Items[i]).AccountsHub := Self;
  //with Ledgers do for i := 0 to Count-1 do
   //if (TLKLedgerAccounts(Items[i]) <> Value) // avoid repeated call
  // then TLKLedgerAccounts(Items[i]).AccountsHub := Self;
  FCostLedger := Value;
end;

procedure TLKHubAccounts.SetJournal(Value: TLKJournal);
begin
  if assigned(Value) then
    Value.FreeNotification(Self);
  FJournal := Value;
end;

procedure TLKHubAccounts.SetNominalLedger(Value: TLKLedgerNominal);
begin
  //use components.AccountsHub to refresh NominalLedger
  //with Daybooks do for i := 0 to Count-1 do TLKDaybookSalesTax(Items[i]).AccountsHub := Self;
  //with Documents do for i := 0 to Count-1 do TLKRecordsAccounts(Items[i]).AccountsHub := Self;
  //with Ledgers do for i := 0 to Count-1 do
   //if (TLKLedgerAccounts(Items[i]) <> Value) // avoid repeated call
  // then TLKLedgerAccounts(Items[i]).AccountsHub := Self;
  FNominalLedger := Value;
  if assigned(Value) then
    Value.FreeNotification(Self);
end;

procedure TLKHubAccounts.SetP11D(Value: TLKDataComponent);
begin
  FP11D := Value;
  if assigned(Value) then
    Value.FreeNotification(Self);
end;

procedure TLKHubAccounts.SetSalesTaxRegistration(const Value: string);
begin
  if ((Value > '') and Assigned(SalesTaxReturns)) then
    SalesTaxReturns.SalesTaxRegistration := Value;
end;

procedure TLKHubAccounts.SetSalesTaxReturns(Value: TLKSalesTaxReturns);
begin
  if assigned(Value) then
    Value.FreeNotification(Self);
  FSalesTaxReturns := Value;
end;


initialization

  RegisterClasses([TLKHubAccounts, TLKHubAccountsDiary,
    TLKLedgerAnalysisAccountsPeriods, TLKLedgerAnalysisDeferredPostings,
      TLKLedgerAnalysisReportLayouts,
      TLKLedgerAnalysisReportLines, TLKLedgerNominalNominalCodes,
      TLKLedgerNominalFixedAssetCategories, TLKLedgerNominalFixedAssets,
      TLKLedgerNominalAccruals, TLKLedgerNominalPrepayments,
      TLKLedgerNominalEnterOpeningBalances, TLKLedgerNominalEnterReserves,
      TLKLedgerNominalManagementAccounts, TLKLedgerNominalBudget,
      TLKLedgerCostCostCentres, TLKLedgerCostManagementAccounts,
      TLKLedgerCash, TLKDaybookPettyCash, TLKDocumentsPettyCash,
      TLKLedgerBank,
      TLKSalesTaxCodes, TLKSalesTaxReturns,
      //TLKDaybookJournal,
    TLKJournalSkeletons,
      TLKSalesManagers, TLKLedgerSales, TLKDaybookSales,
      TLKDocumentsSalesInvoices,
      TLKLedgerPurchase, TLKDaybookPurchase, TLKDocumentsPurchaseInvoices,
      TLKBankTransferPaymentsList]);

end.
